//      -------------------------------------------------------------------------
//
//      ACCESSRY.CPP
//
//      -------------------------------------------------------------------------
//
//      AUTHOR:                 Darren J. Dwyer
//
//      DESCRIPTION:    This file contains code for the Accessory base class as
//                                      well as functions for installing accessory bitmaps, etc.
//                                      eg. InstallGUI() and ExitGUI()
//
//      -------------------------------------------------------------------------

#include "mouse.hpp"
#include "accessry.hpp"
#include "strings.hpp"

//      -------------------------------------------------------------------------
//
//      FUNC:           CONSTRUCTOR for Accessory
//      DESC:       This is the base constructor for _all_ Accessory's.
//
//      PARAMS:         x1,y1,x2,y2                     - the relative (to the handler's inset_r)
//                                                                        coordinates of the accessory
//                              fg,bg                           - the foreground and background colors
//                                                                        when inactive
//                              active_border,
//                              inactive_border         - the border style for active & inactive
//                              options                         - the options for the accessory
//                              retevent                        - the event the accessory returns to
//                                                                        its eventhandler
//
//      -------------------------------------------------------------------------
Accessory::Accessory(int x1,int y1,int x2,int y2,
					 int options,int ret_event)
	{
	// set the coords of the accessory
	r.left = x1; r.right = x2; r.top = y1; r.bottom = y2;
	loc_r.left = x1; loc_r.right = x2; loc_r.top = y1; loc_r.bottom = y2;

	// set default options
	selected_options = options;

	// set the event to be returned
	LMBreturnEvent = ret_event;
	RMBreturnEvent = ret_event;

	// by default, all accessories are selectable...
	isActive = true;
	// we are not currently selected...
	isSelected = false;
	}

Accessory::Accessory(int x1,int y1,int x2,int y2,
					 int options,int lmb_event,int rmb_event)
	{
	// set the coords of the accessory
	r.left = x1; r.right = x2; r.top = y1; r.bottom = y2;
	loc_r.left = x1; loc_r.right = x2; loc_r.top = y1; loc_r.bottom = y2;

	// set default options
	selected_options = options;

	// set the event to be returned
	LMBreturnEvent = lmb_event;
	RMBreturnEvent = rmb_event;

	// by default, all accessories are selectable...
	isActive = true;
	// we are not currently selected...
	isSelected = false;
	}

//      -------------------------------------------------------------------------
//
//      FUNC:           DESTRUCTOR for Accessory
//      DESC:       This is the base destructor for _all_ Accessory's.
//
//      -------------------------------------------------------------------------
Accessory::~Accessory()
	{
//	HEAPCHK("by Accessory DESTRUCTOR");
	}

//      -------------------------------------------------------------------------
//
//      PARAMS:         dx      - +ve or -ve value to adjust coords (no bounds checking)
//                              dy      - +ve or -ve value to adjust coords (no bounds checking)
//
//      -------------------------------------------------------------------------
void 	Accessory::AdjustCoords(int dx,int dy)
	{
	// adjust global coords
	r.left += dx; r.right += dx;
	r.top  += dy; r.bottom += dy;
	}

//      -------------------------------------------------------------------------
//
//      FUNC:           Accessory::MouseInside()
//      DESC:           Returns true if the mouse is inside the accessory
//      NOTE:           Can be used for any general point....
//
//      -------------------------------------------------------------------------
bool 	Accessory::MouseInside(grPoint& p)
	{
	if (isActive)
		return ((p.x>=r.left)&&(p.x<=r.right)&&
				(p.y>=r.top)&&(p.y<=r.bottom));
	else
		return false;
	}

void	Accessory::SetEvents(int lmb_event,int rmb_event)
	{
	LMBreturnEvent = lmb_event;
	RMBreturnEvent = rmb_event;
	}

//      -------------------------------------------------------------------------
//
//      FUNC:           Accessory::Select()
//      DESC:       This function selects the accessory.
//
//      -------------------------------------------------------------------------
void Accessory::Select()
	{
	if (!isActive) return;

	if (!isSelected)
		{
		isSelected = true;
		Refresh();
		}
	}

//      -------------------------------------------------------------------------
//
//      FUNC:           Accessory::DeSelect()
//      DESC:       This function deselects the accessory.
//
//      -------------------------------------------------------------------------
void 	Accessory::DeSelect()
	{
	if (!isActive) return;

	if (isSelected)
		{
		isSelected = false;
		Refresh();
		}
	}

void	Accessory::Activate()
	{
	isActive = true;
	}

void	Accessory::DeActivate()
	{
	isActive = false;
	}

//      -------------------------------------------------------------------------
//
//      FUNC:           Accessory::Draw()
//      DESC:       A simply draw function which fills the area underneath
//                              the accessory and then draws a pretty border around this.
//
//      -------------------------------------------------------------------------
void 	Accessory::Draw()
	{
	// if the accessory is active
	if (isActive)
		{
		// and it's selected, draw the selected accessory
		if (isSelected) ;
		// otherwise draw it as unselected
		else			;
		}
	else
		// draw the accessory as inactive
		{
		}
	}

void	Accessory::Draw(grVirtualBuffer* vb)
	{
	if (!vb) return;

	if (isActive)
		{
		if (isSelected)	;
		else			;
		}
	else
		{
		}
	}

void 	Accessory::Refresh()
	{
	Draw();
	grBlitRegion(r.left,r.top,r.right,r.bottom);
	}

//      -------------------------------------------------------------------------
//
//      FUNC:           Accessory::HandleEvent()
//      DESC:           This is the event-interpretation function for all Accessories.
//                              Any event that is meant to go to an accessory arrives & is
//                              interpreted here.
//
//      -------------------------------------------------------------------------
void 	Accessory::HandleEvent(Event &e)
	{
	// does nothing by default
	}


