//	-------------------------------------------------------------------------
//
//	BUTTONS.HPP
//
//	-------------------------------------------------------------------------
//
//	AUTHOR:			Darren J. Dwyer
//
//	DESCRIPTION:	This file contains defines for Accessory-derived classes
//					such as TextButton and PictureButton. See GUI.HPP for
//					TextButton & PictureButton derived objects such as
//					OKButton, UpArrowButton, etc.
//
//	-------------------------------------------------------------------------

#ifndef __buttons_hpp
#define __buttons_hpp

#include "accessry.hpp"
#include "graphics.hpp"
#include "fonts.hpp"

//	-------------------------------------------------------------------------
//
//	CLASS: 		Button
//	DESC: 		This is a button which is 'depressed' while the mouse
//				is down and is 'released' while the mouse is up.
//
//				The button responds as per:
//
//					MODE			ACTION			RESPONSE
//
//					default &		mb_leftup		lmb_event
//					toggle_mode		mb_rightup		rmb_event
//
//					auto_mode		mb_leftauto		lmb_event
//									mb_rightauto	rmb_event
//
//
//	-------------------------------------------------------------------------

// button options

#define	ac_AutoMode				0x0004
#define	ac_ToggleMode			0x0008


class Button : public Accessory
	{
	public:
		Button(int x1,int y1,int x2,int y2,
				int options,int lmb_event,int rmb_event )
			: Accessory(x1,y1,x2,y2,options,lmb_event,rmb_event) {}

		Button(int x1,int y1,int x2,int y2,
				int options,int ret_event)
			: Accessory(x1,y1,x2,y2,options,ret_event,ret_event) {}

		virtual	void	Select();
		virtual	void	DeSelect();

		virtual	void	Toggle();

		virtual void 	HandleEvent(Event &e);
		virtual	void	HandleToggleEvent(Event& e);
	};

//	-------------------------------------------------------------------------
//
//	CLASS:		TextButton
//	DESC:		A simple Button() which displays a text string
//
//	-------------------------------------------------------------------------

class TextButton : public Button
	{
	private:
		char* 		text_ptr;
		char		xofs;
		char		yofs;

		int			font;

		char		active_text;
		char		inactive_text;
		char		active_bg;
		char		inactive_bg;
		int			active_bs;
		int			inactive_bs;

	public:
		TextButton(int x,int y,
					int	font_id,char* text,
					int options,int ret_event);

		TextButton(int x,int y,
					int font_id,char* text,
					int options,int lmb_event,int rmb_event);

		TextButton(int x1,int y1,
					int font_id,char* text,
					char active_txt,char active_bkg,int active_border,
					char inactive_txt,char inactive_bkg,int inactive_border,
					int options,int ret_event);

		TextButton(int x1,int y1,
					int font_id,char* text,
					char active_txt,char active_bkg,int active_border,
					char inactive_txt,char inactive_bkg,int inactive_border,
					int options,int lmb_event,int rmb_event);

		virtual void 	Draw();
		virtual	void 	Draw(grVirtualBuffer* vb);
	};

//	-------------------------------------------------------------------------
//
//	CLASS:		PictureButton
//	DESC:		A simple button displaying two pics
//
//	-------------------------------------------------------------------------

class PictureButton : public Button
	{
	protected:
		grImage*	active_pic;
		grImage*	inactive_pic;

		int			trans;

	public:
		// assuming no transparency
		PictureButton(int x1,int y1,
						grImage* active_image,
						grImage* inactive_image,
						int options,
						int ret_event);

		PictureButton(int x1,int y1,
						grImage* active_image,
						grImage* inactive_image,
						int options,
						int lmb_event,int rmb_event);

		// assumes transparency
		PictureButton(int x1,int y1,
						int transparent_color,
						grImage* active_image,
						grImage* inactive_image,
						int options,
						int ret_event);

		PictureButton(int x1,int y1,
						int transparent_color,
						grImage* active_image,
						grImage* inactive_image,
						int options,
						int lmb_event,int rmb_event);

		// a quick check to see whether or not the mouse is inside 'r'
		virtual bool 	MouseInside(grPoint& p);

		// sets the appropriate picture 'on the fly' using same draw mode
		virtual void 	SetActivePic(grImage* active_image);
		virtual void 	SetInactivePic(grImage* inactive_image);

		virtual void 	SetPics(grImage* active_image,grImage* inactive_image);

		// draws the button
		virtual void 	Draw();
		virtual	void 	Draw(grVirtualBuffer* vb);
		};


#endif