//      -------------------------------------------------------------------------
//
//      ACCESSRY.CPP
//
//      -------------------------------------------------------------------------
//
//      AUTHOR:                 Darren J. Dwyer
//
//      DESCRIPTION:    This file contains code for the Accessory base class as
//                                      well as functions for installing accessory bitmaps, etc.
//                                      eg. InstallGUI() and ExitGUI()
//
//      -------------------------------------------------------------------------

#include "mouse.hpp"
#include "accessry.hpp"
#include "strings.hpp"

//      -------------------------------------------------------------------------
//
//      FUNC:           CONSTRUCTOR for Accessory
//      DESC:       This is the base constructor for _all_ Accessory's.
//
//      PARAMS:         x1,y1,x2,y2                     - the relative (to the handler's inset_r)
//                                                                        coordinates of the accessory
//                              fg,bg                           - the foreground and background colors
//                                                                        when inactive
//                              active_border,
//                              inactive_border         - the border style for active & inactive
//                              options                         - the options for the accessory
//                              retevent                        - the event the accessory returns to
//                                                                        its eventhandler
//
//      -------------------------------------------------------------------------
Accessory::Accessory(int x1,int y1,int x2,int y2,uchar active_clr,uchar inactive_clr,
					 int active_border,int inactive_border,
					 int options,int retevent)
	{
	// set the active & inactive fill colors
	active_color = active_clr; inactive_color = inactive_clr;

	// set the border styles
	active_bs = active_border; inactive_bs = inactive_border;

	// set the coords of the accessory
	r.left = x1; r.right = x2; r.top = y1; r.bottom = y2;
	loc_r.left = x1; loc_r.right = x2; loc_r.top = y1; loc_r.bottom = y2;

	// calculate max size required for the border
	int bs = max(grSizeOfBorder[active_border],grSizeOfBorder[inactive_border]);

	// and define our inset_r rectangle
	inset_r.left = r.left + bs; inset_r.right = r.right - bs;
	inset_r.top  = r.top + bs; inset_r.bottom = r.bottom - bs;

	loc_inset_r.left = inset_r.left;
	loc_inset_r.top  = inset_r.top;
	loc_inset_r.right  = inset_r.right;
	loc_inset_r.bottom = inset_r.bottom;

	// set default options
	selected_options = options;

	// set the event to be returned
	returnEvent = retevent;

	// by default, all accessories are selectable...
	isActive = true;
	// we are not currently selected...
	isSelected = false;
	}

//      -------------------------------------------------------------------------
//
//      FUNC:           DESTRUCTOR for Accessory
//      DESC:       This is the base destructor for _all_ Accessory's.
//
//      -------------------------------------------------------------------------
Accessory::~Accessory()
	{
//	HEAPCHK("by Accessory DESTRUCTOR");
	}

//      -------------------------------------------------------------------------
//
//      PARAMS:         dx      - +ve or -ve value to adjust coords (no bounds checking)
//                              dy      - +ve or -ve value to adjust coords (no bounds checking)
//
//      -------------------------------------------------------------------------
void Accessory::AdjustCoords(int dx,int dy)
	{
	// adjust global coords
	r.left += dx; r.right += dx;
	r.top  += dy; r.bottom += dy;

	inset_r.left += dx; inset_r.right += dx;
	inset_r.top  += dy; inset_r.bottom += dy;
	}

//      -------------------------------------------------------------------------
//
//      FUNC:           Accessory::MouseInside()
//      DESC:           Returns true if the mouse is inside the accessory
//      NOTE:           Can be used for any general point....
//
//      -------------------------------------------------------------------------
bool Accessory::MouseInside(grPoint& p)
	{
	return ((p.x>=r.left)&&(p.x<=r.right)&&
			(p.y>=r.top)&&(p.y<=r.bottom));
	}

//      -------------------------------------------------------------------------
//
//      FUNC:           Accessory::Select()
//      DESC:       This function selects the accessory.
//
//      -------------------------------------------------------------------------
void Accessory::Select()
	{
	if (!isActive) return;

	if (!isSelected)
		{
		isSelected = true;
		Refresh();
		}
	}

//      -------------------------------------------------------------------------
//
//      FUNC:           Accessory::DeSelect()
//      DESC:       This function deselects the accessory.
//
//      -------------------------------------------------------------------------
void Accessory::DeSelect()
	{
	if (!isActive) return;

	if (isSelected)
		{
		isSelected = false;
		Refresh();
		}
	}

void	Accessory::Activate()
	{
	isActive = true;
	}

void	Accessory::DeActivate()
	{
	isActive = false;
	}

//      -------------------------------------------------------------------------
//
//      FUNC:           Accessory::Draw()
//      DESC:       A simply draw function which fills the area underneath
//                              the accessory and then draws a pretty border around this.
//
//      -------------------------------------------------------------------------
void Accessory::Draw()
	{
	// if the accessory is currently selected AND it is active, display
	// the active piccy
	if (isActive)
		{
		if (isSelected)
			grDrawWindow(r.left,r.top,r.right,r.bottom,active_color,active_bs);
		else
			grDrawWindow(r.left,r.top,r.right,r.bottom,inactive_color,inactive_bs);
		}
	else
		{
		grDrawWindow(r.left,r.top,r.right,r.bottom,active_color,active_bs);
		grDrawShadedRect(r.left,r.top,r.right,r.bottom,active_color,2);
		}
	}

void	Accessory::Draw(grVirtualBuffer* vb)
	{
	if (!vb) return;

	// if the accessory is currently selected AND it is active, display
	// the active piccy
	if (isActive)
		{
		if (isSelected)
			grDrawWindowVirtual(vb,loc_r.left,loc_r.top,loc_r.right,loc_r.bottom,active_color,active_bs);
		else
			grDrawWindowVirtual(vb,loc_r.left,loc_r.top,loc_r.right,loc_r.bottom,inactive_color,inactive_bs);
		}
	else
		{
		grDrawWindowVirtual(vb,loc_r.left,loc_r.top,loc_r.right,loc_r.bottom,active_color,active_bs);
		grDrawShadedRectVirtual(vb,loc_r.left,loc_r.top,loc_r.right,loc_r.bottom,active_color,2);
		}
	}

void Accessory::Refresh()
	{
	Draw();
	grBlitRegion(r.left,r.top,r.right,r.bottom);
	}

//      -------------------------------------------------------------------------
//
//      FUNC:           Accessory::HandleEvent()
//      DESC:           This is the event-interpretation function for all Accessories.
//                              Any event that is meant to go to an accessory arrives & is
//                              interpreted here.
//
//      -------------------------------------------------------------------------
void Accessory::HandleEvent(Event &e)
	{
	// does nothing by default
	}


