//	-------------------------------------------------------------------------
//
//	BUTTONS.CPP
//
//	-------------------------------------------------------------------------
//
//	AUTHOR:			Darren J. Dwyer
//
//	DESCRIPTION:	This file contains code for Button-derived classes
//					such as TextButton and PictureButton. It also has code for
//					TextButton & PictureButton derived objects such as
//					OKButton, UpArrowButton, etc.
//
//	-------------------------------------------------------------------------

#include "buttons.hpp"
#include "keyboard.hpp"
#include "keydefns.hpp"
#include "mouse.hpp"
#include "eventmgr.hpp"

#include "resource.hpp"

#include "stdlib.h"
#include "dos.h"

// the file spec, etc used to store GUI bitmaps...
#define 	GUI_BITMAP_MAX		256

ResourceFile*		GUI_BITMAP_FILE	= NULL;   	// the file for the bitmaps
grImage*  			GUI_BITMAP[GUI_BITMAP_MAX];	// the bitmap storage area
int					GUI_BITMAP_CNT	= 0;		// how many bitmaps we've loaded

//	-------------------------------------------------------------------------
//
//	FUNC:		InstallGUI
//	DESC:       A routine used to load the default bitmaps used by any of the
//				predefined system Accessories.
//
//	NOTES:		(1) _Before_ using any PictureButton accessory you MUST call
//					this function!
//
//				(2)	To modify the "look & feel" of the GUI, you can edit the
//					bitmaps in the file ACCESSRY.BTM _but_ you must keep the
//					same image order. ie. BLANK1, BLANK2, UP_ARR1, UP_ARR2,
//					DN_ARR1, DN_ARR2, etc.
//
//	-------------------------------------------------------------------------

void InstallGUI(char* path,char* gui_res)
	{
	// the install simply loads relevant bitmaps for the GUI into memory
	GUI_BITMAP_FILE = new ResourceFile(path,gui_res);
	if (!GUI_BITMAP_FILE)
		{
		DBG(470,"ERROR: Could not find GUI resourcefile %s::%s",path,gui_res);
		delay(5000);
		return;
		}

//	DBG(400,"GUI_BITMAP_FILE:     Resource Type:  %d",GUI_BITMAP_FILE->ResourceType());
//	DBG(410,"                     Max Records  :  %d",GUI_BITMAP_FILE->MaxRecords());
//	DBG(420,"                     Num Records  :  %d",GUI_BITMAP_FILE->NumRecords());

	// load all relevant images into memory
	for (int i=0;i<GUI_BITMAP_FILE->NumRecords();i++)
		{
//		DBG(470,"Attempting to install GUI record %d",i);
//		delay(500);
//		if (i>15) return;

		GUI_BITMAP[GUI_BITMAP_CNT] = (grImage*) GUI_BITMAP_FILE->ReadRecord(GUI_BITMAP_CNT);
		GUI_BITMAP_CNT++;

		}
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		ExitGUI
//	DESC:       This routine releases all resources used by InstallGUI!
//
//	-------------------------------------------------------------------------

void ExitGUI()
	{
	// blow away all loaded pictures
	if (GUI_BITMAP_CNT>0)
		for (int i=GUI_BITMAP_CNT-1;i>=0;i--)
			{
			free(GUI_BITMAP[i]);
			}

	// and delete the file
	delete GUI_BITMAP_FILE; GUI_BITMAP_FILE = NULL;
	}


//	-------------------------------------------------------------------------
//
//	FUNC:		CONSTRUCTOR for Button
//	DESC:		creates a button which is inactive displaying text
//
//	-------------------------------------------------------------------------
Button::Button(int x1,int y1,int x2,int y2,
				uchar active_clr,uchar inactive_clr,
				int active_border,int inactive_border,
				int options,int retevent)
				: Accessory(x1,y1,x2,y2,active_clr,inactive_clr,
							active_border,inactive_border,options,retevent)
	{
	isSelected = false;

//	HEAPCHK("in Button CONSTRUCTOR");
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		Button::HandleEvent()
//	DESC:		Interprets ENTER,SPACE,LEFTMB to toggle the button
//				and return an event when the button is released
//
//	-------------------------------------------------------------------------
void Button::HandleEvent(Event &e)
	{
	// a flag to specify whether or not we have handled the event
	bool handled = false;

	switch (e.what)
		{
		// if its a mouse event
		case	ev_mouse:		switch (e.mouse.event)
									{
									case	mb_rightdn:	;
									case	mb_leftdn:	if (isActive)
															{
															Select();
															EVENTMGR->ClearEvent(e);
															}
														break;
									case	mb_rightauto:	;
									case	mb_leftauto:if (isActive)
															{
															Select();
															if ((selected_options&ac_AutoRepeats)!=0)
																handled = true;
															}
														break;
									case	mb_rightup:	;
									case	mb_leftup:  if (isActive)
															{
															DeSelect();
//															if ((selected_options&ac_AutoRepeats)==0)
															handled = true;
															}
														break;
									}
								break;
		// if its a keyboard event
		case	ev_keyboard:	switch (e.keyboard.event)
									{
									case	ENTER_DN:	if (isActive)
															{
															Select();
															EVENTMGR->ClearEvent(e);
															}
														break;
									case	ENTER_UP:	if (isActive)
															{
															DeSelect();
															handled = true;
															}
														break;
									}
								break;

		}

	// if the event was handled, generate our return event
	if (handled)
		{
		CommandEvent c = { returnEvent, NULL };
		EVENTMGR->ModEvent(e,e.where,ev_command,&c);
		}
	}

TextButton::TextButton(int x,int y,char* text,int options,int retevent)
	: Button(x,y,
			 x+2*grSizeOfBorder[bs_RAISED]+grFontWidth(text)+4,
			 y+2*grSizeOfBorder[bs_RAISED]+grFontHeight()+4,
			 grACCESSORY_FG,grACCESSORY_BG,bs_LOWERED,bs_RAISED,
			 options,retevent)
	{
	textptr = text;
	}

TextButton::TextButton(int x,int y,char* text,
						uchar active_clr,uchar inactive_clr,
						int active_border,int inactive_border,
						int options,int retevent )
	: Button(x,y,
			 x+2*grSizeOfBorder[active_border]+grFontWidth(text)+4,
			 y+2*grSizeOfBorder[active_border]+grFontHeight()+4,
			 active_color,inactive_color,active_border,inactive_border,
			 options,retevent)
	{
	textptr = text;
	}

//	-------------------------------------------------------------------------
//
//	FUNC: 		TextButton::Draw
//	DESC:		Draw function for a Text Button
//
//	-------------------------------------------------------------------------
void TextButton::Draw()
	{
	if (isActive)
		{
		if (isSelected)
			{
			grDrawWindow(r.left,r.top,r.right,r.bottom,active_color,active_bs);
			grDrawString(inset_r.left+2,inset_r.top+2,"%s",textptr);
			}
		else
			{
			grDrawWindow(r.left,r.top,r.right,r.bottom,inactive_color,inactive_bs);
			grDrawString(inset_r.left+2,inset_r.top+2,"%s",textptr);
			}
		}
	else
		{
		grDrawWindow(r.left,r.top,r.right,r.bottom,active_color,active_bs);
		grDrawString(inset_r.left+2,inset_r.top+2,"%s",textptr);
		grDrawShadedRect(r.left,r.top,r.right,r.bottom,active_color,2);
		}
	}


void TextButton::Draw(grVirtualBuffer* vb)
	{
	if (!vb) return;

	if (isActive)
		{
		if (isSelected)
			{
			grDrawWindowVirtual(vb,loc_r.left,loc_r.top,loc_r.right,loc_r.bottom,active_color,active_bs);
			grDrawStringVirtual(vb,loc_inset_r.left+2,loc_inset_r.top+2,"%s",textptr);
			}
		else
			{
			grDrawWindowVirtual(vb,loc_r.left,loc_r.top,loc_r.right,loc_r.bottom,inactive_color,inactive_bs);
			grDrawStringVirtual(vb,loc_inset_r.left+2,loc_inset_r.top+2,"%s",textptr);
			}
		}
	else
		{
		grDrawWindowVirtual(vb,loc_r.left,loc_r.top,loc_r.right,loc_r.bottom,active_color,active_bs);
		grDrawStringVirtual(vb,loc_inset_r.left+2,loc_inset_r.top+2,"%s",textptr);
		grDrawShadedRectVirtual(vb,loc_r.left,loc_r.top,loc_r.right,loc_r.bottom,active_color,2);
		}
	}

//	-------------------------------------------------------------------------
//
//	FUNC: 		CONSTRUCTORs for PictureButtons
//	DESC:
//
//	-------------------------------------------------------------------------

// assumes color 0 is transparent, no borders
PictureButton::PictureButton(int x1,int y1,grImage* active_pic,grImage* inactive_pic,
							 int options,int retevent)
	: Button( x1,y1,x1+active_pic->hdr.width-1,y1+active_pic->hdr.height-1,
			  0,0,bs_NONE,bs_NONE,options,retevent)
	{
	active_image = active_pic;
	inactive_image = inactive_pic;
	transparent_color = 0;
	isSelected = false;
	}

// assumes color supplied is transparent, no borders
PictureButton::PictureButton(int x1,int y1,grImage* active_pic,grImage* inactive_pic,
							 int tr_color,int options,int retevent)
	: Button( x1,y1,x1+active_pic->hdr.width-1,y1+active_pic->hdr.height-1,
			  0,0,bs_NONE,bs_NONE,options,retevent)
	{
	active_image = active_pic;
	inactive_image = inactive_pic;
	transparent_color = tr_color;
	isSelected = false;
	}

// assumes borders, no transparency
PictureButton::PictureButton(int x1,int y1,int active_border,int inactive_border,
							 grImage* active_pic,grImage* inactive_pic,
							 int options,int retevent)
	: Button( x1,y1,
			  x1+active_pic->hdr.width-1 + 2 * grSizeOfBorder[active_border],
			  y1+active_pic->hdr.height-1 + 2 * grSizeOfBorder[active_border],
			  0,0,active_border,inactive_border,options,retevent)
	{
	active_image = active_pic;
	inactive_image = inactive_pic;
	transparent_color = -1;
	isSelected = false;
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		PictureButton::Draw()
//	DESC:		Draws the picture button
//
//	-------------------------------------------------------------------------
void PictureButton::Draw()
	{
	if (isActive)
		{
		if (isSelected)
			{
			if (active_bs) grDrawWindow(r.left,r.top,r.right,r.bottom,active_color,active_bs);
			if (active_image)
				{
				if (transparent_color==-1) grPutImage(inset_r.left,inset_r.top,inset_r.right,inset_r.bottom,active_image);
				else grOverlayImage(inset_r.left,inset_r.top,inset_r.right,inset_r.bottom,transparent_color,active_image);
				}
			}
		else
			{
			if (inactive_bs) grDrawWindow(r.left,r.top,r.right,r.bottom,inactive_color,inactive_bs);
			if (inactive_image)
				{
				if (transparent_color==-1) grPutImage(inset_r.left,inset_r.top,inset_r.right,inset_r.bottom,inactive_image);
				else grOverlayImage(inset_r.left,inset_r.top,inset_r.right,inset_r.bottom,transparent_color,inactive_image);
				}
			}
		}
	else
		{
		if (active_bs) grDrawWindow(r.left,r.top,r.right,r.bottom,active_color,active_bs);
		if (active_image)
			{
			if (transparent_color==-1) grPutImage(inset_r.left,inset_r.top,inset_r.right,inset_r.bottom,active_image);
			else grOverlayImage(inset_r.left,inset_r.top,inset_r.right,inset_r.bottom,transparent_color,active_image);
			}
		grDrawShadedRect(r.left,r.top,r.right,r.bottom,active_color,2);
		}
	}

void PictureButton::Draw(grVirtualBuffer* vb)
	{
	if (!vb) return;

	if (isActive)
		{
		if (isSelected)
			{
			if (active_bs) grDrawWindowVirtual(vb,loc_r.left,loc_r.top,
										loc_r.right,loc_r.bottom,
										active_color,active_bs);

			if (active_image)
				{
				if (transparent_color==-1)
					grPutImageVirtual(vb,loc_inset_r.left,loc_inset_r.top,
							   loc_inset_r.right,loc_inset_r.bottom,
							   active_image);
				else
					grOverlayImageVirtual(vb,loc_inset_r.left,loc_inset_r.top,
								   loc_inset_r.right,loc_inset_r.bottom,
								   transparent_color,active_image);
				}
			}
		else
			{
			if (inactive_bs) grDrawWindowVirtual(vb,loc_r.left,loc_r.top,
											loc_r.right,loc_r.bottom,
											inactive_color,inactive_bs);

			if (inactive_image)
				{
				if (transparent_color==-1)
					grPutImageVirtual(vb,loc_inset_r.left,loc_inset_r.top,
								loc_inset_r.right,loc_inset_r.bottom,
								inactive_image);
				else
					grOverlayImageVirtual(vb,loc_inset_r.left,loc_inset_r.top,
									loc_inset_r.right,loc_inset_r.bottom,
									transparent_color,inactive_image);
				}
			}
		}
	else
		{
		if (active_bs) grDrawWindowVirtual(vb,loc_r.left,loc_r.top,
									loc_r.right,loc_r.bottom,
									active_color,active_bs);

		if (active_image)
			{
			if (transparent_color==-1)
				grPutImageVirtual(vb,loc_inset_r.left,loc_inset_r.top,
							loc_inset_r.right,loc_inset_r.bottom,active_image);
			else
				grOverlayImageVirtual(vb,loc_inset_r.left,loc_inset_r.top,
								loc_inset_r.right,loc_inset_r.bottom,
								transparent_color,active_image);
			}

		grDrawShadedRectVirtual(vb,loc_r.left,loc_r.top,
							loc_r.right,loc_r.bottom,active_color,2);
		}
	}

// a quick check to see whether or not the mouse is inside 'r'
bool PictureButton::MouseInside(grPoint& p)
	{
	// find out if we're inside the button quickly....
	bool isInside = Button::MouseInside(p);

	// if we're checking for transparent colors
	if (transparent_color!=-1)
		{
		// if we are inside...
		if (isInside)
			{
			// convert global 'p' to local coordinates
			p.x -= r.left;
			p.y -= r.top;

			// return true if the pixel selected is not the transparent color
			grImage* pic = inactive_image;
			if (isSelected) pic = active_image;

			return (pic->data[p.y * pic->hdr.width + p.x] != transparent_color );
			}
		}

	// return if mouse is inside the coordinates
	return isInside;
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		CONSTRUCTOR for SmallButton
//	DESC:		creates a small blank button
//
//	-------------------------------------------------------------------------
SmallButton::SmallButton(int x,int y,int options,int retevent)
	: PictureButton(x,y,bs_LOWERED,bs_RAISED,0,0,options,retevent)
	{
	int borderspace = grSizeOfBorder[bs_DBLLOWERED]*2;

	r.right = r.left + 8 + borderspace;
	r.bottom = r.top + 8 + borderspace;

	inset_r.right = inset_r.left + 8;
	inset_r.bottom = inset_r.top + 8;

	active_color = grACCESSORY_FG;
	inactive_color = grACCESSORY_BG;
//	HEAPCHK("in SmallButton CONSTRUCTOR");
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		CONSTRUCTOR for MeduimButton
//	DESC:		create a medium-size blank button
//
//	-------------------------------------------------------------------------
MediumButton::MediumButton(int x,int y,int options,int retevent)
	: PictureButton(x,y,bs_DBLLOWERED,bs_DBLRAISED,0,0,options,retevent)
	{
	int borderspace = grSizeOfBorder[bs_DBLLOWERED]*2;

	r.right = r.left + 16 + borderspace;
	r.bottom = r.top + 16 + borderspace;

	inset_r.right = inset_r.left + 16;
	inset_r.bottom = inset_r.top + 16;

	active_color = grACCESSORY_FG;
	inactive_color = grACCESSORY_BG;
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		CONSTRUCTOR for LargeButton
//	DESC:		create a large-size blank button
//
//	-------------------------------------------------------------------------
LargeButton::LargeButton(int x,int y,int options,int retevent)
	: PictureButton(x,y,bs_DBLLOWERED,bs_DBLRAISED,0,0,options,retevent)
	{
	int borderspace = grSizeOfBorder[bs_DBLLOWERED]*2;

	r.right = r.left + 28 + borderspace;
	r.bottom = r.top + 28 + borderspace;

	inset_r.right = inset_r.left + 28;
	inset_r.bottom = inset_r.top + 28;

	active_color = grACCESSORY_FG;
	inactive_color = grACCESSORY_BG;
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		FussyButton::HandleEvent()
//	DESC:		Interprets ENTER,SPACE,LEFTMB to toggle the button
//				and return an event when the button is released
//
//	-------------------------------------------------------------------------
void FussyButton::HandleEvent(Event &e)
	{
	// a flag to specify whether or not we have handled the event
	bool handled = false;

	CommandEvent c;
	c.data = NULL;

	switch (e.what)
		{
		// if its a mouse event
		case	ev_mouse:

			switch (e.mouse.event)
				{
				case	mb_rightup:
				case	mb_leftup:

					if (isActive)
						{
						if (isSelected)
							DeSelect();
						else
							Select();

						handled = true;
						}
					break;
				}
			switch (e.mouse.event)
				{
				case	mb_rightup:
					c.event = secondEvent;
					break;
				case	mb_leftup:
					c.event = returnEvent;
					break;
				}
			break;

		// if its a keyboard event
		case	ev_keyboard:

			switch (e.keyboard.event)
				{
				case	ENTER_UP:

					if (isActive)
						{
						if (isSelected)
							DeSelect();
						else
							Select();

						handled = true;
						}

					if (keyboard_GetState() == keyboard_SHIFT)
						c.event = secondEvent;
					else
						c.event = returnEvent;
					break;
					}
				break;
		}

	// if the event was handled, generate our return event
	if (handled)
		if (returnEvent!=cmdNothing)
			{
			EVENTMGR->ModEvent(e,e.where,ev_command,&c);
			}
//	HEAPCHK("in FussyButton::HandleEvent()");
	}