//	-------------------------------------------------------------------------
//
//	BUTTONS.HPP
//
//	-------------------------------------------------------------------------
//
//	AUTHOR:			Darren J. Dwyer
//
//	DESCRIPTION:	This file contains defines for Button-derived classes
//					such as TextButton and PictureButton. It also defines
//					TextButton & PictureButton derived objects such as
//					OKButton, UpArrowButton, etc.
//
//	-------------------------------------------------------------------------

#ifndef __buttons_hpp
#define __buttons_hpp

#include "accessry.hpp"
#include "graphics.hpp"
#include "fonts.hpp"

#include "string.h"

//	-------------------------------------------------------------------------
//
//	FUNC:		InstallGUI, ExitGUI
//
//	DESC:		Installs or exits the basic GUI, including a number of
//				pre-defined buttons, vertical & horizontal scroll bars,
//				a text input bar, a vertical scrolling text box, and so on.
//
//	NOTES:		ExitGUI _must_ be called if you use InstallGUI
//
//	-------------------------------------------------------------------------

void InstallGUI(char* path,char* gui_res);
void ExitGUI();

// images declared globally
extern grImage*  	GUI_BITMAP[];	// the bitmap storage area

// default TextButton return events

#define cmdOKButton				401
#define cmdCANCELButton			402

// default PictureButton return events

#define cmdUpArrowButton 		502
#define cmdDnArrowButton 		503
#define cmdLtArrowButton 		504
#define cmdRtArrowButton 		505
#define cmdDblUpArrowButton 	506
#define cmdDblDnArrowButton 	507
#define cmdDblLtArrowButton 	508
#define cmdDblRtArrowButton 	509
#define cmdUpDnArrowButton		510
#define cmdLtRtArrowButton		511
#define cmdTickButton			512
#define cmdCrossButton			513
#define cmdQuestionMarkButton 	514
#define cmdSmallButton			515
#define	cmdMediumButton			516
#define cmdLargeButton			517

// default BlankButton return event

#define	cmdBlankButton			501

//	-------------------------------------------------------------------------
//
//	CLASS: 		Button
//	DESC: 		This is a generic button which features the following:
//                  (1)	The ability to 'toggle' between two states
//					(2) A smart toggle which is 'depressed' while the mouse
//						is down and is 'released' while the mouse is up.
//
//				Events are generated as follows:
//					(1) on mb_leftup, mb_rightup
//					(2) on mb_leftup, mb_rightup (by default)
//						on mb_leftauto, mb_rightauto (if options==ac_AutoRepeats)
//
//	-------------------------------------------------------------------------

// button options

#define	ac_AutoRepeats			0x0004
#define ac_ToggleMode			0x0008


class Button : public Accessory
	{
	public:
		Button(int x1,int y1,int x2,int y2,
				uchar active_clr,uchar inactive_clr,
				int active_border,int inactive_border,
				int options,int retevent );

		virtual void HandleEvent(Event &e);
	};

//	-------------------------------------------------------------------------
//
//	CLASS:		TextButton
//	DESC:		A simple button which displays a text string on a background
//
//	PARAMS:		x,y		- the top-left coords of the button
//				text	- the text string to display
//
//	-------------------------------------------------------------------------

class TextButton : public Button
	{
	private:
		char* textptr;
	protected:
	public:
		TextButton(int x,int y,char* text,int options,int retevent);
		TextButton(int x,int y,char* text,
					uchar active_clr,uchar inactive_clr,
					int active_border,int inactive_border,
					int options,int retevent );

		virtual void 	Draw();
		virtual	void 	Draw(grVirtualBuffer* vb);
	};

//	-------------------------------------------------------------------------
//
//	CLASS:		PictureButton
//	DESC:		A simple button which displays one of two pictures
//				depending on the state of the button.
//
//	PARAMS:		x1,y1		- the top-left coordinates of the picturebutton
//				bs,bs1		- the active & inactive borders
//				pic1,pic2	- pointers to the active & inactive pictures
//
//	-------------------------------------------------------------------------

class PictureButton : public Button
	{
	protected:
		grImage*	active_image;		// "active" picture
		grImage*	inactive_image;		// "inactive" picture

		int			transparent_color;

	public:
		// assumes color 0 is transparent, no borders
		PictureButton(int x1,int y1,grImage* active_pic,grImage* inactive_pic,
					  int options,int retevent);

		// assumes color supplied is transparent, no borders
		PictureButton(int x1,int y1,grImage* active_pic,grImage* inactive_pic,
					  int tr_color,int options,int retevent);

		// assumes borders, no transparency
		PictureButton(int x1,int y1,int active_border,int inactive_border,
					  grImage* active_pic,grImage* inactive_pic,
					  int options,int retevent);


		// a quick check to see whether or not the mouse is inside 'r'
		virtual bool MouseInside(grPoint& p);

		// sets the appropriate picture 'on the fly' using same draw mode
		virtual void SetActivePic(grImage* active_pic) { active_image = active_pic; }
		virtual void SetInactivePic(grImage* inactive_pic) { inactive_image = inactive_pic; }
		virtual void Set(grImage* active_pic,grImage* inactive_pic,int retevent)
							{ active_image = active_pic;
							  inactive_image = inactive_pic;
							  returnEvent = retevent; }

		// draws the button
		virtual void 	Draw();
		virtual	void 	Draw(grVirtualBuffer* vb);
		};


//	-------------------------------------------------------------------------
//
//	CLASS:		BlankButton
//
//	DESC:		A simple button which works the same as a normal 'button'
//				as far as selection goes, but it displays nothing!!!
//
//	-------------------------------------------------------------------------

class BlankButton : public Button
	{
		BlankButton(int x1,int y1,int x2,int y2,int options,int retevent)
			:	Button(x1,y1,x2,y2,0,0,0,0,options,retevent) {}

		virtual	void	Draw() {}
		virtual	void 	Draw(grVirtualBuffer* vb) {}
	};

//	-------------------------------------------------------------------------
//
//	PRE-DEFINED BUTTON TYPES
//
//  Most of these buttons require that you InstallGUI() & ExitGUI()
//
//	-------------------------------------------------------------------------


//	-------------------------------------------------------------------------
//
//	CLASS:		OKButton
//	DESC:		A button displaying the text "OK"
//
//	-------------------------------------------------------------------------

class OKButton : public TextButton
	{
	public:
		OKButton(int x,int y,int options,int retevent)
			: TextButton( x,y,"   OK   ",options,retevent) {}
	};

//	-------------------------------------------------------------------------
//
//	CLASS:		CANCELButton
//	DESC:		A button displaying the text "CANCEL"
//
//	-------------------------------------------------------------------------

class CANCELButton: public TextButton
	{
	public:
		CANCELButton(int x,int y,int options,int retevent)
			: TextButton( x,y," CANCEL ",options,retevent) {}
	};


//	-------------------------------------------------------------------------
//
//	CLASS:		UpArrowButton
//	DESC:		A button displaying an Up Arrow
//
//	-------------------------------------------------------------------------

class UpArrowButton : public PictureButton
	{
	public:
		UpArrowButton(int x,int y,int options,int retevent)
			   : PictureButton(x,y,GUI_BITMAP[2],GUI_BITMAP[3],options,retevent) {}

//		UpArrowButton(int x,int y,
//						int active_bs,int inactive_bs,
//						grImage* active_pic,grImage* inactive_pic,
//						int transparent_color,int options,int retevent)
//			   : PictureButton(x,y,active_bs,inactive_bs,active_pic,inactive_pic,transparent,options,retevent) {}
	};

//	-------------------------------------------------------------------------
//
//	CLASS:		DnArrowButton
//	DESC:		A button displaying a Down Arrow
//
//	-------------------------------------------------------------------------

class DnArrowButton : public PictureButton
	{
	public:
		DnArrowButton(int x,int y,int options,int retevent)
			   : PictureButton(x,y,GUI_BITMAP[4],GUI_BITMAP[5],options,retevent) {}
	};

//	-------------------------------------------------------------------------
//
//	CLASS:		LtArrowButton
//	DESC:		A button displaying a Left Arrow
//
//	-------------------------------------------------------------------------

class LtArrowButton : public PictureButton
	{
	public:
		LtArrowButton(int x,int y,int options,int retevent)
				: PictureButton(x,y,GUI_BITMAP[6],GUI_BITMAP[7],options,retevent) {}
	};

//	-------------------------------------------------------------------------
//
//	CLASS:		RtArrowButton
//	DESC:		A button displaying a Right Arrow
//
//	-------------------------------------------------------------------------

class RtArrowButton : public PictureButton
	{
	public:
		RtArrowButton(int x,int y,int options,int retevent)
				: PictureButton(x,y,GUI_BITMAP[8],GUI_BITMAP[9],options,retevent) {}
	};

//	-------------------------------------------------------------------------
//
//	CLASS:		DblUpArrowButton
//	DESC:		A button displaying two Up Arrows. The top half allows
//				large increments while the lower half allows single increments
//
//	-------------------------------------------------------------------------

class DblUpArrowButton : public PictureButton
	{
	public:
		DblUpArrowButton(int x,int y,int options,int retevent)
				: PictureButton(x,y,GUI_BITMAP[0],GUI_BITMAP[0],options,retevent) {}
	};

//	-------------------------------------------------------------------------
//
//	CLASS:		DblDnArrowButton
//	DESC:		A button displaying two Down Arrows. The lower half allows
//				large increments while the top half allows single increments
//
//	-------------------------------------------------------------------------

class DblDnArrowButton : public PictureButton
	{
	public:
		DblDnArrowButton(int x,int y,int options,int retevent)
				: PictureButton(x,y,GUI_BITMAP[0],GUI_BITMAP[0],options,retevent) {}
	};

//	-------------------------------------------------------------------------
//
//	CLASS:		DblLtArrowButton
//	DESC:		A button displaying two Left Arrows. The left half allows
//				large increments while the right half allows single increments
//
//	-------------------------------------------------------------------------

class DblLtArrowButton : public PictureButton
	{
	public:
		DblLtArrowButton(int x,int y,int options,int retevent)
				: PictureButton(x,y,GUI_BITMAP[0],GUI_BITMAP[0],options,retevent) {}
	};

//	-------------------------------------------------------------------------
//
//	CLASS:		DblRtArrowButton
//	DESC:		A button displaying two Right Arrows. The right half allows
//				large increments while the left half allows single increments
//
//	-------------------------------------------------------------------------

class DblRtArrowButton : public PictureButton
	{
	public:
		DblRtArrowButton(int x,int y,int options,int retevent)
				: PictureButton(x,y,GUI_BITMAP[0],GUI_BITMAP[0],options,retevent) {}
	};

//	-------------------------------------------------------------------------
//
//	CLASS:		UpDnArrowButton
//	DESC:		A button displaying an Up Arrow and a Down Arrow.
//				The up arrow decrements & the down arrow increments
//
//	-------------------------------------------------------------------------

class UpDnArrowButton : public PictureButton
	{
	private:
	protected:
	public:
		UpDnArrowButton(int x,int y,int options,int retevent)
				: PictureButton(x,y,GUI_BITMAP[0],GUI_BITMAP[0],options,retevent) {}
	};

//	-------------------------------------------------------------------------
//
//	CLASS:		LtRtArrowButton
//	DESC:		A button displaying a Left Arrow and a Right Arrow.
//				The left arrow decrements & the right arrow increments
//
//	-------------------------------------------------------------------------

class LtRtArrowButton : public PictureButton
	{
	private:
	protected:
	public:
		LtRtArrowButton(int x,int y,int options,int retevent)
				: PictureButton(x,y,GUI_BITMAP[0],GUI_BITMAP[0],options,retevent) {}
	};

//	-------------------------------------------------------------------------
//
//	CLASS:		TickButton
//	DESC:		A button displaying a Tick.
//
//	-------------------------------------------------------------------------

class TickButton : public PictureButton
	{
	private:
	protected:
	public:
		TickButton(int x,int y,int options,int retevent)
				: PictureButton(x,y,GUI_BITMAP[0],GUI_BITMAP[0],options,retevent) {}
	};

//	-------------------------------------------------------------------------
//
//	CLASS:		CrossButton
//	DESC:		A button displaying a Cross
//
//	-------------------------------------------------------------------------

class CrossButton : public PictureButton
	{
	private:
	protected:
	public:
		CrossButton(int x,int y,int options,int retevent)
				: PictureButton(x,y,GUI_BITMAP[0],GUI_BITMAP[0],options,retevent) {}
	};

//	-------------------------------------------------------------------------
//
//	CLASS:		QuestionMarkButton
//	DESC:		A button displaying a Question Mark
//
//	-------------------------------------------------------------------------

class QuestionMarkButton : public PictureButton
	{
	public:
		QuestionMarkButton(int x,int y,int options,int retevent)
				: PictureButton(x,y,GUI_BITMAP[0],GUI_BITMAP[0],options,retevent) {}
	};

//	-------------------------------------------------------------------------
//
//	CLASS:		SmallButton
//	DESC:		A small-size blank button
//
//	-------------------------------------------------------------------------

class SmallButton : public PictureButton
	{
	public:
		SmallButton(int x,int y,int options,int retevent);
	};

//	-------------------------------------------------------------------------
//
//	CLASS:		MediumButton
//	DESC:		A medium-size blank button
//
//	-------------------------------------------------------------------------

class MediumButton : public PictureButton
	{
	public:
		MediumButton(int x,int y,int options,int retevent);
	};

//	-------------------------------------------------------------------------
//
//	CLASS:		LargeButton
//	DESC:		A large-size blank button
//
//	-------------------------------------------------------------------------

class LargeButton : public PictureButton
	{
	public:
		LargeButton(int x,int y,int options,int retevent);
	};

//	-------------------------------------------------------------------------
//
//	CLASS:		FussyButton
//	DESC:		A button that knows the difference between left and right mb
//
//	-------------------------------------------------------------------------

class FussyButton : public PictureButton
	{
	private:
	protected:
		// if this command is received then the button will toggle automatically
		int		secondEvent;
	public:
		// assumes color 0 is transparent, no borders
		FussyButton(int x1,int y1,grImage* active_pic,grImage* inactive_pic,
					  int options,int retevent,int secondevent) :
			PictureButton(x1,y1,active_pic,inactive_pic,options,retevent)
				{ secondEvent = secondevent; }

		// assumes color supplied is transparent, no borders
		FussyButton(int x1,int y1,grImage* active_pic,grImage* inactive_pic,
					  int tr_color,int options,int retevent,int secondevent) :
			PictureButton(x1,y1,active_pic,inactive_pic,tr_color,options,retevent)
				{ secondEvent = secondevent; }

		// assumes borders, no transparency
		FussyButton(int x1,int y1,int active_border,int inactive_border,
					  grImage* active_pic,grImage* inactive_pic,
					  int options,int retevent,int secondevent) :
			PictureButton(x1,y1,active_border,inactive_border,active_pic,inactive_pic,options,retevent)
				{ secondEvent = secondevent; }

		virtual void HandleEvent(Event &e);
	};
#endif