//	-------------------------------------------------------------------------
//
//	DIALOG.CPP
//
//	-------------------------------------------------------------------------
//
//	AUTHOR:			Darren J. Dwyer
//
//	DESCRIPTION:	this file contains the implementation for a standard
//					Dialog Box
//
//	-------------------------------------------------------------------------

#include "dialog.hpp"
#include "stddefns.hpp"
#include "eventmgr.hpp"
#include "keydefns.hpp"

#include "stdlib.h"
#include "dos.h"

//	-------------------------------------------------------------------------
//
//	FUNC:		CONSTRUCTOR for DialogBox
//	DESC:		The default constructor which defines the behaviour for
//				all derived DialogBoxes
//
//	-------------------------------------------------------------------------
DialogBox::DialogBox(int x1,int y1,int x2,int y2,	// the coords
					uchar bg,					// the background color
					uchar active_border,		// and border styles
					uchar inactive_border,
					int* evhno,					// the event-handler number
					int parent_handler_num,		// the parent's evhno
					int options,				// the dialog box options
					int zfact,					// its z-factor
					int okretevent,				// command event returned on OK
					int cancelretevent)         // command event returned on CANCEL
	: OnScreenHandler(x1,y1,x2,y2,bg,active_border,inactive_border,evhno,
						options|evh_DIALOG_BOX,zfact)
	{
	// modify selected_options because we are derived from an onscreen handler
	selected_options -= evh_ONSCREEN_HANDLER;

	isDialogBox = true;
	isOnScreenHandler = false;

	parent_handler = parent_handler_num;
	OKreturnEvent  = okretevent;
	CANCELreturnEvent = cancelretevent;

	isOpen = false;

	// make sure we've selected a default response (CANCEL)
	if (selected_options&dlg_CANCEL_DEFAULT);
	else
		if (selected_options&dlg_OK_DEFAULT);
		else
			selected_options |= dlg_CANCEL_DEFAULT;

//	DBG(460,"IN DIALOGBOX CONSTRUCTOR");
//	DBG(450,"evhno == %p,    parent handler == %d",evhno,parent_handler);
//	delay(10000);
	// create space to store the image (where the dialog box is to be placed)
	savedImage = grCreateImage(r.left,r.top,r.right,r.bottom);
	}

DialogBox::DialogBox(int x1,int y1,int x2,int y2,	// the coords
					grImage* active_bitmap,		// the bitmaps
					grImage* inactive_bitmap,
					int* evhno,					// the event-handler number
					int parent_handler_num,		// the parent's evhno
					int options,				// the dialog box options
					int zfact,					// its z-factor
					int okretevent,				// command event returned on OK
					int cancelretevent)         // command event returned on CANCEL
	: OnScreenHandler(x1,y1,x2,y2,active_bitmap,inactive_bitmap,evhno,
						options|evh_DIALOG_BOX,zfact)
	{
	// modify selected_options because we are derived from an onscreen handler
	selected_options -= evh_ONSCREEN_HANDLER;

	isDialogBox = true;
	isOnScreenHandler = false;

	parent_handler = parent_handler_num;
	OKreturnEvent  = okretevent;
	CANCELreturnEvent = cancelretevent;

	isOpen = false;

	// make sure we've selected a default response (CANCEL)
	if (selected_options&dlg_CANCEL_DEFAULT);
	else
		if (selected_options&dlg_OK_DEFAULT);
		else
			selected_options |= dlg_CANCEL_DEFAULT;

	// create space to store the image (where the dialog box is to be placed)
	savedImage = grCreateImage(r.left,r.top,r.right,r.bottom);
	}

DialogBox::~DialogBox()
	{
	// and free up the memory used
	if (savedImage!=NULL) grDestroyImage(savedImage);
	}

void DialogBox::Draw()
	{
	if (!isOpen) return;
	OnScreenHandler::Draw();
	}

//	-------------------------------------------------------------------------
//
//	FUNC:	OpenDialog()
//	DESC:	Opens the dialog box at defined zfactor
//
//	-------------------------------------------------------------------------
void DialogBox::OpenDialog()
	{
	// skip out if dialog is already open
	if (isOpen) return;

	// if there was enough memory in the heap, save the image
	if (savedImage!=NULL) grGetImage(r.left,r.top,r.right,r.bottom,savedImage);

	// we are now about to be onscreen
	isOpen = true;
	isOnScreenHandler = true;

	// save & deselect currently selected handler
	prev_handler = EVENTMGR->currentHandler;
//	DBG(470,"DIALOGBOX::OPENDIALOG() PREVIOUSLY SELECTED HANDLER %d",prev_handler);
	EVENTMGR->DeSelectEventHandler( prev_handler );
//	DBG(460,"DIALOGBOX::OPENDIALOG() AFTER DESELECTING HANDLER   %d",EVENTMGR->currentHandler);

	// save the previous z-factor & tell eventmgr to update zfactor
	prev_zfactor = EVENTMGR->currentZFactor;
	EVENTMGR->currentZFactor = zfactor;

	// draw and then select our handler
	Draw();

	// select our dialog box as the current handler
//	DBG(450,"DIALOGBOX::OPENDIALOG() ABOUT TO SELECT DIALOG BOX  %d",handler_number);
	EVENTMGR->SelectEventHandler( handler_number );
//	DBG(440,"DIALOGBOX::OPENDIALOG() AFTER SELECTION, HANDLER    %d",EVENTMGR->currentHandler);

//	DBG(470,"OUT DIALOGBOX::OPENDIALOG()");
	}

//	-------------------------------------------------------------------------
//
//	FUNC:	CloseDialog()
//	DESC:	Closes the dialog box, restoring everything as on OpenDialog()
//
//	-------------------------------------------------------------------------
void DialogBox::CloseDialog()
	{
	if (!isOpen) return;

	// deselect our dialog box
	EVENTMGR->DeSelectEventHandler( handler_number );

//	DBG(470,"DESELECTED EVENT HANDLER %d",handler_number );
//	delay(5000);
	// iff we didn't run out of heap
	if (savedImage!=NULL)
		// restore image under the dialog box
		grPutImage(r.left,r.top,r.right,r.bottom,savedImage);

//	DBG(460,"RESTORED WHAT WAS UNDERNEATH");
//	delay(5000);

	// we are no longer on the screen
	isOpen = false;
	isOnScreenHandler = false;

	// reset eventmgr's z-factor
	EVENTMGR->currentZFactor = prev_zfactor;

	// and select the previous handler
	EVENTMGR->SelectEventHandler( prev_handler );

//	DBG(450,"SELECTED PREVIOUS HANDLER %d",prev_handler);
//	DBG(440,"CURRENTLY SELECTED HANDLER %d",EVENTMGR->currentHandler);

//	delay(5000);
	HEAPCHK("by DialogBox::CloseDialog()");
	}

//	-------------------------------------------------------------------------
//
//	FUNC:	HandleEvent(e);
//	DESC:	Checks for the ESC key which will generate the returnEvent
//
//	-------------------------------------------------------------------------
void DialogBox::HandleEvent(Event &e)
	{
	// tell the onscreen handler to process events as normal
	OnScreenHandler::HandleEvent(e);

	// post-process dialog box specific events
	switch (e.what)
		{
		case	ev_keyboard:		// check for special keys and if they were
									// pressed, simply modify the event & skip out
									switch (e.keyboard.event)
										{
										case	ENTER_UP:		e.what  = ev_command;
																e.where = parent_handler;
																e.command.event = OKreturnEvent;
																e.command.data  = NULL;
																return;

										case	ESC_UP:			e.what  = ev_command;
																e.where = parent_handler;
																e.command.event = CANCELreturnEvent;
																e.command.data  = NULL;
																return;
										}
									break;
		}

	HEAPCHK("by DialogBox::HandleEvent()");
	}
