//	-------------------------------------------------------------------------
//
//	INPUTBAR.CPP
//
//	-------------------------------------------------------------------------
//
//	AUTHOR:			Darren J. Dwyer
//
//	DESCRIPTION:	This file contains code for a number of InputBars
//					including TextInputBar, FormattedInputBar,
//					NumericalInputBar, & DollarInputBar.
//
//	-------------------------------------------------------------------------

#include "inputbar.hpp"
#include "keyboard.hpp"
#include "keydefns.hpp"
#include "eventmgr.hpp"
#include "mouse.hpp"

#include "dos.h"

//	-------------------------------------------------------------------------
//
//	FUNC: 		CONSTRUCTOR for TextInputBar
//	DESC:		A simple input bar allowing limited editting of the text
//				within the bar.
//
//	USES:		ENTER			- Activates/Deactivates the input bar
//				ESC				- Deactivates the input bar, resetting the
//								  input text
//				LT,RT			- moves the cursor left/right
//				CTRL_LT,CTRL_RT	- moves the cursor one word left/right
//				HOME,END		- moves the cursor to the start/end of text
//
//	PARAMS:		x1,y1			- the top,left coordinates of the input bar
//				cols			- the number of columns to display text in
//				st				- the string to modify
//
//	NOTES:		The input string is not modified until we have finished
//				editting and press ENTER!
//
//				If the maximum length of the string (st) is greater than
//				the number of cols, a simple scrolling function is
//				automatically used.
//
//	-------------------------------------------------------------------------
TextInputBar::TextInputBar(int x1,int y1,int num_cols,
							String* input_str,int max_len,
							int active_border,int inactive_border,
							int options,
							int OKretevent,CANCELretevent)
	// note: this returns the maximum width of any displayed
	//		 string using the current font....
	//		 	(numcols+1) so that we can include the cursor
	: Accessory(x1,y1,
				x1+6+2*grSizeOfBorder[active_border]+grFontWidthMax(num_cols+1),
				y1+6+2*grSizeOfBorder[active_border]+grFontHeight(),
				BGCOLOR,FGCOLOR,active_border,inactive_border,
				options,OKretevent)
	{
	// save all of the original variables
	orig_str 		= input_str;

	num_columns 	= num_cols;
	max_columns 	= max_len;

	// calculate what we have to
	col_offset		= 0;
	max_col_offset	= max_columns - num_columns;

	cursor_pos		= 0;

	edit_pos		= 0;
	max_edit_pos	= max_col_offset - 1;

	edit_str		= new String(orig_str->Contents(),max_columns);
	}

TextInputBar::~TextInputBar()
	{
	delete edit_str;
	}

void TextInputBar::SetEditStr(String* st)
	{
	edit_str->Set(st->Contents());
	}

String* TextInputBar::GetEditStr()
	{
	return edit_str;
	}

void TextInputBar::SetEditPos(int offset,int pos)
	{
	// make sure the values are in range
	CLIP(offset,0,max_col_offset-1);
	CLIP(pos,0,num_columns-1);

	// set the new values
	col_offset 	= offset;
	cursor_pos  = pos;
	edit_pos 	= col_offset+cursor_pos;

	// and refresh the display
	Refresh();
	}

void TextInputBar::SetEditPos(grPoint& mouse_pt)
	{
	// convert the global coords to local coords
	mouse_pt.x -= (global_r.left + grSizeOfBorder[active_border]);

	// now figure out which character was clicked on (if any)
	int mouse_pos = -1;
	int lhs = 0,rhs = 0;
	int charwidth = 0;

	// scan through each column displayed
	for (int col=0;col<num_columns,mouse_pos==-1;col++)
		{
//		String *a = *another_string;
//		*a = "THIS IS A TEST";			(1)
//		*a += " WITH A BIT APPENDED";	(2)
//		*a -= "THIS";					(3)

//		(1)	// a == "THIS IS A TEST"
//		(2)	// a == "THIS IS A TEST WITH A BIT APPENDED"
//		(3)	// a == " IS A TEST WITH A BIT APPENDED"

		// get the width of the nth displayed char
//		charwidth = grFontCharWidth(edit_str[col_offset+col]);
		charwidth = grFontCharWidth(edit_str->Contents()[col_offset+col]);

		// add this to the lhs to get the rhs boundaries for the char
		rhs = lhs + charwidth-1;

		// check to see if the mouse is inside the boundaries
		if ((mouse_pt.x>=lhs)&&(mouse_pt.x<=rhs)) mouse_pos = col;

		// now the lhs becomes the rhs+1
		lhs = rhs + 1;
		}

	// if the mouse was clicked on any particular character
	if (mouse_pos!=-1)
		{
		SetEditPos(col_offset,mouse_pos);
		}
	}

void TextInputBar::IncEditPosn()
	{
	// move the cursor along in the displayed portion iff possible
	if (cursor_pos<num_columns) cursor_pos++;
	else
		// see if we can scroll the input bar
		if (col_offset<max_col_offset) col_offset++;

	// increment our edit position
	if (edit_pos<max_edit_pos) edit_pos++;
	}

void TextInputBar::DecEditPosn()
	{
	if (cursor_pos>0) cursor_pos--;
	else
		if (col_offset>0) col_offset--;

	if (edit_pos>0) edit_pos--;
	}

//	-------------------------------------------------------------------------
//
//	FUNC:	Draw()
//	DESC:   refreshes the contents of the TextInputBar
//
//	-------------------------------------------------------------------------
void TextInputBar::Draw(grVirtualBuffer* vb)
	{
	if (isSelected)
		{
		// draw the window frame
		grDrawWindowVirtual(vb,r.left,r.top,r.right,r.bottom,active_color,active_bs);

		// draw the contents...

		String* tmpstr;

		// calculate the length of the string at the current edit_pos
		// so that we can calculate the new cursor pos...
		int cursor_start;
		if (edit_pos>0)
			{
			tmpstr			= edit_str->SubStr(col_offset,col_offset+cursor_pos-1);
			cursor_start 	= inset_r.left+2+grFontWidth(tmpstr->Contents());
			delete tmpstr;
			}
		else
			cursor_start = inset_r.left+2;

		grDrawFilledRectVirtual(vb,cursor_start,inset_r.top+2,
								   cursor_start+grFontCharWidth(str->Contents()[edit_pos]),
								   inset_r.top+2+grFontCharHeight(str->Contents()[edit_pos]),
								   inactive_color);

		// draw the input string
		tmpstr = edit_str->SubStr(col_offset,col_offset+num_columns);
		grDrawStringVirtual(vb,inset_r.left+2,inset_r.top+2,"%s",tmpstr->Contents());
		delete tmpstr;
		}
	else
		{
		grDrawWindowVirtual(vb,r.left,r.top,r.right,r.bottom,inactive_color,inactive_bs);
		String* tmpstr = edit_str->SubStr(col_offset,col_offset+num_columns);
		grDrawStringVirtual(vb,inset_r.left+2,inset_r.top+2,"%s",tmpstr->Contents());
		delete tmpstr;
		}
	}

//	-------------------------------------------------------------------------
//
//	FUNC:	HandleEvent()
//	DESC:	Handles left mouse button, the arrows, home & end, to position
//			the cursor.
//
//			Right mouse button, ENTER or TAB mean accept input.
//
//			ESC means reset input
//
//			BS,DEL,other keys edit input
//
//	-------------------------------------------------------------------------
void TextInputBar::HandleEvent(Event &e)
	{
	// a flag to specifiy 'event has been handled'
	bool handled = false;

	switch (e.what)
		{
		case	ev_mouse:       switch (e.mouse.event)
									{
									// lmb up sets the edit position
									case	mb_leftup:	if (isSelectable)
															{
															SetEditPos(e.mouse.where);
															EVENTMGR->ClearEvent(e);
															}
														break;

									case	mb_rightup:	if (isSelectable)
															{
															if (isSelected)
																DeSelect();
															EVENTMGR->ClearEvent(e);
															}
														break;

									// ignore everything else
									default:    		EVENTMGR->ClearEvent(e);
														break;
									}
								break;


		// handle the keyboard events
		case	ev_keyboard:	switch (e.keyboard.event)
									{
									// handle left arrow
									case	LT_ARROW_UP:	if (!isSelected)
																{
																isSelected = true;
																SetEditPosn(0);
																Refresh();
																}
															else
																{
																// check for ctrl_lt
																if (e.keyboard.modifiers&keyboard_CTRL)
																	{
																	do
																		{
																		DecEditPosn();
																		}
																	while ((edit_pos>0)&&
																		   (str->Contents()[edit_pos-1]!=' '));
																	}
																else
																	DecEditPosn();
																Refresh();
																}
															EVENTMGR->ClearEvent(e);
															break;
									case	LT_ARROW_DN:	EVENTMGR->ClearEvent(e);
															break;

									// handle right arrow
									case	RT_ARROW_UP:	if (!isSelected)
																{
																isSelected = true;
																SetEditPosn(max_columns);
																Refresh();
																}
															else
																{
																// check for ctrl_rt
																if (e.keyboard.modifiers&keyboard_CTRL)
																	{
																	int len = str->Length();
																	do
																		{
																		IncEditPosn();
																		}
																	while ((edit_pos<len)&&
																		   (str->Contents()[edit_pos-1]!=' '));
																	}
																else
																	IncEditPosn();
																Refresh();
																}
															EVENTMGR->ClearEvent(e);
															break;

									case	RT_ARROW_DN:	EVENTMGR->ClearEvent(e);
															break;

									// handle return
									case	ENTER_DN:		EVENTMGR->ClearEvent(e);
															break;
									case	ENTER_UP:   	if (!isSelected)
																{
																isSelected = true;
																SetEditPosn(0);
																Refresh();
																}
															else
																{
																isSelected = false;
																Refresh();
																handled = true;
																}
															EVENTMGR->ClearEvent(e);
															break;

									// handle esc
									case	ESC_UP:			if (isSelected)
																{
																isSelected = false;
																str->Set(origstr->Contents());
																SetEditPosn(0);
																Refresh();
																EVENTMGR->ClearEvent(e);
																}
															break;

									// handle backspace
									case	BS_DN:		EVENTMGR->ClearEvent(e);
														break;
									case    BS_UP:      if (edit_pos>0)
															{
															Event e1;
															e1.what = ev_keyboard;
															e1.keyboard.event  = DEL_UP;
															DecEditPosn();
															HandleEvent(e1);
															}
														EVENTMGR->ClearEvent(e);
														break;

									// handle home
									case	HOME_DN:	EVENTMGR->ClearEvent(e);
														break;
									case 	HOME_UP:	SetEditPosn(0);

														if (!isSelected)
															{
															isSelected = true;
															Refresh();
															}
														else
															Refresh();

														EVENTMGR->ClearEvent(e);
														break;

									// handle end
									case	END_DN:		EVENTMGR->ClearEvent(e);
														break;
									case 	END_UP:     SetEditPosn(str->Length());

														if (!isSelected)
															{
															isSelected = true;
															Refresh();
															}
														else
															Refresh();

														EVENTMGR->ClearEvent(e);
														break;

									// handle del
									case	DEL_DN:		EVENTMGR->ClearEvent(e);
														break;
									case	DEL_UP:		if (!isSelected)
															{
															str->Set("");
															isSelected = true;
															SetEditPosn(0);
															Refresh();
															}
														else
															{
															if (edit_pos>=0)
																if (str->Length()>0) str->Delete(edit_pos,1);
															if (edit_pos>str->Length()) SetEditPosn(str->Length());
															Refresh();
															}
														EVENTMGR->ClearEvent(e);
														break;

									case	SPACE_DN:	EVENTMGR->ClearEvent(e);
														break;

									default:	// check if we should clear the input field,
												// or if we should simply manipulate the input..
												if (keyboard_IsReleased(e.keyboard.event))
													{
													int keyval = keyboard_ASCII(e.keyboard.event,e.keyboard.modifiers);
													if ((keyval>=32)&&(keyval<=127))
														{
														if (isSelected)
															{
															int len = origstr->MaxLength();
															if (edit_pos<len)
																{
																int tmp = str->Length();
																str->Insert(edit_pos,(char)keyval);
																if (str->Length()!=tmp) IncEditPosn();
																}
															Refresh();
															}
														else
															{
															isSelected = true;
															str->Set((char)keyval);
															SetEditPosn(1);
															Refresh();
															}
														EVENTMGR->ClearEvent(e);
														}
													}
												break;
									}
								break;
		}

	// if we have handled the event, modify the event to contain our returnEvent
	if (handled)
		{
		CommandEvent cmd = { returnEvent, NULL };
		EVENTMGR->ModEvent(e,0,ev_command,&cmd);
		// and set the returning string
		origstr->Set(str->Contents());
		}
	}
