//	-------------------------------------------------------------------------
//
//	SCROLBAR.CPP
//
//	-------------------------------------------------------------------------
//
//	AUTHOR:			Darren J. Dwyer
//
//	DESCRIPTION:	This file contains code for Vertical & Horizontal
//					scroll bars (VScrollBar & HScrollBar).
//
//	-------------------------------------------------------------------------

#include "scrolbar.hpp"
#include "keyboard.hpp"
#include "keydefns.hpp"
#include "eventmgr.hpp"
#include "mouse.hpp"

#include "dos.h"

//	-------------------------------------------------------------------------
//
//	FUNC: 		CONSTRUCTOR for VScrollBar
//	DESC: 		A vertical scroll bar class.
//
//	PARAMS:		xx				- the position of the left side of the bar
//				y1,y2			- the start/end positions
//				position		- the int to modify up/dn
//				maxpositions	- the maximum number of positions to scroll
//
//	-------------------------------------------------------------------------
VScrollBar::VScrollBar(int xx,int y1,int y2,int position,int maxpositions,
						int options,int retevent)
	: Accessory(xx,y1,xx,y2,grDARKGRAY,grLIGHTGRAY,bs_NONE,bs_NONE,options,retevent)
	{
	// save a pointer to our position & maxposition indicators
	pos = position;
	maxpos = maxpositions;

	// and save what was the last recorded position
	lastpos = pos;

	// check our options to make sure arrows do something!!!
	if (selected_options&scrollbar_SINGLE_CLICK) ;
	else
		if (selected_options&scrollbar_DOUBLE_CLICK) ;
		else
			selected_options |= scrollbar_AUTO_REPEAT;

	// add an UpArrow button
	UpArrow = new UpArrowButton(inset_r.left,inset_r.top,0,cmdUpArrowButton);

	// figure out the width of the uparrow (including borders)
	buttonsize = UpArrow->r.bottom-UpArrow->r.top;

	// modify our coordinates
	r.right += buttonsize;
	inset_r.right += buttonsize;

	// calculate the area in which to draw the blankbutton
	scalestart = inset_r.top+buttonsize;
	scaleend = inset_r.bottom-buttonsize-1;
	scaleheight = scaleend-scalestart+1;

	// add a DnArrow button
	DnArrow = new DnArrowButton(inset_r.left,inset_r.bottom-buttonsize,0,cmdDnArrowButton);

	// and place the blank button at its starting coords.
	blankpos = (int) ((float) scalestart + (float) pos * (float) scale + 0.5);
	Blank = new BlankButton(inset_r.left,blankpos,0,cmdBlankButton);

	// we are always active
	isSelected = true;
	}

//	-------------------------------------------------------------------------
//
//	FUNC: 		DeSelect
//	DESC:		Deselects the VScrollBar
//
//	-------------------------------------------------------------------------
void VScrollBar::DeSelect()
	{
	Accessory::DeSelect();

	UpArrow->isSelected = false;
	UpArrow->Refresh();
	DnArrow->isSelected = false;
	DnArrow->Refresh();
	isSelected = false;
	Refresh();
	}

//	-------------------------------------------------------------------------
//
//	FUNC: 		AdjustCoords
//	DESC:		Adjusts the coordinates of the VScrollBar (used internally)
//
//	-------------------------------------------------------------------------
void VScrollBar::AdjustCoords(int dx,int dy)
	{
	// we adjust our own coordinates, then the buttons
	Accessory::AdjustCoords(dx,dy);
	scalestart += dy;
	scaleend += dy;
	UpArrow->AdjustCoords(dx,dy);
	DnArrow->AdjustCoords(dx,dy);
	Blank->AdjustCoords(dx,dy);
	}

//	-------------------------------------------------------------------------
//
//	FUNC: 		Draw()
//	DESC:		Draws the VScrollBar
//
//	-------------------------------------------------------------------------
void VScrollBar::Draw()
	{
	// draw the 3 v-bars
	grDrawVBarVirtual(r.left,scalestart,scaleend,bs_LOWERED);
	grDrawVBarVirtual(vb,r.right,scalestart,scaleend,bs_RAISED);
	grDrawVBarVirtual(vb,r.left+(r.right-r.left)/2,
						 scalestart,scaleend,bs_OUTLINE);

	// and draw the 3 buttons
	UpArrow->Draw();
	DnArrow->Draw();
	Blank->Draw();
	}

//	-------------------------------------------------------------------------
//
//	FUNC: 		HandleEvent
//	DESC:		Handles up,down,pgup,pgdn,home and end.
//				Also responds to the left mouse button
//
//	-------------------------------------------------------------------------
void VScrollBar::HandleEvent(Event &e)
	{
	// a flag for 'the event has been handled'
	bool handled = false;

	switch (e.what)
		{
		// if its a mouse event
		case	ev_mouse:	// if the mouse was clicked inside the uparrow
							if (UpArrow->isSelectable)
								{
								if (UpArrow->MouseInside(e.mouse.where))
									{
									if (selected_options&scrollbar_SINGLE_CLICK)
										{
										bool leftup = (e.mouse.event==mb_leftup);
										UpArrow->HandleEvent(e);
										if (leftup)
											if (pos>0) { (pos)--; RefreshBlank(); }
										}
									else
										if (selected_options&scrollbar_AUTO_REPEAT)
											{
											UpArrow->HandleEvent(e);
											if (pos>0) { (pos)--; RefreshBlank(); }
											}
									handled = true;
									}
								}

							if (!handled)
								// if it was clicked inside the dnarrow
								if (DnArrow->MouseInside(e.mouse.where))
									{
									if (selected_options&scrollbar_SINGLE_CLICK)
										{
										bool leftup1 = (e.mouse.event==mb_leftup);
										DnArrow->HandleEvent(e);
										if (leftup1)
											if (pos<maxpos) { (pos)++; RefreshBlank(); }
										}
									else
										if (selected_options&scrollbar_AUTO_REPEAT)
											{
											DnArrow->HandleEvent(e);
											if (pos<maxpos) { (pos)++; RefreshBlank(); }
											}
									handled = true;
									}

							if (!handled)
							// the mouse was clicked somewhere on the bar
									{
									// if the UpArrow is currently active, deactivate it
									if (UpArrow->isSelectable)
										{
										UpArrow->isSelectable = false;
										UpArrow->Refresh();
										}

									// if the DnArrow is currently active, deactivate it
									if (DnArrow->isSelectable)
										{
										DnArrow->isSelectable = false;
										DnArrow->Refresh();
										}

									// calc a new position based upon the mouse coordinates
									if (maxpos>0)
										{
										scale  = (float) (scaleheight-buttonsize) / (float) (maxpos);
										pos = (int) (((float)e.mouse.where.y-(float)scalestart-(float)(buttonsize/2)) / (float)scale);
										}
									else pos = 0;

									// check bounds
									if (pos<0) pos = 0;
									if (pos>maxpos) pos = maxpos;

									// and whether or not to redraw the blank button
									if (pos!=lastpos)
										{
										RefreshBlank();
										lastpos=pos;
										}

									handled = true;
									}
							break;

		case	ev_keyboard:    switch (e.keyboard.event)
									{
									// handle uparrow
									case	UP_ARROW_DN:  	if (!UpArrow->isSelectable)
																{
																UpArrow->isSelectable = true;
																UpArrow->Refresh();
																}
															if (!(selected_options&scrollbar_SINGLE_CLICK))
																if (pos>0) { (pos)--; RefreshBlank(); }
															handled = true;
															break;
									case	UP_ARROW_UP:	UpArrow->isSelectable = false;
															UpArrow->Refresh();
															if (selected_options&scrollbar_SINGLE_CLICK)
																if (pos>0) { (pos)--; RefreshBlank(); }
															EVENTMGR->ClearEvent(e);
															break;

									// handle downarrow
									case	DN_ARROW_DN: 	if (!DnArrow->isSelectable)
																{
																DnArrow->isSelectable = true;
																DnArrow->Refresh();
																}
															if (!(selected_options&scrollbar_SINGLE_CLICK))
																if (pos<maxpos) { (pos)++; RefreshBlank(); }
															handled = true;
															break;
									case	DN_ARROW_UP:	DnArrow->isSelectable = false;
															DnArrow->Refresh();
															if (selected_options&scrollbar_SINGLE_CLICK)
																if (pos<maxpos) { (pos)++; RefreshBlank(); }
															EVENTMGR->ClearEvent(e);
															break;

									// handle page_up
									case	PGUP_DN:  		if (!UpArrow->isSelectable)
																{
																UpArrow->isSelectable = true;
																UpArrow->Refresh();
																}
															if (!(selected_options&scrollbar_SINGLE_CLICK))
																{
																if (pos>(maxpos/10)) (pos)-=(maxpos/10);
																else pos=0;
																RefreshBlank();
																}
															EVENTMGR->ClearEvent(e);
															break;
									case	PGUP_UP:		UpArrow->isSelectable = false;
															UpArrow->Refresh();
															if (selected_options&scrollbar_SINGLE_CLICK)
																{
																if (pos>(maxpos/10)) (pos)-=(maxpos/10);
																else pos=0;
																RefreshBlank();
																}
															handled = true;
															EVENTMGR->ClearEvent(e);
															break;

									// handle page_dn
									case	PGDN_DN: 		if (!DnArrow->isSelectable)
																{
																DnArrow->isSelectable = true;
																DnArrow->Refresh();
																}
															if (!(selected_options&scrollbar_SINGLE_CLICK))
																{
																if (pos<maxpos-(maxpos/10)) (pos)+=(maxpos/10);
																else pos = maxpos;
																RefreshBlank();
																}
															EVENTMGR->ClearEvent(e);
															break;
									case	PGDN_UP:		DnArrow->isSelectable = false;
															DnArrow->Refresh();
															if (selected_options&scrollbar_SINGLE_CLICK)
																{
																if (pos<maxpos-(maxpos/10))  (pos)+=(maxpos/10);
																else pos = maxpos;
																RefreshBlank();
																}
															handled = true;
															EVENTMGR->ClearEvent(e);
															break;


									// handle home
									case	HOME_DN:	if (!UpArrow->isSelectable)
															{
															UpArrow->isSelectable = true;
															UpArrow->Refresh();
															}
														EVENTMGR->ClearEvent(e);
														break;
									case	HOME_UP:	UpArrow->isSelectable = false;
														UpArrow->Refresh();
														if (pos!=0)
															{
															pos = 0;
															RefreshBlank();
															}
														handled = true;
														break;

									// handle end
									case	END_DN:		if (!DnArrow->isSelectable)
															{
															DnArrow->isSelectable = true;
															DnArrow->Refresh();
															}
														EVENTMGR->ClearEvent(e);
														break;
									case	END_UP:		DnArrow->isSelectable = false;
														DnArrow->Refresh();
														if (pos!=maxpos)
															{
															pos = maxpos;
															RefreshBlank();
															}
														handled = true;
														break;
									}
								break;
		}

	// if we have handled the event, modify the event to contain our returnEvent
	if (handled)
		{
		CommandEvent ce = { returnEvent, &pos };
		EVENTMGR->ModEvent(e,0,ev_command,&ce);
		}
	}


//	-------------------------------------------------------------------------
//
//	FUNC: 		CONSTRUCTOR for HScrollBar
//	DESC: 		A horizontal scroll bar class.
//
//	PARAMS:		yy				- the position of the left side of the bar
//				x1,x2			- the start/end positions
//				position		- the int to modify lt/rt
//				maxpositions	- the maximum number of positions to scroll
//
//
//	-------------------------------------------------------------------------
HScrollBar::HScrollBar(int yy,int x1,int x2,int position,int maxpositions,
						int options,int retevent)
	: Accessory(x1,yy,x2,yy,grDARKGRAY,grLIGHTGRAY,bs_NONE,bs_NONE,options,retevent)
	{
	// save a pointer to our position & maxposition indicators
	pos = position;
	maxpos = maxpositions;

	// check our options to make sure arrows do something!!!
	if (selected_options&scrollbar_SINGLE_CLICK) ;
	else
		if (selected_options&scrollbar_DOUBLE_CLICK) ;
		else
			selected_options |= scrollbar_AUTO_REPEAT;

	// and save what was the last recorded position
	lastpos = pos;

	// add a LtArrow button
	LtArrow = new LtArrowButton(inset_r.left,inset_r.top,0,cmdLtArrowButton);

	// figure out the height of the ltarrow (including borders)
	buttonsize = LtArrow->r.right-LtArrow->r.left;

	// modify our coordinates
	r.bottom += buttonsize;
	inset_r.bottom += buttonsize;

	// calculate the area in which to draw the blankbutton
	scalestart = inset_r.left+buttonsize;
	scaleend = inset_r.right-buttonsize-1;
	scalewidth = scaleend-scalestart+1;

	// add a RtArrow button
	RtArrow = new RtArrowButton(inset_r.right-buttonsize,inset_r.top,0,cmdLtArrowButton);

	// and place the blank button at its starting coords.
	blankpos = (int) ((float) scalestart + (float) pos * (float) scale + 0.5);
	Blank = new BlankButton(blankpos,inset_r.top,0,cmdBlankButton);

	// we are always active
	isSelectable = true;
	}

//	-------------------------------------------------------------------------
//
//	FUNC: 	    AdjustGlobalCoords
//	DESC:		Adjusts the coordinates of the HScrollBar (used internally)
//
//	-------------------------------------------------------------------------
void HScrollBar::AdjustGlobalCoords(int dx,int dy)
	{
	// we adjust our own coordinates, then the buttons
	Accessory::AdjustGlobalCoords(dx,dy);
	scalestart += dx;
	scaleend += dx;
	LtArrow->AdjustGlobalCoords(dx,dy);
	RtArrow->AdjustGlobalCoords(dx,dy);
	Blank->AdjustGlobalCoords(dx,dy);
	}

//	-------------------------------------------------------------------------
//
//	FUNC: 		RefreshBlank
//	DESC:		Draws the 'button' which slides up and down the HScrollBar
//
//	-------------------------------------------------------------------------
void HScrollBar::RefreshBlank()
	{
	// store the coords for the last blank box
	grRect oldblank = {Blank->r.left,Blank->r.top,Blank->r.right,Blank->r.bottom};
	grSetColor(inactive_color);

	// calculate a new starting position (0.5 added due to rounding errors)
	// by calc'ng a scaling factor, we can then simply mult (pos * scale)
	int max = maxpos;
	if (max>0)
		{
		scale  = (float) (scalewidth-buttonsize) / (float) max;
		blankpos = (int) ((float) scalestart + (float) pos * (float) scale + 0.5);
		}
	else blankpos = scalestart;

	// check....
	int dx = blankpos-Blank->r.left;
//	if (dx==0) return;

	// and set the new coordinates
	Blank->AdjustGlobalCoords(dx,0);

	// we calculate area to redraw, ignoring where the new button overwrites.
	if (oldblank.left<Blank->r.left)
		{
		// we are going right.
		if (oldblank.right>Blank->r.left) oldblank.right = Blank->r.left;
		if (oldblank.left>Blank->r.right) oldblank.left  = Blank->r.right;
		}
	else
		{
		// we are going left.
		if (oldblank.right<Blank->r.left) oldblank.right = Blank->r.left;
		if (oldblank.left<Blank->r.right) oldblank.left = Blank->r.right;
		}

	// clear the resized area
	grDrawFilledRect(oldblank.left,oldblank.top,oldblank.right,oldblank.bottom);

	// seems to be a bug, so adjust for it!
	while (oldblank.right>scaleend) { oldblank.left--; oldblank.right--; }

	// redraw the vertical bars over the filled area
	grDrawHBar(oldblank.top,oldblank.left,oldblank.right,bs_LOWERED);
	grDrawHBar(oldblank.bottom-grSizeOfBorder[bs_RAISED],oldblank.left,oldblank.right,bs_RAISED);
	grDrawHBar(oldblank.top+(oldblank.bottom-oldblank.top)/2 - grSizeOfBorder[bs_OUTLINE],
			 oldblank.left,oldblank.right,bs_OUTLINE);

	// and redraw the blank (at its new coordinates)
	Blank->Refresh();
	}

//	-------------------------------------------------------------------------
//
//	FUNC: 		DeSelect
//	DESC:		Deselects the HScrollBar
//
//	-------------------------------------------------------------------------
void HScrollBar::DeSelect()
	{
	Accessory::DeSelect();
	LtArrow->isSelectable = false;
	LtArrow->Refresh();
	RtArrow->isSelectable = false;
	RtArrow->Refresh();
	isSelectable = false;
	Refresh();
	}

//	-------------------------------------------------------------------------
//
//	FUNC: 		Draw
//	DESC:		Draws the HScrollBar
//
//	-------------------------------------------------------------------------
void HScrollBar::Refresh()
	{
	grDrawHBar(inset_r.top,scalestart,scaleend,bs_LOWERED);
	grDrawHBar(inset_r.bottom-grSizeOfBorder[bs_RAISED],scalestart,scaleend,bs_RAISED);
	grDrawHBar(inset_r.top+(inset_r.bottom-inset_r.top)/2-grSizeOfBorder[bs_OUTLINE],
				scalestart,scaleend,bs_OUTLINE);

	LtArrow->Refresh();
	RtArrow->Refresh();

	RefreshBlank();
	}

//	-------------------------------------------------------------------------
//
//	FUNC: 		HandleEvent
//	DESC:		Handles up,down,pgup,pgdn,home and end.
//				Also responds to the left mouse button
//
//	-------------------------------------------------------------------------
void HScrollBar::HandleEvent(Event &e)
	{
	// a flag for 'the event has been handled'
	bool handled = false;

	switch (e.what)
		{
		// if its a mouse event
		case	ev_mouse:   // if the mouse was clicked inside the ltarrow
							if (LtArrow->MouseInside(e.mouse.where))
								{
								if (selected_options&scrollbar_SINGLE_CLICK)
									{
									bool leftup = (e.mouse.event==mb_leftup);
									LtArrow->HandleEvent(e);
									if (leftup)
										if (pos>0) { (pos)--; RefreshBlank(); }
									}
								else
									if (selected_options&scrollbar_AUTO_REPEAT)
										{
										LtArrow->HandleEvent(e);
										if (pos>0) { (pos)--; RefreshBlank(); }
										}
								handled = true;
								}

							if (!handled)
								// if it was clicked inside the rtarrow
								if (RtArrow->MouseInside(e.mouse.where))
									{
									if (selected_options&scrollbar_SINGLE_CLICK)
										{
										bool leftup1 = (e.mouse.event==mb_leftup);
										RtArrow->HandleEvent(e);
										if (leftup1)
											if (pos<maxpos) { (pos)++; RefreshBlank(); }
										}
									else
										if (selected_options&scrollbar_AUTO_REPEAT)
											{
											RtArrow->HandleEvent(e);
											if (pos<maxpos) { (pos)++; RefreshBlank(); }
											}
									handled = true;
									}

							if (!handled)
							// the mouse was clicked somewhere on the bar
									{
									// if the LtArrow is currently active, deactivate it
									if (LtArrow->isSelectable)
										{
										LtArrow->isSelectable = false;
										LtArrow->Refresh();
										}

									// if the RtArrow is currently active, deactivate it
									if (RtArrow->isSelectable)
										{
										RtArrow->isSelectable = false;
										RtArrow->Refresh();
										}

									// calc a new position based upon the mouse coordinates
									if (maxpos>0)
										{
										scale  = (float) (scalewidth-buttonsize) / (float) (maxpos);
										pos = (int) (((float)e.mouse.where.x-(float)scalestart-(float)(buttonsize/2)) / (float)scale);
										}
									else pos = 0;

									// check bounds
									if (pos<0) pos = 0;
									if (pos>maxpos) pos = maxpos;

									// and whether or not to redraw the blank button
									if (pos!=lastpos)
										{
										RefreshBlank();
										lastpos=pos;
										}

									handled = true;
									}
							break;

		case	ev_keyboard:    switch (e.keyboard.event)
									{
									// handle ltarrow
									case	LT_ARROW_DN:  	if (!LtArrow->isSelectable)
																{
																LtArrow->isSelectable = true;
																LtArrow->Refresh();
																}
															if (!(selected_options&scrollbar_SINGLE_CLICK))
																if (pos>0) { (pos)--; RefreshBlank(); }
															handled = true;
															break;
									case	LT_ARROW_UP:	LtArrow->isSelectable = false;
															LtArrow->Refresh();
															if (selected_options&scrollbar_SINGLE_CLICK)
																if (pos>0) { (pos)--; RefreshBlank(); }
															EVENTMGR->ClearEvent(e);
															break;

									// handle rtarrow
									case	RT_ARROW_DN: 	if (!RtArrow->isSelectable)
																{
																RtArrow->isSelectable = true;
																RtArrow->Refresh();
																}
															if (!(selected_options&scrollbar_SINGLE_CLICK))
																if (pos<maxpos) { (pos)++; RefreshBlank(); }
															handled = true;
															break;
									case	RT_ARROW_UP:	RtArrow->isSelectable = false;
															RtArrow->Refresh();
															if (selected_options&scrollbar_SINGLE_CLICK)
																if (pos<maxpos) { (pos)++; RefreshBlank(); }
															EVENTMGR->ClearEvent(e);
															break;

									// handle page_up
									case	PGUP_DN:  		if (!LtArrow->isSelectable)
																{
																LtArrow->isSelectable = true;
																LtArrow->Refresh();
																}
															if (!(selected_options&scrollbar_SINGLE_CLICK))
																{
																if (pos>(maxpos/10)) (pos)-=(maxpos/10);
																else pos=0;
																RefreshBlank();
																}
															EVENTMGR->ClearEvent(e);
															break;
									case	PGUP_UP:		LtArrow->isSelectable = false;
															LtArrow->Refresh();
															if (selected_options&scrollbar_SINGLE_CLICK)
																{
																if (pos>(maxpos/10)) (pos)-=(maxpos/10);
																else pos=0;
																RefreshBlank();
																}
															handled = true;
															EVENTMGR->ClearEvent(e);
															break;

									// handle page_dn
									case	PGDN_DN: 		if (!RtArrow->isSelectable)
																{
																RtArrow->isSelectable = true;
																RtArrow->Refresh();
																}
															if (!(selected_options&scrollbar_SINGLE_CLICK))
																{
																if (pos<maxpos-(maxpos/10)) (pos)+=(maxpos/10);
																else pos = maxpos;
																RefreshBlank();
																}
															EVENTMGR->ClearEvent(e);
															break;
									case	PGDN_UP:		RtArrow->isSelectable = false;
															RtArrow->Refresh();
															if (selected_options&scrollbar_SINGLE_CLICK)
																{
																if (pos<maxpos-(maxpos/10))  (pos)+=(maxpos/10);
																else pos = maxpos;
																RefreshBlank();
																}
															handled = true;
															EVENTMGR->ClearEvent(e);
															break;


									// handle home
									case	HOME_DN:	if (!LtArrow->isSelectable)
															{
															LtArrow->isSelectable = true;
															LtArrow->Refresh();
															}
														EVENTMGR->ClearEvent(e);
														break;
									case	HOME_UP:	LtArrow->isSelectable = false;
														LtArrow->Refresh();
														if (pos!=0)
															{
															pos = 0;
															RefreshBlank();
															}
														handled = true;
														break;

									// handle end
									case	END_DN:		if (!RtArrow->isSelectable)
															{
															RtArrow->isSelectable = true;
															RtArrow->Refresh();
															}
														EVENTMGR->ClearEvent(e);
														break;
									case	END_UP:		RtArrow->isSelectable = false;
														RtArrow->Refresh();
														if (pos!=maxpos)
															{
															pos = maxpos;
															RefreshBlank();
															}
														handled = true;
														break;
									}
								break;
		}

	// if we have handled the event, modify the event to contain our returnEvent
	if (handled)
		{
		CommandEvent ce = { returnEvent, &pos };
		EVENTMGR->ModEvent(e,0,ev_command,&ce);
		}
	}


