//	-------------------------------------------------------------------------
//
//	TEXTBOX.CPP
//
//	-------------------------------------------------------------------------
//
//	AUTHOR:			Darren J. Dwyer
//
//	DESCRIPTION:	This file contains code for a TextBox class which allows
//					simple scrolling of text lists using scrollbars.
//
//					Also containes code for a DirBox, FileBox & DiskBox.
//
//	-------------------------------------------------------------------------

#include "textbox.hpp"
#include "eventmgr.hpp"
#include "mouse.hpp"
#include "keydefns.hpp"
//#include "dir.hpp"
#include "buttons.hpp"

//#include "dir.h"

//	-------------------------------------------------------------------------
//
//	FUNC:		CONSTRUCTOR for TextBox
//
//	DESC:		A simple "list box" containing multiple rows of Strings
//				with vertical scroll bars to scan through them
//
//	PARAMS:		x1,y1		- the coordinates of the top-left of the box
//				numrows		- the number of rows to display
//				numcols		- the number of cols to display
//				rownum		- the starting position
//				str			- the string containing the rows delimeted by '/'
//							  chars. eg. "JAN/FEB/MAR/APR/MAY"
//
//	-------------------------------------------------------------------------
TextBox::TextBox(int x1,int y1,int numrows,int numcols,
				 int rownum,String* str,
				 int options,int retevent)
		: Accessory(x1,y1,x1+grSizeOfBorder[bs_LOWERED]*2+numcols*grFontWidth("M")+6,
						  y1+grSizeOfBorder[bs_LOWERED]*2+numrows*(grFontHeight()+4),
						  FGCOLOR,BGCOLOR,bs_LOWERED,bs_LOWERED,
						  options,retevent)
	{
	// save our row string
	rowstr = str;

	// calc some values
	row_height = 2 + grFontHeight() + 2;

	num_rows = numrows;
	num_cols = numcols;

	row_posn = rownum;
	last_row_posn = rownum+1;

	timesclicked = 0;

	vbar = NULL;
	vbar_result = row_posn;

	row_offset = 0;
	last_row_offset = 0;

	// calc sizes of rows (speedy access later)
	CalcRowSizes();

	// create a standard button and get its width
	DnArrowButton* tmparr = new DnArrowButton(0,0,0,0);
	vbar_width = tmparr->r.right-tmparr->r.left-1;
	delete tmparr;

	// modify our scrollbox to include a vertical scroll bar
	inset_r.right += vbar_width;
	r.right += vbar_width;

	vbar = new VScrollBar(inset_r.right-vbar_width,inset_r.top,inset_r.bottom,
							row_posn,max_rows-1,0,cmdVScrollBar);

	// and make sure we have selected either single or double click
	if ((selected_options&textbox_SINGLECLICK)==0)
		if ((selected_options&textbox_DOUBLECLICK)==0)
			selected_options |= textbox_SINGLECLICK;

	// we are always active
	isActive = true;
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		DESTRUCTOR for TextBox
//
//	-------------------------------------------------------------------------
TextBox::~TextBox()
	{
	if (vbar!=NULL) { delete vbar; vbar = NULL; }
	}

//	-------------------------------------------------------------------------
//
//	FUNC:	AdjustCoords()
//	DESC:	adjusts the coordinates of the textbox & the vscrollbar
//
//	-------------------------------------------------------------------------
void TextBox::AdjustCoords(int dx,int dy)
	{
	Accessory::AdjustCoords(dx,dy);
	vbar->AdjustCoords(dx,dy);
	}

//	-------------------------------------------------------------------------
//
//	FUNC:	CalcRowSizes()
//	DESC:   Calculates the starting position (char*) and length (int)
//			of each row located in the current rowstr
//
//	-------------------------------------------------------------------------
void TextBox::CalcRowSizes()
	{
	// skip out if no calculating required
	if (selected_options&textbox_DONT_CALC) return;

	// set some values
	max_rows = 0;
	max_row_offset = 0;
	row_posn = 0;
	row_offset = 0;

	// skip out if we have a null string
	if (rowstr->Length()==0)
		{
		vbar->SetPos(0);
		vbar->SetMaxPos(0);
		return;
		}

	// calc max rows
	max_rows = rowstr->NumElems('/')+1;
	if (max_rows<=0) return;
	if (max_rows>TEXTBOX_MAX_ELEMS) max_rows = TEXTBOX_MAX_ELEMS;

	// and scan through the rowstr, calculating start positions
	// and lengths of each row
	int start = 0;
	int end = 0;
	for (int i=0;i<max_rows;i++)
		{
		start = end;
		rowstart[i] = &rowstr->Contents()[end];
		end = rowstr->ScanRight(end,'/')-1;
		rowsize[i] = end-start+1;
		end+=2;
		}

	// calc the maximum row offset (saves recalculating later)
	max_row_offset = max_rows - num_rows;
	if (max_row_offset<0) max_row_offset = 0;

	// and set the vbar position
	if (max_rows>0)
		{
		vbar->SetPos( row_posn,false );
		vbar->SetMaxPos( max_rows-1,false );
		}
	}

//	-------------------------------------------------------------------------
//
//	FUNC:	IncRowPosn()
//	DESC:   Increments the row pointer, updating offset if required &
//			redrawing if required
//
//	-------------------------------------------------------------------------
void TextBox::IncRowPosn(bool refresh)
	{
	int last_row_posn = row_offset + num_rows - 1;

	if (row_posn<last_row_posn)
		{
		if (row_posn<max_rows)
			{
			row_posn++;
			if (refresh) Refresh();
			}
		}
	else
		if (row_offset<max_row_offset)
			{
			row_posn++;
			row_offset++;
			if (refresh) Refresh();
			}
	}

//	-------------------------------------------------------------------------
//
//	FUNC:	DecRowPosn()
//	DESC:   Decrements the row pointer, updating offset if required &
//			redrawing if required
//
//	-------------------------------------------------------------------------
void TextBox::DecRowPosn(bool refresh)
	{
	if (row_posn>row_offset)
		{
		row_posn--;
		if (refresh) Refresh();
		}
	else
		if (row_offset>0)
			{
			row_posn--;
			row_offset--;
			if (refresh) Refresh();
			}
	}

//	-------------------------------------------------------------------------
//
//	FUNC:	SetRowPosn()
//	DESC:   Sets the row pointer to posn, updating offset if required &
//			updating the vertical bar if required. Automatically refreshes.
//
//	-------------------------------------------------------------------------
void TextBox::SetRowPosn(int posn, bool update_vbar)
	{
	if (row_posn<posn)
		{
		while (row_posn<posn)
			IncRowPosn(false);
		if (update_vbar) vbar->SetPos( row_posn );
		Refresh();
		}

	if (row_posn>posn)
		{
		while (row_posn>posn)
			DecRowPosn(false);
		if (update_vbar) vbar->SetPos( row_posn );
		Refresh();
		}
	}

//	-------------------------------------------------------------------------
//
//	FUNC:	GetRowPosn()
//	DESC:   Returns the row selected based upon absolute mouse coordinates
//
//	-------------------------------------------------------------------------
int  TextBox::GetRowPosn(grPoint& pt)
	{
	int posn = ((pt.y-inset_r.top) / row_height) + row_offset;

	if (posn>max_rows-1) posn = max_rows-1;
	if (posn<0) posn = 0;

	return posn;
	}

//	-------------------------------------------------------------------------
//
//	FUNC:	DrawRow()
//	DESC:   Draws the specified row, highlighting if required
//
//	-------------------------------------------------------------------------
void TextBox::DrawRow(int rownum)
	{
	String* tmpstr;

	if (rownum<max_rows)
		{
		int siz = rowsize[rownum];
		if (siz>12) siz = 12;
		tmpstr = new String(rowstart[rownum],siz);
		}
	else
		tmpstr = new String("");

	// save some time
	int row_number = rownum - row_offset;

	int yy1 = inset_r.top + row_number * row_height;
	int yy2 = yy1 + row_height;

	int fg = fgcolor;
	int bg = bgcolor;
	if (rownum == row_posn)
		{
		fg = bgcolor;
		bg = fgcolor;
		}

	// 'undraw' the last selected row...
	grSetColor(bg);
	grDrawFilledRect(inset_r.left,yy1,inset_r.right-vbar_width,yy2);

	// and draw the newly highlighted row iff its allright
	if (tmpstr->Length()>0)
		{
		grSetColor(fg);
		grDrawString(inset_r.left+3,yy1+3,"%s",tmpstr->Contents());
		}

	delete tmpstr;
	}

//	-------------------------------------------------------------------------
//
//	FUNC:	DrawAllRows()
//	DESC:   Draws all rows in the textscrollbox
//
//	-------------------------------------------------------------------------
void TextBox::DrawAllRows()
	{
	for (int i=0;i<num_rows;i++)
		DrawRow(row_offset+i);
	}

//	-------------------------------------------------------------------------
//
//	FUNC:	Draw()
//	DESC:   Draws the textscrollbox
//
//	-------------------------------------------------------------------------
void TextBox::Draw()
	{
	if (isActive)
		grDrawWindow(r.left,r.top,r.right,r.bottom,bgcolor,active_bs);
	else
		grDrawWindow(r.left,r.top,r.right,r.bottom,fgcolor,inactive_bs);

	vbar->Draw();

	Refresh();
	}

//	-------------------------------------------------------------------------
//
//	FUNC: 		Refresh
//	DESC:		refreshes the textscrollbox
//
//	-------------------------------------------------------------------------
void TextBox::Refresh()
	{
	if (vbar!=NULL) vbar->DrawBlank();
	DrawAllRows();
	}

//	-------------------------------------------------------------------------
//
//	FUNC:	HandleEvent()
//	DESC:   Handles all events pertaining to a vertical scroll bar as
//			well as:
//
//			left mouse button selects a specified row. if DOUBLE_CLICK then
//			must click same row twice in succession.
//
//			ENTER selects the currently highlighted row
//
//	-------------------------------------------------------------------------
void TextBox::HandleEvent(Event &e)
	{
	if (vbar!=NULL)
		if (e.what==ev_mouse)
		{
		if (vbar->MouseInside(e.mouse.where))
			{
			vbar->HandleEvent(e);
			vbar_result = *((int*) e.command.data);

			if (vbar_result==row_posn-1) 	DecRowPosn();
			else
				if (vbar_result==row_posn+1) 	IncRowPosn();
				else
					SetRowPosn(vbar_result);

			EVENTMGR->ClearEvent(e);
			return;
			}
		}

	// see if we should return the row number now or not...
	bool ret = false;
	int lastrow;

	switch (e.what)
		{
		case	ev_mouse:		switch (e.mouse.event)
									{
									case	mb_leftup: 		lastrow = row_posn;

															// calculate & set the row the mouse was clicked on
															SetRowPosn(GetRowPosn(e.mouse.where));

															// draw the row iff it has changed or we are checking for single clicks
															if (row_posn!=lastrow) timesclicked = 0;
															else timesclicked++;

															// check for double click
															if (selected_options&textbox_DOUBLECLICK)
																{
																// if it was double clicked
																if (timesclicked==1)
																	{
																	// we want to return some info
																	ret = true;
																	// reset our click counter
																	timesclicked = 0;
																	}
																}

															// if we have selected single click, we must return a value now!
															if (selected_options&textbox_SINGLECLICK)
																{
																ret = true;
																timesclicked = 0;
																}

															EVENTMGR->ClearEvent(e);
															break;
									case	mb_rightup:		EVENTMGR->ClearEvent(e);
															break;
									}
								break;

		case	ev_keyboard:	switch (e.keyboard.event)
									{
									case	UP_ARROW_UP:	;
									case	DN_ARROW_UP:	;
									case	PGUP_DN:		;
									case	PGDN_DN:		;
									case	HOME_DN:		;
									case	END_DN:			vbar->HandleEvent(e);
															EVENTMGR->ClearEvent(e);
															break;
									case	UP_ARROW_DN:	;
									case	DN_ARROW_DN:	;
									case	PGUP_UP:		;
									case	PGDN_UP:		;
									case	HOME_UP:		;
									case	END_UP:			vbar->HandleEvent(e);
															vbar_result = *((int*) e.command.data);

															if (vbar_result==row_posn-1) 	DecRowPosn();
															else
																if (vbar_result==row_posn+1) 	IncRowPosn();
																else
																	SetRowPosn(vbar_result,false);

															EVENTMGR->ClearEvent(e);
															break;

									case	ENTER_UP:		ret = true;
															break;
									}
								break;
		}

	if (ret)
		{
		// and return the element number which was selected
		CommandEvent cmd = { returnEvent, &row_posn};
		EVENTMGR->ModEvent(e,0,ev_command,&cmd);
		}
	}


// save this for later..... draws an interesting scale!!
/*	MGL_pushViewport();
	rect r1 = {0,0,639,479};
	MGL_setViewport(r1);
	savecolor();
	MGL_setColor(LIGHTBLUE);
	MGL_lineCoord(r.right+2,r.top+BUTTONSIZE,
			  r.right+2,r.top+BUTTONSIZE+scalewidth);
	MGL_setColor(LIGHTGREEN);
	for (int ii=0;ii<maxpos;ii++)
		MGL_lineCoord(r.right-2,ii*scale,r.right+2,ii*scale);

	getch();
	restorecolor();
	MGL_popViewport();
*/

