//	-------------------------------------------------------------------------
//
//	SCROLBOX.CPP
//
//	-------------------------------------------------------------------------
//
//	AUTHOR:			Darren J. Dwyer
//
//	DESCRIPTION:	This file contains the definition for the basic
//					ScrollBox accessory
//
//	-------------------------------------------------------------------------

#include "scrolbox.hpp"
#include "mouse.hpp"
#include "eventmgr.hpp"
#include "borders.hpp"

//	-------------------------------------------------------------------------
//
//	CLASS: 		ScrollBox
//	DESC: 		This is a simple 2-way scrollbox
//
//	NOTES:		Assumes ODD number of rows
//
//	-------------------------------------------------------------------------

ScrollBox::ScrollBox(int x1,int y1,int x2,int y2,
						int rows,int maxrows,
						grImage* title_tile, int title_bstyle,
						grImage* normal_tile, int normal_bstyle,
						grImage* highlighted_tile, int highlighted_bstyle,
						int options,
						int hit_row_event,int hit_hi_row_event)
	: Accessory(x1,y1,x2,y2,options,hit_row_event,hit_hi_row_event)
	{
	ti_tile = title_tile;
	lo_tile = normal_tile;
	hi_tile = highlighted_tile;

	ti_bstyle = title_bstyle;
	lo_bstyle = normal_bstyle;
	hi_bstyle = highlighted_bstyle;

	num_rows 	= rows;
	max_rows	= maxrows;

	centre_pos  	= rows / 2;
	max_row_pos 	= rows-1;
	max_row_offset 	= max_rows-rows;

	current_row	= centre_pos;
	row_pos		= centre_pos;
	row_offset	= 0;

	title_bg	= grRED;
	title_text	= grWHITE;
	title_border= grRED;
	lo_bg		= grDARKGRAY;
	lo_text		= grLIGHTGRAY;
	lo_border	= grLIGHTGRAY;
	hi_bg		= grLIGHTGRAY;
	hi_text		= grDARKGRAY;
	hi_border	= grDARKGRAY;
	}


ScrollBox::~ScrollBox()
	{
	}

//	--- data manip functions

int		ScrollBox::GetCurrentRow()
	{
	return current_row;
	}

void	ScrollBox::SetRows(int rows)
	{
	num_rows = rows;

	centre_pos  	= rows / 2;
	max_row_pos 	= rows-1;
	max_row_offset 	= max_rows-rows;

	row_offset		= current_row-centre_pos;
	if (row_offset<0) row_offset = 0;
	if (row_offset>max_row_offset) row_offset = max_row_offset;

	row_pos			= current_row - row_offset;
	}

void	ScrollBox::SetMaxRows(int maxrows)
	{
	max_rows = maxrows-1;
	}

void	ScrollBox::SetEvents(int hit_row_event,int hit_hi_row_event)
	{
	LMBreturnEvent = hit_row_event;
	RMBreturnEvent = hit_hi_row_event;
	}

//	---	scroll functions

void	ScrollBox::ScrollUp(int rows)
	{
	// skip if at top
	if ((row_offset==0) && (row_pos==0)) return;

	// loop through, refreshing constantly
	int cnt = 0;

	if (selected_options&ac_CentreScroll)
		{
		// CENTRE SCROLL

		// if the highlight is below the centre position,
		// scroll it to the centre
		while ((cnt<rows) && (row_pos>centre_pos))
			{
			int orig_pos = row_pos;

			row_pos--; current_row--;

			DrawRow(orig_pos);
			DrawHighlightedRow(row_pos);
			RefreshRows(orig_pos,row_pos);

//			DBG(0,"  CENTRE SCROLL UP SINGLE:   OFFSET: %d     POS: %d    ACTUAL: %d",row_offset,row_pos,current_row);
			cnt++;
			}

		// scroll the whole display, keeping the highlighted row in the
		// centre until we reach the top
		while ((cnt<rows) && (row_pos>0))
			{
			if (row_offset>0)
				{
				row_offset--;
				current_row--;

				Refresh();
				}
			else
				if (row_pos>0)
					{
					int orig_pos = row_pos;

					row_pos--;
					current_row--;

					DrawRow(orig_pos);
					DrawHighlightedRow(row_pos);
					RefreshRows(orig_pos,row_pos);
					}

//			DBG(0,"  CENTRE SCROLL UP SINGLE:   OFFSET: %d     POS: %d    ACTUAL: %d",row_offset,row_pos,current_row);
			cnt++;
			}
		}
	else
		{
		// ENDPOINT SCROLL

		// if the highlight is below the centre position,
		// scroll it to the centre
		while ((cnt<rows) && (row_pos>0))
			{
			int orig_pos = row_pos;

			row_pos--; current_row--;

			DrawRow(orig_pos);
			DrawHighlightedRow(row_pos);
			RefreshRows(orig_pos,row_pos);

//			DBG(0,"ENDPOINT SCROLL UP SINGLE:   OFFSET: %d     POS: %d    ACTUAL: %d",row_offset,row_pos,current_row);
			cnt++;
			}

		// scroll the whole display, keeping the highlighted row at the
		// top until we reach the first row
		while ((cnt<rows) && (row_offset>0))
			{
			row_offset--;
			current_row--;

			Refresh();
//			DBG(0,"ENDPOINT SCROLL UP ALL:      OFFSET: %d     POS: %d    ACTUAL: %d",row_offset,row_pos,current_row);
			cnt++;
			}
		}
	}

void	ScrollBox::ScrollDn(int rows)
	{
	// skip if at top
	if ((row_offset==max_row_offset) && (row_pos==max_row_pos)) return;

	// loop through, refreshing constantly
	int cnt = 0;

	if (selected_options&ac_CentreScroll)
		{
		// CENTRE SCROLL

		// if the highlight is above the centre position,
		// scroll it to the centre
		while ((cnt<rows) && (row_pos<centre_pos))
			{
			int orig_pos = row_pos;

			row_pos++; current_row++;

			DrawRow(orig_pos);
			DrawHighlightedRow(row_pos);
			RefreshRows(orig_pos,row_pos);

//			DBG(0,"  CENTRE SCROLL DN SINGLE:   OFFSET: %d     POS: %d    ACTUAL: %d",row_offset,row_pos,current_row);
			cnt++;
			}

		// scroll the whole display, keeping the highlighted row in the
		// centre until we reach the bottom
		while ((cnt<rows) && (row_pos<max_row_pos))
			{
			if (row_offset<max_row_offset)
				{
				row_offset++;
				current_row++;
				Refresh();
				}
			else
				if (row_pos<max_row_pos)
					{
					int orig_pos = row_pos;

					row_pos++;
					current_row++;

					DrawRow(orig_pos);
					DrawHighlightedRow(row_pos);
					RefreshRows(orig_pos,row_pos);
					}

//			DBG(0,"  CENTRE SCROLL DN SINGLE:   OFFSET: %d     POS: %d    ACTUAL: %d",row_offset,row_pos,current_row);
			cnt++;
			}
		}
	else
		{
		// ENDPOINT SCROLL

		// if the highlight is below the centre position,
		// scroll it to the centre
		while ((cnt<rows) && (row_pos<max_row_pos))
			{
			int orig_pos = row_pos;

			row_pos++; current_row++;

			DrawRow(orig_pos);
			DrawHighlightedRow(row_pos);
			RefreshRows(orig_pos,row_pos);

//			DBG(0,"ENDPOINT SCROLL DN ALL:      OFFSET: %d     POS: %d    ACTUAL: %d",row_offset,row_pos,current_row);
			cnt++;
			}

		// scroll the whole display, keeping the highlighted row at the
		// top until we reach the first row
		while ((cnt<rows) && (row_offset<max_row_offset))
			{
			row_offset++;
			current_row++;

			Refresh();
//			DBG(0,"ENDPOINT SCROLL DN ALL:      OFFSET: %d     POS: %d    ACTUAL: %d",row_offset,row_pos,current_row);
			cnt++;
			}
		}
	}

void	ScrollBox::ScrollTo(int row)
	{
	if (row==current_row) return;

	int rows = row-current_row;
	if (rows>0) ScrollDn(rows);
	else if (rows<0) ScrollUp(-rows);
	}

void	ScrollBox::ScrollTop()
	{
	ScrollTo(0);
	}

void	ScrollBox::ScrollBottom()
	{
	ScrollTo(max_rows);
	}


//	--- draw functions

void	ScrollBox::DrawTitle()
	{
	if (!isActive) return;

	grTileRect(r.left,r.top,r.right,r.top+19,-1,ti_tile,NO_ROTATE);
	grDrawBorder(r.left,r.top,r.right,r.top+19,ti_bstyle);
	}

void	ScrollBox::DrawRow(int row,bool auto_fill)
	{
	if (!isActive) return;

	int y = r.top + 20 + 20*row;	// title + row

	if (auto_fill)
		{
		grTileRect(r.left,y,r.right,y+19,-1,lo_tile,NO_ROTATE);
		grDrawBorderSection(r.left,r.top+20,r.right,r.bottom,
						r.left,y,r.right,y+19,lo_bstyle);
		}
	}

void	ScrollBox::DrawHighlightedRow(int row,bool auto_fill)
	{
	if (!isActive) return;

	int y = r.top + 20 + 20*row;	// title + row

	if (auto_fill)
		{
		grTileRect(r.left,y,r.right,y+19,-1,hi_tile,NO_ROTATE);
//		grDrawBorderSection(r.left,r.top+20,r.right,r.bottom,
//						r.left,y,r.right,y+19,lo_bstyle);
		grDrawBorder(r.left,y,r.right,y+19,hi_bstyle);
		}
	}

void	ScrollBox::DrawAllRows()
	{
	if (!isActive) return;

	// draw all rows background & border now
	grTileRect(r.left,r.top+20,r.right,r.bottom,-1,lo_tile,NO_ROTATE);
	grDrawBorder(r.left,r.top+20,r.right,r.bottom,lo_bstyle);

	// for each row
	for (int row=0;row<num_rows;row++)
		{
		// highlight the appropriate row, filling the background
		if (row+row_offset==current_row) DrawHighlightedRow(row,true);
		// draw the row, but don't fill it
		else DrawRow(row,false);
		}
	}

void	ScrollBox::RefreshRows(int row1,int row2)
	{
	if (!isActive) return;

	int first = min(row1,row2); //first = min(first,0);
	int last  = max(row1,row2); //last  = max(last,max_row_pos);

	int scroll_top = r.top + 20;

	int y1 = scroll_top + 20*first;
	int y2 = scroll_top + 20*last + 19;

	grBlitRegion(r.left,y1,r.right,y2);
	}

//	---	accessory functions

int		ScrollBox::WhichRowSelected(grPoint& ms_pt)
	{
	// clip mouse coords to our scrollbox (excluding title area)
	int y = ms_pt.y - r.top;

	// return -1 for the title area
	if (y<20) return -1;

	// return absolute row number
	return row_offset + ((y-20) / 20);
	}

void	ScrollBox::Draw()
	{
	if (!isActive) return;

	DrawTitle();
	DrawAllRows();
	}

void	ScrollBox::Draw(grVirtualBuffer* vb)
	{
	if (!isActive) return;
	}

void	ScrollBox::HandleEvent(Event& e)
	{
	int row;
	switch (e.what)
		{
		case ev_mouse:

			switch (e.mouse.event)
				{
				case mb_leftup:
				case mb_rightup:

					row = WhichRowSelected(e.mouse.where);

					if (row!=-1)
						{
						ScrollTo(row);

						int evt;
						if (row!=current_row) evt = LMBreturnEvent;
						else				  evt = RMBreturnEvent;

						EVENTMGR->ModCommandEvent(e,e.where,evt,(void*) row);
						return;
						}

					EVENTMGR->ClearEvent(e);
					break;
				}
			return;
		}
	}
