//	-------------------------------------------------------------------------
//
//	EVENTMGR.CPP
//
//	-------------------------------------------------------------------------
//
//	AUTHOR:			Darren J. Dwyer
//
//	DESCRIPTION:	This file contains code for the EVENT MANAGER.
//					The EVENTMGR maintains internally an EVENT QUEUE, which
//					can be used through the calls GenEvent() & ModEvent()
//
//	-------------------------------------------------------------------------

#include "eventmgr.hpp"
#include "mouse.hpp"
#include "keyboard.hpp"
#include "timer.hpp"

#include "backgrnd.hpp"
#include "onscreen.hpp"

#include "dos.h"
#include "mem.h"


//#define	__DEBUG_TAB

// the GLOBAL declaration for our EVENTMGR!!!

EventMgr* EVENTMGR = NULL;

// a GLOBAL flag to tell whether or not the EVENTMGR is installed
bool	eventmgr_INSTALLED = false;


// different display method in GRAPHICS.CPP

//	-------------------------------------------------------------------------
//
//  FUNC:		The CONSTRUCTOR
//	DESC:		Initializes the event manager, setting no event handlers,etc.
//
//	-------------------------------------------------------------------------

EventMgr::EventMgr()
	{
	numEventHandlers = 0;
	numEventsPending = 0;

	nextPending = 0;
	lastPending = 0;

	isActive = true;
	isTerminating = false;
	eventIsPending = false;
	eventQueueIsFull = false;
	eventHandlersAtMax = false;

	// note: the first event handler is assumed to respond to keyboard input
	// 		 always!!! (May be a background handler)
	currentHandler = -1;
	currentZFactor = -1;

	HEAPCHK("by EventMgr CONSTRUCTOR");
	}

//	-------------------------------------------------------------------------
//
//  FUNC:		The DESTRUCTOR
//	DESC:		Terminates the event manager, removing all event handlers
//
//	-------------------------------------------------------------------------
EventMgr::~EventMgr()
	{
	isActive = false;
	HEAPCHK("by EventMgr DESTRUCTOR");
	}

//	-------------------------------------------------------------------------
//
//  FUNC:		SelectNextHandler()
//	DESC:		Selects the next available handler (which is responding to
//				keyboard input) to be the selected keyboard input handler!
//
//	-------------------------------------------------------------------------
void EventMgr::SelectNextHandler()
	{
	if (numEventHandlers==0) return;
	if (!handler[currentHandler]->isOnScreenHandler) return;

#ifdef __DEBUG_TAB
	DBG(450,"IN EVENTMGR          : SELECT_NEXT_HANDLER()");
	DBG(460,"NUM HANDLERS         : %d",numEventHandlers);
#endif

	OnScreenHandler* hand = (OnScreenHandler*) handler[currentHandler];

	// deselect the current event handler
	hand->DeSelect();

	// save the current handler
	int current = currentHandler;

	// and do at least once
	bool quit = false;
	do
		{
		// increment the current handler to the next
		current++;
		// wrap if out of bounds
		if (current>=numEventHandlers) current = 0;

		// if its an onscreen handler, we can select it
		if (handler[current]->isOnScreenHandler)
			{
			hand = (OnScreenHandler*) handler[current];
			if ((hand->isOnScreen) && (hand->isRespondingKbd))
				quit = true;
			}
		// note: automatically ignores all BackgroundTasks
		}
	while (!quit);

	// set the new keyboard input handler
	currentHandler = current;

	// select the new event handler
	hand->Select(evh_FIRST_ACCESSORY);


#ifdef __DEBUG_TAB
	DBG(470,"SELECTED HANDLER     : %d",current);
	grDrawRect(hand->r.left,hand->r.top,
			   hand->r.right,hand->r.bottom,
			   grWHITE);
	grBlitRegion(hand->r.left,hand->r.top,
			   hand->r.right,hand->r.bottom);
#endif
//	HEAPCHK("by EventMgr::SelectNextHandler()");
	}

//	-------------------------------------------------------------------------
//
//  FUNC:		SelectPrevHandler()
//	DESC:		Selects the previously available handler (which is responding
//				to keyboard input) to be the selected keyboard input handler!
//
//	-------------------------------------------------------------------------
void EventMgr::SelectPrevHandler()
	{
	if (numEventHandlers==0) return;
	if (!handler[currentHandler]->isOnScreenHandler) return;

#ifdef __DEBUG_TAB
	DBG(450,"IN EVENTMGR          : SELECT_PREV_HANDLER()");
	DBG(460,"NUM HANDLERS         : %d",numEventHandlers);
#endif

	OnScreenHandler* hand = (OnScreenHandler*) handler[currentHandler];

	// deselect the current event handler
	hand->DeSelect();

	// save the current handler
	int current = currentHandler;

	// and do at least once
	bool quit = false;
	do
		{
		// increment the current handler to the next
		current--;
		// wrap if out of bounds
		if (current<0) current = numEventHandlers-1;

		// if its an onscreen handler, we can select it
		if (handler[current]->isOnScreenHandler)
			{
			hand = (OnScreenHandler*) handler[current];
			if ((hand->isOnScreen) && (hand->isRespondingKbd))
				quit = true;
			}
		// note: automatically ignores all BackgroundTasks
		}
	while (!quit);

	// set the new keyboard input handler
	currentHandler = current;

	// select the new event handler
	hand->Select(evh_LAST_ACCESSORY);

#ifdef __DEBUG_TAB
	DBG(470,"SELECTED HANDLER     : %d",current);
	grDrawRect(hand->r.left,hand->r.top,
			   hand->r.right,hand->r.bottom,
			   grWHITE);
	grBlitRegion(hand->r.left,hand->r.top,
			   hand->r.right,hand->r.bottom);
#endif
//	HEAPCHK("by EventMgr::SelectPrevHandler()");
	}

//	-------------------------------------------------------------------------
//
//  FUNC:		SelectEventHandler( int hand )
//	DESC:		Selects a specified event handler (which is responding to
//				mouse & keyboard input) as the current handler.
//
//	NOTES:		This function does not DeSelect() the current event handler!!
//
//
//	-------------------------------------------------------------------------
void EventMgr::SelectEventHandler( int hand )
	{
	// skip out if dummy values
	if ((hand<0) || (hand>=numEventHandlers)) return;

	// act if its an onscreen handler
	if (handler[hand]->isOnScreenHandler)
		{
		OnScreenHandler* handlr = (OnScreenHandler*) handler[hand];
		if ((handlr->isOnScreen) && (handlr->isRespondingMouse))
			{
			handlr->Select(evh_CURRENT_ACCESSORY);
			currentHandler = hand;
//			DBG(100,"SELECTED ONSCREENHANDLER FOR HANDLER %d",hand);
	//			delay(500);
			}
		}

//	DBG(50,"SELECTED EVENT HANDLER #%d",currentHandler);
//	HEAPCHK("by EventMgr::SelectEventHandler()");
	}

//	-------------------------------------------------------------------------
//
//  FUNC:		DeSelectEventHandler( int hand )
//	DESC:		DeSelects a specified event handler (which is responding to
//				mouse & keyboard input) so its not the current handler.
//
//	NOTES:		This function does not Select() the previously selected handler!!
//
//	-------------------------------------------------------------------------
void EventMgr::DeSelectEventHandler( int hand )
	{
//	HEAPCHK("IN DESELECT EVENT HANDLER");
	// skip out if dummy values
	if (hand>=numEventHandlers) return;

	// act if its an onscreen handler
	if (handler[hand]->isOnScreenHandler)
		{
		OnScreenHandler* handlr = (OnScreenHandler*) handler[hand];
		if (handlr->isOnScreen)
			{
			handlr->DeSelect();
	//			DBG(100,"DESELECTED ONSCREENHANDLER FOR HANDLER %d",hand);
	//			delay(500);
			}
		}
//	DBG(60,"DESELECTED EVENT HANDLER #%d",currentHandler);
//	HEAPCHK("by EventMgr::DeSelectEventHandler()");
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		WhichHandlerSelected()
//	DESC:		Calculates which event handler was selected based upon a mouse
//				location point.
//
//	PARAMS:		ms_point	- the point to check
//								eg. int which = WhichHandlerSelected(e.mouse.where);
//
//	RETURNS		>=0 - the number of the handler selected
//				 -1 - no event handler was selected
//
//	-------------------------------------------------------------------------
int EventMgr::WhichHandlerSelected(grPoint& ms_point)
	{
	// if there are no accessories, return -1
	if (numEventHandlers==0) return -1;

	// by default we assume no handler selected
	int handlernum = -1;
	int zfact = 0;

	// scan through each accessory
	for (int i=0;i<numEventHandlers;i++)
		if (handler[i]->isOnScreenHandler)
			{
			OnScreenHandler* hand = (OnScreenHandler*) handler[i];
			// if the point is inside the bounds of the handler
			if ((hand->isOnScreen)&&(hand->isRespondingMouse)&&
				(hand->MouseInside(ms_point)) )
				{
				// if the handler's zfactor is the highest
				if (hand->zfactor>=zfact)
					{
					// then thats the handler on top
					handlernum = i;
					zfact = hand->zfactor;
					}
				}
			}

	// if the zfactor of the handler found was less than the current
	// zfactor, we ignore the result
	if (zfact<currentZFactor) handlernum = -1;


//	HEAPCHK("by EventMgr::WhichHandlerSelected()");

	// return the result of our checking
	return handlernum;
	}

//	-------------------------------------------------------------------------
//
//  FUNC:		DrawAllHandlers()
//	DESC:		Scans through the event handler list and tells all onscreen
//				handlers to Draw()
//
//	-------------------------------------------------------------------------
void EventMgr::DrawAllHandlers()
	{
	// scan through all of the handlers
	for (int i=0;i<numEventHandlers;i++)
		// and if they're on screen
		if (handler[i]->isOnScreenHandler)
			// draw them
			{
//			DBG(470,"DRAWALLHANDLERS:: ABOUT TO START DRAWING HANDLER #%d",i);
			OnScreenHandler* hand = (OnScreenHandler*) handler[i];
			if (hand->isOnScreen) hand->Draw();
//			grBlitScreen();
//			if (hand->isOnScreen)
//				DBG(470,"HANDLER #%d  ATTRIBUTES ARE: %x   ISONSCREEN",i,hand->selected_options);
//			else
//				DBG(470,"HANDLER #%d  ATTRIBUTES ARE: %x",i,handler[i]->selected_options);
//			delay(2000);
			}

//	HEAPCHK("by EventMgr::DrawAllHandlers()");
	}

//	-------------------------------------------------------------------------
//
//  FUNC:		EventInQueue()
//	DESC:		scans the event queue, returning true if the specified event
//				was found in the queue
//
//	PARAMS:		what 	- the type of event to scan for
//				e 		- the specific event to scan for
//
//	-------------------------------------------------------------------------
bool EventMgr::EventInQueue(int what,void* e)
	{
	// scan through the number of events pending
	for (int i=0;i<numEventsPending;i++)
		{
		// figure out the appropriate index into the queue
		int index = i + nextPending;
		// wrapping iff required
		if (index>=EVENTMGR_DEFAULT_QUEUE_SIZE)
			index -= EVENTMGR_DEFAULT_QUEUE_SIZE;

		// if the pending event is of the appropriate type
		if (queue[index].what == what)
			// compare the events
			switch (what)
				{
				case	ev_timer: 		return (memcmp(e,&queue[index].timer,sizeof(TimerEvent))==0);
				case	ev_mouse:		return (memcmp(e,&queue[index].mouse,sizeof(MouseEvent))==0);
				case	ev_keyboard:	return (memcmp(e,&queue[index].keyboard,sizeof(KeyboardEvent))==0);
				case	ev_command:		return (memcmp(e,&queue[index].command,sizeof(CommandEvent))==0);
				case	ev_broadcast:	return (memcmp(e,&queue[index].broadcast,sizeof(BroadcastEvent))==0);
				}
		}


//	HEAPCHK("by EventMgr::EventInQueue()");
	// we didn't find it, return false...
	return false;
	}

//	-------------------------------------------------------------------------
//
//  FUNC:		GenEvent()
//	DESC:		Generate an event, placing it at the end of the Event Queue.
//
//	PARAMS:		where 	- the event handler where the event is going
//				what	- the type of event
//				data	- any relevant data accompanying the event
//
//	RETURNS:	true 	- if event was generated correctly.
//
//	-------------------------------------------------------------------------
bool 	EventMgr::GenEvent(int where,int what,void* data)
	{
	if (EVENTMGR==NULL) return false;

	// if the event queue is full return false
	if (eventQueueIsFull) return false;
	if (lastPending>=EVENTMGR_DEFAULT_QUEUE_SIZE)
		{
		DBG(470,"ERROR: ATTEMPT TO GENERATE EVENT IS OUT OF BOUNDS");
//		delay(5000);
		return false;
		}

	// place appropriate bits in queue
	queue[lastPending].where = where;
	queue[lastPending].what  = what;

	// copy the event data to the new spot
	char* srce = (char*) data;
	char* dest;

	int siz = 0;
	switch (what)
		{
		case ev_mouse:		siz = sizeof(MouseEvent);
							dest = (char*) &queue[lastPending].mouse;
							break;
		case ev_keyboard:	siz = sizeof(KeyboardEvent);
							dest = (char*) &queue[lastPending].keyboard;
							break;
		case ev_command:  	siz = sizeof(CommandEvent);
							dest = (char*) &queue[lastPending].command;
							break;
		case ev_broadcast:	siz = sizeof(BroadcastEvent);
							dest = (char*) &queue[lastPending].broadcast;
							break;
		case ev_timer:		siz = sizeof(TimerEvent);
							dest = (char*) &queue[lastPending].timer;
							break;
		}

	for (int cnt=0;cnt<siz;cnt++) *dest++ = *srce++;

	// figure out the next available slot
	lastPending++;
	if (lastPending>=EVENTMGR_DEFAULT_QUEUE_SIZE) lastPending = 0;

	// tell the manager that an event is pending
	eventIsPending = true;

	// increment our events pending counter
	if (numEventsPending<EVENTMGR_DEFAULT_QUEUE_SIZE) numEventsPending++;

	// the event queue is full when the first == last
	eventQueueIsFull = (numEventsPending>=EVENTMGR_DEFAULT_QUEUE_SIZE-1);

//	HEAPCHK("by EventMgr::GenEvent()");

	// event was created correctly
	return true;
	}

// --------------------

bool 	EventMgr::GenCommandEvent(int where,int event,void* data)
	{
	CommandEvent ce = { event, data };
	return GenEvent(where,ev_command,&ce);
	}

bool 	EventMgr::GenBroadcastEvent(int event,void* data)
	{
	BroadcastEvent be = { event,data };
	return GenEvent(0,ev_broadcast,&be);
	}

//	-------------------------------------------------------------------------
//
//  FUNC:		InsEvent()
//	DESC:		Generate an event, placing it at the start of the Event Queue.
//
//	PARAMS:		where 	- the event handler where the event is going
//				what	- the type of event
//				data	- any relevant data accompanying the event
//
//	RETURNS:	true 	- if event was generated correctly.
//
//	-------------------------------------------------------------------------
//$$$ int last_what;
bool EventMgr::InsEvent(int where,int what,void* data)
	{
	if (EVENTMGR==NULL) return false;

	// if the event queue is full return false
	if (eventQueueIsFull) return false;

	// figure out the next available slot
	nextPending--;
	if (nextPending<0) nextPending = EVENTMGR_DEFAULT_QUEUE_SIZE - 1;

	// place appropriate bits in queue
	queue[nextPending].where = where;
	queue[nextPending].what  = what;

	// copy the event data to the new spot
	char* srce = (char*) data;
	char* dest = (char*) &queue[nextPending].mouse;

	int siz = 0;
	switch (what)
		{
		case ev_mouse:		siz = sizeof(MouseEvent); 		break;
		case ev_keyboard:	siz = sizeof(KeyboardEvent); 	break;
		case ev_command:  	siz = sizeof(CommandEvent); 	break;
		case ev_broadcast:	siz = sizeof(BroadcastEvent); 	break;
		case ev_timer:		siz = sizeof(TimerEvent);		break;
		}

	for (int cnt=0;cnt<siz;cnt++) *dest++ = *srce++;

	// tell the manager that an event is pending
	eventIsPending = true;

	// increment our events pending counter
	if (numEventsPending<EVENTMGR_DEFAULT_QUEUE_SIZE) numEventsPending++;

	// the event queue is full when the first == last
	eventQueueIsFull = (numEventsPending>=EVENTMGR_DEFAULT_QUEUE_SIZE-1);

//	HEAPCHK("by EventMgr::InsEvent()");

	// event was created correctly
	return true;
	}

bool	EventMgr::InsCommandEvent(int where,int event, void* data)
	{
	CommandEvent ce = { event, data };
	return InsEvent(where,ev_command,&ce);
	}

bool	EventMgr::InsBroadcastEvent(int event,void* data)
	{
	BroadcastEvent be = { event, data };
	return InsEvent(0,ev_broadcast,&be);
	}

//	-------------------------------------------------------------------------
//
//  FUNC:		ModEvent()
//	DESC:		Modify a specified event (generally the current one).
//
//	PARAMS:     event	- the event to be modified
//				where 	- the event handler where the event is going
//				what	- the type of event
//				data	- any relevant data accompanying the event
//
//	RETURNS:	true if event was modified correctly.
//
//	-------------------------------------------------------------------------
void EventMgr::ModEvent(Event& e,int where,int what,void* data)
	{
	if (EVENTMGR==NULL) return;

	// replace appropriate bits in event
	e.where = where;
	e.what  = what;

	// copy the event data to the new spot
	char* srce = (char*) data;
	char* dest;

	int siz = 0;
	switch (what)
		{
		case ev_mouse:		siz = sizeof(MouseEvent);
							dest = (char*) &e.mouse;
							break;
		case ev_keyboard:	siz = sizeof(KeyboardEvent);
							dest = (char*) &e.keyboard;
							break;
		case ev_command:  	siz = sizeof(CommandEvent);
							dest = (char*) &e.command;
							break;
		case ev_broadcast:	siz = sizeof(BroadcastEvent);
							dest = (char*) &e.broadcast;
							break;
		case ev_timer:		siz = sizeof(TimerEvent);
							dest = (char*) &e.timer;
							break;
		}

	for (int cnt=0;cnt<siz;cnt++) *dest++ = *srce++;

//	HEAPCHK("by EventMgr::ModEvent()");
	}

void 	EventMgr::ModCommandEvent(Event& e,int where,int event, void* data)
	{
	CommandEvent ce = { event, data };
	ModEvent(e,where,ev_command,&ce);
	}

void 	EventMgr::ModBroadcastEvent(Event& e,int event,void* data)
	{
	BroadcastEvent be = { event, data };
	ModEvent(e,0,ev_broadcast,&be);
	}

//	-------------------------------------------------------------------------
//
//  FUNC:		ClearEvent()
//	DESC:		Clears the current event, by simply skipping to the next
//				pending event
//
//	PARAMS		e	- the event to be cleared
//
//	-------------------------------------------------------------------------

void EventMgr::ClearEvent(Event &e)
	{
	// clear the event
	e.what = ev_nothing;
	}

//	-------------------------------------------------------------------------
//
//  FUNC:		ClearAllEvents()
//	DESC:		Clears all pending events by resetting the queue.
//
//	-------------------------------------------------------------------------

void EventMgr::ClearAllEvents()
	{
	// clear all pending events
	nextPending = 0;
	lastPending = 0;
	eventIsPending = false;
	}

//	-------------------------------------------------------------------------
//
//  FUNC:		AddEventHandler()
//	DESC:		Adds an event handler to the (internal) list of event handlers
//
//	PARAMS:		eventhandler	- the result of a new() call.
//						  			eg. MenuWindow* menu =
//										EVENTMGR->AddEventHandler(new(MenuWindow));
//
//	RETURNS:	a pointer to the resulting event handler
//
//	-------------------------------------------------------------------------
EventHandler* EventMgr::AddEventHandler(EventHandler* eventhandler)
	{
	// skip out if new didn't work!
	if (eventhandler==NULL) return NULL;

	// skip out upon reaching max event handlers
	if (eventHandlersAtMax)
		{
		// remove floating copy of the eventhandler
		delete eventhandler; eventhandler = NULL;
		// and return a null
		return NULL;
		}

	// set the event handler
	handler[numEventHandlers] = eventhandler;
	eventhandler->SetHandlerNumber(numEventHandlers);

	// check to see if we need to select our first available OnScreenHandler
	if (currentHandler==-1)
		if (handler[numEventHandlers]->isOnScreenHandler)
			{
			// if this handler is responding to keyboard
			if (handler[numEventHandlers]->isRespondingKbd)
				currentHandler = numEventHandlers;
			currentZFactor = handler[numEventHandlers]->zfactor;
			}

	// increment the number of event handlers
	numEventHandlers++;

	// and check to see if we are at max...
	if (numEventHandlers==EVENTMGR_MAX_EVENT_HANDLERS-1) eventHandlersAtMax = true;

//	HEAPCHK("by EventMgr::AddEventHandler()");

	// and return the pointer for later access
	return eventhandler;
	}

//	-------------------------------------------------------------------------
//
//  FUNC:		DelEventHandler()
//	DESC:		Deletes the most recently Add'ed event handler.
//
//	-------------------------------------------------------------------------
void EventMgr::DelEventHandler()
	{
	// skip out if there are no eventhandlers
	if (numEventHandlers==0) return;

	// decrement the number of event handlers
	numEventHandlers--;

	// and delete the event handler
	delete handler[numEventHandlers]; handler[numEventHandlers] = NULL;

//	HEAPCHK("by EventMgr::DelEventHandler()~ERROR DELETING EVENT HANDLER #%d",numEventHandlers);
	}

//	-------------------------------------------------------------------------
//
//  FUNC:		DelAllEventHandlers()
//	DESC:		Deletes all of the event handlers...
//
//	NOTE:		do not use outside of PROGRAM.CPP
//
//	-------------------------------------------------------------------------
void EventMgr::DelAllEventHandlers()
	{
	// delete all of the event handlers using DelEventHandler
	while (numEventHandlers>0)
		{
//		DBG(470,"Deleting EVENT HANDLER #%d",numEventHandlers);
//		delay(1000);
		DelEventHandler();
		}
	}

//	-------------------------------------------------------------------------
//
//  FUNC:		HandleEvent()
//	DESC:		Handles the next pending event.
//
//				If the pending event's _where_ field is set to an event
//				handler then that event handler is passed the event immediately
//
//				If the pending event's _where_ field == ALLHANDLERS then all
//				EventHandlers are called in order of allocation till one of
//				the handlers handles the event!
//
//	-------------------------------------------------------------------------
void EventMgr::HandleEvent(Event &e)
	{
	// create this here
	int nexthandler = 0;
	int hand;
	bool polled;

	int	lastwhat = e.what;
	int lastwhere = e.where;
	int iter = 0;

	// do the switch (in order of speed preference!!)
	switch (e.what)
		{
		// if its a command event, go to its destination (allowing chaining!)
		// exception: if e.command.event == cmdExitProgram, terminate!
		case	ev_command:		//DBG(460,"EVENTMGR Handling CommandEvent");
								// skim through, chaining command events!
								while (e.what==ev_command)
									{
									if (e.command.event==cmdExitProgram)
										{
										ClearEvent(e);

										// finished executing event loop
										isExecuting = false;

										// so skip out
										return;
										}
									if (e.command.event==cmdTerminate)
										{
										ClearEvent(e);

										// finished executing event loop
										isTerminating = true;

										// so skip out
										return;
										}

//									DBG(470,"Sending CommandEvent %d to handler #%d",e.command.event,e.where);
									if (iter==0) handler[e.where]->HandleEvent(e);

									if ((e.what==lastwhat) &&
										(e.where==lastwhere)) iter++;
									else
										{
										lastwhat = e.what;
										lastwhere = e.where;
										iter = 0;
										}

									if (iter>10)
										{
										DBG(470,"ERROR: EVENT TYPE %d,  %d WAS NEVER USED!!!!",e.what,e.command.event);
//										delay(1000);
										EVENTMGR->ClearEvent(e);
										break;
										}
									}
								break;

		// if its a timer or a broadcast event
		case	ev_timer:		polled = false;
								switch (e.timer.event)
									{
/*									case	cmdPollMouse:   mouse_GetEvent();
															EVENTMGR->ClearEvent(e);
															polled = true;
*/
/*															pt ms = mouse_GetPos();
															hand = WhichHandlerSelected(ms);
															acc = hand->WhichAccessorySelected(ms);
															if (acc!=-1)
																hc = handler[hand]->accessory[acc]->helpcontext;
															else hc = handler[hand]->helpcontext;
															DisplayHelpContext(ms,hc);
*/
/*															break;
									case	cmdPollKbd:		keyboard_GetEvent();
															EVENTMGR->ClearEvent(e);
															polled = true;
															break;
*/									}
								if (polled) break;
								else; // note: continue as per broadcast events

		case	ev_broadcast:	// start at the first event handler
								nexthandler = 0;

								// loop through all the event handlers
								do
									{
									// sending the event to the next handler
//									DBG(460,"HANDLER #%d is handling event type %d, event %d",
//										nexthandler,e.what,e.command.event);
									handler[nexthandler]->HandleEvent(e);

									// and increment
									nexthandler++;
									}
								while (nexthandler<numEventHandlers);

								// broadcasts & timers are once-only, so clear
								ClearEvent(e);
								break;

		// if its a mouse event, transfer the event to the appropriate handler
		case	ev_mouse:       // DJD TEMPORARY

								if ((e.mouse.modifiers & keyboard_LEFT_CTRL) &&
									(e.mouse.modifiers & keyboard_LEFT_ALT))
									{
									grScaleScreen(e.mouse.where.x-16,
												  e.mouse.where.y-16,
												  e.mouse.where.x+16,
												  e.mouse.where.y+16,
												  320 * (e.mouse.where.x<320),
												  240 * (e.mouse.where.y<240),
												  4);
									}

								// figure out which handler was selected
								hand = WhichHandlerSelected(e.mouse.where);

								// if a handler was selected
								if (hand>=0)
									{
									// if we have selected a different handler, tell that handler to deselect
									if (hand!=currentHandler)
										{
										DeSelectEventHandler(currentHandler);
										SelectEventHandler(hand);
										}

									// and tell the handler to handle the event
									handler[currentHandler]->HandleEvent(e);
									}

								// if no handler has interpreted the info, clear it
								if (e.what==ev_mouse) ClearEvent(e);
								break;

		// if its a keyboard event, send it to (firstly) the last selected
		// handler,
		// ... then test for special keys
		// ... then all handlers in construction priority,

		case	ev_keyboard:    // try the current selected handler
								handler[currentHandler]->HandleEvent(e);

								// skip if it's been handled
								if (e.what!=ev_keyboard) return;

								// if the event wasn't handled, check for special keys
								if (e.what==ev_keyboard)
									{
									// SPECIAL CASE: CHECK FOR ALT_ESC OR CTRL_ESC
									//				 TO QUIT PROGRAM!!!
									switch (e.keyboard.event)
										{
										// handle the tab being released
										case TAB_DN	:	// if its shift-tab
														if (e.keyboard.modifiers&keyboard_SHIFT)
															SelectPrevHandler();
														else
															// its normal tab
															SelectNextHandler();

														// clear the event
														ClearEvent(e);
														return;

										case CAPS_LOCK_DN:

											SelectPrevHandler();
											ClearEvent(e);
											return;

										case ESC_UP: //DBG(470,"ESC key was released");
													 //if (e.keyboard.modifiers&keyboard_ALT) DBG(460,"ALT key was pressed also");
													 //if (e.keyboard.modifiers&keyboard_CTRL) DBG(450,"CTRL key was pressed also");
													 if ((e.keyboard.modifiers&keyboard_ALT) ||
														 (e.keyboard.modifiers&keyboard_CTRL))
														// turn off Execute()!
														{
														isExecuting = false;
														ClearEvent(e);
														return;
														}
													 break;

										case V_UP:	if (e.keyboard.modifiers&keyboard_RIGHT_CTRL)
														{
														DBG(0,"SWITCHING BLIT MODE");
														grToggleBlitMode();
														ClearEvent(e);
														}
													break;
										}
									}

								// start at the first handler
								nexthandler = 0;

								// loop through all event handlers
								do
									{
									// see if the handler can handle the event
									if (nexthandler!=currentHandler)
										{
//										DBG(470,"Sending keyboard event %d to handler %d",e.keyboard.event,nexthandler);
										handler[nexthandler]->HandleEvent(e);
										}

									// if its still a keyboard event, try the next handler
									if (e.what==ev_keyboard) nexthandler++;
									else
										// skip out as it's a new event
										break;
									}
								while (nexthandler<numEventHandlers);

								// clear event regardless as it wasn't used
								if (e.what==ev_keyboard) ClearEvent(e);
								break;

		// if its a non-event, skip out after clearing the event!
		default:				ClearEvent(e);
								break;
		}


	// if the event was handled, take it out of the queue!
//	if (e.what==ev_nothing)
		if (numEventsPending>0)
			{
			if (e.what==ev_nothing)
				{
				// update the range of events that are pending
				nextPending++;
				if (nextPending>=EVENTMGR_DEFAULT_QUEUE_SIZE) nextPending = 0;

				// event queue is no longer full
				eventQueueIsFull = false;

				// decrement our number of events pending
				numEventsPending--;

				// set the global pending flag
				eventIsPending = (numEventsPending!=0);
				}
			}
		else
			{
			DBG(470,"MAJOR ERROR:: AT END OF HANDLEEVENT,  NUMEVENTS PENDING == %d",numEventsPending);
//			delay(5000);
			}

//	HEAPCHK("by EventMgr::HandleEvent()");
	}

//	-------------------------------------------------------------------------
//
//  FUNC:		Execute()
//	DESC:		Performs the main loop of the event-handling mechanism.
//				It continuously loops performing the following algorithm:
//
//				DO
//					IF (event_is_pending) THEN handle_the_pending_event
//					ELSE perform_an_idle_function
//              UNTIL end_of_program
//
//				In the current version of the library, PERFORM_AN_IDLE_FUNCTION
//				checks for mouse input!
//
//	-------------------------------------------------------------------------
void EventMgr::Execute()
	{
	// if there are no event handlers, skip out!!!
	if (numEventHandlers==0) return;
	if (isTerminating) return;

	// initialize the event loop
	numEventsPending = 0;
	eventIsPending = false;
	eventQueueIsFull = false;

	nextPending = 0;
	lastPending = 0;

	isExecuting = true;

	if (currentHandler==-1) currentHandler = 0;

	// loop through the event mechanism till we get a "finished" signal
	while ((isExecuting) && (!isTerminating))
		{
		// if there is an event pending
		if (eventIsPending)
			{
//			DBG(470,"HANDLING EVENT TYPE %d, event == %d",
//				queue[nextPending].what,queue[nextPending].command.event);
//				delay(200);
			// handle the next pending event
			HandleEvent(queue[nextPending]);
			}
		else
			// poll the relevent input routines
			{
			keyboard_GetEvent();
			mouse_GetEvent();
			timer_GetEvent();
			}
//		mouse_UpdateCursor();
		}
	}


//	-------------------------------------------------------------------------
//
//	FUNC:		eventmgr_InstallEVENTMGR()
//	DESC:		creates our event manager & stores it as the global EVENTMGR
//
//	RETURNS:	true 	- if installed correctly
//				false	- error occurred during installation
//
//	-------------------------------------------------------------------------
bool eventmgr_InstallEVENTMGR()
	{
	// if we aren't already installed
	if (!eventmgr_INSTALLED)
		{
		// attempt to create our event manager
		EVENTMGR = new EventMgr();
		// if it was created okay
		if (EVENTMGR!=NULL)
			{
			// set flag
			eventmgr_INSTALLED = true;
			// and return true
			return true;
			}
		}

	// otherwise(s) return false
	return false;
	}


//	-------------------------------------------------------------------------
//
//	FUNC:		eventmgr_RemoveEVENTMGR()
//	DESC:		deletes the EVENTMGR, thus closing down all eventhandlers.
//
//	-------------------------------------------------------------------------
void eventmgr_RemoveEVENTMGR()
	{
	// if its actually installed
	if (eventmgr_INSTALLED)
		{
		// delete the event manager
		delete EVENTMGR; EVENTMGR = NULL;

		// and set our flag to 'not installed'
		eventmgr_INSTALLED = false;

//		HEAPCHK("by EventMgr::RemoveEVENTMGR()");
		}
	}
