//	-------------------------------------------------------------------------
//
//	EVENTMGR.HPP
//
//	-------------------------------------------------------------------------
//
//	AUTHOR:			Darren J. Dwyer
//
//	DESCRIPTION:	This file contains defines for the EVENT MANAGER.
//					The EVENTMGR maintains internally an EVENT QUEUE, which
//					can be used through the calls Gen???Event() & Mod??Event()
//
//	-------------------------------------------------------------------------

#ifndef __eventmgr_hpp
#define __eventmgr_hpp

#include "event.hpp"
#include "handler.hpp"

// at most the event queue contains 32 events
// at least contains NUMBER_TIMER_TICKERS

#define EVENTMGR_DEFAULT_QUEUE_SIZE 		32

// at most the event manager can contain 64 event handlers (more == spaghetti ???)

#define EVENTMGR_MAX_EVENT_HANDLERS		64


// THE EVENT MANAGER CLASS

class EventMgr
	{
	private:
		// INTERNAL storage of the event queue (for speed!!!)
		int		nextPending;						// the next pending event
		int 	lastPending;						// the last pending event
		Event	queue[EVENTMGR_DEFAULT_QUEUE_SIZE];	// the event queue

		// INTERNAL storage of the event handlers (for speed!!!)
		EventHandler* handler[EVENTMGR_MAX_EVENT_HANDLERS];	// event handlers

		// (internal) Generic HandleEvent which handles the next pending event
		void HandleEvent(Event& e);

	public:
		EventMgr();		// initialize the event manager
		~EventMgr();  	// terminate the event manager (as well as all handlers)

		int		currentHandler;		// the currently selected handler
		int		currentZFactor;		// the current z dimension

		int		numEventsPending;	// how many events are pending
		int		numEventHandlers;	// the number of event handlers

		bool 	isExecuting;		// true if executing. Set to false to stop
									// Execute()'ing
		bool	isTerminating;		// true if Terminating the program!!!!

		bool 	isActive;  			// true if EventMgr is active
		bool 	eventIsPending; 	// true if an event is pending in the queue
		bool 	eventQueueIsFull;  	// true if event queue is completely full
		bool 	eventHandlersAtMax;	// true if cannot have any more event handlers

		void 	SelectNextHandler();		// select the next handler (internal)
		void 	SelectPrevHandler();		// select the previous handler (internal)
		void	SelectEventHandler( int hand );	// select a specified handler
		void	DeSelectEventHandler( int hand );	// deselect a specified handler
		int		WhichHandlerSelected(grPoint& mouse_pos);	// which handler selected?

		// manually tells all onscreen handlers to draw
		void	DrawAllHandlers();

		// scans the event queue, returning true if the specified event is in the queue
		bool	EventInQueue(int what,void* e);

		// generate an event, placing it at the end of the event queue
		bool 	GenEvent(int where, int what, void* data);
//		bool	GenMouseEvent(int event,int modifiers,grPoint& where);
//		bool	GenKeyboardEvent(int event,int modifiers,grPoint& where);
//		bool	GenTimerEvent(int event,int time_delay);
		bool	GenCommandEvent(int where,int event,void* data);
		bool	GenBroadcastEvent(int event,void* data);

		// generate an event, placing it at the start of the event queue
		bool 	InsEvent(int where, int what, void* data);
//		bool	InsMouseEvent(int event,int modifiers,grPoint& where);
//		bool	InsKeyboardEvent(int event,int modifiers,grPoint& where);
//		bool	InsTimerEvent(int event,int time_delay);
		bool	InsCommandEvent(int where,int event,void* data);
		bool	InsBroadcastEvent(int event,void* data);

		// modify the current event, leaving the event at the current point
		// in the queue (chaining events)
		void 	ModEvent(Event& e,int where, int what, void* data);
//		void	ModMouseEvent(Event& e,int event,int modifiers,grPoint& where);
//		void	ModKeyboardEvent(Event& e,int event,int modifiers,grPoint& where);
//		void	ModTimerEvent(Event& e,int event,int time_delay);
		void 	ModCommandEvent(Event& e,int where,int event, void* data);
		void	ModBroadcastEvent(Event& e,int event,void* data);

		// clear a specific event
		void 	ClearEvent(Event& e);

		// clear all pending events
		void 	ClearAllEvents();

		// add an new event handler which (automatically) can then handle events
		EventHandler* AddEventHandler(EventHandler* handler);

		// remove the most recently Add'ed event handler
		void 	DelEventHandler();

		// remove ALL of the event handlers!
		void	DelAllEventHandlers();

		// the main loop of the event handling mechanism!!!!!
		void 	Execute();
	};

// globally declare the event manager!!!

extern EventMgr* EVENTMGR;

// a GLOBAL flag to tell whether or not the EVENTMGR is installed
extern	bool	eventmgr_INSTALLED;

// global functions to Install/Remove the EVENTMGR

bool eventmgr_InstallEVENTMGR();
void eventmgr_RemoveEVENTMGR();

// and make it easier to access

//#define	GenEvent			EVENTMGR->GenEvent
//#define ModEvent			EVENTMGR->ModEvent
//#define ClearEvent			EVENTMGR->ClearEvent
//#define AddEventHandler		EVENTMGR->AddEventHandler
//#define DelEventHandler		EVENTMGR->DelEventHandler


#endif