//	-------------------------------------------------------------------------
//
//	FONTS.CPP
//
//	-------------------------------------------------------------------------
//
//	AUTHOR:			Andrew C. Evans & Darren J. Dwyer
//
//	DESCRIPTION:	This file defines the font functions used
//					by the game development library.
//
//	-------------------------------------------------------------------------
#include	"fonts.hpp"
#include 	"resource.hpp"
#include 	"rects.hpp"

#include 	"graphics.hpp"
#include 	"resmgr.hpp"
//#include	"vesadrv.hpp"
#include	"grmisc.hpp"

#include 	"stdarg.h"

// --- bits for cached fonts

ResourceFile*	FontFile[MAX_CACHED_FONTS];
int				FontId[MAX_CACHED_FONTS];
int				FontsLoaded = 0;

int				CurrentFont	= -1;

//	-------------------------------------------------------------------------
//
//	FUNC:		grInitFont
//	DESC:		load a font into memory and return a tag value
//
//	PARAMS:		fontname = path and name of the font in DOS
//
// --------------------------------------------------------------------------
int		grInitFont(char* fontname)
	{
	return grInitFont("",fontname);
	}

int		grInitFont(char* path,char* fontname)
	{
//	grDrawFilledRect(0,0,9,9,grBLUE);
//	grBlitRegion(0,0,9,9);

	if (FontsLoaded>=MAX_CACHED_FONTS) return font_TooManyFonts;

//	grDrawFilledRect(10,0,19,9,grLIGHTBLUE);
//	grBlitRegion(10,0,19,9);

	FontFile[FontsLoaded] = new ResourceFile(path,fontname);

//	grDrawFilledRect(20,0,29,9,grYELLOW);
//	grBlitRegion(20,0,29,9);

	// check the cached bit's validity
	if (!FontFile[FontsLoaded]) return font_InitFontFailed;
	if (!FontFile[FontsLoaded]->Exists()) return font_NoSuchFont;

//	grDrawFilledRect(30,0,39,9,grCYAN);
//	grBlitRegion(30,0,39,9);

	int id = RESOURCE_MGR->RegisterFont(FontFile[FontsLoaded]);

	if (id>=0) FontId[FontsLoaded++] = id;

//	grDrawFilledRect(40,0,49,9,grGREEN);
//	grBlitRegion(40,0,49,9);

	return id;
	}

void	grExitFont()
	{
	if (FontsLoaded<0) return;

	FontsLoaded--;
	if (FontsLoaded==0) CurrentFont = -1;

	RESOURCE_MGR->DeRegisterFont(FontId[FontsLoaded]);
	delete FontFile[FontsLoaded];
	}

// --------------------------------------------------------------------------
//
//	FUNC:		grExitAllFonts
//	DESC:		delete all installed fonts from memory
//
// --------------------------------------------------------------------------
void 	grExitAllFonts()
	{
	// free all resources of all fonts until font_present is false
	while (FontsLoaded>0) grExitFont();
	CurrentFont = -1;
	}

// --------------------------------------------------------------------------
//
//	FUNC:		grGetFont
//	DESC:       returns the currently active font number
//
// --------------------------------------------------------------------------
int		grGetFont()
	{
	// return the current font number
	return CurrentFont;
	}

// --------------------------------------------------------------------------
//
//	FUNC:		grSetFont
//	DESC:       set the current font
//	PARAMS:		font_id 	- tag returned from grInitFont()
//
// --------------------------------------------------------------------------
int		grSetFont(int font_id)
	{
	if (FontsLoaded==0) return font_NoFontsLoaded;

	CLIP1(font_id,0,FontsLoaded-1);

	CurrentFont = font_id;

	return font_NoError;
	}

// --------------------------------------------------------------------------
//
//	FUNC:	grFontHeight
//	DESC:	returns the height of the first character in the current font
//
// --------------------------------------------------------------------------
int		grFontHeight()
	{
	// we don't have any fonts loaded then return
	if (FontsLoaded==0) return font_NoFontsLoaded;

	// we don't have a current font then return
	if (CurrentFont==-1) return font_NoFontSet;

	return grFontImagePtr(CurrentFont,0)->hdr.height;
	}

// --------------------------------------------------------------------------
//
//	FUNC:	grFontCharHeight
//	DESC:	returns the height of the ch in the current font
//
// --------------------------------------------------------------------------
int		grFontCharHeight(char ch)
	{
	// we don't have any fonts loaded then return
	if (FontsLoaded==0) return font_NoFontsLoaded;

	// we don't have a current font then return
	if (CurrentFont==-1) return font_NoFontSet;

	// if ch is out of bounds then return
	if ((ch<font_START_ASCII) || (ch>=font_START_ASCII+font_NUM_ASCII)) return font_NoSuchCharacter;

	return grFontImagePtr(CurrentFont,ch-font_START_ASCII)->hdr.height;
	}

// --------------------------------------------------------------------------
//
//	FUNC:	grFontHeight
//	DESC:	returns the height of the string in the current font
//
// --------------------------------------------------------------------------
int		grFontHeight(char* st)
	{
	int i, ch_height, max_height = 0;

	for (i=0;st[i]!='\0';i++)
		{
		// get width of each character
		ch_height = grFontCharHeight(st[i]);

		// if no error then add it to the height
		if (ch_height > max_height)
			max_height = ch_height;
		}

	return max_height;
	}

// --------------------------------------------------------------------------
//
//	FUNC:	grFontCharWidth
//	DESC:	returns the width of the ch in the current font
//
// --------------------------------------------------------------------------
int		grFontCharWidth(char ch)
	{
	// we don't have any fonts loaded then return
	if (FontsLoaded==0) return font_NoFontsLoaded;

	// we don't have a current font then return
	if (CurrentFont==-1) return font_NoFontSet;

	// if ch is out of bounds then return
	if ((ch<font_START_ASCII) || (ch>=font_START_ASCII+font_NUM_ASCII)) return font_NoSuchCharacter;

	return grFontImagePtr(CurrentFont,ch-font_START_ASCII)->hdr.width;
	}

// --------------------------------------------------------------------------
//
//	FUNC:	grFontWidth
//	DESC:	returns the width of the st string in the current font
//
// --------------------------------------------------------------------------
int		grFontWidth(char* st)
	{
	int i, ch_width, total_width = 0;

	for (i=0;st[i]!='\0';i++)
		{
		// get width of each character
		ch_width = grFontCharWidth(st[i]);

		// if no error then add it to the width
		if (ch_width > 0)
			total_width += ch_width;
		}

	return total_width;
	}

// draw a string in a virtual buffer at the specified coordinates using variable args
void grDrawString(int x,int y,char* fmt,...)
	{
	va_list	argptr;			// pointer to argument list
	char	buf[255];		// buffer to store transposed string

	va_start(argptr,fmt);	// figure out the argument starting address

	vsprintf(buf,fmt,argptr);	// shove it all into a buffer

	grDrawStringVirtual(grSCREEN,x,y,buf);

	va_end(argptr);				// to perform a normal return from this func
	}

// draw a 2-color string in a virtual buffer in specified color
void grDraw2CString(int x,int y,int color,char* fmt,...)
	{
	va_list	argptr;			// pointer to argument list
	char	buf[255];		// buffer to store transposed string

	va_start(argptr,fmt);	// figure out the argument starting address

	vsprintf(buf,fmt,argptr);	// shove it all into a buffer

	grDraw2CStringVirtual(grSCREEN,x,y,color,buf);

	va_end(argptr);				// to perform a normal return from this func
	}

void grBlitString(int x,int y,int bgcolor,char* fmt,...)
	{
	va_list	argptr;			// pointer to argument list
	char	buf[255];		// buffer to store transposed string

	va_start(argptr,fmt);	// figure out the argument starting address

	vsprintf(buf,fmt,argptr);	// shove it all into a buffer

	int x2 = x + grFontWidth(buf) -1;
	int y2 = y + grFontHeight(buf) -1;

	grDrawFilledRect(x,y,x2,y2,bgcolor);
	grDrawString(x,y,buf);
	grBlitRegion(x,y,x2,y2);

	va_end(argptr);				// to perform a normal return from this func
	}

void grBlitString(int x,int y,char* fmt,...)
	{
	va_list	argptr;			// pointer to argument list
	char	buf[255];		// buffer to store transposed string

	va_start(argptr,fmt);	// figure out the argument starting address

	vsprintf(buf,fmt,argptr);	// shove it all into a buffer

	grDrawStringVirtual(grSCREEN,x,y,buf);

	grBlitRegion(x,y,x+grFontWidth(buf)-1,y+grFontHeight()-1);

	va_end(argptr);				// to perform a normal return from this func
	}

void grBlit2CString(int x,int y,int bgcolor,int txtcolor,char* fmt,...)
	{
	va_list	argptr;			// pointer to argument list
	char	buf[255];		// buffer to store transposed string

	va_start(argptr,fmt);	// figure out the argument starting address

	vsprintf(buf,fmt,argptr);	// shove it all into a buffer

	int x2 = x + grFontWidth(buf) -1;
	int y2 = y + grFontHeight(buf) -1;

	grDrawFilledRect(x,y,x2,y2,bgcolor);
	grDraw2CString(x,y,txtcolor,buf);

	grBlitRegion(x,y,x2,y2);

	va_end(argptr);				// to perform a normal return from this func
	}

// --------------------------------------------------------------------------
//
//	FUNC:		grDrawStringVirtual()
// 	DESC:		draws a string in a virtual buffer at location x,y
//
// --------------------------------------------------------------------------
void grDrawStringVirtual(grVirtualBuffer* vb,int x,int y,char* fmt,...)
	{
	if (vb==NULL) return;
	if (CurrentFont==-1) return;

	// make sure coords are clipped!!!
	CLIP2(x,y,0,0,vb->hdr.width-1,vb->hdr.height-1);

	va_list	argptr;			// pointer to argument list
	char	buf[255];		// buffer to store transposed string

	va_start(argptr,fmt);	// figure out the argument starting address
	vsprintf(buf,fmt,argptr);	// shove it all into a buffer

	// display text onto the screen

	int		tmpwidth;		// temporary width
	int 	last = font_START_ASCII + font_NUM_ASCII - 1;
	char 	ch;
	grImage* img;

	// display using cached font
	for (int i=0;buf[i]!='\0';i++)
		{
		ch = buf[i];
		if ((ch<font_START_ASCII) || (ch>last)) continue;

		// *NOTE:: the font is stored in the same order as the ascii set
		// but the first character is the space key 32, the list is contiguous
		// after this for 95 characters. See top of file for full list

//		grDrawFilledRect(grMaxx-20,grMaxy-20,grMaxx,grMaxy,ch-font_START_ASCII);
		img = grFontImagePtr(CurrentFont,ch-font_START_ASCII);

		// set our temp width
		tmpwidth = img->hdr.width;

		// put image into the virtual buffer
		grOverlayImageVirtual(vb,x,y,x+tmpwidth-1,y+img->hdr.height-1,grBLACK,img);

		// increment currentx by last char's image width
		x += tmpwidth;
		}

	va_end(argptr);				// to perform a normal return from this func
	}

// --------------------------------------------------------------------------
//
//	FUNC:		grDraw2CStringVirtual()
// 	DESC:		draws a string in a virtual buffer at location x,y
//				in the specified color
//
// --------------------------------------------------------------------------
void grDraw2CStringVirtual(grVirtualBuffer* vb,int x,int y,int color,char* fmt,...)
	{
	if (vb==NULL) return;
	if (CurrentFont==-1) return;

	// make sure coords are clipped!!!
	CLIP2(x,y,0,0,vb->hdr.width-1,vb->hdr.height-1);

	va_list	argptr;			// pointer to argument list
	char	buf[255];		// buffer to store transposed string

	va_start(argptr,fmt);	// figure out the argument starting address
	vsprintf(buf,fmt,argptr);	// shove it all into a buffer

	// display text onto the screen

	int		tmpwidth;		// temporary width
	int 	len = strlen(buf);
	int 	last = font_START_ASCII + font_NUM_ASCII - 1;
	char 	ch;
	grImage* img;

	// display using cached font
	for (int i=0;i<len;i++)
		{
		ch = buf[i];
		if ((ch<font_START_ASCII) || (ch>last)) continue;

		// *NOTE:: the font is stored in the same order as the ascii set
		// but the first character is the space key 32, the list is contiguous
		// after this for 95 characters. See top of file for full list

		img = grFontImagePtr(CurrentFont,ch-font_START_ASCII);

		// set our temp width
		tmpwidth = img->hdr.width;

		// put image into the virtual buffer
		grOverlay2CImageVirtual(vb,x,y,x+tmpwidth-1,y+img->hdr.height-1,color,img);

		// increment currentx by last char's image width
		x += tmpwidth;
		}

	va_end(argptr);				// to perform a normal return from this func
	}

