// -------------------------------------------------------------------------
//
//		FRACTAL.CPP
//
//		Andrew Evans - September 95.
//
// -------------------------------------------------------------------------
#include	"fractal.hpp"
#include	"graphics.hpp"

#include	"stdlib.h"
#include	"math.h"
#include	"conio.h"
#include	"time.h"
#include	"dos.h"

//#define fractal_DRAW

//--------------------------------------------------------------------------
// GLOBAL variables from outside this file

// the_map memory and location are allocated outside the fractal_ functions
// and passed through the fractal_Init() function
grVirtualBuffer*	the_map = NULL;

// position of the map on the screen in rectangle format
grRect				the_map_pos;

// type of land being placed on map
uchar				land_type;

// -------------------------------------------------------------------------
//
//	FUNC:	fractal_Init
//	DESC:	point the_map at any virtual buffer, this is where the data goes.
//			locate the_map at any rectangular position, this is where the map
//			is displayed for debugging or show.
//          randomize the random number generator with a seed or a time value
//
// -------------------------------------------------------------------------
void fractal_Init(grVirtualBuffer* map, grRect position, int seed)
	{
	// randomize
	if (seed!=0) srand(seed);
	else
		{
		time_t t;
		srand(time(&t));
		}

	the_map = map;
	the_map_pos = position;
	}

// -------------------------------------------------------------------------
//
//	FUNC:	fractal_Exit
//	DESC:	points the map at nothing
//
// -------------------------------------------------------------------------
void fractal_Exit()
	{
	the_map = NULL;
	}

//--------------------------------------------------------------------------
// GLOBAL variables used by most functions within this file

// generally the rate of change should be a fraction so that as you get
// deeper into recursion there are more tight/small/close values
float	rate_of_change;

// elasticity or eccentricity or something like that
float	bendiness;

// measurements
float	length_squared,
		minimum_length_squared;

// -------------------------------------------------------------------------
//
//	FUNC:	fractal_Line
//	DESC:	will bend a line x1,y1 x2,y2 in a fractal manner
//
//	PARAMS:	pointer to where the map is located
//			x & y values for the line
//			seed for random number generator
//
// -------------------------------------------------------------------------
void fractal_Line(int x1,int y1, int x2,int y2, uchar land)
	{
	// set the land_type
	land_type = land;

	// how far the midsection point may move initially
	float	standard_deviation = 0.4;

	// bendiness ranges from -1.0 to 0.5 approx
	float	bendiness = -0.4;

	// rate of change for how far the midsection point may move
	rate_of_change = exp ( bendiness * log(2.0) );

	// smallest length allowed to be worked with
	minimum_length_squared = 36;

	// start the iterations
	fractal_IterateLine (x1,y1, x2,y2, standard_deviation);
	}

// -------------------------------------------------------------------------
//
//	FUNC:	fractal_IterateLine
//	DESC:	recursively will move randomly the midpoint of a line until
//			the length of the line we are working on is below the minimum
//			length
//
//	PARAMS:	x & y values for the line
//			how far the line is allowed to move
//
// -------------------------------------------------------------------------
void fractal_IterateLine (int x1,int y1, int x2,int y2, float standard_deviation)
	{
	// get the length of the current line
	length_squared = fabs((x1-x2)*(x1-x2)) + fabs((y1-y2)*(y1-y2));

	// if the line length is shorter than the minimum length required
	if (length_squared < minimum_length_squared)
		{
		// then draw it onto the map
		grDrawLineVirtual(the_map,x1,y1,x2,y2,land_type);

#ifdef fractal_DRAW
		grDrawLine(the_map_pos.left+x1,the_map_pos.top+y1,the_map_pos.left+x2,the_map_pos.top+y2,land_type);
#endif
		}
	else
		{
		// calculate how far the midsection point is ALLOWED to move
		standard_deviation *= rate_of_change;
		// calculate how far the midsection point WILL move
		float movement = (50 - (float)random(100)) / 50 * standard_deviation;

		// calculate new x value for the next midsection point
		int newx = 0.5*(x1+x2) - movement*(y2-y1);
		// if its out of range move it the other way
		if ((newx<0) || (newx>the_map->hdr.width-1))
			newx = 0.5*(x1+x2) + movement*(y2-y1);

		// calculate new y value for the next midsection point
		int newy = 0.5*(y1+y2) + movement*(x2-x1);
		// if its out of range move it the other way
		if ((newy<0) || (newy>the_map->hdr.height-1))
			newy = 0.5*(y1+y2) - movement*(x2-x1);

		// get what's at the chosen control point
		uchar spot = grGetPixelVirtual(the_map,newx,newy);

		if (spot == 0)
			{
			newx = 0.5*(x1+x2) + movement*(y2-y1);
			newy = 0.5*(y1+y2) - movement*(x2-x1);
			}

		// iterate the line on each side of the midsection
		fractal_IterateLine (x1,y1, newx,newy, standard_deviation);
		fractal_IterateLine (newx,newy, x2,y2, standard_deviation);
		}
	}

/**************************************************************************/

//--------------------------------------------------------------------------
// AREA variables

// percent of area covered
int		density;

// -------------------------------------------------------------------------
//
//	FUNC:	fractal_Area
//	DESC:
//
//	PARAMS:
//
// -------------------------------------------------------------------------
void fractal_Area(int x,int y, int radius, uchar land)
	{
	// set the land_type
	land_type = land;

	// percent of area covered
	density = 90;

	// bendiness ranges from -1.0 to 0.0 approx
	bendiness = -0.4;

	// the rate of change for each successive radius
	rate_of_change = exp(bendiness * log(2.0));

	// smallest area's radius allowed to be worked with
	minimum_length_squared = 3;

	// start the iterations
	fractal_IterateArea(x,y, radius);
	}

// -------------------------------------------------------------------------
//
//	FUNC:	fractal_IterateArea
//	DESC:	recursively will create control points for an area until the
//			radius of the area we are working on is below the minimum length
//
//	PARAMS:	x & y values for the area
//			how far the radius is allowed to reach
//
// -------------------------------------------------------------------------
void fractal_IterateArea(int x,int y, int radius)
	{
	int newx,newy, new_radius;
	int	radius_squared = radius * radius;

	// calculate new radius for next recursion
	new_radius = radius * rate_of_change;

	// if we are working on a large enough area
	if (radius_squared > minimum_length_squared)
		{
		// create a number of control points
		for (int i=0;i<density/10;i++)
			{
			bool done = false;
			int count = 0;
			int max = 50;

			while (!done)
				{
				count++;
				// only try for a point max times
				if (count>=max) done = true;

				// create a control point
				newx = random(2*radius+1) + x - radius;
				newy = random(2*radius+1) + y - radius;

				// calculate distance the point newx,newy from centre x,y
				length_squared = abs((newx-x)*(newx-x)) + abs((newy-y)*(newy-y));

				// and while the point isn't within the radius try again
				if (length_squared > radius_squared) continue;

				// get what's at the chosen control point
				uchar spot = grGetPixelVirtual(the_map,newx,newy);

				if (spot == 0) continue;			// sea level has no land
				if (spot == land_type) continue;	// stop wasting time

				done = true;
				}

			if (count<max)
				{
				// make sure the point is within the_map's dimensions
				CLIP2(newx,newy, 0,0, the_map->hdr.width,the_map->hdr.height);

				// put the point on the_map
				grPutPixelVirtual(the_map, newx,newy, land_type);

#ifdef fractal_DRAW
				grPutPixel(the_map_pos.left+newx,the_map_pos.top+newy, 0);
				grPutPixel(the_map_pos.left+newx,the_map_pos.top+newy, land_type);
#endif

				// iterate the area around the new control point
				fractal_IterateArea(newx,newy,new_radius);
				}
			}
		}
	else
		{
		}
	}
