//	-------------------------------------------------------------------------
//
//	HANDLER.HPP
//
//	-------------------------------------------------------------------------
//
//	AUTHOR:			Darren J. Dwyer
//
//	DESCRIPTION:	This file contains defines for the Event Handler object
//					used by the EVENT MANAGER.  Note that some routines
//					directly call event manager functions!
//
//	-------------------------------------------------------------------------

#ifndef __handler_hpp
#define __handler_hpp

#include "event.hpp"		// event structures

// event handler types 			  - 0x0000000X
#define evh_BACKGROUND_TASK			0x00000001
#define evh_ONSCREEN_HANDLER		0x00000002

// basic handler options 		  - 0x000000X0
#define evh_NO_KBD_RESPONSE 		0x00000010
#define evh_NO_MOUSE_RESPONSE		0x00000020


// the maximum number of accessories in any specific event handler

#define MAX_ACCESSORIES_IN_HANDLER		64

//	-------------------------------------------------------------------------
//
//	CLASS:	EventHandler
//
//	DESC:   This is the base object which can respond to events. There are
//			three objects derived from this base class:
//
//				onscreen			- handler is displayed on the screen
//									  and (by default) responds to all event types.
//				background_task		- handler is "offscreen" and (by default)
//									  does not interpret mouse & keyboard input.
//				dialog_box			- this type of handler is similar to a
//									  background task that waits for an
//									  "opendialogbox" command and then
//									  becomes an onscreen task with exclusive
//									  access to mouse & kbd. Upon receiving
//									  a "closedialogbox" command it returns
//									  to a background task.
//
//			Onscreen & dialog eventhandlers may contain "accessories" which
//			can be any type of input device responding to mouse & kbd. These
//			can range from a simple TextInputButton through to a TextScrollBox
//			through to a spaceship! :-)
//
//	-------------------------------------------------------------------------

class EventHandler
	{
	protected:
	public:
		int		selected_options;	// the options selected for the EventHandler
		int 	zfactor;			// the z dimension (x,y,z)
		int		handler_number;		// the handler number
		int*	ext_handler_number;	// the external handler number for global use
									// (evhno_???)

		bool	isOnScreenHandler;	// true-> handler is OnScreenHandler derived
		bool	isBackgroundTask;	// true-> handler is BackgroundTask derived

		bool	isRespondingMouse;	// true-> responds to mouse
									// false-> does not respond to mouse

		bool	isRespondingKbd;    // true-> responds to keyboard
									// false-> does not respond to keyboard



		EventHandler(int* evhno,int options,int zfact);
		virtual ~EventHandler() {}

		void	SetHandlerNumber(int handler_num);
//		void	SetZFactor(int zfact) { zfactor = zfact; }
//		int		GetZFactor()		  { return zfactor; }

		// note: the following methods are called by the event manager
		//		 automatically before calling the generic HandleEvent
		//		 (for backwards compatibility)

/*		virtual void	OnLeftMouseDn(Event &e);
		virtual	void	OnLeftMouseUp(Event &e);
		virtual	void	OnLeftMouseAuto(Event &e);
		virtual	void	OnRightMouseDn(Event &e);
		virtual	void	OnRightMouseUp(Event &e);
		virtual	void	OnRightMouseAuto(Event &e);
		virtual	void	OnMiddleMouseDn(Event &e);
		virtual	void	OnMiddleMouseUp(Event &e);
		virtual	void	OnMiddleMouseAuto(Event &e);
		virtual	void	OnMouseMove(Event &e);

		virtual	void	OnKeyDn(Event &e);
		virtual	void	OnKeyUp(Event &e);
		virtual	void	OnKeyAuto(Event &e);

		virtual	void	OnTimer(Event &e);
		virtual	void	OnCommand(Event &e);
		virtual	void	OnBroadcast(Event &e);

*/		virtual void 	HandleEvent(Event& e);		// handles the current event
	};


#endif

