//	-------------------------------------------------------------------------
//
//	KEYBOARD.CPP
//
//	-------------------------------------------------------------------------
//
//	AUTHOR:			Darren J. Dwyer
//
//	DESCRIPTION:	This file contains code for a keyboard handler (ISR)
//					which generates events into the EVENTQUEUE
//
//	-------------------------------------------------------------------------

#include "keyboard.hpp"
#include "mouse.hpp"
#include "event.hpp"
#include "stddefns.hpp"
#include "timer.hpp"
#include "eventmgr.hpp"

#include "stdlib.h"
#include "conio.h"
#include "stdio.h"
#include "mem.h"
#include "dos.h"

// enable debugging of program

//#define __debugging

#ifdef __debugging
bool 	Quit = false;		// used in debugging
#endif

// the size of the keyboard buffer

#define KEYBOARD_TABLE_SIZE 	132

// some interrupt & port numbers

#define	KEYBOARD_INT			0x09	// keyboard interrupt no.
#define	KEY_BUFFER				0x60	// location of keyboard buffer
#define	KEY_CONTROL				0x61	// location of keyboard controller

// storage of previous keyboard ISR (Interrupt Service Routine)

#ifdef __cplusplus
#ifdef __WATCOMC__
#define	CPPARGS
#else
#define CPPARGS		...
#endif
#else
#define	CPPARGS
#endif

void _interrupt (_far *Old_Keyboard_Isr)(CPPARGS);

// a KeyboardEvent variable used when generating new events

KeyboardEvent	keyboard_EventInfo;

// other variables used by the keyboard functions

uchar	raw_key;							// the raw keypress
int		keys_in_buffer = 0;					// number of keys in keyboard buffer
char	key_table[KEYBOARD_TABLE_SIZE];	// storage of current state of up to 132 keys!!
bool	prefixed;							// keypress is prefixed
int		keyboard_IntCnt = 0;				// used to count interrupts

// the status of a key_table index

#define key_neither		0		// special use for non-auto-repeating
#define key_released	1		// key is released
#define key_pressed		2		// key is pressed

// a flag used to test whether or not the handler is installed!

bool	keyboard_HANDLER_INSTALLED = false;

// the tiemr ticker value used to generate machine-independant keyboard response

int		keyboard_ticker = 0;

// and the keyboard auto repeat rate

int		keyboard_auto_repeat_rate = 10;	// 0 = fastest, n = slowest

// the scankey-to-ASCII tables

const uchar	ascii_table[] =
	{ 0,27,'1','2','3','4','5','6','7','8','9','0','-','=',8,9,
	  'q','w','e','r','t','y','u','i','o','p','[',']',13,0,
	  'a','s','d','f','g','h','j','k','l',';',39,'`',0,92,
	  'z','x','c','v','b','n','m',',','.','/' };

const uchar	shift_ascii_table[] =
	{ 0,27,'!','@','#','$','%','^','&','*','(',')','_','+',0,0,
	  'Q','W','E','R','T','Y','U','I','O','P','{','}',13,0,
	  'A','S','D','F','G','H','J','K','L',':','"','~',0,'|',
	  'Z','X','C','V','B','N','M','<','>','?' };

// for quick-quit mode

#ifndef __debugging
extern void grResetMode(int mode);
#endif

//	-------------------------------------------------------------------------
//
//	FUNC:		keyboard_Pause()
//	DESC:		waits for a keypress and then returns
//
//	-------------------------------------------------------------------------
void keyboard_Pause()
	{
	while (!keyboard_IsHit()) ;
	keyboard_Flush();
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		keyboard_IsReleased()
//	DESC:		returns true if the specified keypress is a key_released event
//
//	-------------------------------------------------------------------------
bool keyboard_IsReleased( int keyvalue )
	{
	return (keyvalue>=1000);
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		keyboard_IsPressed()
//	DESC:		returns true if the specified keypress is a key_pressed event
//
//	-------------------------------------------------------------------------
bool keyboard_IsPressed( int keyvalue )
	{
	return (keyvalue<1000);
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		keyboard_ASCII()
//	DESC:		Returns the ascii value of a keyboard return value. Also
//				takes into account the status of the SHIFT keys.
//
//				eg. keyboard_ASCII( RET_DN, keyboard_SHIFT ) == 13
//					keyboard_ASCII( A_UP, keyboard_SHIFT ) == 'A'
//
//	NOTE:		I cannot see why we should bother with the ascii CTRL chars
//				as they have no particular use in this engine. DJD
//
//	-------------------------------------------------------------------------
uchar keyboard_ASCII( int keyvalue, int modifiers )
	{
	int keyval = keyvalue;

	// convert from a "_UP" value to a "_DN" value
	if (keyval>1000) keyval -= 1000;

	// handle special case for space bar
	if (keyval==SPACE_DN) return 32;

	// skip out if we shouldn't check the keyboard value
	if (keyval>53) return 0;

	// quickly compare and return any shift-modified key values
	if (modifiers & keyboard_SHIFT)
		keyval = shift_ascii_table[keyval];
	else
		// return any normal key values
		keyval = ascii_table[keyval];

	// if we have found an incorrect translation, set to original input!
	if (keyval==0) keyval = keyvalue;

	// return the converted value
	return keyval;
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		keyboard_GetState()
//	DESC:		returns the current state of the shift/alt/ctrl keys
//
//	-------------------------------------------------------------------------
int keyboard_GetState()
	{
	// assume no modifiers
	int modifiers = 0;

	// scan the appropriate key_table indexes and reflect whats down
	if (key_table[LEFT_SHIFT_DN]==key_pressed) 	   modifiers |= keyboard_LEFT_SHIFT;
	if (key_table[RIGHT_SHIFT_DN]==key_pressed)    modifiers |= keyboard_RIGHT_SHIFT;
	if (key_table[LEFT_ALT_DN]==key_pressed)       modifiers |= keyboard_LEFT_ALT;
	if (key_table[RIGHT_ALT_DN]==key_pressed)      modifiers |= keyboard_RIGHT_ALT;
	if (key_table[LEFT_CTRL_DN]==key_pressed)      modifiers |= keyboard_LEFT_CTRL;
	if (key_table[RIGHT_CTRL_DN]==key_pressed)     modifiers |= keyboard_RIGHT_CTRL;

	// and return our modifiers flag
	return modifiers;
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		keyboard_SetAutoRepeat()
//	DESC:		sets the delay for the auto repeat on key_dn
//
//	-------------------------------------------------------------------------
void keyboard_SetAutoRepeat( int rate )
	{
	keyboard_auto_repeat_rate = rate;
	if (keyboard_auto_repeat_rate<0) keyboard_auto_repeat_rate = 0;

#ifndef __debugging
//	timer_SetTicker( keyboard_ticker, rate );
#endif
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		keyboard_GetAutoRepeat()
//	DESC:		returns the delay for the auto repeat on key_dn
//
//	-------------------------------------------------------------------------
int	keyboard_GetAutoRepeat()
	{
	return keyboard_auto_repeat_rate;
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		keyboard_GetEvent()
//	DESC:		checks to see if any keys have been pressed & generates
//				an appropriate keyboard event into the EVENT QUEUE
//
//	-------------------------------------------------------------------------
void keyboard_GetEvent()
	{
/*
	// skip out if there are no keystrokes in the buffer
	if (keys_in_buffer==0) return;

	// get the keyboard modifiers
	keyboard_EventInfo.modifiers = keyboard_GetState();

	// get where the mouse is
#ifndef __debugging
	keyboard_EventInfo.where = mouse_GetPos();
#endif

	// scan through our keystroke buffer & generate events as required

	for (int index=0;index<keys_in_buffer;index++)
		{
		// get the appropriate (pre-processed) event
		keyboard_EventInfo.event = key_buffer[index];

#ifdef __debugging
		// display keystroke information
		if (keyboard_EventInfo.event!=-1)
			{
			char buf[200];
			char* bufptr = (char*) &buf;
			strcpy(bufptr,"MODIFIERS:  ");
			if (key_table[LEFT_SHIFT_DN]==key_pressed) 	strcat(bufptr,"LSHIFT ");
			else										strcat(bufptr,"       ");
			if (key_table[LEFT_ALT_DN]==key_pressed) 	strcat(bufptr,"LALT ");
			else										strcat(bufptr,"     ");
			if (key_table[LEFT_CTRL_DN]==key_pressed) 	strcat(bufptr,"LCTRL   ");
			else										strcat(bufptr,"        ");
			if (key_table[RIGHT_SHIFT_DN]==key_pressed) strcat(bufptr,"RSHIFT ");
			else										strcat(bufptr,"       ");
			if (key_table[RIGHT_ALT_DN]==key_pressed) 	strcat(bufptr,"RALT ");
			else										strcat(bufptr,"     ");
			if (key_table[RIGHT_CTRL_DN]==key_pressed) 	strcat(bufptr,"RCTRL ");
			else										strcat(bufptr,"      ");

			if (keyboard_EventInfo.event>1000)
				printf("KEY RELEASED = %d  %s\n",keyboard_EventInfo.event%1000,buf);
			else
				printf("KEY PRESSED  = %d  %s\n",keyboard_EventInfo.event%1000,buf);
			keyboard_EventInfo.event = -1;
			}
#else
			// if we don't already have this event in the queue
//			if (!EVENTMGR->EventInQueue(ev_keyboard,&keyboard_EventInfo))
				// place it onto the queue
				EVENTMGR->GenEvent(0,ev_keyboard,&keyboard_EventInfo);
#endif
		}
*/

	// we've extracted all keystrokes, so reset
//	keys_in_buffer = 0;
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		keyboard_IsHit()
//	DESC:		returns true if the keyboard is waiting to be processed
//
//	-------------------------------------------------------------------------
bool keyboard_IsHit()
	{
	return (keys_in_buffer>0);
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		keyboard_Flush()
//	DESC:		flushes the keyboard input buffer
//
//	-------------------------------------------------------------------------
void keyboard_Flush()
	{
	keys_in_buffer = 0;
	for (int i=0;i<KEYBOARD_TABLE_SIZE;i++)
		key_table[i] = key_neither;
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		keyboard_ISR()
//	DESC:		this is the keyboard ISR function which interprets most of
//				the keyboard and generates appropriate responses.
//
//	NOTES:		Upon pressing or releasing any particular key, the keyboard
//				seems to generate either ONE interrupt, TWO interrupts, or
//				FOUR interrupts.
//
//				For ONE interrupt the keyboard generates a pressed (hi bit clear)
//				or released (hi bit set) keycode. (Refer TWO interrupts)
//
//				For TWO interrupts, a prefix of 96 is sent, followed by
//				the actual keycode
//
//				For FOUR interrupts, a prefix of 96 42 96 is sent, followed
//				by the actual keycode
//
//				SCAN CODES:
//				----------
//
//				The following keys are treated as normal keys
//
//				LEFT_SHIFT		42
//				RIGHT_SHIFT		54
//				LEFT_CTRL		29		// refer right ctrl
//				LEFT_ALT		56		// refer right alt
//
//				Other keys generate a prefix to their scan code:
//
//				RIGHT_CTRL		96 29
//				RIGHT_ALT		96 56
//				KEYPAD_ENTER	96 28
//				KEYPAD_SLASH	96 53
//
//				Yet more wierdo keys generate three-byte prefixes as follows:
//
//				INS				96 42 96 82
//				DEL             96 42 96 83
//				HOME            96 42 96 71
//				END             96 42 96 79
//				PGUP            96 42 96 73
//				PGDN            96 42 96 81
//				UP              96 42 96 72
//				DN              96 42 96 80
//				LT              96 42 96 95
//				RT              96 42 96 77
//
//	-------------------------------------------------------------------------
void _interrupt _far keyboard_ISR(CPPARGS)
	{
	// disable interrupts while in this ISR
	_disable();

	// talk to the keyboard & get the keypress
	raw_key = inp(KEY_BUFFER);			// get the raw keypress
	char tmp = inp(KEY_CONTROL) | 0x82;	// read the control register
	tmp |= 0x82;						// set bits to reset flip-flop
	outp(KEY_CONTROL,tmp);				// reset the flip-flop
	tmp &= 0x7F;
	outp(KEY_CONTROL,tmp);

	// --- start processing here

	int		scan_code = -1;				// set our scan code for later check
	int 	key = raw_key & 0x7F;		// get actual key
	bool	key_is_pressed = (raw_key<128);	// and whether or not it is pressed

	switch (keyboard_IntCnt)
		{
		case 0:		switch (key)		// first byte
						{

						// check for prefixed keys

						case 96:        prefixed = true;
										keyboard_IntCnt++;		// check the next byte
										break;

						// check for special case keys

						case 42:		scan_code = -LEFT_SHIFT_DN;	break;
						case 29:		scan_code = -LEFT_CTRL_DN;	break;
						case 56:		scan_code = -LEFT_ALT_DN;	break;
						case 54:		scan_code = -RIGHT_SHIFT_DN;break;

						// and now normal keys

						default:		scan_code = key;      		break;
						}
					break;

		case 1:		switch (key)
						{
						// check for redundant prefix (96 42 96 ??)

						case 42:		if (prefixed)
											{
											prefixed = false;
											keyboard_IntCnt = 0;	// ignore prefix
											}
										break;

						// check for special keys

						case 29:		scan_code = -RIGHT_CTRL_DN;	break;
						case 56:		scan_code = -RIGHT_ALT_DN;	break;
						case 28:		scan_code = ENTER_DN;		break;
						case 53:		scan_code = DIVIDE_DN;		break;


						// and now for normal keypresses

						default:		prefixed = false;
										scan_code = key;
										break;
						}

					keyboard_IntCnt = 0;
					break;
		}

	if (scan_code != -1)
		{
		bool okay;
		if (scan_code<0)
			{
			okay = false;
			scan_code = -scan_code;
			}
		else
			okay = true;

		if (key_is_pressed)
			{
			// update key_table to reflect current keyboard state
			if (key_table[scan_code] != key_pressed) keys_in_buffer++;
			key_table[scan_code] = key_pressed;

			// if key is not negative & eventmgr okay, generate event
			if ((okay) && (eventmgr_INSTALLED))
				{
//				keyboard_EventInfo.where = mouse_GetPos();
				keyboard_EventInfo.modifiers = keyboard_GetState();
				keyboard_EventInfo.event = scan_code;

//				if (!EVENTMGR->EventInQueue(ev_keyboard,&keyboard_EventInfo))
					EVENTMGR->GenEvent(0,ev_keyboard,&keyboard_EventInfo);
				}
			}
		else
			{
			if (key_table[scan_code] == key_pressed) keys_in_buffer--;
			key_table[scan_code] = key_released;

			if ((okay) && (eventmgr_INSTALLED))
				{
//				keyboard_EventInfo.where = mouse_GetPos();
				keyboard_EventInfo.modifiers = keyboard_GetState();
				keyboard_EventInfo.event = scan_code + 1000;
//				if (!EVENTMGR->EventInQueue(ev_keyboard,&keyboard_EventInfo))
					EVENTMGR->GenEvent(0,ev_keyboard,&keyboard_EventInfo);
				}
			}
		}

	// check for special case ABNORMAL PROGRAM TERMINATION - RT_CTRL_ALT_DEL
	// generally used to quit an infinite loop (or to get out quick!!)
	if ((key_table[DEL_DN]==key_pressed)&&
		(key_table[RIGHT_CTRL_DN]==key_pressed)&&
		(key_table[RIGHT_ALT_DN]==key_pressed))
		{
//		mouse_RemoveHandler();
//		timer_RemoveHandler();
		outp(0x20,0x20);
		_enable();
		keyboard_RemoveHandler();
//		grResetMode(3);
		exit(-1);	// -1 means abnormal program termination
		}

	// --- end processing here

	outp(0x20,0x20);	// signal end-of-interrupt ?? needed??
	// re-enable interrupts
	_enable();
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		keyboard_InstallHandler()
//	DESC:		installs our custom keyboard handler.
//
//	-------------------------------------------------------------------------

void keyboard_InstallHandler()
	{
	if (!keyboard_HANDLER_INSTALLED)
		{
		// flush the keyboard buffer caJUSTse
		while (kbhit()) getch();

		// clear keypress table (assumes all keys released)
		memset(&key_table,key_neither,132);

		// save old keyboard ISR
		Old_Keyboard_Isr = _dos_getvect(KEYBOARD_INT);

		// and replace it with ours!
		_dos_setvect(KEYBOARD_INT,keyboard_ISR);

		// set a flag for check later
		keyboard_HANDLER_INSTALLED = true;

		// set up the keyboard timer
		// used for generating events independently of system speed

#ifndef __debugging
//		keyboard_ticker = timer_AddTicker(keyboard_auto_repeat_rate,cmdPollKbd);
#else
		keyboard_EventInfo.event = -1;
#endif
		}
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		keyboard_RemoveHandler()
//	DESC:		Removes our custom keyboard handler & sets the handler back
//				to the original (BIOS/DOS/???) keyboard handler.
//
//	-------------------------------------------------------------------------
void keyboard_RemoveHandler()
	{
	if (keyboard_HANDLER_INSTALLED)
		{
		// remove our keyboard polling
#ifndef __debugging
//		timer_DelTicker(keyboard_ticker);
#endif

		// restore old keyboard ISR
		_dos_setvect(KEYBOARD_INT,Old_Keyboard_Isr);

		// and clear our handler_installed flag!!!
		keyboard_HANDLER_INSTALLED = false;
		}
	}


#ifdef __debugging

// a dummy MAIN()

void main()
	{
	keyboard_InstallHandler();

	do
		{
//		delay(100);
		keyboard_GetEvent();

		Quit = (raw_key==129);	// ESC == 1 == dn, 	ESC == 129 == up
		}
	while (!Quit);

	keyboard_RemoveHandler();
	}
#endif
