//	-------------------------------------------------------------------------
//
//	MOUSE.HPP
//
//	-------------------------------------------------------------------------
//
//	DESCRIPTION:	This file contains defines for the mouse ISR which is
//					capable of auto-detection & direct hardware response.
//
//	-------------------------------------------------------------------------

#ifndef __mouse_hpp
#define __mouse_hpp

#include "graphics.hpp"
#include "stddefns.hpp"

// button status

#define mouse_PRESSED		1
#define mouse_RELEASED		0

// the mouse information structure
struct mouse_Status
	{
	int			leftBut;
	int 		rightBut;
	int			middleBut;

	int			x;
	int			y;
	};

// note: a mouse cursor is based on a grImage, but it _must_ always be a
//		 32 * 32 pixel (or less) image!!

#define MouseCursor		grImage

// defines used to interpret the mouse
//
// eg. 		if (e.what==ev_mouse)
//          	switch (e.mouse.event) {
//					case mb_leftdn:		printf("left mouse was pressed\n"); break;
//					case mb_leftup:		printf("left mouse was released\n"); break;
//					case mb_move:		printf("mouse is moving!!!\n"); break; }

#define		mb_none			0
#define		mb_leftdn		1
#define		mb_leftauto		2
#define		mb_leftup		3
#define		mb_rightdn		4
#define 	mb_rightauto	5
#define		mb_rightup		6
#define 	mb_middledn		7
#define		mb_middleauto	8
#define		mb_middleup		9
#define		mb_move			10

// some pre-defined cursors
extern		MouseCursor		ARROW_CURSOR;
//extern 	MouseCursor		HOURGLASS_CURSOR;
extern		MouseCursor*	DEFAULT_CURSOR;

// this is the current mouse cursor
extern		MouseCursor*	mouse_Cursor;

//	-------------------------------------------------------------------------
//
//	FUNC:		mouse_IsInstalled()
//	DESC:		returns true if mouse driver is currently installed
//
//	-------------------------------------------------------------------------
bool mouse_IsInstalled();

//	-------------------------------------------------------------------------
//
//	FUNC:		mouse_Reset()
//	DESC:		resets the mouse to its original state
//
//	-------------------------------------------------------------------------
void mouse_Reset();

//	-------------------------------------------------------------------------
//
//	FUNC:		mouse_EvableMoveEvents()
//	DESC:		upon calling this function, mouse_GetEvent() will generate
//				mb_move events. (Note: this may cause a large number of
//				events to be placed in the event queue!!!)
//
//	-------------------------------------------------------------------------
void mouse_EnableMoveEvents();

//	-------------------------------------------------------------------------
//
//	FUNC:		mouse_DisableMoveEvents()
//	DESC:		stops mouse_GetEvent from generating mouse move events into
//				the EVENT QUEUE.
//
//	-------------------------------------------------------------------------
void mouse_DisableMoveEvents();

//	-------------------------------------------------------------------------
//
//	FUNC:		mouse_GetPos()
//	DESC:		returns the current position of the mouse as a point
//
//	-------------------------------------------------------------------------
grPoint mouse_GetPos();

//	-------------------------------------------------------------------------
//
//	FUNC:		mouse_SetPos()
//	DESC:		sets the current position of the mouse given a point
//
//	-------------------------------------------------------------------------
void 	mouse_SetPos(grPoint& pt);
void    mouse_SetPos(int x1,int y1);

//	-------------------------------------------------------------------------
//
//	FUNC:		mouse_GetStatus()
//	DESC:		gets the position and the button's status
//
//	-------------------------------------------------------------------------
void	mouse_GetStatus(mouse_Status* ms_status);

//	-------------------------------------------------------------------------
//
//	FUNC:		mouse_SetRange()
//	DESC:		sets the range of the mouse cursor
//
//	-------------------------------------------------------------------------
void mouse_SetRange(int x1,int y1,int x2,int y2);

//	-------------------------------------------------------------------------
//
//	FUNC:		mouse_GetEvent()
//	DESC:		checks to see if the mouse has changed status at all
//				and generates an appropriate event into the EVENT QUEUE
//
//	-------------------------------------------------------------------------
void mouse_GetEvent();

//	-------------------------------------------------------------------------
//
//	FUNC:		mouse_IsHit()
//	DESC:		returns true if mouse is waiting to be processed
//
//	-------------------------------------------------------------------------
bool mouse_IsHit();

//	-------------------------------------------------------------------------
//
//	FUNC:		mouse_Flush()
//	DESC:		flushes the mouse input buffer
//
//	-------------------------------------------------------------------------
void mouse_Flush();

//	-------------------------------------------------------------------------
//
//	FUNC:		mouse_InRect()
//	DESC:		Checks to see whether the mouse cursor is in the specified
//				coordinates. Returns true if it is.
//
//	-------------------------------------------------------------------------
bool mouse_InRect(int x1,int y1,int x2,int y2);

//	-------------------------------------------------------------------------
//
//	FUNC:		mouse_DrawCursor()
//	DESC:		Draws the mouse on the currently active page
//
//	-------------------------------------------------------------------------

#define SIMPLE_DRAW				0
#define	SAVE_UNDERNEATH_CURSOR	1

void mouse_DrawCursor( int options = SIMPLE_DRAW );

//	-------------------------------------------------------------------------
//
//	FUNC:		mouse_UnDrawCursor()
//	DESC:		Erases the mouse on the currently active page
//
//	-------------------------------------------------------------------------
void mouse_UndrawCursor();

//	-------------------------------------------------------------------------
//
//	FUNC:		mouse_DrawCursorRow()
//	DESC:		Draws a row of the mouse cursor
//
//	-------------------------------------------------------------------------

void mouse_DrawCursorRow( int row_y, int row_x1, int row_x2, int options = SIMPLE_DRAW );

//	-------------------------------------------------------------------------
//
//	FUNC:		mouse_GetUnderCursor()
//	DESC:		Saves what is underneath mouse on the currently active page
//
//	-------------------------------------------------------------------------
void mouse_GetUnderCursor();

//	-------------------------------------------------------------------------
//
//	FUNC:		mouse_Hide()
//	DESC:		Hides the mouse if required. NOT REQUIRED TO USE GRAPHICS
//
//	-------------------------------------------------------------------------
void mouse_Hide();

//	-------------------------------------------------------------------------
//
//	FUNC:		mouse_Show()
//	DESC:		Shows the mouse if required. NOT REQUIRED TO USE GRAPHICS
//
//	-------------------------------------------------------------------------
void mouse_Show();

//	-------------------------------------------------------------------------
//
//	FUNC:		mouse_IsHidden()
//	DESC:		returns true if mouse is currently hidden
//
//	-------------------------------------------------------------------------
bool mouse_IsHidden();

//	-------------------------------------------------------------------------
//
//	FUNC:		mouse_UpdateCursor()
//	DESC:		Updates cursor on screen iff the mouse moves.
//	NOTE:		This function is to be called during the main program loop
//
//	-------------------------------------------------------------------------
void mouse_UpdateCursor();

//	-------------------------------------------------------------------------
//
//	FUNC:		mouse_SetCursor()
//	DESC:		Sets the mouse cursor to 'curs' as well as the mouse hotspot
//
//	-------------------------------------------------------------------------
void mouse_SetCursor(int hotx,int hoty,MouseCursor* curs);

//	-------------------------------------------------------------------------
//
//	FUNC:		mouse_InstallHandler()
//	DESC:		checks to see if a mouse driver is active and if so will
//				install an interrupt handler to respond to mouse input
//
//	-------------------------------------------------------------------------
void mouse_InstallHandler();

//	-------------------------------------------------------------------------
//
//	FUNC:		mouse_RemoveHandler()
//	DESC:		removes previously installed mouse handler
//
//	-------------------------------------------------------------------------
void mouse_RemoveHandler();

#endif