//	-------------------------------------------------------------------------
//
//	ONSCREEN.CPP
//
//	-------------------------------------------------------------------------
//
//	AUTHOR:			Darren J. Dwyer
//
//	DESCRIPTION:	This file contains code for the Event Handler object
//					used by the EVENT MANAGER. Note that some routines
//					directly call event manager functions!
//
//	-------------------------------------------------------------------------

#include "onscreen.hpp"
#include "eventmgr.hpp"
#include "keyboard.hpp"
#include "mouse.hpp"

#include "malloc.h"
#include "dos.h"

//#define __DEBUG_TAB

//	-------------------------------------------------------------------------
//
//	FUNC:		CONSTRUCTOR for OnScreenHandler
//	DESC:		The default constructor which defines the behaviour for
//				all derived OnScreenHandler()'s
//
//	PARAMS:		x1,y1,x2,y2			- the coordinates of the OnScreenHandler
//									  (required for dialog_boxes & onscreen
//				fg,bg				- the foreground & background colors
//									  used when active, swapped when inactive
//				active_border,
//				inactive_border		- the borders used when active/inactive
//				handler_num			- the handler number. THIS MUST BE THE
//									  EXACT ORDER OF CONSTRUCTION THROUGH
//									  CALLS TO EVENTMGR::ADDOnScreenHandler()!!!
//              options				- the type of the event handler & any
//									  required options
//				zfact				- the zfactor of the handler. When
//									  interpreting mouse & kbd events for
//									  overlapping handlers, the highest
//									  zfactor is selected to handle the input
//				retevent			- the event to return on destruction
//									  NOTE: NOT YET IMPLEMENTED
//
//	-------------------------------------------------------------------------
OnScreenHandler::OnScreenHandler(int x1,int y1,int x2,int y2,
						int* evhno,
						int options,
						int zfact)
	: EventHandler(evhno,options|evh_ONSCREEN_HANDLER,zfact)
	{
	// signal that we're an onscreen handler
	isOnScreenHandler = true;

	// automatically reset cursor iff !DEFAULT_CURSOR
	AutoResetCursor = true;

	// by default we are onscreen
	isOnScreen = true;

	// set our colors, borders, coordinates
	r.left = x1; r.top = y1; r.right = x2; r.bottom = y2;

	// calc our inset rectangle
	inset_r.left = x1; inset_r.top = y1; inset_r.right = x2; inset_r.bottom = y2;

	// set the accessorySelected to the first one
	accessorySelected = 0;

	// and set the number of accessories to 0
	numAccessories 	  = 0;
	accessoriesAtMax  = false;

//	HEAPCHK("by OnScreenHandler CONSTRUCTOR");
	}

OnScreenHandler::~OnScreenHandler()
	{
	DelAllAccessories();
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		ACCESSORY FUNCTIONS
//
//	-------------------------------------------------------------------------
//
//	FUNC:		AddAccessory()
//	DESC:		Adds an accessory to the window to be used automatically by
//				OnScreenHandler::HandleEvent()
//
//	PARAMS:		accessry - the result from a  new Accessory()
//						   eg. int OK = AddAccessory( new OKButton(,,,) );
//
//	RETURNS:			- a pointer to the accessory.
//				-1		- cannot create accessory (too many)
//
//	-------------------------------------------------------------------------
int OnScreenHandler::AddAccessory(Accessory* accessry)
	{
	// if we are already at max delete the new'd accessory & return
	if (accessoriesAtMax)
		{
		delete accessry; accessry = NULL;
		return -1;
		}

	// tell the accessory it's global coordinates
	accessry->AdjustCoords(r.left,r.top);

	// add an accessory
	accessory[numAccessories] = accessry;

	// and increment the number of accessories
	numAccessories++;

	// check if we are at max.
	accessoriesAtMax = (numAccessories >= MAX_ACCESSORIES_IN_HANDLER-1);

//	HEAPCHK("by OnScreenHandler::AddAccessory()");
	// and return our accessory index
	return numAccessories-1;
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		DelAccessory()
//	DESC:		Deletes the last created accessory from the window.
//
//	-------------------------------------------------------------------------
void OnScreenHandler::DelAccessory()
	{
	// if there are no accessories, skip out
	if (numAccessories==0) return;

	// decrement the number of accessories
	numAccessories--;

	// and delete the accessory
	delete accessory[numAccessories]; accessory[numAccessories] = NULL;

	// check if we are at max.
	accessoriesAtMax = (numAccessories >= MAX_ACCESSORIES_IN_HANDLER-1);

//	HEAPCHK("by OnScreenHandler::DelAccessory()");
//	HEAPCHK("ERROR DELETING ACCESSORY #%d, HANDLER #%d",numAccessories,handler_number);
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		DelAllAccessories()
//	DESC:		Deletes all accessories
//
//	-------------------------------------------------------------------------
void OnScreenHandler::DelAllAccessories()
	{
	// use DelAccessory to delete all accessories
	while (numAccessories>0) DelAccessory();
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		WhichAccessorySelected()
//	DESC:		Calculates which accessory was selected based upon a mouse
//				location point.
//
//	PARAMS:		ms_point	- the point to check
//								eg. int which = WhichAccessorySelected(e.mouse.where);
//
//	RETURNS		>=0 - the number of the accessory selected
//				 -1 - the event handler (window) was selected
//
//	-------------------------------------------------------------------------
int OnScreenHandler::WhichAccessorySelected(grPoint& ms_point)
	{
	// if there are no accessories, return -1
	if (numAccessories==0) return -1;

	// scan through each accessory
	for (int i=0;i<numAccessories;i++)
		// and if mouse point is inside the bounds of the accessory
		if (accessory[i]->MouseInside(ms_point))
			// if the point is inside, return the accessory number
			return i;

//	HEAPCHK("by OnScreenHandler::WhichAccessorySelected()");

	// otherwise the point was not inside any of the accessories
	return -1;
	}

void	OnScreenHandler::ActivateAccessory(int acc)
	{
	if ((acc>=0) && (acc<numAccessories)) accessory[acc]->Activate();
	}

void	OnScreenHandler::DeActivateAccessory(int acc)
	{
	if ((acc>=0) && (acc<numAccessories)) accessory[acc]->DeActivate();
	}

void	OnScreenHandler::DrawAccessory(int acc)
	{
	if ((acc>=0) && (acc<numAccessories)) accessory[acc]->Draw();
	}

void	OnScreenHandler::DrawAccessory(grVirtualBuffer* vb,int acc)
	{
	if (!vb) return;

	if ((acc>=0) && (acc<numAccessories)) accessory[acc]->Draw(vb);
	}

void	OnScreenHandler::ActivateAllAccessories()
	{
	// tell all accessories to draw in order of creation
	for (int acc=0;acc<numAccessories;acc++) accessory[acc]->Activate();
	}

void	OnScreenHandler::DeActivateAllAccessories()
	{
	// tell all accessories to draw in order of creation
	for (int acc=0;acc<numAccessories;acc++) accessory[acc]->DeActivate();
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		DrawAllAccessories()
//	DESC:		draws all accessories (into specified virtual buffer)
//
//	-------------------------------------------------------------------------
void 	OnScreenHandler::DrawAllAccessories()
	{
	// skip out if no accessories
	if (numAccessories==0) return;

	// tell all accessories to draw in order of creation
	for (int acc=0;acc<numAccessories;acc++)
		{
//		DBG(400,"IN HANDLER %p, DRAWING ACCESSORY %d",this,acc);
		accessory[acc]->Draw();
		}

//	HEAPCHK("by OnScreenHandler::DrawAllAccessories()");
	}

void	OnScreenHandler::DrawAllAccessories(grVirtualBuffer* vb)
	{
	if (!vb) return;

	// skip out if no accessories
	if (numAccessories==0) return;

	// tell all accessories to draw in order of creation
	for (int acc=0;acc<numAccessories;acc++) accessory[acc]->Draw(vb);
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		SelectNextAccessory()
//	DESC:		Selects the next (selectable) accessory.
//
//	-------------------------------------------------------------------------
void 	OnScreenHandler::SelectNextAccessory()
	{
	if (numAccessories==0) return;

#ifdef __DEBUG_TAB
	DBG(450,"IN ONSCREENHANDLER %d:: SELECT_NEXT_ACCESSORY()",handler_number);
	DBG(460,"NUM ACCESSORIES       : %d",numAccessories);
#endif

	// tell the previous accessory to change to deselected mode
	accessory[accessorySelected]->DeSelect();

	// use temp value for currently selected accessory
	int acc = accessorySelected;

	do
		{
		// get the next accessory
		acc++;

		// skip iff at the last accessory
		if (acc>=numAccessories) return;
		}
	// keep looping until we find an active accessory
	// or we find the same accessory
	while ((!accessory[acc]->isActive)&&(acc!=accessorySelected));

	// update our selected accessory
	accessorySelected = acc;

	// tell the accessory that it is selected
	accessory[accessorySelected]->Select();

#ifdef __DEBUG_TAB
	DBG(470,"SELECTED ACCESSORY    : %d",accessorySelected);
	grDrawRect(accessory[acc]->r.left,accessory[acc]->r.top,
			   accessory[acc]->r.right,accessory[acc]->r.bottom,
			   grWHITE);
	grBlitRegion(accessory[acc]->r.left,accessory[acc]->r.top,
			   accessory[acc]->r.right,accessory[acc]->r.bottom);
#endif

//	HEAPCHK("by OnScreenHandler::SelectNextAccessory()");
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		SelectPrevAccessory()
//	DESC:		Selects the previous (selectable) accessory.
//
//	-------------------------------------------------------------------------
void 	OnScreenHandler::SelectPrevAccessory()
	{
	if (numAccessories==0) return;

#ifdef __DEBUG_TAB
	DBG(450,"IN ONSCREENHANDLER %d:: SELECT_PREV_ACCESSORY()",handler_number);
	DBG(460,"NUM ACCESSORIES       : %d",numAccessories);
#endif

	// tell the previous accessory to change to deselected mode
	accessory[accessorySelected]->DeSelect();

	// use temp value for currently selected accessory
	int acc = accessorySelected;

	do
		{
		// get the previous accessory
		acc--;

		// skip if at the first accessory
		if (acc<0) return;
		}
	// keep looping until we find an active accessory
	// or we find the same accessory
	while ((!accessory[acc]->isActive)&&(acc!=accessorySelected));

	// update our selected accessory
	accessorySelected = acc;

	// tell the accessory that it is selected
	accessory[accessorySelected]->Select();

#ifdef __DEBUG_TAB
	DBG(470,"SELECTED ACCESSORY    : %d",accessorySelected);
	grDrawRect(accessory[acc]->r.left,accessory[acc]->r.top,
			   accessory[acc]->r.right,accessory[acc]->r.bottom,
			   grWHITE);
	grBlitRegion(accessory[acc]->r.left,accessory[acc]->r.top,
			   accessory[acc]->r.right,accessory[acc]->r.bottom);
#endif

//	HEAPCHK("by OnScreenHandler::SelectPrevAccessory()");
	}

bool	OnScreenHandler::AtFirstAccessory()
	{
	if (numAccessories==0) return true;
	return (accessorySelected==0);
	}

bool	OnScreenHandler::AtLastAccessory()
	{
	if (numAccessories==0) return true;
	return (accessorySelected==numAccessories-1);
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		MouseInside()
//	DESC:		returns true if mouse is inside 'r'
//
//	-------------------------------------------------------------------------
bool 	OnScreenHandler::MouseInside(grPoint& p)
	{
	if (isOnScreen)
		return ((p.x>=r.left)&&(p.x<=r.right)&&(p.y>=r.top)&&(p.y<=r.bottom));
	else return false;
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		Select()
//	DESC:		This function is called by the Event Manager upon its
//				selecting a new event handler for keyboard/mouse input...
//
//	PARAMS:		which	- which accessory to select
//						  evh_FIRST_ACCESSORY,
//						  evh_LAST_ACCESSORY or
//						  evh_CURRENT_ACCESSORY
//
//	-------------------------------------------------------------------------
void OnScreenHandler::Select( int which )
	{
//	grDrawRect(r.left,r.top,r.right,r.bottom,grLIGHTRED);

	if (numAccessories>0)
		{

#ifdef __DEBUG_TAB
	DBG(450,"IN ONSCREENHANDLER %d:: SELECT_NEXT_ACCESSORY()",handler_number);
	DBG(460,"NUM ACCESSORIES       : %d",numAccessories);
#endif

		int acc = accessorySelected;

		switch (which)
			{
			case	evh_FIRST_ACCESSORY:

				accessorySelected = 0;
				break;

			case	evh_CURRENT_ACCESSORY:

				break;

			case	evh_LAST_ACCESSORY:
			default:

				accessorySelected = numAccessories-1;
				break;
			}

		if (accessorySelected!=acc)
			{
#ifdef __DEBUG_TAB
	DBG(470,"SELECTED ACCESSORY    : %d",accessorySelected);
	grDrawRect(accessory[acc]->r.left,accessory[acc]->r.top,
			   accessory[acc]->r.right,accessory[acc]->r.bottom,
			   grWHITE);
	grBlitRegion(accessory[acc]->r.left,accessory[acc]->r.top,
			   accessory[acc]->r.right,accessory[acc]->r.bottom);
#endif

			if (!accessory[accessorySelected]->isSelected)
				accessory[accessorySelected]->Select();

			}
		}

	isSelected = true;
//	HEAPCHK("by OnScreenHandler::Select()");
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		DeSelect()
//	DESC:		This function is called by the Event Manager upon its
//				selecting a new event handler for keyboard/mouse input...
//
//	-------------------------------------------------------------------------
void OnScreenHandler::DeSelect()
	{
//	grDrawRect(r.left,r.top,r.right,r.bottom,grBLUE);

	if (numAccessories>0)
		if (accessory[accessorySelected]->isSelected)
			accessory[accessorySelected]->DeSelect();

	isSelected = false;
//	HEAPCHK("by OnScreenHandler::DeSelect()");
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		Animate()
//	DESC:		This function is called upon receiving a cmdAnimate event
//				This occurs automatically by calling
//				OnScreenHandler::HandleEvent() in derived HandleEvent()'s
//
//	-------------------------------------------------------------------------
void OnScreenHandler::Animate()
	{
	// does nothing
//	HEAPCHK("by OnScreenHandler::Animate()");
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		Draw()
//	DESC:		This function does all of the drawing of the OnScreenHandler
//
//	-------------------------------------------------------------------------
void OnScreenHandler::Draw()
	{
	if (!isOnScreen) return;
	if (isSelected)	;	// do something
	else			;	// do something else

	// add the accessories
	DrawAllAccessories();
	}

void	OnScreenHandler::Refresh()
	{
	Draw();

	// we've finished drawing everything, blit our buffer to the screen
	grBlitRegion(r.left,r.top,r.right,r.bottom);
	}

void	OnScreenHandler::Open()
	{
	if (!isOnScreen)
		{
		isOnScreen = true;
		Refresh();
		}
	}

void	OnScreenHandler::Close()
	{
	isOnScreen = false;
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		HandleEvent()
//	DESC:		This is the generic HandleEvent() function which should be
//              called by any derived objects HandleEvent() function as follows
//
//				eg. MyWindow::HandleEvent(Event &e)
//						{
//
//						(interpret events here to override OnScreenHandler)
//
//						// call parent HandleEvent() to handle accessories!!
//						OnScreenHandler::HandleEvent(e);
//
//						(interpret events here after accessories handle event)
//
//						}
//
//				by default this HandleEvent interprets the following keys:
//					TAB 	- move to next accessory
//					SH_TAB 	- move to previous accessory
//					RET 	- select current accessory
//					SPACE   - select current accessory
//
//	PARAMS:		e - the specified event to be handled
//
//	-------------------------------------------------------------------------
void OnScreenHandler::HandleEvent(Event &e)
	{
//	DBG(440,"IN OnScreenHandler::HANDLEEVENT()   %p",this);
	// skip out if we aren't meant to be here....
	// handle the event!
	int acc;
	switch (e.what)
		{
		// handle mouse events
		case ev_mouse:

			if (!isOnScreen) return;

			// if its a mouse move event
			if (e.mouse.event==mb_move)
				// and we're not manually doing cursor bits
				if (AutoResetCursor) mouse_SetCursor(0,0,DEFAULT_CURSOR);

			// find selected accessory
			acc = WhichAccessorySelected(e.mouse.where);
//							DBG(470,"WHICH ACCESSORY SELECTED FOR %d,%d IS %d",e.mouse.where.x,e.mouse.where.y,acc);

			// if an accessory was selected
			if (acc>=0)
				{
				// and it isn't the currently selected accessory
				if (acc!=accessorySelected)
					{
					// deselect the current accessory
					if (accessory[accessorySelected]->isSelected)
						accessory[accessorySelected]->DeSelect();
					}

				// we've now selected the accessory (as such)
				accessorySelected = acc;

				// and tell it to handle the event
				accessory[accessorySelected]->HandleEvent(e);
				}
			else
				// no accessory was selected
				// so if there are any accessories, deselect the current one
				if (numAccessories>0)
					if (accessory[accessorySelected]->isSelected)
						accessory[accessorySelected]->DeSelect();

			break;


		// handle keyboard events
		case ev_keyboard:

			if (!isOnScreen) return;

//			DBG(400,"IN ONSCREENHANDLER %d :: NUM ACCESSORIES == %d",handler_number,numAccessories);
			if (numAccessories==0) return;

			// try selected accessory first!!!
			accessory[accessorySelected]->HandleEvent(e);

			// skip out if handled
			if (e.what!=ev_keyboard) return;

			// check for keyevents not handled by the selected accessory
			if (e.what==ev_keyboard)
				{
				switch (e.keyboard.event)
					{
					case TAB_DN:

						// if its shift-tab
						if (e.keyboard.modifiers&keyboard_SHIFT)
							{
							if (!AtFirstAccessory())
								{
								SelectPrevAccessory();
								EVENTMGR->ClearEvent(e);
								}
							}
						else
							{
							if (!AtLastAccessory())
								{
								SelectNextAccessory();
								EVENTMGR->ClearEvent(e);
								}
							}
						return;

					case CAPS_LOCK_DN:

						if (!AtFirstAccessory())
							{
							SelectPrevAccessory();
							EVENTMGR->ClearEvent(e);
							}
						return;
					}
				}

			// loop through all accessories
			acc = 0;
			do
				{
				// see if the accessry can handle the event
				if (acc!=accessorySelected)
					{
//					DBG(410,"ACCESSORY[%d]->Handling event",acc);
					accessory[acc]->HandleEvent(e);
					}

				// if its still a keyboard event, try the next accessory
				if (e.what==ev_keyboard) acc++;
				else
					// skip out as it's a new event
					break;
				}
			while (acc<numAccessories);

			// let anything else fall back to eventmgr
			break;

		// handle command events
		case ev_command:

			switch (e.command.event)
				{
				case cmdAnimate:

					if (isOnScreen) Animate();
					EVENTMGR->ClearEvent(e);
					break;

				case cmdRedraw:

					if (isOnScreen) Draw();
					EVENTMGR->ClearEvent(e);
					break;

				case cmdRefresh:

					if (isOnScreen) Refresh();
					EVENTMGR->ClearEvent(e);
					break;

				case cmdOpen:

					Open();
					EVENTMGR->ClearEvent(e);
					break;

				case cmdClose:

					Close();
					EVENTMGR->ClearEvent(e);
					break;
				}
			break;

		// handle broadcast events
		case ev_broadcast:

			switch (e.broadcast.event)
				{
				case cmdAnimate:

					if (isOnScreen) Animate();
					break;
				}
			break;

		// handle timer events
		case ev_timer:

			switch (e.timer.event)
				{
				case cmdAnimate:

					if (isOnScreen) Animate();
					break;
				}
			break;
		}
//	HEAPCHK("by OnScreenHandler::HandleEvent()");
	}
