//	-------------------------------------------------------------------------
//
//	PALETTES.CPP
//
//	-------------------------------------------------------------------------
//
//	AUTHOR:			Darren J. Dwyer
//
//	DESCRIPTION:	This file defines the palette functions used
//					by the game development library.
//
//	-------------------------------------------------------------------------

#include "palettes.hpp"
#include "palfile.hpp"

#include "conio.h"

grPalette*  		grPALETTE;		// the current palette

// Standard VGA palette ports

#define 	PAL_MASK		0x03C6
#define		PAL_READ		0x03C7
#define		PAL_WRITE		0x03C8
#define		PAL_DATA		0x03C9

// set palette entry #num to r,g,b

float 		red_step[256];		// used by grFadeOutPalette & grFadeInPalette
float 		green_step[256];	// ditto
float 		blue_step[256];		// ditto

grPalette* grCreatePalette(int numentries)
	{
	if (numentries>256) numentries = 256;

	grPalette* p = (grPalette*) SYS_Malloc(grPaletteSize(numentries));

	p->hdr.numentries = numentries;

	return p;
	}

void grDestroyPalette(grPalette* pal)
	{
	if (pal!=NULL) { SYS_Free(pal); pal=NULL; }
	}

int grPaletteSize(int numentries)
	{
	if (numentries>256) numentries = 256;
	return numentries * 3 + sizeof(grPaletteHeader);
	}

void grSetPaletteEntry(uchar entry,uchar r,uchar g,uchar b)
	{
	// set the palette mask register (why?? dunno)
	outp(PAL_MASK,0xFF);

	// tell card to update specific register
	outp(PAL_WRITE,entry);

	// and write the R,G,B values
	outp(PAL_DATA,r & 63);
	outp(PAL_DATA,g & 63);
	outp(PAL_DATA,b & 63);
	}

void grGetPaletteEntry(uchar entry,uchar* r,uchar* g,uchar* b)
	{
	// set the palette mask register
	outp(PAL_MASK,0xFF);

	// tell card to update specific register
	outp(PAL_READ,entry);

	// and write the R,G,B values
	*r = inp(PAL_DATA);
	*g = inp(PAL_DATA);
	*b = inp(PAL_DATA);
	}

void grSetPaletteEntry(grPalette* pal,int index,int r,int g,int b)
	{
	if (index>pal->hdr.numentries) index = pal->hdr.numentries;
	int pos = index*3;
	pal->data[pos++] = r;
	pal->data[pos++] = g;
	pal->data[pos] = b;
	}

void grSetStandardPalette(int mult,int red,int green,int blue)
	{
	grPalette* pal = grCreatePalette(256);
	if (pal)
		{
		char r,g,b;
		int r2,g2,b2;

		if (mult>12) mult=12;
		if (mult<0) mult=0;
		for (r=0; r<6; r++)
			for (g=0; g<6; g++)
				for (b=0; b<6; b++)
					{
					r2=r*mult+red;
					if (r2>63) r2=63;
					else if (r2<0) r2=0;

					g2=g*mult+green;
					if (g2>63) g2=63;
					else if (g2<0) g2=0;

					b2=b*mult+blue;
					if (b2>63) b2=63;
					else if (b2<0) b2=0;

					grSetPaletteEntry(pal,r+6*g+36*b,r2,g2,b2);
					}

		for (r2=216; r2<255; r2++) grSetPaletteEntry(pal,r2,0,0,0);
		grSetPaletteEntry(pal,255,63,63,63);

		grSetPalette(pal,0,255);
		grDestroyPalette(pal);
		}
	}


void grSetPalette(grPalette* pal,uchar start,uchar end)
	{
	int index = 0;
	for (int i=start;i<=end;i++)
		{
		grSetPaletteEntry(i,pal->data[index],pal->data[index+1],pal->data[index+2]);
		index += 3;
		}
	}

// gets palette entry #num
void grGetPalette(grPalette* pal,uchar start,uchar end)
	{
	int index = 0;
	for (int i=start;i<=end;i++)
		{
		grGetPaletteEntry(i,(uchar*) &pal->data[index],(uchar*) &pal->data[index+1],(uchar*) &pal->data[index+2]);
		index += 3;
		}
	}

void grFadeOutPalette(grPalette* pal,uchar start,uchar end,uchar timesin)
	{
	uchar r,g,b;
	int index;
	float f_iter = 64 - timesin;

	// fade out the palette
	for (int i=start;i<=end;i++)
		{
		grGetPaletteEntry(i,&r,&g,&b);

		// if its the first iteration, save the palette for grFadeInPalette()
		// and calculate the steps required for even fading...
		if (timesin==0)
			{
			index = i * 3;
			pal->data[index]   = r & 63;
			pal->data[index+1] = g & 63;
			pal->data[index+2] = b & 63;

			red_step[i]   = (float) r / 64.0;
			green_step[i] = (float) g / 64.0;
			blue_step[i]  = (float) b / 64.0;
			}

		r = (uchar) (f_iter * red_step[i]);
		g = (uchar) (f_iter * green_step[i]);
		b = (uchar) (f_iter * blue_step[i]);

		grSetPaletteEntry(i,r,g,b);
		}
	}

void grFadeInPalette(uchar start,uchar end,uchar timesin)
	{
	uchar r,g,b;
	float f_iter = timesin;

	// fade in the palette
	for (int i=start;i<=end;i++)
		{
		r = (uchar) (f_iter * red_step[i]);
		g = (uchar) (f_iter * green_step[i]);
		b = (uchar) (f_iter * blue_step[i]);

		grSetPaletteEntry(i,r,g,b);
		}
	}

void grRotatePalette(uchar start,uchar end)
	{
	// some vars
	uchar r,g,b,r1,g1,b1;

	// save the start entry
	grGetPaletteEntry(start,&r,&g,&b);

	// and set all based on the next
	for (int i=start+1;i<=end;i++)
		{
		grGetPaletteEntry(i,&r1,&g1,&b1);
		grSetPaletteEntry(i-1,r1,g1,b1);
		}

	// and set the last entry to the saved one
	grSetPaletteEntry(end,r,g,b);
	}
