//	-------------------------------------------------------------------------
//
//	PCX.CPP
//
//	-------------------------------------------------------------------------
//
//	DESCRIPTION:	PCX utilities
//
//	-------------------------------------------------------------------------

#include		"pcx.hpp"

//	-------------------------------------------------------------------------
//
//	FUNC:		PCX_Decompress
//	DESC:		converts a run of PCX-compressed bytes to a run of
//				decompressed (raw) bytes.
//
//	PARAMS:		dest		- the address of the decompressed run
//				srce		- the address of the compressed run
//				decompressed_size 	- the number of bytes that were compressed
//				decompressed_width	- the number of bytes to decompress to (max)
//
//	RETURNS:	long		- the size of the _compressed_ data
//
//	-------------------------------------------------------------------------
long 	PCX_Decompress(char* dest,char* srce,long decompressed_size,long decompressed_width)
	{
	char	thecolor;
	char	run_length;

	long	srce_index = 0;
	long	dest_index = 0;

	// decompress till we have finished the complete run....
	for (dest_index=0; dest_index<decompressed_size; )
		{
		// the run length is at least 1 byte long
		run_length = 1;

		// get the next color
		thecolor = srce[srce_index++];

		// iff thebyte is greater than or equal to 192, the
		// color is RLE
		if (thecolor > 192)
			{
			// then the run length is the byte - 192
			run_length = thecolor - 192;

			// and get the color which was encoded
			thecolor = srce[srce_index++];
			}

		// now, we either have an RLE run or a single byte....
		for (int run_cnt=0; run_cnt<run_length; run_cnt++)
			{

	// note: * 	Standard pcx images seem to be compressed on a row-by-row
	//		   	basis.
	//
	//		 *	Odd width images are compressed using an even bytes_per_line
	//			count (rounded up)
	//			ie. a 23x23 image has a bytes per line of 24!!! (23+1)
	//

			// make sure the bytes are decompressed correctly
			// due to the funny bytes_per_row rounding on odd-width images
			if (decompressed_width>0)
				{
				if (dest_index<decompressed_width)	dest[dest_index] = thecolor;
				}
			else
				{
				// normal decompression....
				dest[dest_index] = thecolor;
				}

			// choose the next color byte
			dest_index++;
			}
		}

	// return the size of the run that was decompressed
	return srce_index;
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		PCX_Compress
//	DESC:		converts a row of decompressed (raw) bytes to a run of
//				PCX-compressed bytes.
//
//	PARAMS:		dest		- the address of the compressed run
//				srce		- the address of the decompressed run
//				decompressed_size 	- the number of bytes to be compressed
//				decompressed_width	- the number of bytes to decompress to (max)
//
//	RETURNS:	long		- the size of the _compressed_ data
//
//	-------------------------------------------------------------------------
long	PCX_Compress(char* dest, char* srce, long decompressed_size, long decompressed_width)
	{
	long	srce_index = 0;
	long	dest_index = 0;

	char	cur_color;
	// note: save the first available color as it saves one iteration!!!?
	char	last_color = srce[srce_index];
	char	run_length = 1;

	// scan through the uncompressed data byte-by-byte
	for (srce_index=1; srce_index<decompressed_size; srce_index++ )
		{
		// get the current color
		if (decompressed_width)
			{
			// if we are inbounds get current color
			if (srce_index < decompressed_width)
				cur_color = srce[srce_index];

			// else just use last color for maximum compression
			else cur_color = last_color;
			}
		else
			{
			// we don't care about bounds
			cur_color = srce[srce_index];
			}

		// iff the current color is the same as the last color
		if (cur_color == last_color)
			{
			// count how many times it's the same...
			run_length++;

			// if we have a run count of 63 (63 + 192 = 255)
			if (run_length >= 63)
				{
				// store the run length
				dest[dest_index++] = run_length + 192;

				// store the run color
				dest[dest_index++] = last_color;

				// we're now ready for the next run...
				run_length = 0;
				}
			}
		else
			// the color has changed....
			{
			// if we are in the middle of a run
			if (run_length)
				{
				// if run count is only 1 and last color < 192
				if ((run_length == 1) && (last_color < 192))
					{
					// save the color as is
					dest[dest_index++] = last_color;
					}
				else
					// we have to store the data as an RLE
					{
					dest[dest_index++] = run_length + 192;
					dest[dest_index++] = last_color;
					}
				}

			// we're now starting a new run on the current color
			run_length = 1;

			// set last color to the current color
			last_color = cur_color;
			}

		}

	// save the last run now!!!
	if (run_length)
		{
		// iff run count is only 1 and last color < 192
		if ((run_length == 1) && (last_color<192))
			{
			// store the last color without RLE
			dest[dest_index++] = last_color;
			}
		else
			{
			// store the data as an RLE
			dest[dest_index++] = run_length + 192;
			dest[dest_index++] = last_color;
			}
		}

	// return the size of the compressed data
	return dest_index;
	}

