//
//  RESMGR.HPP
//
//
//	Simple Resource Manager using Cached Objects
//
//	DJD
//


#ifndef __resmgr_hpp
#define __resmgr_hpp

/*

USAGE:
-----

	ResourceFile*	map_file 	= new ResourceFile("RESOURCE\\MAPS.RES");
	ResourceFile*	font_file 	= new ResourceFile("RESOURCE\\TEST.FNT");
	ResourceFile*	pal_file	= new ResourceFile("RESOURCE\\TEST.PAL");
	ResourceFile*	bstyle_file = new ResourceFile("RESOURCE\\TEST.BRD");

	.
	.
	.

	// note: * order of creation is _irrelevant_

	int		pic1 = grInitCachedImage(map_file,0);            // RM->RegisterImage()
	int		pic3 = grInitCachedImage(map_file,2);			 // RM->RegisterImage()
	int		pic2 = grInitCachedImage(map_file,1);            // RM->RegisterImage()

	int		pal1 = grInitCachedPalette(pal_file);            // RM->RegisterPalette()
	int		fnt1 = grInitCachedFont(font_file);              // RM->RegisterFont()

	int		bs1  = grInitCachedBorder(bstyle_file,0);		 // RM->RegisterBorder()
	int		bs2	 = grInitCachedBorder(bstyle_file,1);		 // RM->RegisterBorder()

	int		x1 = 0, y1 = 0, x2 = grMaxx, y2 = grMaxy;

	// note: * order of usage is _irrelevant_

	grSetCachedPalette(pal1,0,255);

	grPutCachedImage(x1,y1,x2,y2,pic1);
	grOverlayCachedImage(x1,y1,x2,y2,pic2);

	grDrawCachedBorder(x1,y1,x2,y2,bs1);

	char*	buf 	= "Test String Number 99");

	int		width  	= grFontCachedWidth(buf);
	int 	height 	= grFontCachedHeight(buf);

	grDrawCachedBorder(x1+90,y1+90,x1+90+width+10,y1+90+10*height+10,bs2);
	grSetCachedFont(fnt1);
	for (int i=0;i<10;i++)
		grDrawCachedString(x1+100,y1+100+i*height,i,"Test String Number %d",i);

	.
	.
	.

	// note: * order of deletion is _irrelevant_
	//       * if about to terminate, don't even need to grExit...() anything!

	grExitCachedBorder(bs1);
	grExitCachedPalette(pal1);
	grExitCachedImage(pic3);
	grExitCachedBorder(bs2);
	grExitCachedImage(pic1);
	grExitCachedFont(fnt1);
	grExitCachedImage(pic2);

	.
	.
	.

	delete pal_file;
	delete font_file;
	delete map_file;

*/

// -------------

#include "graphics.hpp"
#include "fonts.hpp"
#include "borders.hpp"
#include "resource.hpp"
#include "pcxfile.hpp"
#include "palfile.hpp"

// -------------

#define	err_NoAvailableId		-1
#define	err_TooManyCachedFonts	-2
#define	err_UnknownCachedFont	-3

// -------------

#define	MAX_CACHE_ENTRIES		5000
#define	MAX_HITS				1000000

// ------------- different object types that are implemented

#define	C_AVAILABLE		0
#define	C_IMAGE			1
#define	C_FONT			2
#define	C_PALETTE		3
#define	C_BSTYLE		4
#define	C_TILE          5

// -------------

struct	MemBlockHeader
	{
	long	prev;
	long	size;
	char	used;
	};

// -------------

struct 	CacheItemStruct
	{
	char			type; 		// C_IMAGE, etc.
	int				size;		// size of stored data in bytes
	long			hits;		// total count of times accessed
	ResourceFile*	file;		// resourcefile where item is located
	int				ofs;		// record number of item
	void*			data;		// pointer to first byte of data, NULL = not loaded
	};

// ------------- CACHEMGR STORAGE

class 	ResourceMgr
	{
	private:

		// ---------------

		char*				CacheMem;

		long				CacheUsed;
		long				CacheSize;

		long				LastBlockAddr;
		long				BiggestBlockSize;
		long				BiggestBlockAddr;

		int					CacheIdsUsed;

		CacheItemStruct*	CacheItem;

		// ---------------

		void		FindBiggestFreeBlock();

		void		SplitBlock(MemBlockHeader* block,int new_size);

		void		MergeBlocks(MemBlockHeader* new_block,
								MemBlockHeader* dead_block);

		void*		AllocateBlock(int nbytes);
		void		DeAllocateBlock(void* block_data_ptr);

		int			FirstAvailId();
		int			LeastUsedId();

		// ---------------

	public:

		ResourceMgr();
		~ResourceMgr();

		int		RegisterImage(ResourceFile* file,int file_ofs);
		int		RegisterPCXImage(PCXFile* file,int file_ofs);
		int		RegisterFont(ResourceFile* file);
		int		RegisterPalette(PALFile* file,int file_ofs);
		int		RegisterBorder(ResourceFile* file);
		int		RegisterTile(ResourceFile* file,int file_ofs);

		void	DeRegisterImage(int id);
		void	DeRegisterPCXImage(int id);
		void	DeRegisterFont(int id);
		void	DeRegisterPalette(int id);
		void	DeRegisterBorder(int id);
		void	DeRegisterTile(int id);

		grImage*		ImagePtr(int id);
		grImage*		FontImagePtr(int font_id,int ch);

		grImage*		PCXImagePtr(int id);
		grPalette*		PalettePtr(int id);
		grBorder*		BorderPtr(int id);
		grImage*		TilePtr(int id);

		void	Hit(int id);
	};

// ---------------

#define	MAX_CACHED_FONTS	10

// ---------------

extern	ResourceMgr*		RESOURCE_MGR;
extern	int					CachedFont[MAX_CACHED_FONTS][font_NUM_ASCII];
extern	int					NumCachedFonts;

// ---------------

extern	void				InitResourceManager();
extern	void				ExitResourceManager();

// ---------------

#define	grInitCachedImage		RESOURCE_MGR->RegisterImage
#define	grInitCachedPCXImage	RESOURCE_MGR->RegisterPCXImage
#define	grInitCachedPalette		RESOURCE_MGR->RegisterPalette
#define	grInitCachedBorder		RESOURCE_MGR->RegisterBorder
#define	grInitCachedTile		RESOURCE_MGR->RegisterTile

#define	grImagePtr				RESOURCE_MGR->ImagePtr
#define	grPCXImagePtr			RESOURCE_MGR->PCXImagePtr
#define	grFontImagePtr			RESOURCE_MGR->FontImagePtr
#define	grPalettePtr			RESOURCE_MGR->PalettePtr
#define grBorderPtr				RESOURCE_MGR->BorderPtr
#define	grTilePtr				RESOURCE_MGR->TilePtr

#define	grExitCachedImage		RESOURCE_MGR->DeRegisterImage
#define	grExitCachedPCXImage	RESOURCE_MGR->DeRegisterPCXImage
#define	grExitCachedPalette		RESOURCE_MGR->DeRegisterPalette
#define	grExitCachedBorder		RESOURCE_MGR->DeRegisterBorder
#define	grExitCachedTile		RESOURCE_MGR->DeRegisterTile

#define	HIT						RESOURCE_MGR->Hit

// ---------------

/*#define	grSetCachedPalette(PAL,START,END) \
			grSetPalette(grCachedPalettePtr((PAL)),(START),(END)); \
			HIT(PAL)

#define	grPutCachedImage(X1,Y1,X2,Y2,IMAGE) \
			grPutImage((X1),(Y1),(X2),(Y2),grCachedImagePtr(IMAGE)); \
			HIT((IMAGE))

#define	grOverlayCachedImage(X1,Y1,X2,Y2,TR_COLOR,IMAGE) \
			grOverlayImage((X1),(Y1),(X2),(Y2),(TR_COLOR),grCachedImagePtr(IMAGE)); \
			HIT((IMAGE))

//#define	grDrawCachedBorder(X1,Y1,X2,Y2,TR_COLOR,BORDER) \
//			grDrawBorder((X1),(Y1),(X2),(Y2),grCachedBorderPtr(BORDER)); \
//			HIT((BORDER))
*/

// ---------------

#endif

