//      -------------------------------------------------------------------------
//
//      RESOURCE.HPP
//
//      -------------------------------------------------------------------------
//
//      DESCRIPTION:    This file implements the Resource object which is used to
//                                      access a pseudo-random-access file containing variable
//                                      length records (resources)
//
//      -------------------------------------------------------------------------

#include "resource.hpp"
#include "pcx.hpp"

#include "graphics.hpp"
#include "fonts.hpp"
#include "palettes.hpp"
#include "sprites.hpp"
#include "program.hpp"

#include "process.h"
#include "fcntl.h"
#include "dos.h"
#include "sys\stat.h"
#include "errno.h"
#include "malloc.h"

ResourceFile::ResourceFile(char* filename) : File()
	{
	file_path = new String("");
	file_name = new String(filename);
	file_isopen = false;

	Open();
	}

ResourceFile::ResourceFile(String* filename) : File()
	{
	file_path = new String("");
	file_name = new String(filename->Contents());
	file_isopen = false;

	Open();
	}


ResourceFile::ResourceFile(char* path,char* filename) : File()
	{
	file_path = new String(path);
	file_name = new String(filename);
	file_isopen = false;

	Open();
	}

ResourceFile::ResourceFile(String* path,String* filename) : File()
	{
	file_path = new String(path->Contents());
	file_name = new String(filename->Contents());
	file_isopen = false;

	Open();
	}


ResourceFile::~ResourceFile()
	{
	Close();

	if (file_name) { delete file_name; file_name = NULL; }
	if (file_path) { delete file_path; file_path = NULL; }
	}

int		ResourceFile::Create(long max_records,char type_of_resource)
	{
//	printf("Creating new resource file!\n");

	// if we can successfully create a new file
	if (File::Create() != file_NoError) return file_CreateFailed;

	// open it
//	printf("Open() status == %d\n",File::Open());
	Open();

	header.resource_type = type_of_resource;
	header.max_records = max_records;
	header.num_records = 0;

	// and store it in the file
//	printf("Writing resource header\n");
	WriteResourceHeader();

	// create all required indexes
	memset(&index,0,sizeof(RecordIndex));
	for (int i=0;i<header.max_records;i++)
		{
//		printf("Writing blank record index #%d\n",i);
		WriteRecordIndex(i);
		}

	// and close it
	return File::Close();
	}

int		ResourceFile::Open()
	{
	// attempt to open the file
	File::Open();

	// if it worked, simply read the resource header information
	return ReadResourceHeader();
	}

//      -------------------------------------------------------------------------
//
//      FUNC:           CompressedRecordSize()
//      DESC:           returns the size of the compressed record
//
//      -------------------------------------------------------------------------
long	ResourceFile::CompressedRecordSize(long record_num)
	{
	// read the resource index
	if (ReadRecordIndex(record_num) != file_NoError) return resource_RecordNotFound;

	// and return its compressed size
	return index.sizeofcompressed;
	}

//      -------------------------------------------------------------------------
//
//      FUNC:           DecompressedRecordSize()
//      DESC:           returns the size of the decompressed record
//
//      -------------------------------------------------------------------------
long	ResourceFile::DecompressedRecordSize(long record_num)
	{
	// read the resource index
//	DBG(300,"READING DECOMPRESSED RECORD SIZE FOR RECORD #%d",record_num);
	if (ReadRecordIndex(record_num) != file_NoError) return resource_RecordNotFound;

	// and return its decompressed size
	return index.sizeofdecompressed;
	}

//      -------------------------------------------------------------------------
//
//      FUNC:           ReadHeader()
//      DESC:           reads the resource file header
//
//      -------------------------------------------------------------------------
int ResourceFile::ReadResourceHeader()
	{
	// skip out if file isn't open
	if (!file_isopen) return file_FileNotOpen;

	// seek to the start of the file
	if (SeekTOF() != file_NoError) return file_SeekFailed;

	// and read the header
	return Read((char*) &header,sizeof(ResourceHeader));
	}

//      -------------------------------------------------------------------------
//
//      FUNC:           WriteHeader()
//      DESC:           writes the resource file header
//
//      -------------------------------------------------------------------------
int ResourceFile::WriteResourceHeader()
	{
	// skip out if file isn't open
	if (!file_isopen) return file_FileNotOpen;

	// seek to the start of the file
	if (SeekTOF() != file_NoError) return file_SeekFailed;

	// and read the header
	return Write((char*) &header,sizeof(ResourceHeader));
	}

//      -------------------------------------------------------------------------
//
//      FUNC:           ReadResourceIndex()
//      DESC:           reads the location & size pertaining to a particular resource
//
//      -------------------------------------------------------------------------
int ResourceFile::ReadRecordIndex(long record_num)
	{
	// skip out if file isn't open
	if (!file_isopen)  return file_FileNotOpen;

	// skip out if the resource is out of bounds
	if (record_num>header.num_records) return resource_InvalidIndexNum;

	// seek to the start of the file + offsets
	if (Seek( sizeof(ResourceHeader) + sizeof(RecordIndex) * record_num) != file_NoError)
		return file_SeekFailed;

	// and read the header
	return Read((char*) &index,sizeof(RecordIndex));
	}

//      -------------------------------------------------------------------------
//
//      FUNC:           WriteResourceIndex()
//      DESC:           reads the location & size pertaining to a particular resource
//
//      -------------------------------------------------------------------------
int ResourceFile::WriteRecordIndex(long record_num)
	{
	// skip out if file isn't open
	if (!file_isopen) return file_FileNotOpen;

	// skip out if the resource is out of bounds
	if (record_num>header.max_records) return resource_InvalidIndexNum;

	// seek to the start of the file + offsets
	if (Seek( sizeof(ResourceHeader) + sizeof(RecordIndex) * record_num) != file_NoError)
		return file_SeekFailed;

	// and write the header
	return Write((char*) &index,sizeof(RecordIndex));
	}

// -------------------------------------------------------------------------
//
// FUNC:       ReadRecord()
// DESC:       reads a record
//
// -------------------------------------------------------------------------
char* 	ResourceFile::ReadRecord(long record_num)
	{
	// read the resource index
	if (ReadRecordIndex(record_num) != file_NoError) return NULL;

	// if the size is zero, there is no record available
	if (index.sizeofdecompressed==0) return NULL;

	// create the buffer to be returned
	char* decompression_buffer = (char*) SYS_Malloc(index.sizeofdecompressed);

	// skip out if we couldn't create it...
//	if (!decompression_buffer) DBG(270,"ERROR: COULD NOT MALLOC DECOMPRESSION BUFFER");
	if (!decompression_buffer) return NULL;

	// seek to the start of the specified record
	if (Seek(index.location) != file_NoError) return NULL;

	// create a buffer to read the record
	char* compression_buffer = (char*) SYS_Malloc(index.sizeofcompressed);

	// skip out if we couldn't create it...
//	if (!compression_buffer) DBG(280,"ERROR: COULD NOT MALLOC COMPRESSION BUFFER");
	if (!compression_buffer) { SYS_Free(decompression_buffer); return NULL; }

	// read the compressed data from the file
	if (Read(compression_buffer,index.sizeofcompressed) == file_NoError)
		{
		// and decompress it!!!
		DecompressRecord(decompression_buffer,compression_buffer,index.sizeofdecompressed);
		}

	// get rid of the buffer for the compressed data
	SYS_Free(compression_buffer);

	// and return the pointer to the decompressed data
	return decompression_buffer;
	}

// -------------------------------------------------------------------------
//
// FUNC:       ReadRecord()
// DESC:       reads a record into an existing buffer
//
// -------------------------------------------------------------------------
char*	ResourceFile::ReadRecord(long record_num,char* buffer)
	{
	// read the resource index
	if (ReadRecordIndex(record_num) != file_NoError) return NULL;

	// if the size is zero, there is no record available
	if (index.sizeofdecompressed==0) return NULL;

	// setup the decompression buffer
	char* decompression_buffer = buffer;

	// seek to the start of the specified record
	if (Seek(index.location) != file_NoError) return NULL;

	// create a buffer to read the record
	char* compression_buffer = (char*) SYS_Malloc(index.sizeofcompressed);

	// skip out if we couldn't create it...
//	if (!compression_buffer) DBG(280,"ERROR: COULD NOT MALLOC COMPRESSION BUFFER");
	if (!compression_buffer) { SYS_Free(decompression_buffer); return NULL; }

	// read the compressed data from the file
	if (Read(compression_buffer,index.sizeofcompressed) == file_NoError)
		{
		// and decompress it!!!
		DecompressRecord(decompression_buffer,compression_buffer,index.sizeofdecompressed);
		}

	// and return the pointer to the decompressed data
	return decompression_buffer;
	}

// -------------------------------------------------------------------------
//
// FUNC:       ReadAllRecords()
// DESC:       read all the records in the resource into memory, returning
//			   an array of pointers to the records
//
// -------------------------------------------------------------------------
char** ResourceFile::ReadAllRecords()
	{
//HEAPCHK("RESOURCEFILE::ReadAllRecords() corrupt at start");

	// get the current resource header caJUSTse
	if (ReadResourceHeader() != file_NoError) return NULL;

	char** records_ptr = (char**) SYS_Malloc(sizeof(char*) * header.num_records);

//HEAPCHK("RESOURCEFILE::ReadAllRecords()~~\
//Corrupted just after attempting to SYS_Malloc()~\
//records_ptr, attempted malloc: %d * %d == %d",
//sizeof(char*),header.num_records,
//sizeof(char*)*header.num_records);

//	if (!records_ptr)
//		{
//		DBG(250,"ERROR: COULD NOT MALLOC RECORDS POINTER, BYTES REQUESTED == %lu, NUM RECORDS == %u",
//					sizeof(char*) * header.num_records,header.num_records);
//		if (file_isopen) DBG(260,"NOTE: FILE IS OPEN");
//		else DBG(260,"NOTE: FILE IS --NOT-- OPEN");
//		HEAPCHK("ERROR: HEAP CORRUPT ON MALLOC RECORDS POINTER");
//		}

	if (records_ptr)
		{
		for (int i=0;i<header.num_records;i++)
			{
			// data[i]'s memory is allocated within ReadRecord()
			records_ptr[i] = ReadRecord(i);

//HEAPCHK("RESOURCEFILE::ReadAllRecords()~~\
//Corrupted just after attempting to ReadRecord(%d)",i);

//			if (!records_ptr[i])
//				{
//				DBG(290,"ERROR: COULD NOT READ RECORD #%d",i);
//				}
			}
		}

	return records_ptr;
	}

// -------------------------------------------------------------------------
//
// FUNC:       FreeAllRecords()
// DESC:       free each record of a resource and then free the resource
//
// -------------------------------------------------------------------------
char** 		ResourceFile::FreeAllRecords(char** records_ptr)
	{
	if (ReadResourceHeader() != file_NoError) return records_ptr;

	if (records_ptr)
		{
		for (int i=header.num_records-1;i>=0;i--)
			{
			SYS_Free(records_ptr[i]);
			records_ptr[i] = NULL;
			}
		SYS_Free(records_ptr);
		records_ptr = NULL;
		}
	return records_ptr;
	}

// -------------------------------------------------------------------------
//
// FUNC:       WriteRecord()
// DESC:       writes a record
//
// -------------------------------------------------------------------------
int	ResourceFile::WriteRecord(long record_num,char* data,long size_of_data)
	{
	// read the resource index
	if (ReadRecordIndex(record_num) != file_NoError) return NULL;

	// create the buffer to compress to
	char* compression_buffer = (char*) SYS_Malloc(size_of_data);

	// skip out if we couldn't create it...
	if (!compression_buffer) return NULL;

	// compress the data
	long sizeofcompressed = CompressRecord(compression_buffer,data,size_of_data);

	// if the size is different, skip out
	if (sizeofcompressed != index.sizeofcompressed) { SYS_Free(compression_buffer); return NULL; }

	// seek to the start of the specified record
	if (Seek(index.location) != file_NoError) return NULL;

	// write the compressed data to the file
	if (Write(compression_buffer,index.sizeofcompressed) != file_NoError) return file_WriteFailed;

	// get rid of the buffer for the compressed data
	SYS_Free(compression_buffer);

	// everything went smoothly.... :-)
	return file_NoError;
	}

//      -------------------------------------------------------------------------
//
//      FUNC:           AppendResource
//      DESC:           append a resource to the resource file
//                              using compression if specified
//
//      -------------------------------------------------------------------------
int		ResourceFile::AppendRecord(char* data,long size_of_data)
	{
//	printf("IN APPEND RECORD()\n");

	// if we can't read the resource header (caJUSTse), skip out
//	printf("READRESOURCEHEADER() : %d\n",ReadResourceHeader());
	if (ReadResourceHeader() != file_NoError) return file_ReadFailed;

//	printf("HERE1\n");

	// make sure the resource isn't full already
	if (header.num_records >= header.max_records) return resource_OutOfRecords;

	// create space for the compressed data
	char* compression_buffer = (char*) SYS_Malloc(size_of_data*2);

	// skip out if we couldn't create it...
	if (!compression_buffer) return NULL;

	// compress the raw data
	long sizeofcompressed = CompressRecord(compression_buffer,data,size_of_data);

	// create the appropriate record_index data
	index.location = FileSize();
	index.sizeofcompressed = sizeofcompressed;
	index.sizeofdecompressed = size_of_data;

	// seek to the end of the file
	SeekEOF();

	// write the compressed data to the file
	if (Write(compression_buffer,index.sizeofcompressed) != file_NoError) return file_WriteFailed;

	// get rid of the buffer for the compressed data
	SYS_Free(compression_buffer);

	// write our record index
	WriteRecordIndex(header.num_records);

	// update our resource header
	header.num_records++;
	WriteResourceHeader();

	// everything went smoothly.... :-)
	return file_NoError;
	}

long	ResourceFile::CompressRecord(char* compressed,char* decompressed,long sizeofdecompressed)
	{
	// we wish to return the size of the decompressed record at most!!!
	long sizeofcompressed = sizeofdecompressed;

	switch (header.resource_type)
		{
		case	DATA_RESOURCE:		memcpy(compressed,decompressed,sizeofdecompressed);
									break;
		case	TEXT_RESOURCE:		memcpy(compressed,decompressed,sizeofdecompressed);
									break;
		case	BITMAP_RESOURCE:	sizeofcompressed = PCX_Compress(compressed,decompressed,sizeofdecompressed);
									break;
		case	FONT_RESOURCE:		sizeofcompressed = PCX_Compress(compressed,decompressed,sizeofdecompressed);
									break;
		case	PALETTE_RESOURCE:	memcpy(compressed,decompressed,sizeofdecompressed);
									break;
		case	SPRITE_RESOURCE:	sizeofcompressed = PCX_Compress(compressed,decompressed,sizeofdecompressed);
									break;
		case	SOUND_RESOURCE:		memcpy(compressed,decompressed,sizeofdecompressed);
									break;
		}
	return sizeofcompressed;
	}

long	ResourceFile::DecompressRecord(char* decompressed, char* compressed,long sizeofdecompressed)
	{
	switch (header.resource_type)
		{
		case	DATA_RESOURCE:		memcpy(decompressed,compressed,sizeofdecompressed);
									break;
		case	TEXT_RESOURCE:		memcpy(decompressed,compressed,sizeofdecompressed);
									break;
		case	BITMAP_RESOURCE:	PCX_Decompress(decompressed,compressed,sizeofdecompressed);
									break;
		case	FONT_RESOURCE:		PCX_Decompress(decompressed,compressed,sizeofdecompressed);
									break;
		case	PALETTE_RESOURCE:	memcpy(decompressed,compressed,sizeofdecompressed);
									break;
		case	SPRITE_RESOURCE:	PCX_Decompress(decompressed,compressed,sizeofdecompressed);
									break;
		case	SOUND_RESOURCE:		memcpy(decompressed,compressed,sizeofdecompressed);
									break;
		}

	return sizeofdecompressed;
	}

