//	-------------------------------------------------------------------------
//
//	RESOURCE.HPP
//
//	-------------------------------------------------------------------------
//
//	DESCRIPTION:	This file defines the Resource object which is used to
//					access a pseudo-random-access file containing variable
//					length records (resources)
//
//	-------------------------------------------------------------------------

#ifndef __resource_hpp
#define __resource_hpp

#include "file.hpp"

#include "stdio.h"
#include "io.h"

#define resource_InvalidIndexNum			-200
#define resource_RecordNotFound				-201
#define resource_RecordAlreadyUsed			-202
#define resource_OutOfRecords				-203
#define resource_CompressionMismatch		-204
#define resource_DiffRecordSize				-205
#define resource_HeapError					-206
#define resource_IndexCorrupt				-207

// the types of resources
#define DATA_RESOURCE				0	// used to store anything
#define TEXT_RESOURCE				1
#define BITMAP_RESOURCE			    2
#define FONT_RESOURCE				3
#define PALETTE_RESOURCE			4
#define SPRITE_RESOURCE				5
#define SOUND_RESOURCE				6

// the types of records
#define FONT_RECORD					1
#define BITMAP_RECORD				2
#define SPRITE_RECORD				3
#define PALETTE_RECORD				4
#define TEXT_RECORD					5
#define SOUND_RECORD				6

// the types of compression used for each record in a resource
#define NONE						0
#define PCX							1

// more examples 	#define TIFF	2
//					#define JPEG	3

// this is the header of a resource file
struct ResourceHeader
	{
	char			resource_type;	// the type of resource
	long 			max_records;	// the maximum number of records in the resource
	long			num_records;	// the number of records in the resource
	};

// this is the information for a specified record
struct RecordIndex
	{
	long 	location; 				// the absolute start location of the record
	long	sizeofcompressed;		// size of data on disk
	long	sizeofdecompressed;		// size of data in memory
	};

class ResourceFile : public File
	{
	public:
				ResourceFile(char* filename);
				ResourceFile(String* filename);
				ResourceFile(char* path,char* filename);
				ResourceFile(String* path,String* filename);

		virtual	~ResourceFile();

		// overloaded Create() function, constructs the resource!!!
		virtual	int		Create(long max_records = 256,char type_of_resource = DATA_RESOURCE);
		virtual	int		Open();

		// maintenance functions
		char			ResourceType() 	{ return header.resource_type; }
		long			NumRecords() { return header.num_records; }
		long			MaxRecords() { return header.max_records; }

		long			CompressedRecordSize(long record_num);
		long			DecompressedRecordSize(long record_num);

		// input/output functions
		char*			ReadRecord(long record_num);
		char*			ReadRecord(long record_num,char* buffer);

		char**			ReadAllRecords();
		char**			FreeAllRecords(char** records_ptr);

		int				WriteRecord(long  record_num,
									char* data,
									long  size_of_data);

		int				AppendRecord(char* data,
									 long  size_of_data);

	protected:
		ResourceHeader	header;
		RecordIndex		index;

		// whole resource file

		int		ReadResourceHeader();		// read the resource file header
		int		WriteResourceHeader();		// write the resource file header

		int 	ReadRecordIndex(long record_num);	// read a specified record's index
		int		WriteRecordIndex(long record_num);  // write a specified record's index

		long	CompressRecord(char* compressed,char* decompressed,long sizeofdecompressed);
		long	DecompressRecord(char* decompressed, char* compressed,long sizeofdecompressed);
	};

#endif
