//	-------------------------------------------------------------------------
//
//	SPRITES
//
//	-------------------------------------------------------------------------
//
//	AUTHOR:			Andrew C. Evans
//
//	DESCRIPTION:	This file declares the sprite functions used
//					by the game development library.
//
//	-------------------------------------------------------------------------

#include "sprites.hpp"
#include "grmisc.hpp"
#include "sys.hpp"

//	=========================================================================
//
//	SPRITE FUNCTIONS
//
//	=========================================================================
//
//	FUNC:		grCreateSprite
//	DESC:		SYS_Malloc all memory associated with a sprite and
//				initialise data
//
//	PARAMS:		number_frames - number of images to be stored as animation
//				width, height - width & height of each image
//
//	NOTE:
//			For an ANIMATION sequence a SPRITE after creation should
//			have a:
//					// to give it something for first time in loop
//					grSaveUnderSprite(spr);
//
//					while(!(spr->current_frame != spr->hdr.number_frames))
//						grEraseSprite(spr);
//						grMoveSprite(spr,+4,-8);
//						spr->current_frame++;
//						grSaveUnderSprite(spr);
//						grDrawSprite(spr);
//
//	-------------------------------------------------------------------------
//	FUNC:		grCreateSprite
//	DESC:		SYS_Malloc all memory associated with a sprite and
//				initialise data
//
//	PARAMS:		number_frames - number of images to be stored as animation
//				width, height - width & height of each image
//
//	-------------------------------------------------------------------------
grSprite* grCreateSprite(int number_frames, int width, int height)
	{
	// create the sprite
	grSprite* spr = (grSprite*) SYS_Malloc(sizeof(grSprite)+sizeof(grImage*)*number_frames);

	// create a space for the sprite background
	spr->background = (grImage*) SYS_Malloc(grImageSize(0,0,width-1,height-1));

	// create a space for the sprite frame list
//	spr->frame = (grImage**) SYS_Malloc(sizeof(grImage*)*number_frames);

	// create space for each sprite frame
	for (int i=0; i<number_frames; i++)
		spr->frame[i] = (grImage*) SYS_Malloc(grImageSize(0,0,width-1,height-1));

	// setup header
	spr->hdr.number_frames = number_frames;
	spr->hdr.width = width;
	spr->hdr.height = height;

	// initialise other data
	spr->hdr.x = 0;
	spr->hdr.y = 0;
	spr->hdr.oldx = 0;
	spr->hdr.oldy = 0;
	spr->hdr.current_frame = 0;
	spr->hdr.state = 0;

	// return spr
	return spr;
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		grCreateSprite
//	DESC:		will create a sprite then store a number of grImages
//				into that sprite
//
//	PARAMS:		imagelist - list of images
//				number_frames - how many frames
//
//	-------------------------------------------------------------------------
grSprite* grCreateSprite(grImage** imagelist, int number_frames)
	{
	int		width = imagelist[0]->hdr.width,
			height = imagelist[0]->hdr.height;
	long	frame_size = grImageSize(0,0,width,height);

	grSprite* spr = grCreateSprite(number_frames, width, height);

	if (spr)
		// copy each grImage into each of the sprites frames
		for (int i=0;i<number_frames;i++)
			{
			// if dimensions are the same as the first frames then
			if ((width == imagelist[i]->hdr.width) && (height == imagelist[i]->hdr.height))
				// copy image into frame
				SYS_Memcpy(spr->frame[i],imagelist[i],frame_size);
			else
				// or set the frame to zeros
				SYS_Memset(spr->frame[i],0,frame_size);
			}

	// return the sprite
	return spr;
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		grDestroySprite
//	DESC:		SYS_Free all memory associated with a sprite
//
//	-------------------------------------------------------------------------
void grDestroySprite(grSprite* spr)
	{
	// free each sprite->frame
	for (int i=0; i<spr->hdr.number_frames; i++)
		SYS_Free(spr->frame[i]);

	// free sprite->frame pointer list
	SYS_Free(spr->frame);

	// free sprite->background
	SYS_Free(spr->background);

	// free sprite
	SYS_Free(spr);
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		grEraseSprite
//	DESC:		will grPutImage the background over the sprite's position
//
//	-------------------------------------------------------------------------
void grEraseSprite(grSprite* spr)
	{
	int		x1 = spr->hdr.x,
			y1 = spr->hdr.y,
			x2 = x1 + spr->hdr.width-1,
			y2 = y1 + spr->hdr.height-1;

	// put background in old position
	grPutImage( x1,y1, x2,y2, spr->background );
	grBlitRegion( x1,y1,x2,y2 );
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		grMoveSprite
//	DESC:		saves sprite's old position then increases x & y by dx & dy
//
//	-------------------------------------------------------------------------
void	grMoveSprite( grSprite* spr, int dx, int dy )
	{
	spr->hdr.oldx = spr->hdr.x;
	spr->hdr.oldy = spr->hdr.y;

	spr->hdr.x += dx;
	spr->hdr.y += dy;
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		grBehindSprite
//	DESC:		will save background behind the sprite's position
//
//	-------------------------------------------------------------------------
void grSaveUnderSprite(grSprite* spr)
	{
	int		x1 = spr->hdr.x,
			y1 = spr->hdr.y,
			x2 = x1 + spr->hdr.width-1,
			y2 = y1 + spr->hdr.height-1;

	// get background
	grGetImage( x1,y1, x2,y2, spr->background );
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		grDrawSprite
//	DESC:		should grPutOverlay image of sprite
//				but right now just grPutImage of a merged image
//
//	-------------------------------------------------------------------------
void grDrawSprite(grSprite* spr)
	{
//	if (spr->current_frame >= spr->hdr.number_frames) return;
	int		x1 = spr->hdr.x,
			y1 = spr->hdr.y,
			x2 = x1 + spr->hdr.width-1,
			y2 = y1 + spr->hdr.height-1;

	grOverlayImage( x1,y1, x2,y2, grBLACK,spr->frame[ spr->hdr.current_frame ] );
	grBlitRegion( x1,y1,x2,y2 );
	}

