//	-------------------------------------------------------------------------
//
//	STRINGS.HPP
//
//	-------------------------------------------------------------------------
//
//	AUTHORS:		Darren J. Dwyer  &  Andrew C. Evans
//
//	DESCRIPTION:	This file contains code for a simple string class
//
//	-------------------------------------------------------------------------

#include "strings.hpp"
#include "math.h"

//	-------------------------------------------------------------------------
//
//	FUNC:		SetLength
//	DESC:       set the current length to len, if len is greater than
//				length then fill in the extra bytes with spaces
//
//	-------------------------------------------------------------------------
// set the current length to len
void String::SetLength(long len,bool fill_extra)
	{
	// bounds check
	CLIP1(len,0,maximumlength);

	// iff we are making str longer
	if ((len>length) && fill_extra)
		{
		// fill in extra with spaces
		memset(&str[length],' ',len-length);
		}

	// set new length
	length = len;

	// make sure there is a \0 at the end
	str[length] = 0x00;
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		Insert
//	DESC:       inserts a single char into the String.
//
//	PARAMS:		pos	- the position in the String to insert the char
//				ch	- the character to insert
//
//	NOTE:		This function is used as the basis of all Insert,
//				Set and Append functions.
//
//	-------------------------------------------------------------------------
void String::Insert(long pos,char ch)
	{
	// if string is not full
	if (length<maximumlength)
		{
		// bounds check
		CLIP1(pos,0,length);

		// if we are inserting before the last character
		if (pos<length)
			{
			// shift data at and after insert position to the right one char
			memmove(&str[pos+1],&str[pos],length-pos);
			}

		// insert the character
		str[pos] = ch;

		// increment the length
		length++;
		}

	// check length
	CLIP1(length,0,maximumlength);

	// make sure there is a \0 at the end
	str[length] = 0x00;
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		Insert
//	DESC:       inserts a series of chars into the String.
//
//	PARAMS:		pos	- the position in the string to insert the chars
//				st	- the char array to insert
//
//	-------------------------------------------------------------------------
void String::Insert(long pos,const char* st)
	{
	// if string is not full
	if (length<maximumlength)
		{
		// bounds check
		CLIP1(pos,0,length);

		// get length of st
		long allowable_insert_length = strlen(st);

		// if size of st will overflow length
		if (length+allowable_insert_length > maximumlength)
			{
			// get maximum size of st allowed
			allowable_insert_length = maximumlength - length;
			}

		// if we are actually inserting something
		if (allowable_insert_length > 0)
			{
			// if we are inserting before the last character
			if (pos<length)
				{
				// shift data at and after insert position to the right by allowed length
				memmove(&str[pos+allowable_insert_length],&str[pos],length-pos);
				}

			// make sure we aren't copying onto ourself
			if (st != str)
				// insert the characters
				memcpy(&str[pos],&st[0],allowable_insert_length);

			// increase the length by the allowable insert length
			length += allowable_insert_length;
			}
		}

	// check length
	CLIP1(length,0,maximumlength);

	// make sure there is a \0 at the end
	str[length] = 0x00;
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		Insert
//	DESC:       inserts a number into the String.
//
//	PARAMS:		pos	- the position in the string to insert the chars
//				number - the number to insert
//
//	-------------------------------------------------------------------------
void String::Insert(long pos,long number)
	{
	long work_number = number;
	long stpos = 0;

	// create a temp char array to put the number into
	char* st = (char*) SYS_Malloc(16); if (!st) return;

	// check for negative number
	if (number<0)
		{
		// insert a minus sign and move along
		st[stpos++] = '-';

		// get positive work number
		work_number = -work_number;
		}

	// this will remain true until we find first significant digit
	bool leadingzero = true;
	long tmp = 100000000L;

	// convert number to string, including zeros
	while (tmp!=0)
		{
		// get the next digit (0 <= work_number/tmp <= 9)
		char digit = 48 + work_number/tmp;

		// if digit was not 0 then we have gotten past the last useless 0
		if (digit!=48) leadingzero = false;

		// if not dealing with useless 0s
		if (!leadingzero)
			{
			// insert the digit and move along
			st[stpos++] = digit;
			}

		// and discard the last digit
		work_number = work_number - ((work_number/tmp) * tmp);
		tmp = tmp / 10;
		}

	// if blank string, set to "0"
	if (stpos==0)
		{
		// insert a 0 and move along
		st[stpos++] = '0';
		}

	// make sure there is a \0 at the end
	st[stpos] = 0x00;

	// use normal char array insert
	Insert(pos,st);

	// free temporary char array
	free(st);
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		Delete
//	DESC:       deletes a char from within a String
//
//	PARAMS:		pos			- the position to delete from
//
//	-------------------------------------------------------------------------
void String::Delete(long pos)
	{
	// if string is not empty
	if (length>0)
		{
		// bounds check
		CLIP1(pos,0,length-1);

		// if we are deleting before the last character
		if (pos<length-1)
			{
			// shift data at and after insert position to the right one char
			memmove(&str[pos],&str[pos+1],length-1-pos);
			}

		// decrement the length
		length--;
		}

	// check length
	CLIP1(length,0,maximumlength);

	// make sure there is a \0 at the end
	str[length] = 0x00;
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		Delete
//	DESC:       deletes chars from within a String
//
//	PARAMS:		pos			- the position to delete from
//				numchars	- the number of chars to delete
//
//	-------------------------------------------------------------------------
void String::Delete(long pos,long numchars)
	{
	// if string is not empty
	if (length>0)
		{
		// bounds check
		CLIP1(pos,0,length-1);
		CLIP1(numchars,0,length-pos);

		// get number of characters after the delete region
		long characters_after_delete = length - pos - numchars;

		// if we have to move characters from after delete region to delete position
		if (characters_after_delete > 0)
			{
			// shift data after end of delete position to delete position
			memmove(&str[pos],&str[pos+numchars],characters_after_delete);

			// decrement the length
			length -= numchars;
			}
		else
			{
			// set length to "deletion" point
			length = pos;
			}
		}

	// check length
	CLIP1(length,0,maximumlength);

	// make sure there is a \0 at the end
	str[length] = 0x00;
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		StringToInteger
//	DESC:       converts a string to an integer value, ignoring whitespace
//				eg. "1hello2  3->4.#5,67890" -> `-1234567890'
//
//	RETURNS:	long	- the value of the string
//
//	-------------------------------------------------------------------------
long String::StringToInteger()
	{
	bool minus = false;
	long tmp = 0;
	long cnt = 0;

	// step through whole string
	while (cnt<length)
		{
		// if we find a minus sign anywhere in string
		if (str[cnt]==45) minus = true;

		// if we find a digit
		if ((str[cnt]>=48)&&(str[cnt]<=57))
			{
			// increase temp by factor of 10
			// and put the last digit into the least significant position
			tmp = tmp*10 + (str[cnt]-48);
			}
		// move along
		cnt++;
		}
	if (minus) tmp = -tmp;

	return tmp;
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		StringToInteger
//	DESC:       converts a string to an integer value, ignoring whitespace
//				but only when the value is between min & max.
//
//	PARAMS:		min	- the minimum value
//				max - the maximum value
//
//	RETURNS:	long- the value of the string
//
//	-------------------------------------------------------------------------
long String::StringToInteger(long min,long max)
	{
	long tmp = StringToInteger();

	// if parameters are passed in correct order
	if (min<max)
		{
		// bounds check
		if (tmp<min) tmp=min;
		if (tmp>max) tmp=max;
		}
	else
		{
		// bounds check
		if (tmp<max) tmp=max;
		if (tmp>min) tmp=min;
		}

	return tmp;
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		ScanRight
//	DESC:       scans for a char from a position in the string to the end
//				of the string
//
//	PARAMS:		pos		- the position to search from
//				token	- the char we are searching for
//
//	RETURNS:	long	- the position that the token was found
//							return length if not found
//
//	-------------------------------------------------------------------------
long String::ScanRight(long pos,char tok)
	{
	// bounds check
	if ((length==0) || (pos>=length)) return length;

	// repeat until token is found
	while ((str[pos]!=tok) && (pos<length)) pos++;

	// then return position of first token found after pos
	return pos;
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		ScanLeft
//	DESC:       scans for a char from a position in the string to the start
//				of the string
//
//	PARAMS:		pos		- the position to search from
//				token	- the char we are searching for
//
//	RETURNS:	long	- the position that the token was found
//						  return -1 if not found
//
//	-------------------------------------------------------------------------
long String::ScanLeft(long pos,char tok)
	{
	// bounds check
	if (length==0) return -1;

	// our counter is set to 0 if pos<0 or pos+1 otherwise
	if (pos>length-1) pos = length-1;

//	long cnt = (pos>length-1) ? length-1 : pos-1;

	// repeat until token is found
	while ((str[pos]!=tok)&&(pos>-1)) pos--;

	// then return position of first token found after pos
	return pos;
	}

bool	String::Contains(char ch)
	{
	return (ScanRight(0,ch)!=length);
	}

bool	String::Contains(char* st)
	{
	int stpos = 0;

	// find the first char
	int pos = ScanRight(0,st[stpos]);

	// and skip out if we didn't find it
	if (pos==length) return false;

	// do quick comparison
	while ((str[pos]==st[stpos]) && (st[stpos]!=0x00) && (pos<length))
		{
		pos++; stpos++;
		}

	// return true we found the end of 'st'
	return (st[stpos]==0x00);
	}



//	-------------------------------------------------------------------------
//
//	FUNC:		GetUpper
//	DESC:       converts a string to uppercase
//
//	RETURNS:	char*	- created char* of String's string in uppercase
//
//	-------------------------------------------------------------------------
String* String::GetUpper()
	{
	if (length==0) return NULL;

	String* tmpstr = new String(this);

	tmpstr->ToUpper();

	return tmpstr;
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		GetLower
//	DESC:       converts a string to lowercase
//
//	RETURNS:	char*	- created char* of String's string in lowercase
//
//	-------------------------------------------------------------------------
String* String::GetLower()
	{
	if (length==0) return NULL;

	String* tmpstr = new String(this);

	tmpstr->ToLower();

	return tmpstr;
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		SubStr
//	DESC:       returns a String containing a substring of the string
//
//	PARAMS:		start	- the starting char
//				end		- the end char
//
//	RETURNS:	String*	- a new String containing the substring
//
//	-------------------------------------------------------------------------
String* String::SubStr(long start,long end)
	{
	// make sure we're using real input (caJUSTse)
	CLIP2(start,end,0,0,length-1,length-1);

	long len = end-start+1;

	String* tmp;

	if (len>0)
		{
		tmp = new String("",len);
		long pos = 0;
		while (pos<len)
			tmp->Contents()[pos++] = str[start++];
		tmp->SetLength(len);
		}
	else
		tmp = new String("");

	return tmp;
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		NumElems
//	DESC:       returns the number of elements in the string delimeted by
//				a char token.
//				eg. String* cal = {"JAN/FEB/MAR/APR/.../DEC"}
//					NumElems("/") == 12
//
//	PARAMS:		token	- the token which is used to differentiate elems
//
//	RETURNS:	long	- the number of elements found in the string
//
//	-------------------------------------------------------------------------
long String::NumElems(char tok)
	{
	if (length<1) return 0;

	long tokcnt = 0;
	long strptr = 0;

	// scan through the string counting tokens
	do
		{
		if (str[strptr++]==tok) tokcnt++;
		}
	while (str[strptr]);

	// increment for last element
	tokcnt++;

	// and return the number of tokens found + 1 --> number of elements
	return tokcnt;
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		Elem
//	DESC:       returns a String containing a specified element.
//				eg. String* cal = {"JAN/FEB/MAR/APR/.../DEC"}
//					String* jan = cal->Elem(0,"/");
//					jan == "JAN"
//
//	PARAMS:		n		- the element number to return
//				token	- the token which is used to differentiate elems
//
//	RETURNS:	String*	- a new String containing the element
//
//	-------------------------------------------------------------------------
String* String::Elem(long n,char tok)
	{
	long numelems = NumElems(tok);
	String* tmpstr = NULL;

	if (length<1) return NULL;

	long start;
	long end;

	// if we have an appropriate element!!!
	if ((n>=0)&&(n<numelems))
		{
		// if we are looking for the first element, simple
		if (n==0)
			{
			start = 0;
			end = ScanRight(start,tok) - 1;
			}
		else
			// similar for the last element
			if (n==(numelems-1))
				{
				end = length-1;
				start = ScanLeft(end,tok) + 1;
				}
			else
				// scan through the string from the first position
				// counting tokens as we go
				{
				long pos = -1;
				long tokcnt = 0;

				while (tokcnt<n)
					{
					pos = ScanRight(pos+1,tok);
					tokcnt++;
					}
				start = pos + 1;
				end = ScanRight(start,tok) - 1;
				}
		// create & return appropriate string
		tmpstr = SubStr(start,end);
		}
	else
		{
		// return a NULL string!
		tmpstr = new String("");
		}

	return tmpstr;
	}
