//	-------------------------------------------------------------------------
//
//	TIMER.CPP
//
//
//	-------------------------------------------------------------------------
//
//	AUTHOR:			Darren J. Dwyer
//
//	DESCRIPTION:	This file contains code for a timer ISR which
//					generates events into the EVENTQUEUE.
//
//					The ISR is pretty simple, simply incrementing a counter
//					in a number of different 'tickers'. Upon this counter
//					reaching a maximum value, the counter resets & an event
//					is generated into the EventQueue.
//
//					By doing this we can generate specific events
//					(such as cmdAnimate) at either a fixed rate or an
//					on-the-fly (user-specified) rate.
//
//					NOTE: * currently uses timer0 and updates the system clock
//						    due to inability to set rate for timer1 & timer2
//
//						  * When setting the system timer to 60Hz and then
//					  		exiting without resetting to 18.2Hz, XTREEGOLD
//							seems to run _much_ faster, BORLANDC _much_
//							slower... :-) 			.curious.
//
//	-------------------------------------------------------------------------

#include "timer.hpp"
#include "event.hpp"
#include "eventmgr.hpp"
#include "mouse.hpp"

#include "stdlib.h"
#include "conio.h"
#include "stdio.h"
#include "dos.h"

#define gotoxy(X,Y)			printf("\n")
#define clrscr()			{}

// interrupt & port numbers to access the timer hardware

#define		TIMER_INT			0x08	// the timer interrupt #
#define		INT_CONTROL			0x20   	// the interrupt controller
#define		PIT_FREQ		0x1234DD	// the P.I.T. runs at 0x1234DD Hz

#define		TIMER_0				0x40	// counter 0 -- system timer/disk
#define		TIMER_1				0x41	// counter 1 -- DON'T USE?!!! memory refresh
#define		TIMER_2				0x42	// counter 2 -- tape drive/speaker
#define		TIMER_CONTROL		0x43	// the 8254's control register

// defines for the timer settings

#define		TIMER0_CONTROL_WORD	0x34	// timer 0, mode 2, binary, least/most

// macros to return the LO & HI bytes of a word, as bytes

#define		LO_BYTE(n)			(n & 0x00FF)
#define		HI_BYTE(n)			((n>>8) & 0x00FF)

// --------------------------------------------------------------------------
//
// GLOBAL data used in the timer ticker functions
//
// --------------------------------------------------------------------------

dostime_t	system_time;

// storage of previous timer ISR (Interrupt Service Routine)

#ifdef __cplusplus
#ifdef __WATCOMC__
#define	CPPARGS
#else
#define CPPARGS		...
#endif
#else
#define	CPPARGS
#endif

void _interrupt (_far *Old_Timer_Isr)(CPPARGS);

// storage of our timer event
TimerEvent	timer_EventInfo;

// internal flags

bool	timer_HANDLER_INSTALLED = false;	// a flag to detect if installed
bool	timer_TICKING_ON = false;			// a flag to toggle ticking on/off

// the structure used for each individual timer ticker

struct tickertype
	{
	int		event;			// the event to place in the queue upon completion
	int		max_ticks;		// the maximum number of ticks for the ticker
	uint	tick_counter;	// the tick counter.
	bool	ticking;		// whether or not we are in fact ticking!
	};

// storage area for the ticker bits & pieces

tickertype	ticker[TIMER_MAX_TICKERS];	// storage of the timer tickers
int			timer_numTickers = 0;			// the number of tickers active
int			active_ticker = 0;			// a counter used by the ISR.

// variables used by isr...
long		clock_ticks,counter;

//	-------------------------------------------------------------------------
//
//	FUNC:		timer_ChangeInterruptRate()
//	DESC:		sets the hardware timer to generate interrupts at 'rate' Mhz
//				eg. timer_60Mhz
//
//	-------------------------------------------------------------------------
void timer_ChangeInterruptRate( int rate )
	{
	// set up the rate of chaining to BIOS interrupt handler.... time functions
	clock_ticks = 0;
	counter = PIT_FREQ / rate;

	// send the control word
	outp(TIMER_CONTROL, TIMER0_CONTROL_WORD);
	// write lo byte of the new rate to the counter register
	outp(TIMER_0, counter % 256);
	// write hi byte if the new rate to the counter register
	outp(TIMER_0, counter / 256);
	}


//	-------------------------------------------------------------------------
//
//	FUNC:		timer_AddTicker()
//	DESC:		Adds a new timer ticker to the list of active tickers
//
//	PARAMS:		nticks 	- the number of ticks before the event is generated
//				event  	- the timer event which is generated
//
//	RETURNS:	int		- a handle which may later be used to modify the
//						  state of the ticker. -1 == too many tickers
//
//	-------------------------------------------------------------------------
int timer_AddTicker( int nticks, int event )
	{
	// if we haven't run out of timers
	if (timer_numTickers<TIMER_MAX_TICKERS)
		{
		// turn off timing (temporarily)
		timer_TICKING_ON = false;

		// set our new timer data
		ticker[timer_numTickers].tick_counter = 0;
		ticker[timer_numTickers].max_ticks = nticks;
		ticker[timer_numTickers].event = event;
		ticker[timer_numTickers].ticking = true;

		// increment our max. index
		timer_numTickers++;

		// and turn timing back on
		timer_TICKING_ON = true;

		// and return our timer number
		return timer_numTickers - 1;
		}
	else
		// return an 'invalid' flag
		return -1;
	}


//	-------------------------------------------------------------------------
//
//	FUNC:		timer_DelTicker()
//	DESC:		removes a specified timer ticker from the ticker list
//
//	PARAMS:		ticker	- the ticker handle (from AddTicker()) to delete
//
//	-------------------------------------------------------------------------
void timer_DelTicker( int zticker )
	{
	// skip out if stupid info
	if (zticker>=timer_numTickers) return;

	// turn the timer off (temporarily)
	timer_TICKING_ON = false;

	// start at our first timer
	int pos = 0;
	int pos1 = 0;

	// skip till we find the appropriate timer
	while (pos<zticker)
		pos++;

	// and simply shuffle the rest of the timers!
	do
		{
		pos1 = pos + 1;	// for efficiency
		ticker[pos].tick_counter = ticker[pos1].tick_counter;
		ticker[pos].max_ticks = ticker[pos1].max_ticks;
		ticker[pos].event = ticker[pos1].event;

		pos++;
		}
	while (pos<timer_numTickers);

	// we now have one less timer
	if (timer_numTickers>0) timer_numTickers--;

	// turn timing back on
	timer_TICKING_ON = true;
	}


//	-------------------------------------------------------------------------
//
//	FUNC:		timer_SetTicker()
//	DESC:		modifies a specified ticker, either speeding it up or slowing
//				it down.
//
//	PARAMS:		ticker	- the ticker to set (returned from timer_AddTicker())
//				nticks	- the number of ticks before the event is generated
//
//	-------------------------------------------------------------------------
void timer_SetTicker( int zticker, int nticks )
	{
	// skip out if stupid input
	if (zticker>=timer_numTickers) return;

	// and set the timer
	ticker[zticker].tick_counter = 0;
	ticker[zticker].max_ticks = nticks;
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		timer_ResetTicker()
//	DESC:		resets a specified ticker to 0
//
//	-------------------------------------------------------------------------
void timer_ResetTicker( int zticker )
	{
	// skip out if stupid
	if (zticker>=timer_numTickers) return;

	// reset the timer's counter
	ticker[zticker].tick_counter = 0;
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		timer_StopTicker()
//	DESC:		stops a specified timer ticker from ticking
//
//	-------------------------------------------------------------------------
void timer_StopTicker( int zticker )
	{
	// skip out if stupid
	if (zticker>=timer_numTickers) return;

	// stop the ticker
	ticker[zticker].ticking = false;
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		timer_StartsTicker()
//	DESC:		starts a specified timer ticker ticking
//
//	-------------------------------------------------------------------------
void timer_StartTicker( int zticker )
	{
	// skip out if stupid
	if (zticker>=timer_numTickers) return;

	// start the ticker
	ticker[zticker].ticking = true;
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		timer_GetEvent()
//	DESC:		checks if there are any timer events requiring generation
//				and if so places the appropriate events into the event queue.
//
//	-------------------------------------------------------------------------
void timer_GetEvent()
	{
	if (timer_numTickers==0) return;

	// define our event
	timer_EventInfo.time = 0;

	// start with the first timer
	active_ticker = 0;

	// and update all tickers, generating events if required
	while (active_ticker<timer_numTickers)
		{
		// and if it has reached its threshold
		if (ticker[active_ticker].tick_counter>ticker[active_ticker].max_ticks)
				{
				// define our event
				timer_EventInfo.event = ticker[active_ticker].event;
				// if this event isn't in the queue
				if (!EVENTMGR->EventInQueue(ev_timer,&timer_EventInfo))
					{
					// generate an event
					EVENTMGR->GenEvent(0,ev_timer,&timer_EventInfo);
					// and reset the timer
					ticker[active_ticker].tick_counter = 0;
					}
				}

		// try the next timer
		active_ticker++;
		}
//	HEAPCHK("by timer_GetEvent()");
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		timer_ISR()
//	DESC:		This is our timer interrupt function.
//				For simplicity, this function simply scans through all active
//				"timers" and increments their counters. If the ticks are all
//				completed an event is generated into the event queue!
//
//	-------------------------------------------------------------------------
void _interrupt _far timer_ISR(CPPARGS)
	{
	// disable interrupts while in ISR
//	_disable();

	// if we are worrying about ticking...
	if (timer_TICKING_ON)
		{
		// start with the first timer ticker
		active_ticker = 0;

		// and update all tickers, generating events if required
		while (active_ticker<timer_numTickers)
			{
			// if the ticker is ticking
			if (ticker[active_ticker].ticking)
				// increment the timer's counter
				ticker[active_ticker].tick_counter++;

			// try the next timer
			active_ticker++;
			}
		}

	// adjust the count of clock ticks
	clock_ticks = clock_ticks + counter;

	// is it time to call the BIOS handler??
	if (clock_ticks >= 0x10000)
		{
		// adjust our clock ticks
		clock_ticks -= 0x10000;

//		_enable();

		// chain to BIOS handler
		Old_Timer_Isr();
		}
	else
		{
		// re-enable interrupts
//		_enable();

		outp(0x20,0x20);	// signal end-of-interrupt to hardware
		}
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		timer_InstallHandler()
//	DESC:		installs our timer ISR. Must call timer_RemoveHandler() upon
//				program completion
//
//	-------------------------------------------------------------------------
void timer_InstallHandler()
	{
	if (!timer_HANDLER_INSTALLED)
		{
		// set up the rate of chaining to BIOS interrupt handler.... time functions
		clock_ticks = 0;
		counter = PIT_FREQ / timer_60Hz;

		// save the current system time
		_dos_gettime(&system_time);

		// save the old timer ISR
		Old_Timer_Isr = _dos_getvect( TIMER_INT );

		// and set our own timer ISR
		_dos_setvect( TIMER_INT, timer_ISR );

		// by default, set the rate to 60Hz
		timer_ChangeInterruptRate(timer_60Hz);

		// set a flag to say we are installed
		timer_HANDLER_INSTALLED = true;
		}
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		timer_RemoveHandler()
//	DESC:		removes our timer ISR, restoring the previous handler
//
//	-------------------------------------------------------------------------
void timer_RemoveHandler()
	{
	if (timer_HANDLER_INSTALLED)
		{
		// restore the normal clock frequency
		outp(TIMER_CONTROL, TIMER0_CONTROL_WORD);
		outp(TIMER_0, 0 );
		outp(TIMER_0, 0 );

		// restore the old timer ISR
		_dos_setvect( TIMER_INT, Old_Timer_Isr );

		// and set an appropriate flag
		timer_HANDLER_INSTALLED = false;
		}
	HEAPCHK("by timer_RemoveHandler()");
	}
/*

//#define cmdAnimate		5000
#define	cmdComputerAI	5001

#define evh_computer	6000

void main()
	{
	bool quit = false;

	clrscr();
	timer_InstallHandler();

	timer_AddTicker(50,cmdAnimate);	// animate every 50 ticks
	timer_AddTicker(100,cmdComputerAI);		// process computer AI every 100 ticks

	gotoxy(1,1);
	printf("Animation  ticker == %u\n",ticker[0].max_ticks);
	printf("ComputerAI ticker == %u\n",ticker[1].max_ticks);

//	timer_ChangeInterruptRate(timer_100Hz);

	uint iter = 0;
	uchar quitsec = system_time.second + 10;
	if (quitsec>59) quitsec -= 60;

	dostime_t	t;

	uint ticker0_cnt = 0;
	uint ticker1_cnt = 0;

	while (!quit)
		{
		timer_GetEvent();

		printf(".");
//		iter++;
		quit = (system_time.second==quitsec);

		if (timer_EventInfo.event==cmdAnimate)  { ticker0_cnt++; timer_EventInfo.event = 0; }
		if (timer_EventInfo.event==cmdComputerAI) { ticker1_cnt++; timer_EventInfo.event = 0; }

//		gotoxy(1,5);
//		printf("DOS: %.2u:%.2u.%.2u",t.hour,t.minute,t.second);
//		gotoxy(1,6);
//		printf("PRG: %.2u:%.2u.%.2u",system_time.hour,system_time.minute,system_time.second);

//		gotoxy(28,1);
//		printf("cmdAnimate: %u,    cmdCompAI: %u   ",ticker[0].tick_counter,ticker[1].tick_counter);
//		gotoxy(28,2);
//		printf("%u   ",ticker[1].tick_counter);

//		if (any) delay(1000);

		if (kbhit())
			{
			switch (getch())
				{
				case	27:;	// ESC
				case	'q':;
				case	'Q':		quit = true; break;

				case	'1':		timer_ChangeInterruptRate(timer_18Hz);
									gotoxy(40,2); printf("Timer @ 18Hz ");
									break;
				case	'2':		timer_ChangeInterruptRate(timer_20Hz);
									gotoxy(40,2); printf("Timer @ 20Hz ");
									break;
				case	'3':		timer_ChangeInterruptRate(timer_30Hz);
									gotoxy(40,2); printf("Timer @ 30Hz ");
									break;
				case	'4':		timer_ChangeInterruptRate(timer_40Hz);
									gotoxy(40,2); printf("Timer @ 40Hz ");
									break;
				case	'5':		timer_ChangeInterruptRate(timer_50Hz);
									gotoxy(40,2); printf("Timer @ 50Hz ");
									break;
				case	'6':		timer_ChangeInterruptRate(timer_60Hz);
									gotoxy(40,2); printf("Timer @ 60Hz ");
									break;
				case	'7':		timer_ChangeInterruptRate(timer_70Hz);
									gotoxy(40,2); printf("Timer @ 70Hz ");
									break;
				case	'8':		timer_ChangeInterruptRate(timer_80Hz);
									gotoxy(40,2); printf("Timer @ 80Hz ");
									break;
				case	'9':		timer_ChangeInterruptRate(timer_90Hz);
									gotoxy(40,2); printf("Timer @ 90Hz ");
									break;
				case	'0':		timer_ChangeInterruptRate(timer_100Hz);
									gotoxy(40,2); printf("Timer @ 100Hz");
									break;
				case	'Z':		;
				case	'z':		timer_ChangeInterruptRate(timer_ZOOM);
									gotoxy(40,2); printf("Timer @ ZOOM");
									break;
				default		:;
				}
			}
		}

	printf("\nticker0 cnt == %d",ticker0_cnt);
	printf("\nticker1 cnt == %d",ticker1_cnt);

	timer_RemoveHandler();
	}

*/