//	-------------------------------------------------------------------------
//
//	VESADRV.CPP
//
//	-------------------------------------------------------------------------
//
//	AUTHOR:			Darren J. Dwyer
//
//	DESCRIPTION:	This is the code for the VESA 1.2 driver
//
//	-------------------------------------------------------------------------

#include "vesadrv.hpp"
#include "stdio.h"
#include "conio.h"
#include "dos.h"

//DJD --- TEMPORARILY TILL HARDWARE CURSOR INTERFACE
#include "mouse.hpp"
//DJD --- END

#include "graphics.hpp"
#include "realmode.hpp"

//#define __debugging
//#define	__display_banks

// --------------- internal vars that may be required by externals (ie. mouse)

int		VESA_BankMask;		// the size of each VESA bank -1
int		VESA_BankRotator;	// bank rotator for diff. granularity sizes

bool	VESA_WaitForRetrace;// true--> waiting for retrace
long	VESA_RowPtr[1024];	// addresses for the start of each row
int		VESA_RowBank[1024];	// bank numbers for the start of each row


WORD	VESA_Read_Window; 		// either window A or window B
WORD	VESA_Write_Window;		// either window A or window B
bool	VESA_Using_Same_Window;	// window A is same as window B

int		VESA_Current_Read_Bank = -1; // the current read bank
int		VESA_Current_Write_Bank = -1;// the current write bank
int		VESA_Read_Bank_Offset = 0;	 // bank offset for reading multiple pages
int		VESA_Write_Bank_Offset = 0;	 // bank offset for writing multiple pages

int		VESA_Max_Page = 1;	   		 // the maximum number of display pages
int		VESA_Banks_Per_Page = 8;	 // the number of banks used for 1 page

int		VESA_Visual_Page = 0;		 // the current visual page
int		VESA_Active_Page = 0;		 // the current active page

long 	VESA_Video_Addr	= 0x000A0000L;	// video memory absolute start location
long	VESA_Video_Mask = 0x000AFFFFL;	// mask for video-memory ranges

bool	VESA_Flicker_Free = false;	// disable/enable flicker-free
									// bit blitting

#ifdef __debugging
bool	VESA_Debug_Finished = false; // used in debugging
#endif

// ---------------

VESA_INFO_BLOCK 	VBEInfo;	// storage of vesa information
VESA_MODE_BLOCK		ModeInfo;	// storage of vesa mode information

// --------------- default mode data

struct  modedata
	{
	int			xres; 				// ie. 640x480x256 = 640
	int			yres;				// 				   = 480
	int			colors;             //                 = 256
	WORD		std_mode;			// ie.			   = 0x0101	(std vesa mode)
	};

modedata		VESA_Mode_Data[2] =
				{ 640, 400, 256, 0x0100,
				  640, 480, 256, 0x0101 };

int modeNum = 0;

// =========================================================================
//
// VESA DRIVER PUBLIC FUNCTIONS
//
// =========================================================================
//
// 	FUNC:	grInitVesa()
// 	DESC:	Initialises the graphics display, installs the default SYSTEM
//			font and installs the default palette
//
// 	PARAMS:	mode	- the video mode to set. (ie. gr640x480x256)
//
// 	--------------------------------------------------------------------------
bool grInitVesa(int mode)
	{
#ifdef __debugging
	if (!VESA_Debug_Finished) printf("Checking for VESA BIOS\n");
#endif

	// return if there is no VESA bios installed in system
	if (!VESA_BiosInstalled())
		{
		printf("Note: (1) This program is not able to run correctly as you do not have\n");
		printf("          a VESA BIOS Extension v2.0 or greater installed.\n\n");
		printf("      (2) Scitech Software's UniVBE 5.1a is a shareware VESA driver\n");
		printf("          that should enable you to run this program.\n\n");
		printf("               try: www.scitechsoft.com\n");
		exit(1);
		}

	// set the mode
	if (!VESA_SetMode(mode))
		{
		printf("ERROR: Requested video mode was found but could not be set!  Oops!\n");
		getch();
		exit(1);
		}

	// create our double-buffer
	grSCREEN = grCreateVirtual(0,0,grMaxx,grMaxy,0);
	if (grSCREEN==NULL)
		{
		VESA_ResetMode(grOrigVideoMode);
		printf("ERROR: Could not create double-buffer!\n");
		getch();
		exit(1);
		}

#ifdef __debugging
	VESA_Debug_Finished = true;
	getch();
#endif

	return true;
	}

// --------------------------------------------------------------------------
//
//	FUNC:	grExitVesa
//	DESC:	terminates the display, resetting as before grInitVesa
//
// --------------------------------------------------------------------------
void grExitVesa()
	{
	// remove our double-buffer
	grDestroyVirtual(grSCREEN);

	// clear the video card's memory
	grClearVesa();

	// and reset video mode (to default startup mode)
	VESA_ResetMode(grOrigVideoMode);
	}

// --------------------------------------------------------------------------
//
//	FUNC:	grClearVesa
//	DESC:	clears all banks of the current VESA compatible device,
//			optionally setting the color to 'color'
//
// --------------------------------------------------------------------------
void grClearVesa(int color)
	{
	bool hide = mouse_IsInstalled();
	if (hide) mouse_Hide();

	// create DWORD mask for color
	DWORD fill = color + (color<<8) + (color<<16) + (color<<24);

	DWORD* p;
	int banknum;
	int dwords;

	// for all available banks...
	for (banknum=0;banknum<grNumberOfBanks;banknum++)
		{
		VESA_SwitchWriteBank(banknum);

		// blit DWORDS...
		p = (DWORD*) VESA_Video_Addr;
		for (dwords=0;dwords<16384;dwords++)
			*p++ = fill;
		}

	if (hide) mouse_Show();
	}

// --------------------------------------------------------------------------
//
//	FUNC:	VESA_CheckSpeed()
//	DESC:	Checks the raw graphics speed on the current machine
//			and saves the results for later use...
//
// --------------------------------------------------------------------------
void VESA_CheckSpeed()
	{
	dostime_t 	start_t,end_t;
	DWORD*		p;
	long		dwords,iter = 0;
	int			start_bank,end_bank;

	start_bank = VESA_Write_Bank_Offset;
	end_bank   = start_bank + VESA_Banks_Per_Page;

	// time the video hardware

	_dos_gettime(&start_t);

	for (int bank=start_bank;bank<end_bank;bank++)
		{
		VESA_SwitchWriteBank(bank);

		// fill the write bank with 0
		p = (DWORD*) VESA_Video_Addr;
		for (dwords=0;dwords<16384;dwords++) *p++ = 0;
		}

	_dos_gettime(&end_t);

	// figure out the difference in time

	long start_time = start_t.hour * 360000L +
					  start_t.minute * 6000L +
					  start_t.second * 100L +
					  start_t.hsecond;
	long end_time   = end_t.hour * 360000L +
					  end_t.minute * 6000L +
					  end_t.second * 100L +
					  end_t.hsecond;

	long diff = end_time - start_time;

	// convert this to Hz...

	if (diff>0) grVideoSpeed = 100 / diff;
	else grVideoSpeed = 1;

	// set a flag if we need to sync to the vertical retrace (50Hz or so)

	VESA_WaitForRetrace = (grVideoSpeed>50);
/*
#ifdef __debugging
	if (!VESA_Debug_Finished)
		{
		VESA_ResetMode(grOrigVideoMode);

		printf("\n\nChecking VESA raw speed\n\n");
		printf("\nNumber of banks cleared = %d\n\n",grNumberOfBanks);
		printf("START TIME == %.2d:%.2d:%.2d.%.2d\n",start_t.hour,start_t.minute,start_t.second,start_t.hsecond);
		printf("END   TIME == %.2d:%.2d:%.2d.%.2d\n",end_t.hour,end_t.minute,end_t.second,end_t.hsecond);
		printf("DIFF  TIME == %.2d:%.2d:%.2d.%.2d\n",
				diff/360000L,(diff/6000)%60,(diff/100)%60,diff%100);

		printf("\nVesa Hardware Speed == %d Hz\n",grVideoSpeed);
		if (VESA_WaitForRetrace) printf("Waiting for Retrace!\n");
		else printf("Skipping wait for Retrace!\n");

		getch();

		VESA_Debug_Finished = true;
		VESA_SetMode(mode);
		}
#endif
*/	}

// --------------------------------------------------------------------------
//
//	FUNC:	VESA_BiosInstalled()
//	DESC:	checks to see if there is a VESA 1.2 or better driver installed.
//			If there is a driver, we then get all required information
//			and return true. Otherwise we return false.
//
//	NOTE:	if (#defined __debugging) display vesa information
//
// --------------------------------------------------------------------------
bool VESA_BiosInstalled()
	{
	SREGS 		sregs;
	REGS 		regs;
	short 		selector,
				segment;
	char far*	dos_block;

	// set up video bios call 0x4F00 - Get SuperVGA Information
	memset(&sregs,0,sizeof(sregs));
	memset(&regs,0,sizeof(regs));

	// allocate DOS memory to store the VESA mode information
	dos_block = (char far*) RM_malloc(&selector,&segment,sizeof(VBEInfo));

	// set the VBE Signature to VBE2 as this is the way to get VBE 2.0
	// protected-mode extensions...
	strncpy((char*) VBEInfo.Signature,"VBE2",4);

	// and copy from pmode to DOS memory
	_fmemcpy(dos_block,&VBEInfo,sizeof(VBEInfo));

	// make ES:(E)DI point to our DOS memory
	sregs.es = segment;
	regs.x.edi = 0;
	regs.w.ax = 0x4F00;

	// and get the VESA bios information
	RM_int86x(0x10,&regs,&regs,&sregs);

	// copy our DOS memory block back to pmode
	_fmemcpy(&VBEInfo,dos_block,sizeof(VBEInfo));

	// clean up our dos memory usage
	RM_free(selector);

	// if call failed then skip out
	if ((regs.w.ax!=0x004F) ||
		(strncmp("VESA",(char*)VBEInfo.Signature,4)!=0))
		{
		printf("\n\n=======================\n\n");
		return false;
		}
	else
		{
		printf("\n\n=======================\n\n");
		printf("Graphics Capabilities: VESA VBE v%d.%d Detected\n\n",
				((VBEInfo.OemVersion&0xFF00) >> 8),
				(VBEInfo.OemVersion&0x00FF));
		}

	// check that the VESA VBE version is v2.00
	if (VBEInfo.OemVersion<0x0200) return false;

#ifdef __debugging

	if (!VESA_Debug_Finished)
		{
		printf("\n\n=================\n\n");
		printf("\nFound VESA BIOS!\n\n");

		printf("VESA VBE (Version %x) INFORMATION\n\n",VBEInfo.OemVersion);
		printf("Signature         %.4s\n",VBEInfo.Signature);
		printf("OemVersion        %x\n",VBEInfo.OemVersion);
		printf("OemStringPtr      %p\n",VBEInfo.OemStringPtr);
		printf("Capabilities      %x\n",VBEInfo.Capabilities);
		printf("VideoModePtr      %p\n",VBEInfo.VideoModePtr);
		printf("TotalMemBlocks    %d\n",VBEInfo.TotalMemBlocks);

		if (VBEInfo.OemVersion>=0x0200)
			{
			printf("\nVBE 2.0+ INFORMATION\n\n");
			printf("OemSoftwareRev    %d\n",VBEInfo.OemSoftwareRev);
			printf("OemVendorName     %p\n",VBEInfo.OemVendorName);
			printf("OemProductName    %p\n",VBEInfo.OemProductName);
			printf("OemProductRev     %p\n",VBEInfo.OemProductRev);
			}

		getch();
		}
#endif
	// set up for VESA bios call 0x4F03 - Get Current Video Mode
	memset(&sregs,0,sizeof(sregs));
	memset(&regs,0,sizeof(regs));

	regs.w.ax = 0x4F03;
	RM_int86(0x10,&regs,&regs);

	// if we couldn't get the current video mode, skip out
	if (regs.w.ax!=0x004F)
		{
#ifdef __debugging
	printf("Error: could not figure current video mode, skipping out\n");
	getch();
#endif
		return false;
		}

	// save mode for ResetMode()
	grOrigVideoMode = regs.w.bx;

	// VESA VBE 1.2+ is available
	return true;
	}

// --------------------------------------------------------------------------
//
//	FUNC:	VESA_ModeIsAvailable()
//	DESC:	returns true if the specified mode is available.
//			where 'mode' is the library mode.... ie. gr640x480x256
//
// --------------------------------------------------------------------------
int VESA_ModeIsAvailable(int mode)
	{
	WORD* 	modelist;
	WORD*	modeptr;
	int		retries;

	// create a pointer to the list of modes available (in DOS memory)
	modelist = (WORD*) RM_MapRealPointer(VBEInfo.VideoModePtr);

	// scan till we find 0xFFFF (-1) as this is the end of modes signal
	// taking into account the number of modes we have tested.
	// Some VBE's (version 1.0+) report incorrect mode numbers, so this
	// is simply a patch to try the hardcoded mode numbers
	for (modeptr=modelist,retries=64; *modeptr!=-1 && retries>=0; modeptr++,retries--)
		{
		// test the standard mode number as well....
		if (retries==0) modeptr = &VESA_Mode_Data[mode].std_mode;

#ifdef __debugging
	if (!VESA_Debug_Finished)
		{
		printf("\n\n-------\n");
		printf("Checking validity of mode #%x\n",*modeptr);
		}
#endif

		// if we can get the mode information for the current mode
		if (VESA_GetModeInfo(*modeptr,&ModeInfo))
			{
			// skip if mode is not available, not color, or not graphics
			if ((ModeInfo.ModeAttributes & VESA_ModeAvailable) == 0) continue;
			if ((ModeInfo.ModeAttributes & VESA_ColorMode) == 0) continue;
			if ((ModeInfo.ModeAttributes & VESA_GraphicsMode) == 0) continue;

			// skip if mode is not packed, not 8 bits per pixel, or not 1 plane
			if (ModeInfo.MemoryModel != VESA_Packed) continue;
			if (ModeInfo.BitsPerPixel != 8) continue;
			if (ModeInfo.NumberOfPlanes != 1) continue;

			// make sure the mode is what we want!!!
			if (ModeInfo.XResolution != VESA_Mode_Data[mode].xres ||
				ModeInfo.YResolution != VESA_Mode_Data[mode].yres ) continue;

			// we've found the appropriate mode

			#ifdef __debugging
			if (!VESA_Debug_Finished)
				{
				printf("\nFound valid VESA mode %x\n\n",*modeptr);
				}
			#endif

			return *modeptr;
			}
		}

	return false;
	}

// --------------------------------------------------------------------------
//
//	FUNC:	VESA_GetModeInfo()
//	DESC:	returns true if information for the specified mode is available.
//
// --------------------------------------------------------------------------
bool VESA_GetModeInfo(int mode, VESA_MODE_BLOCK* ModeInfo)
	{
	REGS 		regs;
	SREGS 		sregs;
	short 		selector,
				segment;
	char far*	dos_block;

	// clear regs
	memset(&regs,0,sizeof(regs));
	memset(&sregs,0,sizeof(sregs));

	// allocate a chunk of DOS memory
	dos_block = (char far*) RM_malloc(&selector,&segment,sizeof(VESA_MODE_BLOCK));

	// set up VESA bios call 0x4F01 - Get SuperVGA Mode Information
	sregs.es = segment;
	regs.x.edi = 0;
	regs.w.ax = 0x4F01;
	regs.w.cx = mode;

	// and get the VESA mode information
	RM_int86x(0x10,&regs,&regs,&sregs);

	// copy information from DOS memory to pmode
	_fmemcpy(ModeInfo,dos_block,sizeof(VESA_MODE_BLOCK));

	// and clean up our dos memory usage
	RM_free(selector);

	// skip out if invalid call
	if (regs.w.ax!=0x004F) return false;

	// fix bug in some VBE implementations
	if (ModeInfo->WinGranularity==0) ModeInfo->WinGranularity = 64;

#ifdef __debugging

	if (!VESA_Debug_Finished)
		{
		if ((ModeInfo->ModeAttributes & VESA_ModeAvailable) == 0) return true;

		printf("\n\nVBE 1.2+ MODE %x INFORMATION\n\n",mode);

		printf("ModeAttributes      %x",ModeInfo->ModeAttributes);
		if ((ModeInfo->ModeAttributes & VESA_LinearMode) != 0) printf("LINEAR");
		printf("\nWinAAttributes      %x\n",ModeInfo->WinAAttributes);
		printf("WinBAttributes      %x\n",ModeInfo->WinBAttributes);
		printf("WinGranularity      %d\n",ModeInfo->WinGranularity);
		printf("WinSize             %d\n",ModeInfo->WinSize);
		printf("WinASegment         %x\n",ModeInfo->WinASegment);
		printf("WinBSegment         %x\n",ModeInfo->WinBSegment);
		printf("WinFuncPtr          %p\n",ModeInfo->WinFuncPtr);
		printf("BytesPerScanLine    %d\n",ModeInfo->BytesPerScanLine);
		printf("XResolution         %d\n",ModeInfo->XResolution);
		printf("YResolution         %d\n",ModeInfo->YResolution);

		printf("XCharSize           %d\n",ModeInfo->XCharSize);
		printf("YCharSize           %d\n",ModeInfo->YCharSize);
		printf("NumberOfPlanes      %d\n",ModeInfo->NumberOfPlanes);
		printf("BitsPerPixel        %d\n",ModeInfo->BitsPerPixel);
		printf("NumberOfBanks       %d\n",ModeInfo->NumberOfBanks);
		printf("MemoryModel         %d\n",ModeInfo->MemoryModel);
		printf("BankSize            %d\n",ModeInfo->BankSize);
		printf("NumberOfImagePages  %d\n",ModeInfo->NumberOfImagePages);

		if (VBEInfo.OemVersion>=0x0200)
			{
			printf("\nVBE 2.0+ MODE INFORMATION\n\n");
			printf("PhysBasePtr         %p\n",ModeInfo->PhysBasePtr);
			printf("OffScreenMemOffset  %p\n",ModeInfo->OffScreenMemOffset);
			printf("OffScreenMemSize    %p\n",ModeInfo->OffScreenMemSize);
			}

		getch();
		}
#endif

	return true;
	}

// --------------------------------------------------------------------------
//
//	FUNC:	VESA_SetMode()
//	DESC:	Sets the required video mode iff possible...
//			returns true if the requested mode was set, false otherwise
//
// --------------------------------------------------------------------------
bool VESA_SetMode(int mode)
	{
#ifdef __debugging
	if (!VESA_Debug_Finished)
		{
		printf("\n\n--------------------------- VESADRV v1.2\n\n");
		printf("Setting LIBRARY mode %d\n",mode);
		}
#endif

	// --- check to see if the requested mode is available
	int modeNum = VESA_ModeIsAvailable(mode);

	// and skip out if it doesn't exist
	if (modeNum==0) return false;

	printf("Setting VESA mode %d\n",modeNum);

	// --- Figure out which window we can use for reading & for writing,
	//     trying to use seperate read & write windows if possible...

	// if window B exists & is writable
	if ((ModeInfo.WinBAttributes & 5)==5)
		{
		VESA_Write_Window = 0x0001;	// use window B
#ifdef __debugging
	if (!VESA_Debug_Finished)
		{
		printf("Window B exists and is Writable.\n");
		}
#endif
		}
	else
		{
#ifdef __debugging
	if (!VESA_Debug_Finished)
		{
		printf("Window B is non-existant or non Writeable.\n");
		}
#endif

		// if window A is writeable
		if ((ModeInfo.WinAAttributes & 4)==4)
			{
			VESA_Write_Window = 0x0000;	// use window A
#ifdef __debugging
		if (!VESA_Debug_Finished)
			{
			printf("Window A is Writable\n");
			}
#endif
			}
		else // error???
			{
			VESA_ResetMode(grOrigVideoMode);
			printf("\nERROR: Neither Window A or B are Writeable!\n");
	getch();
			exit(1);
			}
		}

	// if window A is readable.
	if ((ModeInfo.WinAAttributes & 2)==2)
		{
		VESA_Read_Window = 0x0000;	// use window A
#ifdef __debugging
	if (!VESA_Debug_Finished)
		{
		printf("Window A is Readable\n");
		}
#endif
		}
	else
		{
#ifdef __debugging
		if (!VESA_Debug_Finished)
			{
			printf("Window A is non-existant or non Writeable.\n");
			}
#endif
		// if window B exists & is readable
		if ((ModeInfo.WinBAttributes & 3)==3)
			{
			VESA_Read_Window = 0x0001;	// use window B
#ifdef __debugging
		if (!VESA_Debug_Finished)
			{
			printf("Window B exists and is Readable\n");
			}
#endif
			}
		else // error???
			{
			VESA_ResetMode(grOrigVideoMode);
			printf("\nERROR: Neither Window A or B are Readable!\n");
	getch();
			exit(1);
			}
		}

	// --- save the fact if we are using the same window...
	//	   or different windows...

	VESA_Using_Same_Window = (VESA_Read_Window==VESA_Write_Window);

#ifdef __debugging
	if (!VESA_Debug_Finished)
		{
		if (VESA_Write_Window) printf("Write Window is set to Window B\n");
		else printf("Write Window is set to Window A\n");

		if (VESA_Read_Window) printf("Read Window is set to Window B\n");
		else printf("Read Window is set to Window A\n");

		getch();
		}
#endif

	// --- set the video mode

	// set up VESA bios call 0x4F02 - Set Video Mode
	REGS regs;
	memset(&regs,0,sizeof(regs));

	regs.w.ax = 0x4F02;
	regs.w.bx = modeNum | VESA_DontClear;

	// if we can use a linear mode, then might as well!!! :-)
//	if (ModeInfo.ModeAttributes & VESA_LinearMode)
//		{
//		printf("\n\nNOTICE: About to set video mode TO USE LINEAR FRAME BUFFER!\n\n");
//		getch();
//		regs.w.bx |= VESA_EnableLFB;
//		}

	RM_int86(0x10,&regs,&regs);

	// if we couldn't set VESA mode, return error
	if (regs.w.ax!=0x004F)
		{
		VESA_ResetMode(grOrigVideoMode);
		printf("ERROR:: An error occurred during Set Video Mode!\n");
		printf("        Please contact supplier.\n");

		getch();
		exit(1);
		}

	// --- save the current mode number

	grCurrentVideoMode = mode;

	// --- set the logical scan line length to 1024 pixels for speed later...

	memset(&regs,0,sizeof(regs));
	regs.w.ax = 0x4F06;
	regs.w.bx = 0x0000;
	regs.w.cx = 1024;
	RM_int86(0x10,&regs,&regs);

	// skip out if not successful
	if (regs.w.ax!=0x004F)
		{
		VESA_ResetMode(grOrigVideoMode);
		printf("ERROR: Could not set the virtual scan line to 1024 pixels\n");
		getch();
		exit(1);
		}

	// --- if its the first time in, check the video display speed!!!!
	if (grVideoSpeed==0) VESA_CheckSpeed();

	// --- setup global variables

	grMemoryInKb  		= VBEInfo.TotalMemBlocks << 6;
	grMemoryInBytes 	= grMemoryInKb << 10;

	VESA_BankMask		= ModeInfo.WinGranularity - 1;

	grNumberOfBanks		= grMemoryInKb / ModeInfo.WinGranularity;
	grLogicalWidth 		= 1024;
	grLogicalHeight 	= grMemoryInBytes / grLogicalWidth;

	VESA_Video_Addr		= ModeInfo.WinASegment << 4;
	VESA_Video_Mask		= VESA_Video_Addr | VESA_BankMask;

	// temporarily for 1MB access max.
	if (grLogicalHeight>1024) grLogicalHeight = 1024;

	grScreenWidth  = VESA_Mode_Data[mode].xres;
	grScreenHeight = VESA_Mode_Data[mode].yres;

	grMaxx = grScreenWidth-1;
	grMaxy = grScreenHeight-1;

	// --- note:
	// If we precalculate the start of each row of each bank, then we
	// can (a) ignore the window granularity (b) speed up the draw functions

	// start at the top of video memory
	long 	startofrow = VESA_Video_Addr;
	int     banknum = 0;

	// and for each row of the display (ie. 1Mb card, 1024 width == 1024 rows)
	for (int y=0;y<grLogicalHeight;y++)
		{
		// set the pointer to the start of the row
		VESA_RowPtr[y] = startofrow;

		// set the bank number we want to switch to for the row
		VESA_RowBank[y] = banknum;

		// our row pointer increments by the logical scanline width
		startofrow += grLogicalWidth;

		// if we've hit a bank boundary
		if ((y & VESA_BankMask) == VESA_BankMask)
			{
			startofrow = VESA_Video_Addr;	// reset to start of window
			banknum++;						// and select the next bank
			}
		}

	// --- set pages & clear the video, etc.
	if (VESA_Flicker_Free)
		{
		VESA_SetVisualPage(0);
		VESA_SetActivePage(1);
		}
	else
		{
		VESA_SetVisualPage(0);
		VESA_SetActivePage(0);
		}

	// --- clear the display
	grClearVesa();

	// --- a VESA mode has been successfully set
	return true;
	}

// --------------------------------------------------------------------------
//
//	FUNC:	VESA_ResetMode(grOrigVideoMode)
//	DESC:	Sets the video mode to mode # 'mode'. By default it resets
//			the mode to that on startup...
//
// --------------------------------------------------------------------------
void VESA_ResetMode(int mode)
	{
	REGS regs;
	memset(&regs,0,sizeof(regs));

	// and set the video mode
	regs.w.ax = 0x4F02;
	regs.w.bx = mode;
	int386(0x10,&regs,&regs);
	}

// --------------------------------------------------------------------------
//
//	FUNC:	VESA_SwitchWriteBank()
//	DESC:	Sets the write bank for screen writing routines. If the
//			read/write banks are separate, this uses window B, otherwise
//			it uses window A.
//
//	RETURNS:	on success, the global variable 'VESA_Current_Write_Bank' is set
//
// --------------------------------------------------------------------------
REGS VESA_write_bank_regs;

void VESA_SwitchWriteBank(int newbank)
	{
	if (newbank==VESA_Current_Write_Bank) return;

#ifdef __display_banks

	// blank the display on the current bank
	int i,x,y;
	BYTE* p;
	long addr;

	for (i=0;i<16;i++)
		{
		for (y=0;y<4;y++)
			{
			addr = VESA_Video_Addr + (y * 1024 + 8192) + 20 * (i/8) + (10 * i);
			p = (BYTE*) addr;
			for (x=0;x<8;x++)
				{
				*p = grBLACK;
				p++;
				}
			}
		}
//	delay(30);
#endif

	// set up for VESA mode 0x4F05	- CPU video memory control
	VESA_write_bank_regs.w.ax = 0x4F05;
	VESA_write_bank_regs.w.bx = VESA_Write_Window;
	VESA_write_bank_regs.w.dx = newbank;
	int386(0x10,&VESA_write_bank_regs,&VESA_write_bank_regs);

#ifdef __display_banks

	// display the new lights
	for (i=0;i<16;i++)
		{
		for (y=0;y<4;y++)
			{
			addr = VESA_Video_Addr + (y * 1024 + 8192) + 20 * (i/8) + (10 * i);
			p = (BYTE*) addr;
			for (x=0;x<8;x++)
				{
				if (i<8) *p = grBLUE;
				else *p = grGREEN;

				if (i==newbank) *p = grLIGHTRED;
				p++;
				}
			}
		}
//	delay(30);
#endif


	if (VESA_write_bank_regs.w.ax!=0x004F)		// iff !successful
		{
		VESA_ResetMode(grOrigVideoMode);
		printf("ERROR: could not set WRITE bank #%d\n",newbank);
		exit(1);
		}

	VESA_Current_Write_Bank = newbank;
	if (VESA_Using_Same_Window) VESA_Current_Read_Bank = VESA_Current_Write_Bank;
	}

// --------------------------------------------------------------------------
//
//	FUNC:	VESA_SwitchReadBank()
//	DESC:	Sets the read bank for screen reading routines. This uses
//			window A always.
//
//	RETURNS:	on success, the global variable 'VESA_Current_Read_Bank' is set
//
// --------------------------------------------------------------------------
REGS	VESA_read_bank_regs;

void VESA_SwitchReadBank(int newbank)
	{
	if (newbank==VESA_Current_Read_Bank) return;

#ifdef __display_banks

	// blank the display on the current bank
	int i,x,y;
	BYTE* p;
	long addr;

	for (i=0;i<16;i++)
		{
		for (y=0;y<4;y++)
			{
			addr = VESA_Video_Addr + (y * 1024) + 20 * (i/8) + (10 * i);
			p = (BYTE*) addr;
			for (x=0;x<8;x++)
				{
				*p = grBLACK;
				p++;
				}
			}
		}
//	delay(30);
#endif

	// set up for VESA mode 0x4F05	- CPU video memory control
	VESA_read_bank_regs.w.ax = 0x4F05;
	VESA_read_bank_regs.w.bx = VESA_Read_Window;			// select appropriate window
	VESA_read_bank_regs.w.dx = newbank;
	int386(0x10,&VESA_read_bank_regs,&VESA_read_bank_regs);

#ifdef __display_banks

	// display the new lights
	for (i=0;i<16;i++)
		{
		for (y=0;y<4;y++)
			{
			addr = VESA_Video_Addr + (y * 1024) + 20 * (i/8) + (10 * i);
			p = (BYTE*) addr;
			for (x=0;x<8;x++)
				{
				if (i<8) *p = grBLUE;
				else *p = grGREEN;

				if (i==newbank) *p = grLIGHTRED;
				p++;
				}
			}
		}
//	delay(30);
#endif

	if (VESA_read_bank_regs.w.ax!=0x004F)		// iff !successful
		{
		VESA_ResetMode(grOrigVideoMode);
		printf("ERROR: could not set READ bank #%d\n",newbank);
		exit(1);
		}

	VESA_Current_Read_Bank = newbank;
	if (VESA_Using_Same_Window) VESA_Current_Write_Bank = VESA_Current_Read_Bank;
	}

// --------------------------------------------------------------------------
//
//	FUNC:	VESA_SetActivePage()
//	DESC:	Sets the active page for reading/writing the appropriate video page
//
// --------------------------------------------------------------------------
void VESA_SetActivePage(int page)
	{
	if (page>VESA_Max_Page) return;

	VESA_Active_Page = page;

	VESA_Write_Bank_Offset = VESA_Active_Page * VESA_Banks_Per_Page;
	if (VESA_Using_Same_Window) VESA_Read_Bank_Offset = VESA_Write_Bank_Offset;
	}

// --------------------------------------------------------------------------
//
//	FUNC:	VESA_SetActivePage()
//	DESC:	Sets the visual page for displaying the appropriate video page
//
// --------------------------------------------------------------------------
void VESA_SetVisualPage(int page)
	{
	if (page>VESA_Max_Page) return;

	// set visual display page
	VESA_Visual_Page = page;

	REGS regs;
	memset(&regs,0,sizeof(regs));

	// set up for VESA call 0x4F07 - set/get display start
	regs.w.ax = 0x4F07;
	regs.w.bx = 0x0000;	// 80 waits for vertical retrace!! ?? does this work?
	regs.w.cx = 0;		// first displayed pixel in scanline
	regs.w.dx = (DWORD) (VESA_Visual_Page << 9);

	if (VESA_WaitForRetrace) VESA_WaitRetrace();	// wait for retrace iff required

	int386(0x10,&regs,&regs);

	if (regs.w.ax!=0x004F)
		{
		VESA_ResetMode(grOrigVideoMode);
		printf("ERROR: could not set page #%d\n",VESA_Visual_Page);
		exit(1);
		}
	}

// --------------------------------------------------------------------------
//
//	FUNC:	VESA_PageFlip()
//	DESC:	Switches the current page to the next available page
//			(a simple pageflip)
//
// --------------------------------------------------------------------------
void VESA_PageFlip()
	{
//	if (VESA_Flicker_Free)
//		{
//		if (VESA_Visual_Page==0)
//			{
//			VESA_SetVisualPage(1);			// page flip manually
//			VESA_SetActivePage(0);
//			}
//		else
//			{
//			VESA_SetVisualPage(0);			// page flip manually
//			VESA_SetActivePage(1);
//			}
//		}
//	else
		{
		VESA_SetVisualPage(0);
		VESA_SetActivePage(0);
		}
	}

// --------------------------------------------------------------------------
//
//	FUNC:	VESA_WaitRetrace()
//	DESC:	This function returns at the start of the next vertical retrace
//
// --------------------------------------------------------------------------
void VESA_WaitRetrace()
	{
	// check to see if the hardware is already in retrace
	while (inp(VGA_INPUT_STATUS_1) & VGA_VSYNC_MASK)	;

	// now wait for the next retrace and exit
	while (!(inp(VGA_INPUT_STATUS_1) & VGA_VSYNC_MASK))	;
	}

// ---------------
// set BlitRegion() to produce 'flicker-free' blits
void	grSetFlickerFreeMode()
	{
//	VESA_Flicker_Free = true;
//	VESA_PageFlip();
//	grBlitScreen();
	}

// ---------------
// set BlitRegion() to produce quick blits
void	grSetQuickBlitMode()
	{
//	VESA_Flicker_Free = false;
//	VESA_PageFlip();
//	grBlitScreen();
	}

// ---------------
// switches the display method used by the bit-blitter..
// flicker-free or quick
void	grToggleBlitMode()
	{
	if (VESA_Flicker_Free) grSetQuickBlitMode();
	else grSetFlickerFreeMode();
	}

// ---------------
// disables monitor access to the video card memory
void	grOffVideo()
	{
/*	VESA_WaitRetrace();
	outp(0x3C4,0x01);
	outp(0x3C5,inp(0x3C5) | 0x20);
*/	}

// ---------------
// enables monitor access to the video card memory
void	grOnVideo()
	{
/*	VESA_WaitRetrace();
	outp(0x3C4,0x01);
	delay(2);
	outp(0x3C5,inp(0x3C5) & 0xDF);
*/	}

// ---------------

//DJD --- TEMPORARILY TILL HARDWARE CURSOR INTERFACE
extern	bool	MOUSE_EXTERNAL_UPDATE;
extern	bool	MOUSE_ISR_DISABLED;
//DJD --- END

// bitblit a region of the double-buffer to the physical display
// as fast as possible....  depending on the display mode (single-or-double)
void	grBlitRegion(int x1,int y1,int x2,int y2)
	{
	if (!grSCREEN) return;

	CLIP2(x1,y1,0,0,grMaxx,grMaxy);
	CLIP2(x2,y2,0,0,grMaxx,grMaxy);

	int 	x,y;
	int 	bank_num;

	DWORD* 	dest;
	DWORD* 	srce;
	WORD*  	wdest;
	WORD*  	wsrce;
	BYTE*  	bdest;
	BYTE*  	bsrce;

	int copy_width	   	= (x2-x1)+1;				// total bytes to copy
	int dwords_to_copy 	= copy_width >> 2;			// number of dwords to copy
	int words_to_copy  	= ((copy_width & 2) != 0);	// if word-aligned, copy 1 word
	int bytes_to_copy  	= ((copy_width & 1) != 0);	// if byte-aligned, copy 1 byte

	// save the width of the double-buffer
	long  buffer_width	= (long) grSCREEN->hdr.width;

	// calc the address of the start of the buffer region to be displayed
	long  buffer_addr	= (long) &grSCREEN->data[x1 + y1 * buffer_width];

	// create a pointer to the address of the first row of the screen
	long* screen_addr 	= &VESA_RowPtr[y1];

	// --- blit the image (to active page)

	// create temporary vars
	long  srce_addr		= buffer_addr;
	long* dest_addr		= screen_addr;
	int*  bank_ofs		= &VESA_RowBank[y1];

	// tell mouse handler we want to take control of the screen for a bit

	MOUSE_EXTERNAL_UPDATE = true;

	// scan through each row of the x1,y1,x2,y2 region

	for (y=y1;y<=y2;y++)
		{
		// calculate the screen address of where to blit the row
		dest = (DWORD*) (*dest_addr + x1);

		// calc the address of the start of the image row
		srce = (DWORD*) (srce_addr);

//	MOUSE_ISR_DISABLED = true;
		// figure out the bank number & switchbank if required
//		bank_num = *bank_ofs++ + VESA_Write_Bank_Offset;
		bank_num = VESA_RowBank[y] + VESA_Write_Bank_Offset;
		if (bank_num!=VESA_Current_Write_Bank) VESA_SwitchWriteBank( bank_num );

		// splat all DWORDS across
		for (x=0;x<dwords_to_copy;x++)
			{
			*dest = *srce;
			srce++;
			dest++;
			}

		// if we need to copy a WORD
		if (words_to_copy)
			{
			wdest = (WORD*) dest;
			wsrce = (WORD*) srce;
			*wdest = *wsrce;
			wdest++;
			wsrce++;

			// we may need to copy a BYTE as well
			if (bytes_to_copy)
				{
				bdest = (BYTE*) wdest;
				bsrce = (BYTE*) wsrce;
				*bdest = *bsrce;
				}
			}
		else
			// we don't need to copy a word
			// but we may still need to copy a BYTE
			{
			if (bytes_to_copy)
				{
				bdest = (BYTE*) dest;
				bsrce = (BYTE*) srce;
				*bdest = *bsrce;
				}
			}

		// increment our indexes ready for the next row
		dest_addr ++;
		srce_addr += buffer_width;

		mouse_DrawCursorRow( y, x1, x2, SAVE_UNDERNEATH_CURSOR );

//	MOUSE_ISR_DISABLED = false;
		}

	// tell mouse handler we've finished updating the screen

	MOUSE_EXTERNAL_UPDATE = false;

	if (!VESA_Flicker_Free) { return; }

/*	// --- page flip
	VESA_PageFlip();

	// --- blit the image to (newly) hidden page

	srce_addr	= buffer_addr;
	dest_addr	= screen_addr;

	for (y=y1;y<=y2;y++)
		{
		dest = (DWORD*) (*dest_addr + x1);
		srce = (DWORD*) (srce_addr);

		bank_num = VESA_RowBank[y] + VESA_Write_Bank_Offset;
		if (bank_num!=VESA_Current_Write_Bank) VESA_SwitchWriteBank( bank_num );

		for (x=0;x<dwords_to_copy;x++)
			{
			*dest = *srce;
			srce++;
			dest++;
			}
		if (words_to_copy)
			{
			WORD* wdest = (WORD*) dest;
			WORD* wsrce = (WORD*) srce;
			*wdest = *wsrce;
			wdest++;
			wsrce++;

			if (bytes_to_copy)
				{
				BYTE* bdest = (BYTE*) wdest;
				BYTE* bsrce = (BYTE*) wsrce;
				*bdest = *bsrce;
				}
			}
		else
			{
			if (bytes_to_copy)
				{
				BYTE* bdest1 = (BYTE*) dest;
				BYTE* bsrce1 = (BYTE*) srce;
				*bdest1 = *bsrce1;
				}
			}
		dest_addr++;
		srce_addr += buffer_width;
		}
*/	}

