//	-------------------------------------------------------------------------
//
//	VESADRV.HPP
//
//	-------------------------------------------------------------------------
//
//	AUTHOR:			Darren J. Dwyer
//
//	DESCRIPTION:	This file defines bits & pieces of the VESA 1.2 driver
//
//	-------------------------------------------------------------------------

#ifndef __vesadrv_hpp
#define __vesadrv_hpp

#include "graphics.hpp"

// --------------- VESA modes implemented by this driver

#define gr640x400x256		0
#define gr640x480x256		1

// --------------------------------------------------------------------------
//
//	VESA DRIVER PUBLIC FUNCTIONS
//
// --------------------------------------------------------------------------

// attempts to initialize vesa... should only ever need to call this once.
// returns false if cannot initialize specified vesa mode
bool 	grInitVesa(int mode);

// terminates VESA mode
void 	grExitVesa();

// clears all available banks of video card
void 	grClearVesa(int color = 0);

// set BlitRegion() to produce 'flicker-free' blits
void	grSetFlickerFreeMode();

// set BlitRegion() to produce quick blits
void	grSetQuickBlitMode();

// switches the display method used by the bit-blitter..
// flicker-free or quick
void	grToggleBlitMode();

// bitblit a region of the double-buffer to the physical display
// as fast as possible....  depending on the display mode (single-or-double)
void	grBlitRegion(int x1,int y1,int x2,int y2);

// bitblit the entire double-buffer to the physical display
#define grBlitScreen() 				\
			grBlitRegion(0,0,grMaxx,grMaxy)


// --------------------------------------------------------------------------
//
//	VESA DRIVER INTERNALS
//
// --------------------------------------------------------------------------

// --------------- internal vars that may be required by externals (ie. mouse)

extern int		VESA_BankMask;		// the size of each VESA bank -1

extern bool		VESA_WaitForRetrace;// true--> waiting for retrace
extern long		VESA_RowPtr[1024];	// addresses for the start of each row
extern int		VESA_RowBank[1024];	// bank numbers for the start of each row


extern WORD		VESA_Read_Window; 		// either window A or window B
extern WORD		VESA_Write_Window;		// either window A or window B
extern bool		VESA_Using_Same_Window;	// window A is same as window B

extern int		VESA_Current_Read_Bank; // the current read bank
extern int		VESA_Current_Write_Bank;// the current write bank
extern int		VESA_Read_Bank_Offset;	// bank offset for reading multiple pages
extern int		VESA_Write_Bank_Offset;	// bank offset for writing multiple pages

extern int		VESA_Max_Page;			// the maximum number of display pages
extern int		VESA_Banks_Per_Page;	// the number of banks used for 1 page

extern int		VESA_Visual_Page;		// the current visual page
extern int		VESA_Active_Page;		// the current active page

extern long 	VESA_Video_Addr;		// video memory absolute start location
extern long		VESA_Video_Mask;		// mask for video-memory ranges

extern bool		VESA_Flicker_Free;		// disable/enable flicker free
										// bit-blitting

// --------------- the VESA VBE information structure

struct VESA_INFO_BLOCK
	{
	// VBE 1.2+ information
	BYTE 	Signature[4];	// 'VESA' 4 byte signature.
							// Note: Insert 'VBE2' for pmode extensions
							// 		 and make wasted[236] become wasted[492]
	WORD  	OemVersion;		// VBE version number
	DWORD	OemStringPtr;	// pointer to OEM string
	DWORD 	Capabilities;	// capabilities of video card
	DWORD 	VideoModePtr;	// pointer to supported modes
	WORD	TotalMemBlocks;	// number of 64K memory blocks

	// VBE 2.0 extensions
	WORD	OemSoftwareRev;	// OEM software revision number
	DWORD	OemVendorName;	// pointer to OEM vendor name
	DWORD	OemProductName;	// pointer to OEM product name
	DWORD	OemProductRev;	// pointer to OEM product revision string

	BYTE 	reserved[222];	// pad to 256 byte block size
	BYTE	OemDATA[256];	// pad to 512 byte block size for VBE 2.0+
	};

// --------------- the VESA VBE mode information structure

struct VESA_MODE_BLOCK
	{
	// VBE 1.2+ mode information
	WORD 	ModeAttributes;		// mode attributes
	BYTE	WinAAttributes;		// window A attributes
	BYTE	WinBAttributes;		// window B attributes
								//		bit 0 set	- window exists
								//		bit 1 set	- window is readable
								//		bit 2 set	- window is writeable
	WORD	WinGranularity;		// window granularity in Kb
	WORD 	WinSize;			// window size in Kb
	WORD	WinASegment;		// window A segment
	WORD	WinBSegment;		// window B segment
	DWORD   WinFuncPtr;			// pointer to bank switching function
	WORD	BytesPerScanLine;	// number of bytes per scanline
	WORD	XResolution;		// horizontal resolution
	WORD	YResolution;		// vertical resolution

	// note: the following don't seem to be used in VESA modes
	//		 except for supporting OEM modes. (ie. 0x00 - 0xFF)

	BYTE	XCharSize;			// character cell width
	BYTE	YCharSize;			// character cell height
	BYTE	NumberOfPlanes;		// number of memory planes
	BYTE	BitsPerPixel;		// number of bits per pixel
	BYTE	NumberOfBanks;		// number of CGA style banks
	BYTE	MemoryModel;		// memory model type
	BYTE	BankSize;			// size of CGA style banks
	BYTE	NumberOfImagePages;	// number of images pages
	BYTE	reserved1;			// reserved
	BYTE	RedMaskSize;		// size of direct color red mask
	BYTE	RedFieldPosition;	// Bit posn of lsb of red mask
	BYTE	GreenMaskSize;		// size of direct color green mask
	BYTE	GreenFieldPosition;	// Bit posn of lsb of green mask
	BYTE	BlueMaskSize;		// size of direct color blue mask
	BYTE	BlueFieldPosition;	// Bit posn of lsb of blue mask
	BYTE	RsvdMaskSize;		// size of direct color reserved mask
	BYTE	RsvdFieldPosition;	// Bit posn of lsb of reserved mask
	BYTE	DirectColorModeInfo;// Direct color mode attributes

	// VBE 2.0 extensions
	DWORD	PhysBasePtr;		// physical address of linear frame buffer
	DWORD	OffScreenMemOffset;	// pointer to start of offscreen memory
	WORD	OffScreenMemSize;	// amount of offscreen memory in 1Kb's

	BYTE 	reserved[206];		// pad to 256 byte block size
	};

// --------------- mode attributes flags

#define	VESA_ModeAvailable		0x0001	// video mode is available
#define VESA_ColorMode			0x0008	// mode is a color video mode
#define VESA_GraphicsMode		0x0010	// mode is a graphics mode
#define VESA_NonBanked			0x0040	// bank switching is NOT supported
#define VESA_LinearMode			0x0080	// linear mode IS supported

// --------------- memory models

#define	VESA_Packed				0x0004	// packed memory mode

// --------------- other support

#define VESA_EnableLFB			0x4000	// use linear frame buffer
#define	VESA_DontClear			0x8000	// don't clear on set mode

// ---------------

extern VESA_INFO_BLOCK 	VBEInfo;	// storage of vesa information
extern VESA_MODE_BLOCK	ModeInfo;	// storage of vesa mode information

// --------------- bits and pieces

#define 	VGA_VSYNC_MASK 	  	0x08    // masks off unwanted bit of status reg
#define 	VGA_INPUT_STATUS_1  0x3DA 	// vga status reg 1, bit 3 is the vsync
										// when 1 - retrace in progress
										// when 0 - no retrace

#define 	PAL_MASK		0x03C6    	// used for manipulating the palette
#define		PAL_READ		0x03C7		// registers
#define		PAL_WRITE		0x03C8
#define		PAL_DATA		0x03C9

// ---------------

// checks the raw graphics speed on the current machine
// and saves the results for later use...
void 	VESA_CheckSpeed();

// returns true if VESA 1.2+ compatible bios is currently available
bool 	VESA_BiosInstalled();

// returns true if requested VESA mode is available
int	 	VESA_ModeIsAvailable(int mode);

// returns true if can get mode info for 'mode'. Sets up ModeInfo
bool 	VESA_GetModeInfo(int mode, VESA_MODE_BLOCK* ModeInfo);

// returns true if could set requested VESA mode
bool 	VESA_SetMode(int mode);

// resets the video device to mode on startup
void 	VESA_ResetMode(int mode);

// switch write bank for VESA compatibles
void 	VESA_SwitchWriteBank(int newbank);

// switch read bank for VESA compatibles
void 	VESA_SwitchReadBank(int newbank);

// set active page
void 	VESA_SetActivePage(int page);

// set visual page
void 	VESA_SetVisualPage(int page);

// sync with next vertical retrace
void 	VESA_WaitRetrace();

// page flip.
void 	VESA_PageFlip();

// remap physical palette to 'pal'
void	VESA_SetPalette(int start,int end,char* pal);

#endif

