//	-------------------------------------------------------------------------
//
//	MAPWIN.HPP
//
//	-------------------------------------------------------------------------
//
//	AUTHOR:		Darren J. Dwyer
//
//	DESCRIPTION:	These are the definitions required for the MAP_WINDOW
//
//	-------------------------------------------------------------------------

#ifndef __mapwin_hpp
#define __mapwin_hpp

#include "onscreen.hpp"
#include "resource.hpp"
#include "pcxfile.hpp"
#include "mouse.hpp"
#include "map.hpp"
#include "city.hpp"
#include "route.hpp"

// map scrolling bits

#define map_SCROLL_NONE		0
#define map_SCROLL_UP	  	1
#define map_SCROLL_DN		2
#define map_SCROLL_LT		3
#define map_SCROLL_RT		4

// other definitions

#define		NUM_ZOOM_LEVELS		10
#define		MAX_ZOOM_LEVEL		9


extern	int					num_terrain_bitmaps;
extern	int					num_overlaid_bitmaps;
extern	int					num_train_bitmaps;

extern	grImage**			terrain_bitmaps[NUM_ZOOM_LEVELS];
extern	grImage**			overlaid_bitmaps[NUM_ZOOM_LEVELS];
extern	grImage**			train_bitmaps[NUM_ZOOM_LEVELS];

// info box display modes

#define	NO_INFO				0
#define	TERRAIN_INFO		1
#define	CITY_INFO           2
#define	ROUTE_INFO          3
#define	TRAIN_INFO			4


class MapWin : public OnScreenHandler
	{
	private:
//		ResourceFile*		cursor_file;		// file containing map cursors
//		MouseCursor**		mouse_cursors;		// the mouse cursors used by the map

		int					num_terrain_bitmaps;
		int					num_overlaid_bitmaps;

		ResourceFile* 		terrain_file;		// file containing terrain bitmaps
		ResourceFile* 		overlaid_file;		// file containing overlaid bitmaps
		ResourceFile*		train_file;			// file containing train bitmaps

		// these are used to differentiate the different mouse functionalities
		bool	lmb_auto;
		bool	rmb_auto;

		// map draw() bits
		bool	map_ready;					// true-> map is ready to draw

		int		map_centre_x;
		int		map_centre_y;

		int		map_xpos;
		int		map_ypos;

		int		map_x1;
		int		map_y1;
		int		map_x2;
		int		map_y2;

		int		screen_x1;
		int		screen_y1;
		int		screen_x2;
		int		screen_y2;

		int		displayed_x1;
		int		displayed_y1;
		int		displayed_x2;
		int		displayed_y2;

		int		topleft_x;
		int		topleft_y;

		int		icon_width;
		int		icon_height;
		int		icons_wide;
		int		icons_high;

		int		half_icon_width;
		int		half_icon_height;
		int		half_icons_wide;
		int		half_icons_high;

		// map zoom bits
		int 	zoom_level;					// current level of zoom
		bool 	finished_zoom;

		// map info bits
		int		last_infox;
		int		last_infoy;

		char	info_box_displayed;			// 0 == no info box is displayed

	public:
		MapWin();
		virtual ~MapWin();

		// --- primitive map draw functions

				// draws a map made up of a 2D grid of scaled bitmaps
		void	DrawScaledMap();

				// these are (mostly) optimized functions for drawing
				// the level 0,1,2 maps because they're _heaps_ quicker
				// than trying to draw a scaled map for these regions
		void	DrawLevel0Map();
		void	DrawLevel1Map();
		void	DrawLevel2Map();

				// draws a point in a map of a 2D grid of scaled bitmaps
		void	DrawScaledPoint();

				// draws a point as above but overlays pic in scaled mode
				// or draws the color in levels 0,1,2
		void	DrawScaledPoint(grImage* pic,int color);

				// (mostly) optimized functions for level 0,1,2 points
		void	DrawLevel0Point(int color);
		void	DrawLevel1Point(int color);
		void	DrawLevel2Point(int color);


		// --- user map functions

				// draws a single icon if its in the current map display
				// region, returning true if drawn, false if not drawn
		bool	DrawMapPoint(int x,int y);

				// calls DrawMapPoint() and then blits it as required
		void	RefreshMapPoint(int x,int y);


				// draws a given icon if its in the current map display
				// region, returning true if drawn, false if not drawn
		bool	DrawMapPoint(int x,int y,grImage* pic,int color);

				// calls DrawMapPoint(pic) and then blits it as required
		void	RefreshMapPoint(int x,int y,grImage* pic,int color);


				// draws the complete map
		void	DrawMap();

				// calls DrawMap() and then blits it
		void	RefreshMap();


				// call this function to redraw a specific (rectangular)
				// region of the map using the given map coordinates,
				// returning true if any part of the region is refreshed,
				// false if none of the region was redrawn
		bool	DrawMapRegion(int m_x1,int m_y1,int m_x2,int m_y2);

				// calls DrawMapRegion() and then blits to the screen
		void	RefreshMapRegion(int m_x1,int m_y1,int m_x2,int m_y2);


				// call this function to redraw a specific (rectangular)
				// region of the map using the given screen coordinates
				// returning true if any part of the region is refreshed,
				// false if none of the region was redrawn
		bool	DrawScreenRegion(int s_x1,int s_y1,int s_x2,int s_y2);

				// calls DrawScreenRegion() and then blits to the screen
		void	RefreshScreenRegion(int s_x1,int s_y1,int s_x2,int s_y2);


		// --- map movement functions

				// zoom to specified map level
		void	ZoomTo(int new_level);
		void	ZoomIn();
		void	ZoomOut();

				// scroll map to specified point
		void	ScrollTo(int x,int y);
		void	Scroll(int dx,int dy);


		// --- extended map draw functions

		void	DrawMapInfo(int x,int y,bool auto_refresh = true);
		void	UndrawMapInfo();

		void	DrawTerrainInfo(int x,int y,bool auto_refresh);
		void	DrawCityInfo(int x,int y,bool auto_refresh);
		void	DrawRouteInfo(int x,int y,bool auto_refresh);

		void	DrawCity(int city); 	// draws a particular city
		void	DrawRoute(int route);	// draws a particular route
		void	DrawRouteSection(int route,int section);// draws a section of a route
		void	DrawTrain(int train);	// draws a particular train
		void	DrawTrainRoute(int train);	// draws the route a train takes
		void	DrawAllRoutes(int rr);	// draws all routes owned by RR
		void	DrawAllTrains(int rr);	// draws all trains owned by RR
		void	DrawAllTrainRoutes(int rr);	// draws all routes run by trains
											// owned by RR

		void	HighlightCity(int citynum,int color);
		void	HighlightRoute(int routenum,int color);

		// --- map coordinate functions

		int		WhichCitySelected(grPoint& mouse_pt);	// finds the closest city to pt clicked
		grPoint	WhichMapPos(grPoint& mouse_pt);		// returns x,y coord selected by mouse

		// --- query responses

		int		WhichRRNumber(int x,int y);		// returns rr#, -1 if not RR
		int		WhichRouteNumber(int x,int y);	// returns route#, -1 if not ROUTE
		int		WhichCityNumber(int x,int y);	// returns city#, -1 if not CITY

		// --- miscellaneous functions

		void	MagnifyUnderCursor(grPoint& mouse_pt);

//		int		LoadMap(String* filename);
//		int		SaveMap(String* filename);

		// --- eventhandler functions

		virtual void Draw();				// draws the map at current zoom,position
		virtual void Animate();				// animate the map
		virtual void HandleEvent(Event &e);	// handle event information
	};

extern	MapWin*				MAPWIN;

#endif

