//
//	Program: 		MAKERES.CPP
//
//	Author : 		D.J.Dwyer
//
//	Description: 	This is a utility used to build ResourceFiles.
//
//					It used "Resource Definition Files" which are
//					simple text files containing all relevent details
//					need to build the required resource. (See: RESOURCE.DEF)
//

#include "textfile.hpp"
#include "file.hpp"
#include "resource.hpp"
#include "pcxfile.hpp"
#include "fonts.hpp"

// Resource_Types  :     0 - DATA
//                       1 - TEXT
//                       2 - BITMAP
//                       3 - FONT
//                       4 - PALETTE
//                       5 - SPRITE
//                       6 - SOUND

#define MAX_RESOURCE_TYPES		7
char*	resource_str[MAX_RESOURCE_TYPES] =
	{ "Data","Text","Bitmap","Font","Palette","Sprite","Sound" };


#define		TEXT			0
#define		GRAPHICS		1


int				display_mode = TEXT;
int				wait_for_keypress = false;

TextFile* 		definition_file;
ResourceFile* 	resource_file;
PCXFile*		pcx_file;

String*			defstr;
String*			resstr;

int				type_of_resource = -1;
long			max_records = -1;
int				record_cnt = 0;

int				xscale = -1;
int				yscale = -1;

// --------------------------------------------------------------------------
//
// FUNC: 	ERROR() & PAUSE()
// DESC:	These are general-purpose functions to print an error & pause
//			program execution....
//
// --------------------------------------------------------------------------

void ERROR(char* errmsg)
	{
	printf("ERROR: %s\n",errmsg);
	exit(1);
	}

void PAUSE()
	{
	printf("Press any key to continue: \n");
	getch();
	printf("\n\n");
	}


// --------------------------------------------------------------------------
//
// FUNC:	AddToBitmapResource()
// DESC:	Assuming that the filename specified in 'tmpstr' is a PCX file,
//			this function:   * loads the PCX image
//							 * displays the image
//							 * appends the image to the resourcefile
//							 * reads the image back from the resourcefile
//							 * displays it on the screen
//
// --------------------------------------------------------------------------

void AddToBitmapResource(String* filename,bool display_text)
	{
	// display appropriate information iff required
	if (display_text)
		if (display_mode==TEXT) printf("Adding %s as record %d\n",filename->Contents(),record_cnt);

	// accept PCX files as input only... (at the moment)
	pcx_file = new PCXFile(filename);
	if (!pcx_file) ERROR("Could not create PCXFile()!!");

	// load the PCX image
	grImage* image = pcx_file->LoadImage();

	// load the palette for the PCX image & set it accordingly
	if (display_mode==GRAPHICS)
		{
		grPalette* palette = pcx_file->LoadPalette();
		grSetPalette(palette,0,255);
		free(palette);
		}

	// put the image onto the screen
	if (display_mode==GRAPHICS)
		{
		grDrawFilledRect(0,0,99,99,grBLACK);
		grPutImage(0,0,image->hdr.width-1,image->hdr.height-1,image);
		grBlitRegion(0,0,99,99);
		}

	// if it's not scaled, append it to the bitmap resource
	if ((xscale==-1)&&(yscale==-1))
		{
		long image_size = sizeof(grImageHeader) + image->hdr.width * image->hdr.height;
		resource_file->AppendRecord((char*) image,image_size);
		}
	else
		{
		// scale the image to the screen, get an image from the screen &
		// save the resized image
		int x = image->hdr.width, y = image->hdr.height;
		int numer,denom;
		if (xscale != -1)
			{
			numer = xscale / 1000;
			denom = xscale % 1000;
			if (denom==0) denom = 1;
			x = (x * numer) / denom;
			}
		if (y != -1)
			{
			numer = yscale / 1000;
			denom = yscale % 1000;
			if (denom==0) denom = 1;
			y = (y * numer) / denom;
			}
		x--; y--;
		grPutScaledImage(0,0,x,y,image);
		grBlitRegion(0,0,x,y);

		grImage* newimage = grCreateImage(0,0,x,y);
		if (newimage)
			{
			grGetImage(0,0,x,y,newimage);
			long image_size = sizeof(grImageHeader) + newimage->hdr.width * newimage->hdr.height;
			resource_file->AppendRecord((char*) newimage,image_size);
			grDestroyImage(newimage);
			}
		}


	// and put it onto the screen at a different position
	if (display_mode==GRAPHICS)
		{
		// read the image back from the resource
//		grImage* image1 = (grImage*) resource_file->ReadRecord(record_cnt);

//		grDrawFilledRect(0,200,99,299,grBLACK);
//		grPutImage(0,200,image1->hdr.width-1,200+image1->hdr.height-1,image1);
//		grBlitRegion(0,200,99,299);

//		free(image1);
		}

	if (wait_for_keypress) getch();

	// clean up our images
	free(image);

	// choose the next record
	record_cnt++;

	// we've now finished with the pcx file
	delete pcx_file;
	}



// --------------------------------------------------------------------------
//
// FUNC:	AddToFontResource()
// DESC:    This function: 	* checks for a PCX file
//						    * if it exists, use AddToBitmapResource() to add
//							  it to the resource, otherwise adds "SPACE.PCX"
//
// --------------------------------------------------------------------------

void AddToFontResource(String* filename)
	{
	String* thefile = new String(filename->Contents(),255);

	// accept PCX files as input only...
	pcx_file = new PCXFile(thefile);
	if (!pcx_file) ERROR("Could not create PCXFile()!!");

	// if the relevant file doesn't exist, use space.pcx
	bool usingspace = false;
	if (!pcx_file->Exists()) { usingspace = true; thefile->Set("SPACE.PCX"); }
	delete pcx_file;

	// display information
	if (display_mode==TEXT)
		{
		printf("Adding %s as record %d ",thefile->Contents(),record_cnt);
		if (usingspace) printf("[%s]",filename->Contents());
		printf(" .... ");
		}

	// and append the new record to the resource
	AddToBitmapResource(thefile,false);

	if (display_mode==TEXT) printf("DONE!\n");

	// clean up
	delete thefile;
	}


// --------------------------------------------------------------------------
//
// FUNC:	main()
// DESC:    This function: 	* finds the Resource Definition File
//						    * scans through the RDF, getting all required
//                            data for the (to be created) resource
//							* creates a blank resource
//							* scans the RDF for the files to append to the
//							  resource and appends them in turn
//
// --------------------------------------------------------------------------

void main(int argc,char** argv)
	{
	//
	// --- display program title
	//

	printf("\n\n\nMAKERES.EXE   ResourceFile Make Utility v1.1           (c) 1995 Darren J. Dwyer\n\n");

	//
	// --- figure out the appropriate Resource Definition File to use
	//

	// set up some default file strings
	defstr = new String("makeres.def",255);

	// override with params iff required
	if (argc>1) defstr->Set(argv[1]);

	// convert to uppercase....
	defstr->ToUpper();

	// make sure we have full filename
	if (!defstr->Contains(".")) defstr->Append(".DEF");

	// use the definition file
	definition_file = new TextFile(defstr);
	if (!definition_file) ERROR("Could not create TextFile()!!");

	// make sure its there
	if (!definition_file->Exists())
		{
		printf("Could not find Resource Definition File : %s",defstr->Contents());
		delete definition_file;
		delete defstr;
		exit(1);
		}

	printf("Using Resource Definition File : %s\n\n",defstr->Contents());

	//
	// --- read the appropriate information from the RDF
	//

	// open the definition file
	definition_file->Open();

	// create our initial resource data
	resstr = new String("",255);
	type_of_resource = -1;
	max_records = -1;
	display_mode = TEXT;

	do
		{
		// read the next line from the definition file
		String* tmpstr = definition_file->ReadLine();

		// skip if we've hit EOF
		if (definition_file->AtEOF()) continue;

		// ignore all blank lines
		if (tmpstr->Length()==0) { delete tmpstr; continue; }

		// ignore all comments
		if (tmpstr->Contents()[0]=='#') { delete tmpstr; continue; }

		// ignore all lines that aren't definitions
		if (tmpstr->Contents()[0]!='@') { delete tmpstr; continue; }


		// check for the resource name & make sure it's uppercase
		if (tmpstr->Contains("@RESOURCE"))
			{
			resstr->Set((char*) &tmpstr->Contents()[12]);
			resstr->ToUpper();
			}

		// check for the resource type
		if (tmpstr->Contains("@TYPE"))
			{
			String* tmpstr1 = new String((char*) &tmpstr->Contents()[12]);
			type_of_resource = tmpstr1->StringToInt();
			delete tmpstr1;
			}

		// check for the maximum number of records in the resource
		if (tmpstr->Contains("@RECORDS"))
			{
			String* tmpstr1 = new String((char*) &tmpstr->Contents()[12]);
			max_records = tmpstr1->StringToInt();
			delete tmpstr1;
			}

		// check to see if we want to display information as text or graphics
		if (tmpstr->Contains("@DISPLAY"))
			{
			String* tmpstr1 = new String((char*) &tmpstr->Contents()[12]);
			if (tmpstr1->Contains("GRAPHICS")) display_mode = GRAPHICS;
			else display_mode = TEXT;
			delete tmpstr1;
			}

		if (tmpstr->Contains("@PAUSE"))
			{
			wait_for_keypress = true;
			}

		if (tmpstr->Contains("@XSCALE"))
			{
			String* tmpstr1 = new String((char*) &tmpstr->Contents()[12]);
			xscale = tmpstr1->StringToInt();
			delete tmpstr1;
			}

		if (tmpstr->Contains("@YSCALE"))
			{
			String* tmpstr1 = new String((char*) &tmpstr->Contents()[12]);
			yscale = tmpstr1->StringToInt();
			delete tmpstr1;
			}

		delete tmpstr;
		}

	// continue till we find all appropriate information or we reach end-of-file
	while (!definition_file->AtEOF());

	// skip out if we couldn't find the appropriate information
	if ((resstr->Length()==0) || (type_of_resource==-1) || (max_records==-1))
		{
		printf("Error: Resource Definitions missing!\n\n");

		if (resstr->Length()==0)  	printf("       could not find @RESOURCE\n");
		if (type_of_resource==-1) 	printf("       could not find @TYPE\n");
		if (max_records==-1) 		printf("       could not find @RECORDS\n");

		delete definition_file;
		delete resstr;
		delete defstr;
		exit(1);
		}

	// display information read in from definition file
	printf("Resource Name  :  %s\n",resstr->Contents());
	printf("Resource Type  :  %d\n",type_of_resource);
	printf("Number Records :  %d\n",max_records);

	printf("\nDisplay Mode   :  ");
	if (display_mode==TEXT) printf("text\n");
	else printf("graphics\n");

	//
	// --- create the new resource based upon the data just read
	//

	printf("\n\nCreating %s as a %s Resource with %d records ...",
				resstr->Contents(),resource_str[type_of_resource],max_records);

	resource_file = new ResourceFile(resstr);
	resource_file->Close();
	resource_file->Delete();	// only if Exists()!!!
	resource_file->Create(max_records,type_of_resource);
	resource_file->Open();

	printf("Done!\n\n");

	PAUSE();

	//
	// --- Append new records to the resource file (if possible)
	//

	// set the video display
	if (display_mode==GRAPHICS)
		{
		grInitVesa(gr640x480x256);
		int font = grInitFont("\\GAME\\FONTS\\SYSTEM.FNT");
		}

	// --- scan through the file line-by-line, starting at the top??
	definition_file->SeekTOF();

	do
		{
		// get the next line in the definition file....
		String* tmpstr = definition_file->ReadLine();

		// skip out if we reached EOF
		if (definition_file->AtEOF()) continue;

		// skip the line if it's blank
		if (tmpstr->Length()==0) { delete tmpstr; continue; }

		// skip the line if it's a comment
		if (tmpstr->Contents()[0]=='#') { delete tmpstr; continue; }

		// skip the line if it's a declaration
		if (tmpstr->Contents()[0]=='@') { delete tmpstr; continue; }

		// we have a filename..... add this file to our resource
		switch (type_of_resource)
			{
			case BITMAP_RESOURCE:	AddToBitmapResource(tmpstr,true); break;
			case FONT_RESOURCE:		AddToFontResource(tmpstr); break;
			case DATA_RESOURCE:		;
			case TEXT_RESOURCE:		;
			case PALETTE_RESOURCE:	;
			case SPRITE_RESOURCE:	;
			case SOUND_RESOURCE:	printf("\n\nThis resource type is not yet implemented!\n\n");
									exit(1);
									break;
			}

		delete tmpstr;
		}
	// continue till we have enough records or
	// we hit the end of the definition file
	while ( (record_cnt<max_records) && !definition_file->AtEOF() );

	//
	// --- Clean up everything
	//

	// turn off graphics iff required
	if (display_mode==GRAPHICS)
		{
		grExitFont();
		grExitVesa();
		}

	// done.... display status
	printf("\n\nResource file %s has been created with %d records!!!\n",resstr->Contents(),record_cnt);

	if (record_cnt!=max_records) printf("This is CORRUPT!!! (%d records specified)\n",max_records);
	else printf("This is GOOD.\n");

	getch();

	// and clean up
	delete resource_file;
	delete definition_file;

	delete resstr;
	delete defstr;
	}