//	-------------------------------------------------------------------------
//
//	BMP.CPP
//
//	-------------------------------------------------------------------------
//
//	AUTHOR:			Andrew Carvel Evans
//
//	DESCRIPTION:	This is the Bitmap Editting Window
//
//	-------------------------------------------------------------------------

#include "bmp.hpp"
#include "tooldefn.hpp"

#include "pcxfile.hpp"
#include "deffile.hpp"
#include "palfile.hpp"
#include "palettes.hpp"
#include "grid.hpp"
#include "mouse.hpp"
#include "graphics.hpp"
#include "fonts.hpp"
#include "eventmgr.hpp"
#include "timer.hpp"
#include "keydefns.hpp"
#include "buttons.hpp"
#include "strings.hpp"

#include "dos.h"
#include "math.h"

//----------------------------------------------------------------------------
//
//		CONSTRUCTOR
//
//----------------------------------------------------------------------------
BitmapWindow::BitmapWindow(String* file_name)
	: OnScreenHandler(0,0,grMaxx,grMaxy,grBACKGROUND,bs_NONE,bs_NONE,
					&evhno_BitmapWindow,0,1)
	{
	mouse_EnableMoveEvents();

	InitialiseVariables();

	CreateGrids();

	CreateButtons();

	SetupImages(file_name);

	// make sure the whole thing is onscreen
	isOnScreen = true;
	}

//----------------------------------------------------------------------------
//
//		DESTRUCTOR
//
//----------------------------------------------------------------------------
BitmapWindow::~BitmapWindow()
	{
	FreeImages();

	FreePCXFiles();

	FreeDEFFile();

	DelAllAccessories();

	delete small_tile_grid;
	delete palette;
	delete thumbnails;
	delete edit_region;

	delete message;

	free(clipboard);
	free(store);
	free(working);
	free(undo);
	}

//----------------------------------------------------------------------------
//
//		INITIALISE
//
//----------------------------------------------------------------------------
void BitmapWindow::InitialiseVariables()
	{
	int i,x,y;

	extended_image = grCreateImage(0, 0, MAX_WIDTH*MAX_TILES_ACROSS-1, MAX_HEIGHT*MAX_TILES_DOWN-1, true);

	undo = grCreateImage(0, 0, MAX_WIDTH*MAX_TILES_ACROSS-1, MAX_HEIGHT*MAX_TILES_DOWN-1, true);
	working = grCreateImage(0, 0, MAX_WIDTH*MAX_TILES_ACROSS-1, MAX_HEIGHT*MAX_TILES_DOWN-1, true);
	store = grCreateImage(0, 0, MAX_WIDTH*MAX_TILES_ACROSS-1, MAX_HEIGHT*MAX_TILES_DOWN-1, true);
	clipboard = grCreateImage(0, 0, MAX_WIDTH*MAX_TILES_ACROSS-1, MAX_HEIGHT*MAX_TILES_DOWN-1, true);

	undoUsed = false;
	storeUsed = false;
	clipboardUsed = false;

	mouseisdown = false;
	dragison = false;
	dragging = NONE;

	color1 = 15;
	color2 = 0;

	input_query_mode = NORMAL_INPUT;

	input_grid_mode = SINGLE;

	output_draw_mode = DRAW;

	output_fill_mode = NONE;

	output_tile_mode = SINGLE;

	output_bounds_mode = true;

	message = new String(50);

	ClearTiledImage();

	default_width = MAX_WIDTH;
	default_height = MAX_HEIGHT;

	thumbnail_start_index = 0;
	current_image_index = 0;
	}

//----------------------------------------------------------------------------
//
//		DRAW FUNCTION
//
//----------------------------------------------------------------------------
void BitmapWindow::Draw()
	{
	// draw basic screen things
	DrawEdit();
	DrawThumb();
	DrawTile();
	DrawPalette();

	// write current states to screen
	DrawMode();
	DrawImageDetails();
	message->Set("TOOL (c)Copyright 1996 by Andrew C. Evans");
	DrawMessage();

	// draw "normal" stuff to double buffer
	OnScreenHandler::Draw();

	grBlitRegion(r.left,r.top,r.right,r.bottom);
	}

//----------------------------------------------------------------------------
//
//		HANDLEEVENT FUNCTION
//
//----------------------------------------------------------------------------
void BitmapWindow::HandleEvent(Event& e)
	{
	int				i,x,y;
	bool			handled = false;
	CommandEvent	c = { 0,NULL };

	if (input_grid_mode == SINGLE)
		current_image = image[current_image_index];
	else if ((input_grid_mode == EXTENDED) || (input_grid_mode == EXTENDED_SINGLE))
		current_image = extended_image;

	switch (e.what)
		{
		case ev_mouse:
			switch (e.mouse.event)
				{
				case mb_leftdn:
				case mb_rightdn:
				case mb_leftauto:
				case mb_rightauto:
					mouseisdown = true;
					handled = HandleMouseInput(e);
					break;

				case mb_move:
					if (edit_region->IsSelected(e.mouse.where))
						{
						e.mouse.where = edit_region->WhichCell(e.mouse.where);
						DrawMouseXYPos(e.mouse.where);

						handled = true;
						}
					break;

				case mb_leftup:
				case mb_rightup:
					mouseisdown = false;
					handled = HandleMouseInput(e);

					dragison = false;
					dragging = NONE;
					break;
				}
			break;

		case ev_command:
			switch (e.command.event)
				{
				// other commands
				case cmdLoad:
					if (input_grid_mode == SINGLE)
						pcx_file[current_image_index]->LoadImage(image[current_image_index]);
					else if (input_grid_mode == EXTENDED)
						{
						for (y=0; y<MAX_TILES_DOWN; y++)
							for (x=0; x<MAX_TILES_ACROSS; x++)
								if (tile[x][y])
									pcx_file[tile[x][y]-1]->LoadImage(image[tile[x][y]-1]);
						}
					else if (input_grid_mode == EXTENDED_SINGLE)
						pcx_file[tile[tile_selected_x][tile_selected_y]-1]->LoadImage(image[tile[tile_selected_x][tile_selected_y]-1]);

					MakeTiledImage();
					Draw();

					handled = true;
					break;

				case cmdLoadAll:

					FreeImages();
					LoadImages();

					MakeTiledImage();
					Draw();

					handled = true;
					break;

				case cmdSave:
					if (input_grid_mode == SINGLE)
						pcx_file[current_image_index]->SaveImage(image[current_image_index]);
					else if (input_grid_mode == EXTENDED)
						{
						for (y=0; y<MAX_TILES_DOWN; y++)
							for (x=0; x<MAX_TILES_ACROSS; x++)
								if (tile[x][y])
									pcx_file[tile[x][y]-1]->SaveImage(image[tile[x][y]-1]);
						}
					else if (input_grid_mode == EXTENDED_SINGLE)
						pcx_file[tile[tile_selected_x][tile_selected_y]-1]->SaveImage(image[tile[tile_selected_x][tile_selected_y]-1]);

					handled = true;
					break;

				case cmdSaveAll:
					SaveImages();

					handled = true;
					break;

				case cmdFillImage:
				case cmdClearImage:
					if (e.command.event == cmdFillImage)
						grClearImage(current_image,color1);
					else
						grClearImage(current_image,color2);

					UseTiledImage();
					DrawEdit();
					DrawTile();
					DrawThumb();

					handled = true;
					break;

				case cmdStoreImage:
					ResetStore();
					grCopyImage(store,current_image);
					storeUsed = true;

					handled = true;
					break;

				case cmdGetImage:
					if (storeUsed)
						{
						current_image->hdr.width = store->hdr.width;
						current_image->hdr.height = store->hdr.height;
						grCopyImage(current_image,store);

						MakeTiledImage();

						DrawEdit();
						DrawImageDetails();
						DrawTile();
						DrawThumb();
						}

					handled = true;
					break;

				case cmdScrollUp:
					grScrollImageVertical(current_image,-1);

					UseTiledImage();
					DrawEdit();
					DrawTile();
					DrawThumb();

					handled = true;
					break;

				case cmdScrollUpFive:
					grScrollImageVertical(current_image,-5);

					UseTiledImage();
					DrawEdit();
					DrawTile();
					DrawThumb();

					handled = true;
					break;

				case cmdScrollLeft:
					grScrollImageHorizontal(current_image,-1);

					UseTiledImage();
					DrawEdit();
					DrawTile();
					DrawThumb();

					handled = true;
					break;

				case cmdScrollLeftFive:
					grScrollImageHorizontal(current_image,-5);

					UseTiledImage();
					DrawEdit();
					DrawTile();
					DrawThumb();

					handled = true;
					break;

				case cmdScrollRight:
					grScrollImageHorizontal(current_image,1);

					UseTiledImage();
					DrawEdit();
					DrawTile();
					DrawThumb();

					handled = true;
					break;

				case cmdScrollRightFive:
					grScrollImageHorizontal(current_image,5);

					UseTiledImage();
					DrawEdit();
					DrawTile();
					DrawThumb();

					handled = true;
					break;

				case cmdScrollDown:
					grScrollImageVertical(current_image,1);

					UseTiledImage();
					DrawEdit();
					DrawTile();
					DrawThumb();

					handled = true;
					break;

				case cmdScrollDownFive:
					grScrollImageVertical(current_image,5);

					UseTiledImage();
					DrawEdit();
					DrawTile();
					DrawThumb();

					handled = true;
					break;

				case cmdDecreaseHeightBottom:
				case cmdIncreaseHeightBottom:
				case cmdDecreaseWidthRight:
				case cmdIncreaseWidthRight:
					if ((input_grid_mode != SINGLE) && (input_grid_mode != REPEAT))
						break;

					current_image = image[current_image_index];

					memcpy(undo,current_image,grImageSize(0,0,current_image->hdr.width-1,current_image->hdr.height-1));
					undoUsed = true;

					ResetWorking();
					grCopyImage(working,current_image);

					if (e.command.event == cmdDecreaseHeightBottom)
						current_image->hdr.height--;
					if (e.command.event == cmdIncreaseHeightBottom)
						current_image->hdr.height++;
					if (e.command.event == cmdDecreaseWidthRight)
						current_image->hdr.width--;
					if (e.command.event == cmdIncreaseWidthRight)
						current_image->hdr.width++;

					ResizeImageIfNecessary(current_image);

					grCopyImageData(current_image,working);

					DrawEdit();
					DrawImageDetails();
					DrawTile();
					DrawThumb();

					handled = true;
					break;

				case cmdDecreaseHeightTop:
				case cmdIncreaseHeightTop:
				case cmdDecreaseWidthLeft:
				case cmdIncreaseWidthLeft:
					if ((input_grid_mode != SINGLE) && (input_grid_mode != REPEAT))
						break;

					current_image = image[current_image_index];

					memcpy(undo,current_image,grImageSize(0,0,current_image->hdr.width-1,current_image->hdr.height-1));
                    undoUsed = true;

					// scroll image so it seems we have decreased from top left edge
					if (e.command.event == cmdDecreaseHeightTop)
						grScrollImageVertical(current_image,-1);
					if (e.command.event == cmdDecreaseWidthLeft)
						grScrollImageHorizontal(current_image,-1);

					ResetWorking();
					grCopyImage(working,current_image);

					if (e.command.event == cmdDecreaseHeightTop)
						current_image->hdr.height--;
					if (e.command.event == cmdIncreaseHeightTop)
						current_image->hdr.height++;
					if (e.command.event == cmdDecreaseWidthLeft)
						current_image->hdr.width--;
					if (e.command.event == cmdIncreaseWidthLeft)
						current_image->hdr.width++;

					ResizeImageIfNecessary(current_image);

					grCopyImageData(current_image,working);

					// scroll image so it seems we have increased from top left edge
					if (e.command.event == cmdIncreaseHeightTop)
						grScrollImageVertical(current_image,1);
					if (e.command.event == cmdIncreaseWidthLeft)
						grScrollImageHorizontal(current_image,1);

					DrawEdit();
					DrawImageDetails();
					DrawTile();
					DrawThumb();

					handled = true;
					break;

				case cmdGotoFirstImage:
					current_image_index = 0;

					DrawEdit();
					DrawThumb();
					DrawTile();

					handled = true;
					break;

				case cmdGotoLastImage:
					current_image_index = number_images-1;

					DrawEdit();
					DrawThumb();
					DrawTile();

					handled = true;
					break;

				case cmdFlipHorizontal:
					grFlipImageHorizontal(current_image);

					UseTiledImage();
					DrawEdit();
					DrawTile();
					DrawThumb();

					handled = true;
					break;

				case cmdFlipVertical:
					grFlipImageVertical(current_image);

					UseTiledImage();
					DrawEdit();
					DrawTile();
					DrawThumb();

					handled = true;
					break;

				case cmdRotate90:
					grRotateImage(current_image,NINETY);

					UseTiledImage();
					DrawEdit();
					DrawImageDetails();
					DrawTile();
					DrawThumb();

					handled = true;
					break;

				case cmdRotate180:
					grRotateImage(current_image,ONEEIGHTY);

					UseTiledImage();
					DrawEdit();
					DrawTile();
					DrawThumb();

					handled = true;
					break;

				case cmdRotate270:
					grRotateImage(current_image,TWOSEVENTY);

					UseTiledImage();
					DrawEdit();
					DrawImageDetails();
					DrawTile();
					DrawThumb();

					handled = true;
					break;

				case cmdErase:
					input_query_mode = NORMAL_INPUT;
					output_fill_mode = NONE;
					output_draw_mode = ERASE;

					DrawMode();

					handled = true;
					break;

				case cmdFilledErase:
					input_query_mode = NORMAL_INPUT;
					output_fill_mode = SOLID;
					output_draw_mode = ERASE;

					DrawMode();

					handled = true;
					break;

				case cmdDraw:
					input_query_mode = NORMAL_INPUT;
					output_fill_mode = NONE;
					output_draw_mode = DRAW;

					DrawMode();

					handled = true;
					break;

				case cmdFilledDraw:
					input_query_mode = NORMAL_INPUT;
					output_fill_mode = SOLID;
					output_draw_mode = DRAW;

					DrawMode();

					handled = true;
					break;

				case cmdLine:
					input_query_mode = NORMAL_INPUT;
					output_fill_mode = NONE;
					output_draw_mode = LINE;

					DrawMode();

					handled = true;
					break;

				case cmdFlatLine:
					input_query_mode = NORMAL_INPUT;
					output_fill_mode = NONE;
					output_draw_mode = FLATLINE;

					DrawMode();

					handled = true;
					break;

				case cmdRect:
					input_query_mode = NORMAL_INPUT;
					output_fill_mode = NONE;
					output_draw_mode = RECT;

					DrawMode();

					handled = true;
					break;

				case cmdFilledRect:
					input_query_mode = NORMAL_INPUT;
					output_fill_mode = SOLID;
					output_draw_mode = RECT;

					DrawMode();

					handled = true;
					break;

				case cmdEllipse:
					input_query_mode = NORMAL_INPUT;
					output_fill_mode = NONE;
					output_draw_mode = ELLIPSE;

					DrawMode();

					handled = true;
					break;

				case cmdFilledEllipse:
					input_query_mode = NORMAL_INPUT;
					output_fill_mode = SOLID;
					output_draw_mode = ELLIPSE;

					DrawMode();

					handled = true;
					break;

				case cmdFill:
					output_fill_mode = !output_fill_mode;
					input_query_mode = NORMAL_INPUT;

					DrawMode();

					handled = true;
					break;

				case cmdUndo:
					if (undoUsed)
						{
						memcpy(working,current_image,grImageSize(0,0,current_image->hdr.width-1,current_image->hdr.height-1));
						grCopyImage(current_image,undo);
						memcpy(undo,working,grImageSize(0,0,working->hdr.width-1,working->hdr.height-1));

						DrawEdit();
						DrawThumb();
						DrawTile();
						}

					handled = true;
					break;

				case cmdSingle:
					output_tile_mode = SINGLE;
					input_grid_mode = SINGLE;

					DrawEdit();
					DrawTile();
					handled = true;
					break;

				case cmdRepeat:
					if (output_tile_mode == REPEAT)
						output_tile_mode = SINGLE;
					else
						output_tile_mode = REPEAT;

					DrawTile();
					handled = true;
					break;

				case cmdTile:
					output_tile_mode = SINGLE;
					input_grid_mode = EXTENDED;

					if (!any_tiles_selected)
						tile[0][0] = current_image_index+1;

					MakeTiledImage();

					DrawEdit();
					DrawTile();
					handled = true;
					break;

				case cmdClearTile:
					ClearTiledImage();

					tile[0][0] = current_image_index+1;
					MakeTiledImage();

					DrawEdit();
					DrawTile();
					handled = true;
					break;

				case cmdGotoPalette:
					Close();
					grClearScreen(grBACKGROUND);
//DJD
					grBlitScreen();

					EVENTMGR->GenCommandEvent(evhno_PaletteWindow,cmdOpen,NULL);

					handled = true;
					break;
				}
		}

	// if we handled an event here then clear it HERE!!
	if (handled)
		{
		EVENTMGR->ClearEvent(e);
		}

	OnScreenHandler::HandleEvent(e);
	}

//----------------------------------------------------------------------------
//
//		CREATEGRIDS
//
//----------------------------------------------------------------------------
void BitmapWindow::CreateGrids()
	{
	// Edit Grid
	edit_region = new Grid(gap,gap,
		MAX_WIDTH, MAX_HEIGHT, SIZEOF_EDITREGION/MAX_WIDTH-1,SIZEOF_EDITREGION/MAX_HEIGHT-1,1,grDARKGRAY,1,grDARKGRAY,grBACKGROUND);

	// Palette Select Squares
	palette = new Grid(edit_x1,max_edit_y2+gap,
		32,8,11,8,1,grDARKGRAY,1,grDARKGRAY,grBACKGROUND);

	// Create Thumbnail Range Bar
	thumbnails = new Grid(grMaxx-32-gap,GUIbitmap[GUI_TOP]->hdr.height+gap,
		1,12,32,32,1,grDARKGRAY,1,grDARKGRAY,grBACKGROUND);

	// tile Grid
	small_tile_grid = new Grid(max_edit_x2+gap,edit_y1,
		MAX_TILES_ACROSS,MAX_TILES_DOWN, MAX_WIDTH-1, MAX_HEIGHT-1, 1,grRED,1,grRED,grBACKGROUND);

	// where we display the modes
	mode_x = tile_x1;
	mode_y = tile_y2 + gap;
	}

//----------------------------------------------------------------------------
//
//		CREATBUTTONS
//
//----------------------------------------------------------------------------
void BitmapWindow::CreateButtons()
	{
	int	width = GUIbitmap[0]->hdr.width,
		height = GUIbitmap[0]->hdr.height;

	// ---------------------------------------------------------------------

	// Disk/System Button creation

	int	x1 = max_edit_x2+gap,
		x2 = x1 + width,
		x3 = x2 + width,
		x4 = x3 + width,
		x5 = x4 + width,
		y  = mode_y;

	// first row
	y += grFontHeight();
	AddAccessory(new FussyButton(x1,y,
					GUIbitmap[GUI_SINGLE],GUIbitmap[GUI_SINGLE+1],255,0,cmdSingle,cmdSingle));

	AddAccessory(new FussyButton(x2,y,
					GUIbitmap[GUI_REPEAT],GUIbitmap[GUI_REPEAT+1],255,0,cmdRepeat,cmdRepeat));

	AddAccessory(new FussyButton(x3,y,
					GUIbitmap[GUI_TILE],GUIbitmap[GUI_TILE+1],255,0,cmdTile,cmdTile));

	AddAccessory(new FussyButton(x4,y,
					GUIbitmap[GUI_CLEARTILE],GUIbitmap[GUI_CLEARTILE+1],255,0,cmdClearTile,cmdClearTile));

	y += height;
	AddAccessory(new PictureButton(x1,y,
					GUIbitmap[GUI_LOAD],GUIbitmap[GUI_LOAD+1],255,0,cmdLoad));

	AddAccessory(new PictureButton(x2,y,
					GUIbitmap[GUI_LOADALL],GUIbitmap[GUI_LOADALL+1],255,0,cmdLoadAll));

	AddAccessory(new PictureButton(x3,y,
					GUIbitmap[GUI_SAVE],GUIbitmap[GUI_SAVE+1],255,0,cmdSave));

	AddAccessory(new PictureButton(x4,y,
					GUIbitmap[GUI_SAVEALL],GUIbitmap[GUI_SAVEALL+1],255,0,cmdSaveAll));

	AddAccessory(new PictureButton(x5,y,
					GUIbitmap[GUI_QUIT],GUIbitmap[GUI_QUIT+1],255,0,cmdExitProgram));

	y += height;
	AddAccessory(new FussyButton(x1,y,
					GUIbitmap[GUI_FILL],GUIbitmap[GUI_FILL+1],255,0,cmdFill,cmdFill));

	AddAccessory(new FussyButton(x2,y,
					GUIbitmap[GUI_FULLFILL],GUIbitmap[GUI_FULLFILL+1],255,0,cmdFillImage,cmdClearImage));

	AddAccessory(new PictureButton(x3,y,
					GUIbitmap[GUI_STORE],GUIbitmap[GUI_STORE+1],255,0,cmdStoreImage));

	AddAccessory(new PictureButton(x4,y,
					GUIbitmap[GUI_GET],GUIbitmap[GUI_GET+1],255,0,cmdGetImage));

	AddAccessory(new FussyButton(x5,y,
					GUIbitmap[GUI_UNDO],GUIbitmap[GUI_UNDO+1],255,0,cmdUndo,cmdUndo));

	y += height;
	AddAccessory(new FussyButton(x1,y,
					GUIbitmap[GUI_ERASE],GUIbitmap[GUI_ERASE+1],255,0,cmdErase,cmdFilledErase));

	AddAccessory(new FussyButton(x2,y,
					GUIbitmap[GUI_DRAW],GUIbitmap[GUI_DRAW+1],255,0,cmdDraw,cmdFilledDraw));

	AddAccessory(new FussyButton(x3,y,
					GUIbitmap[GUI_LINE],GUIbitmap[GUI_LINE+1],255,0,cmdLine,cmdFlatLine));

	AddAccessory(new FussyButton(x4,y,
					GUIbitmap[GUI_RECT],GUIbitmap[GUI_RECT+1],255,0,cmdRect,cmdFilledRect));

	AddAccessory(new FussyButton(x5,y,
					GUIbitmap[GUI_ELLIPSE],GUIbitmap[GUI_ELLIPSE+1],255,0,cmdEllipse,cmdFilledEllipse));

	y += height;
	AddAccessory(new FussyButton(x1,y,
					GUIbitmap[GUI_FLIPX],GUIbitmap[GUI_FLIPX+1],0,0,cmdFlipHorizontal,cmdFlipHorizontal));

	AddAccessory(new FussyButton(x2,y,
					GUIbitmap[GUI_FLIPY],GUIbitmap[GUI_FLIPY+1],0,0,cmdFlipVertical,cmdFlipVertical));

	AddAccessory(new FussyButton(x3,y,
					GUIbitmap[GUI_ROTATE90],GUIbitmap[GUI_ROTATE90+1],0,0,cmdRotate90,cmdRotate270));

	AddAccessory(new FussyButton(x4,y,
					GUIbitmap[GUI_ROTATE180],GUIbitmap[GUI_ROTATE180+1],0,0,cmdRotate180,cmdRotate180));

	AddAccessory(new FussyButton(x5,y,
					GUIbitmap[GUI_ROTATE270],GUIbitmap[GUI_ROTATE270+1],0,0,cmdRotate270,cmdRotate90));

	// ---------------------------------------------------------------------

	y += height;
	AddAccessory(new FussyButton(x1,y,
					GUIbitmap[GUI_LEFT],GUIbitmap[GUI_LEFT+1],0,0,cmdScrollLeft,cmdScrollLeftFive));

	AddAccessory(new FussyButton(x2,y,
					GUIbitmap[GUI_TOP],GUIbitmap[GUI_TOP+1],0,0,cmdScrollUp,cmdScrollUpFive));

	AddAccessory(new FussyButton(x3,y,
					GUIbitmap[GUI_BOTTOM],GUIbitmap[GUI_BOTTOM+1],0,0,cmdScrollDown,cmdScrollDownFive));

	AddAccessory(new FussyButton(x4,y,
					GUIbitmap[GUI_RIGHT],GUIbitmap[GUI_RIGHT+1],0,0,cmdScrollRight,cmdScrollRightFive));

	y += height;
	AddAccessory(new FussyButton(x1,y,
					GUIbitmap[GUI_LTARROW],GUIbitmap[GUI_LTARROW+1],0,0,cmdIncreaseWidthLeft,cmdDecreaseWidthLeft));

	AddAccessory(new FussyButton(x2,y,
					GUIbitmap[GUI_UPARROW],GUIbitmap[GUI_UPARROW+1],0,0,cmdIncreaseHeightTop,cmdDecreaseHeightTop));

	AddAccessory(new FussyButton(x3,y,
					GUIbitmap[GUI_DNARROW],GUIbitmap[GUI_DNARROW+1],0,0,cmdIncreaseHeightBottom,cmdDecreaseHeightBottom));

	AddAccessory(new FussyButton(x4,y,
					GUIbitmap[GUI_RTARROW],GUIbitmap[GUI_RTARROW+1],0,0,cmdIncreaseWidthRight,cmdDecreaseWidthRight));

	y += height;
	AddAccessory(new PictureButton(x1,y,
					GUIbitmap[GUI_PALETTE],GUIbitmap[GUI_PALETTE+1],0,0,cmdGotoPalette));
	// ---------------------------------------------------------------------

	// RangeBar Button creation

	x1 = thumb_y1-GUIbitmap[GUI_TOP]->hdr.height;
	x2 = thumb_y2;
	y = thumb_x1;
	AddAccessory(new PictureButton(y,x1,
					GUIbitmap[GUI_TOP],GUIbitmap[GUI_TOP+1],0,0,cmdGotoFirstImage));

	AddAccessory(new PictureButton(y,x2,
					GUIbitmap[GUI_BOTTOM],GUIbitmap[GUI_BOTTOM+1],0,0,cmdGotoLastImage));
	}

//----------------------------------------------------------------------------
//
//		SETUPIMAGES
//
//----------------------------------------------------------------------------
void BitmapWindow::SetupImages(String* name)
	{
	// set up the definition file which lists all the image to work with
	definition_name = new String(name->Contents());
//DJD
	definition_file = new DEFFile(".\\",definition_name->Contents());
//	definition_file = new DEFFile(definition_name);

	// are we working with a pcx image
	if (name->Contains(".PCX"))
		{
		number_images = 1;
		definition_str = new String (name->Contents());
		}
	else if (name->Contains(".DEF"))
		{
		LoadDEFFile();
		}
	else
		{
		grExitVesa();
		printf("\n%s is an unrecognised file type!!!",name->Contents());
		printf("\nFiles must be either .PCX or .DEF.\n\n");
		exit(1);
		}

	LoadPCXFiles();

	LoadImages();
	}

//----------------------------------------------------------------------------
//
//		DISK FUNCTIONS
//
//----------------------------------------------------------------------------

void BitmapWindow::LoadDEFFile()
	{
	if (!definition_file->Exists())
		{
		// quit with an error message
		grExitVesa();
		printf("Error: Definition file %s not found.\n\n",definition_name->Contents());
		exit(1);
		}

	// load in the definition file
	definition_file->Open();
	definition_str = definition_file->RecordList();
	definition_file->Close();

	number_images = definition_str->NumElems();
	}

void BitmapWindow::FreeDEFFile()
	{
	if (definition_name->Length() > 0) delete definition_str;

	delete definition_file;
	delete definition_name;
	}

void BitmapWindow::LoadPCXFiles()
	{
	int i;

	pcx_file = (PCXFile**) malloc(sizeof(PCXFile*)*number_images);

	for (i=0; i<number_images; i++)
		{
//DJD
		pcx_file[i] = new PCXFile(".\\",(definition_str->Elem(i))->Contents());
//		pcx_file[i] = new PCXFile((definition_str->Elem(i))->Contents());

		DBG(i*10,pcx_file[i]->FileName());
		}
	}

void BitmapWindow::FreePCXFiles()
	{
	int i;

	for (i=number_images-1;i>=0;i--)
		{
		DBG(i*10,pcx_file[i]->FileName());

		pcx_file[i]->Close();

		delete pcx_file[i];
		}

	free(pcx_file);
	}

bool BitmapWindow::CheckImageOnDiskIsOK(PCXFile* pf)
	{
	bool isOK = true;

	pf->Open();

	current_image = pf->LoadImage();

	if (current_image)
		{
		// if any of the bounds are out then file is not suitable
		if (current_image->hdr.width > MAX_WIDTH) isOK = false;
		if (current_image->hdr.height > MAX_HEIGHT) isOK = false;

		free(current_image);
		}
	else
		{
		isOK = false;
		}

	pf->Close();

	return isOK;
	}

void BitmapWindow::LoadImages()
	{
	image = (grImage**) malloc(sizeof(grImage*)*number_images);

	for (int i=0;i<number_images;i++)
		{
		// initially make all image as large as they're allowed to be
		image[i] = grCreateImage(0,0, MAX_WIDTH-1,MAX_HEIGHT-1,true);

		// make sure the image is within bounds and present
		if (CheckImageOnDiskIsOK(pcx_file[i]))
			{
			// then read the file off the disk
			pcx_file[i]->Open();
			pcx_file[i]->LoadImage(image[i]);
			pcx_file[i]->Close();

			// since we have found an image and loaded it
			// we'll use this one as a default for all the other ones
			default_width = image[i]->hdr.width;
			default_height = image[i]->hdr.height;
			}
		else
			{
			MakeImageDefaultSize(image[i]);
			}
		}
	}

void BitmapWindow::SaveImages()
	{
	for (int i=0;i<number_images;i++)
		{
		pcx_file[i]->Open();
		pcx_file[i]->SaveImage(image[i]);
		pcx_file[i]->Close();
		}
	}

void BitmapWindow::FreeImages()
	{
	// free all the image
	for (int i=number_images-1;i>=0;i--)
		{
		free(image[i]);
		}

	free(image);
	}

//----------------------------------------------------------------------------
//
//		RESIZE FUNCTIONS
//
//----------------------------------------------------------------------------
void BitmapWindow::ResetCurrentImage()
	{
	current_image->hdr.width = MAX_WIDTH*3-1;
	current_image->hdr.height = MAX_HEIGHT*3-1;
	}

void BitmapWindow::ResetExtendedImage()
	{
	extended_image->hdr.width = MAX_WIDTH*3-1;
	extended_image->hdr.height = MAX_HEIGHT*3-1;

	grClearImage(extended_image,0);
	}

void BitmapWindow::ResetUndo()
	{
	undo->hdr.width = MAX_WIDTH*3-1;
	undo->hdr.height = MAX_HEIGHT*3-1;
	}

void BitmapWindow::ResetWorking()
	{
	working->hdr.width = MAX_WIDTH*3-1;
	working->hdr.height = MAX_HEIGHT*3-1;
	}

void BitmapWindow::ResetStore()
	{
	store->hdr.width = MAX_WIDTH*3-1;
	store->hdr.height = MAX_HEIGHT*3-1;
	}

void BitmapWindow::ResetClipboard()
	{
	clipboard->hdr.width = MAX_WIDTH*3-1;
	clipboard->hdr.height = MAX_HEIGHT*3-1;
	}

void BitmapWindow::MakeImageDefaultSize(grImage* img)
	{
	img->hdr.width = default_width;
	img->hdr.height = default_width;
	}

void BitmapWindow::ResizeImageIfNecessary(grImage* img)
	{
	if (img->hdr.width < 1) img->hdr.width = 1;
	if (img->hdr.height < 1) img->hdr.height = 1;

	if (img->hdr.width > MAX_WIDTH) img->hdr.width = MAX_WIDTH;
	if (img->hdr.height > MAX_HEIGHT) img->hdr.height = MAX_HEIGHT;
	}

void BitmapWindow::ResizeEditRegion()
	{
	int cw,ch;

	if (input_grid_mode == EXTENDED_SINGLE)
		{
		// allow a gap of four on all sides
		edit_region->SetWidth(tile_width+EXTRA_SPACE*2);
		edit_region->SetHeight(tile_height+EXTRA_SPACE*2);

		// unless we only need extra space on one side
		if ((tile_selected_x == 0) || (tile_selected_x == tiles_across-1))
			edit_region->SetWidth(tile_width+EXTRA_SPACE);
		if ((tile_selected_y == 0) || (tile_selected_y == tiles_down-1))
			edit_region->SetHeight(tile_height+EXTRA_SPACE);

		// or even none
		if (tiles_across==1)
			edit_region->SetWidth(tile_width);
		if (tiles_down==1)
			edit_region->SetHeight(tile_height);
		}
	else if (input_grid_mode == EXTENDED)
		{
		current_image = extended_image;

		// no gaps needed
		edit_region->SetWidth(current_image->hdr.width);
		edit_region->SetHeight(current_image->hdr.height);
		}
	else
		{
		current_image = image[current_image_index];

		// allow a gap of four on all sides
		edit_region->SetWidth(current_image->hdr.width);
		edit_region->SetHeight(current_image->hdr.height);
		}

	if ((edit_region->Width()!=0) && (edit_region->Height()!=0))
		{
		cw = SIZEOF_EDITREGION/edit_region->Width() - 1;
		ch = SIZEOF_EDITREGION/edit_region->Height() - 1;

		if (cw>ch) cw = ch;
		else ch = cw;

		edit_region->SetCellWidth(cw);
		edit_region->SetCellHeight(ch);
		}
	}

void BitmapWindow::AdjustCoords(int* x, int* y)
	{
	if (tile_selected_x != 0)
		*x = *x+tile_selected_x*tile_width-EXTRA_SPACE;

	if (tile_selected_y != 0)
		*y = *y+tile_selected_y*tile_height-EXTRA_SPACE;
	}

//----------------------------------------------------------------------------
//
//		DRAW FUNCTIONS
//
//----------------------------------------------------------------------------
void BitmapWindow::DrawMouseXYPos(grPoint m)
	{
	String*	xystr = new String();

	xystr->Append("X-->");
	xystr->Append((long)m.x);
	xystr->Append(" Y-->");
	xystr->Append((long)m.y);
	xystr->Append("  ");

	grBlitString(tile_x1,0,grBACKGROUND,xystr->Contents());

	delete xystr;
	}

void BitmapWindow::DrawMode()
	{
	String*	modestr = new String(50);

	switch (output_fill_mode)
		{
		case NONE:		modestr->Append("No Fill "); break;
		case SOLID:		modestr->Append("Filled  "); break;
		}

	switch (output_draw_mode)
		{
		case ERASE:		modestr->Append(" Erase    "); break;
		case DRAW:		modestr->Append(" Draw     "); break;
		case LINE:		modestr->Append(" Line     "); break;
		case FLATLINE:	modestr->Append(" FlatLine "); break;
		case RECT:		modestr->Append(" Rect     "); break;
		case ELLIPSE:	modestr->Append(" Ellipse  "); break;
		}

	grBlitString(mode_x,mode_y,grBACKGROUND,modestr->Contents());

	delete modestr;
	}

void BitmapWindow::DrawMessage(bool just_clear_it)
	{
	// clear the whole message region
	grDrawFilledRect(edit_x1,max_edit_y2,max_edit_x2,palette_y1-1,grBACKGROUND);
	grBlitRegion(edit_x1,max_edit_y2,max_edit_x2,palette_y1-1);

	if (!just_clear_it)
		grBlitString(edit_x1,max_edit_y2,grBACKGROUND,message->Contents());
	}

void BitmapWindow::DrawImageDetails()
	{
	String*	xystr = new String(50);

	if (input_grid_mode == EXTENDED_SINGLE)
		{
		current_image = extended_image;

		xystr->Append((long)image[tile[tile_selected_x][tile_selected_y]-1]->hdr.width);
		xystr->Append(" by ");
		xystr->Append((long)image[tile[tile_selected_x][tile_selected_y]-1]->hdr.height);

		grBlitString(edit_x1,0,grBACKGROUND,"Extended >%s<          ",pcx_file[tile[tile_selected_x][tile_selected_y]-1]->FileName());
		}
	else if (input_grid_mode == EXTENDED)
		{
		current_image = extended_image;

		xystr->Append((long)current_image->hdr.width);
		xystr->Append(" by ");
		xystr->Append((long)current_image->hdr.height);

		grBlitString(edit_x1,0,grBACKGROUND,"Extended                  ");
		}
	else
		{
		current_image = image[current_image_index];

		xystr->Append((long)current_image->hdr.width);
		xystr->Append(" by ");
		xystr->Append((long)current_image->hdr.height);

		grBlitString(edit_x1,0,grBACKGROUND,"%s           ",pcx_file[current_image_index]->FileName());
		}

	grBlitString(max_edit_x2-grFontWidth(xystr->Contents()),0,grBACKGROUND,xystr->Contents());

	delete xystr;
	}

//----------------------------------------------------------------------------
//
//		DRAWGRID
//
//----------------------------------------------------------------------------
void BitmapWindow::DrawEdit(int x1,int y1,int x2,int y2)
	{
	int		x,y, xx,yy, xx1,yy1, xx2,yy2;
	uchar	color;
	grImage*work;

	// resize the main grid so that it fits the current image
	ResizeEditRegion();

	// bounds and order check
	CLIP2(x1,y1,0,0,edit_region->Width()-1,edit_region->Height()-1);
	CLIP2(x2,y2,0,0,edit_region->Width()-1,edit_region->Height()-1);
	if (x2<x1) { int temp = x1; x1 = x2; x2 = temp; }
	if (y2<y1) { int temp = y1; y1 = y2; y2 = temp; }

	// only draw the full grid if the default parameters came into function
	bool	drawfullgrid = ((x1==0)&&(y1==0)&&(x2==edit_region->Width()-1)&&(y2==edit_region->Height()-1));

	// blank max area the grid can take
	if (drawfullgrid)
		{
		grDrawFilledRect(edit_x1,edit_y1, max_edit_x2,max_edit_y2, grBACKGROUND);

		// draw the GRID for the editting image
		edit_region->Draw();
		}

	if (input_grid_mode == SINGLE)
		{
		current_image = image[current_image_index];

		xx1 = (edit_region->TopLeft()).x;
		yy1 = (edit_region->TopLeft()).y;
		xx2 = (edit_region->BottomRight()).x;
		yy2 = (edit_region->BottomRight()).y;

		grDrawRect(xx1,yy1,xx2,yy2,grLIGHTRED);

		xx1 = 0;
		yy1 = 0;
		}
	else
		{
		current_image = extended_image;

		if (input_grid_mode == EXTENDED_SINGLE)
			{
			if (tile_selected_x == 0)
				{
				xx1 = (edit_region->CellTopLeft(0,0)).x-1;
				if (tiles_across>1)
					xx2 = (edit_region->CellTopLeft(edit_region->Width()-EXTRA_SPACE,0)).x-1;
				else
					xx2 = (edit_region->CellTopLeft(edit_region->Width(),0)).x-1;
				}
			else if (tile_selected_x == tiles_across-1)
				{
				xx1 = (edit_region->CellTopLeft(EXTRA_SPACE,0)).x-1;
				xx2 = (edit_region->CellTopLeft(edit_region->Width(),0)).x-1;
				}
			else
				{
				xx1 = (edit_region->CellTopLeft(EXTRA_SPACE,0)).x-1;
				xx2 = (edit_region->CellTopLeft(edit_region->Width()-EXTRA_SPACE,0)).x-1;
				}

			if (tile_selected_y == 0)
				{
				yy1 = (edit_region->CellTopLeft(0,0)).y-1;
				if (tiles_down>1)
					yy2 = (edit_region->CellTopLeft(0,edit_region->Height()-EXTRA_SPACE)).y-1;
				else
					yy2 = (edit_region->CellTopLeft(0,edit_region->Height())).y-1;
				}
			else if (tile_selected_y == tiles_across-1)
				{
				yy1 = (edit_region->CellTopLeft(0,EXTRA_SPACE)).y-1;
				yy2 = (edit_region->CellTopLeft(0,edit_region->Height())).y-1;
				}
			else
				{
				yy1 = (edit_region->CellTopLeft(0,EXTRA_SPACE)).y-1;
				yy2 = (edit_region->CellTopLeft(0,edit_region->Height()-EXTRA_SPACE)).y-1;
				}

			grDrawRect(xx1,yy1,xx2,yy2,grLIGHTRED);

			xx1 = 0;
			yy1 = 0;
			AdjustCoords(&xx1,&yy1);
			}
		else if (input_grid_mode == EXTENDED)
			{
			for (x=0; x<=tiles_across; x++)
				{
				xx1 = (edit_region->CellTopLeft(tile_width*x,0)).x-1;
				grDrawVLine(xx1,edit_y1+1,edit_y2-1,grLIGHTRED);
				}

			for (y=0; y<=tiles_down; y++)
				{
				yy1 = (edit_region->CellTopLeft(0,tile_height*y)).y-1;
				grDrawHLine(yy1,edit_x1+1,edit_x2-1,grLIGHTRED);
				}
			xx1 = 0;
			yy1 = 0;
			}
		}

	// put all the colours in it
	for (yy=y1; yy<=y2; yy++)
		for (xx=x1; xx<=x2; xx++)
			{
			color = grBLACK;

			if ((xx1+xx>=0) && (yy1+yy>=0) && (xx1+xx<current_image->hdr.width) && (yy1+yy<current_image->hdr.height))
				color = current_image->data[Index(current_image,xx1+xx,yy1+yy)];

			// then draw it
			edit_region->FillCell(xx,yy, color);
			}

	if (drawfullgrid)
		grBlitRegion(edit_x1,edit_y1,max_edit_x2,max_edit_y2);
	else
		grBlitRegion(
					(edit_region->CellTopLeft(x1,y1)).x,	(edit_region->CellTopLeft(x1,y1)).y,
					(edit_region->CellBottomRight(x2,y2)).x,(edit_region->CellBottomRight(x2,y2)).y
					);
	}

//----------------------------------------------------------------------------
//
//		DRAWTHUMBNAILS
//
//----------------------------------------------------------------------------
void BitmapWindow::DrawThumb(bool draw_current_image)
	{
	int cnt, i;

	DrawImageDetails();
	grBlitString(grMaxx-grFontWidth("WWWWWWWW.WWW"),0,grBACKGROUND,pcx_file[current_image_index]->FileName());

	thumbnails->Draw();

	// select the image 2 image before the currently selected one
	thumbnail_start_index = current_image_index - thumbnails->Height()/2;

	// and if that means we'll display blanks on the end then move back
	while (thumbnail_start_index>number_images-thumbnails->Height()) thumbnail_start_index--;

	// and if that is before the first one then move along until it isn't
	while (thumbnail_start_index<0) thumbnail_start_index++;

	i = thumbnail_start_index;

	// select the image that we are currently working on
	grPoint selectedcell = { 0,current_image_index-i };
	thumbnails->SelectCell(selectedcell,grLIGHTRED);

	// put all the images in it
	for (cnt=0; cnt<thumbnails->Height(); cnt++)
		{
		if (i<number_images)
			if (image[i])
				{
				if (!draw_current_image || (i == current_image_index))
					{
					grPoint p1 = thumbnails->CellTopLeft(0,cnt);
					grPoint p2 = thumbnails->CellBottomRight(0,cnt);

					grPutScaledImage(p1.x,p1.y, p2.x,p2.y, image[i]);

					if (draw_current_image)	thumbnails->BlitCell(0,cnt);
					}
				}
		i++;
		}

	if (!draw_current_image) thumbnails->BlitGrid();
	}

//----------------------------------------------------------------------------
//
//		DRAWTILES
//
//----------------------------------------------------------------------------
void BitmapWindow::DrawTile()
	{
	int		x1,y1,x2,y2, width,height;

	// blank the area
	grDrawFilledRect(tile_x1,tile_y1, max_tile_x2,max_tile_y2, grBACKGROUND);

	if (input_grid_mode == SINGLE)
		current_image = image[current_image_index];
	else if ((input_grid_mode == EXTENDED) || (input_grid_mode == EXTENDED_SINGLE))
		{
		current_image = extended_image;

		DrawTileSelect();
		}

	if (output_tile_mode == SINGLE)
		{
		width = current_image->hdr.width;
		height = current_image->hdr.height;

		// put the single/extended image on the screen
		grPutImage(tile_x1,tile_y1,tile_x1+width-1,tile_y1+height-1,current_image);

		// maybe draw a red box around the individual image so we can see it
		if ((input_grid_mode == EXTENDED_SINGLE) && (output_bounds_mode))
			small_tile_grid->SelectCell(tile_selected_x,tile_selected_y,grLIGHTRED);
		}
	else
		{
		width = current_image->hdr.width;
		height = current_image->hdr.height;

		x1 = tile_x1;
		y1 = tile_y1;

		x2 = x1 + width - 1;
		y2 = y1 + height - 1;

		// while we haven't gone past the maximum bounds
		while (y2<max_tile_y2)
			{
			while (x2<max_tile_x2)
				{
				// keep on putting the single/extended image on the screen
				grPutImage(x1,y1, x2,y2, current_image);

				// move along to the next position across
				x1 = x2 + 1;
				x2 = x1 + width - 1;
				}
			// finished moving across so go back to the left
			x1 = tile_x1; x2 = x1 + width - 1;
			// and go down line
			y1 = y2 + 1; y2 = y1 + height - 1;
			}
		}
	grBlitRegion(tile_x1,tile_y1, max_tile_x2,max_tile_y2);
	}

//----------------------------------------------------------------------------
//
//		DRAWPALETTE
//
//----------------------------------------------------------------------------
void BitmapWindow::DrawPalette()
	{
	int		xx,yy;

	// draw the grid for selection of colour for editting with
	palette->Draw();

	// select the colours that we are currently working with
	grPoint selectedcell1 = palette->IndexToCell(color1);
	grPoint selectedcell2 = palette->IndexToCell(color2);
	palette->SelectCell(selectedcell1,grLIGHTRED);
	palette->SelectCell(selectedcell2,grRED);

	// put all the colours in it
	for (yy=0; yy<palette->Height(); yy++)
		for (xx=0; xx<palette->Width(); xx++)
			{
			palette->FillCell(xx,yy,palette->CellToIndex(xx,yy));
			}

	palette->BlitGrid();
	}

//----------------------------------------------------------------------------
//
//		DRAWTILESELECT
//
//----------------------------------------------------------------------------
void BitmapWindow::DrawTileSelect()
	{
	small_tile_grid->SetCellWidth(tile_width-1);
	small_tile_grid->SetCellHeight(tile_height-1);

	small_tile_grid->Draw();
	}

//----------------------------------------------------------------------------
//
//		HANDLEMOUSEINPUT
//
//----------------------------------------------------------------------------
bool BitmapWindow::HandleMouseInput(Event& e)
	{
	if (input_grid_mode == SINGLE)
		current_image = image[current_image_index];
	else if ((input_grid_mode == EXTENDED) || (input_grid_mode == EXTENDED))
		current_image = extended_image;

	if (edit_region->IsSelected(e.mouse.where))
		{
		if ((dragging == BRUSH) || (dragging == NONE))
			{
			HandleEditInput(e);
			return true;
			}
		}
	else if (dragison && (dragging == BRUSH))
		{
		// if we were dragging we would have had an original image to work from
		// get it back now as we have aborted the dragging if we came in here
		memcpy(current_image,working,grImageSize(0,0,working->hdr.width-1,working->hdr.height-1));

		if (input_grid_mode == EXTENDED) UseTiledImage();

		dragison = false;
		dragging = NONE;

		DrawEdit();
		DrawTile();
		DrawThumb();

		return true;
		}

	if (small_tile_grid->IsSelected(e.mouse.where))
		{
		HandleTileInput(e);
		return true;
		}

	if (palette->IsSelected(e.mouse.where))
		{
		HandlePaletteInput(e);
		return true;
		}

	if (thumbnails->IsSelected(e.mouse.where))
		{
		HandleThumbInput(e);
		return true;
		}

	return false;
	}

//----------------------------------------------------------------------------
//
//		HANDLEEDITINPUT
//
//----------------------------------------------------------------------------
void BitmapWindow::HandleEditInput(Event& e)
	{
	int			x,y, x1,y1,x2,y2, oldx1,oldy1,oldx2,oldy2;

	// if left mousebutton use color1
	uchar		color = ((e.mouse.event == mb_leftdn)||(e.mouse.event == mb_leftup)||(e.mouse.event == mb_leftauto))
				? color1 : color2;

	// find the cell currently being pointed at
	grCell		cell = edit_region->WhichCell(e.mouse.where);

	if (input_grid_mode == SINGLE)
		current_image = image[current_image_index];
	else if ((input_grid_mode == EXTENDED) || (input_grid_mode == EXTENDED))
		current_image = extended_image;

	if (!dragison)
		{
		memcpy(undo,current_image,grImageSize(0,0,current_image->hdr.width-1,current_image->hdr.height-1));
		undoUsed = true;
		}

	// then do something with the selection depending on right or left
	switch (output_draw_mode)
		{
		case ERASE:
			color = grBACKGROUND;
		case DRAW:
			PutPoint(cell,color);
			break;

		case LINE:
		case RECT:
		case ELLIPSE:
			if (!dragison)
				{
				memcpy(working,current_image,grImageSize(0,0,current_image->hdr.width-1,current_image->hdr.height-1));

				first_x = cell.x;
				first_y = cell.y;

				dragison = true;
				dragging = BRUSH;
				}
			else
				{
				memcpy(current_image,working,grImageSize(0,0,working->hdr.width-1,working->hdr.height-1));

				// store the old values for the second control
				old_second_x = second_x;
				old_second_y = second_y;

				// get the new ones
				second_x = cell.x;
				second_y = cell.y;

				// and put it on the grid, image and tiles
				switch (output_draw_mode)
					{
					case LINE:
						PutLine(color);
						break;

					case RECT:
						PutRect(color);
						break;

					case ELLIPSE:
						PutEllipse(color);
						break;
					}
				}
			break;
		}

	// draw the real life image, tile or repeat, etc... actual sized
	DrawTile();

	// copy the extended image to each of its component images
	if ((input_grid_mode == EXTENDED) || (input_grid_mode == EXTENDED_SINGLE))
		{
		UseTiledImage();
//		MakeTiledImage();
		DrawThumb();
		}
	else
		DrawThumb(true);
	}

//----------------------------------------------------------------------------
//
//		HANDLETILEINPUT
//
//----------------------------------------------------------------------------
void BitmapWindow::HandleTileInput(Event &e)
	{
	// make sure the function doesn't respond to the mouse down events
	// otherwise single click makes a double jump
	if ((e.mouse.event == mb_leftdn)||(e.mouse.event == mb_leftauto)||
			(e.mouse.event == mb_rightdn)||(e.mouse.event == mb_rightauto))
		return;

	if ((input_grid_mode != EXTENDED) && (input_grid_mode != EXTENDED_SINGLE))
		return;

	int x,y;
	grCell cell = small_tile_grid->WhichCell(e.mouse.where);
	bool Ok;

	// for the left mouse button which selects/deselects a single image from an extended image
	if ((e.mouse.event == mb_leftdn)||(e.mouse.event == mb_leftup)||(e.mouse.event == mb_leftauto))
		{
		if ((tile_selected_x == cell.x) && (tile_selected_y == cell.y))
			{
			if (input_grid_mode == EXTENDED_SINGLE)
				input_grid_mode = EXTENDED;
			else
				input_grid_mode = EXTENDED_SINGLE;
			}
		else
			input_grid_mode = EXTENDED_SINGLE;
		}
	else
		{
		// show that a tile position has been selected
		small_tile_grid->SelectCell(cell,grLIGHTRED);
		small_tile_grid->BlitCell(cell);

		// if there is a tile where we have clicked the right mouse button
		if (tile[cell.x][cell.y])
			{
			// then get rid of it
			tile[cell.x][cell.y] = 0;
			}
		else
			{
			// all images have to be the same size
			if ((!any_tiles_selected) || ((image[current_image_index]->hdr.width==tile_width) && (image[current_image_index]->hdr.height==tile_height)))
				{
				Ok = true;

				// if this tile is somewhere else within the extended image
				for (y=0; y<tiles_down; y++)
					for (x=0; x<tiles_across; x++)
						if (tile[x][y] == current_image_index+1)
							// then can't move it
							Ok = false;
//							// then get rid of it
//							tile[x][y] = 0;

				if (Ok)
					tile[cell.x][cell.y] = current_image_index+1;
				}
			else
				{
				message->Set("All tiles must be the same size");
				DrawMessage();
				}
			}
		}

	if (tile[cell.x][cell.y])
		{
		tile_selected_x = cell.x;
		tile_selected_y = cell.y;
		}

	MakeTiledImage();

	DrawTile();
	DrawEdit();
	DrawImageDetails();
	}

//----------------------------------------------------------------------------
//
//		HANDLEPALETTEINPUT
//
//----------------------------------------------------------------------------
void BitmapWindow::HandlePaletteInput(Event &e)
	{
	// get feedback from the base1 grid
	grCell cell = palette->WhichCell(e.mouse.where);

	// then if its the left mouse button
	if ((e.mouse.event == mb_leftdn)||(e.mouse.event == mb_leftup)||(e.mouse.event == mb_leftauto))
		{
		palette->DeselectCell(palette->IndexToCell(color1));
		palette->BlitCell(palette->IndexToCell(color1));

		palette->SelectCell(cell,grLIGHTRED);
		color1 = palette->CellToIndex(cell);
		}
	else // or right mouse button
		{
		palette->DeselectCell(palette->IndexToCell(color2));
		palette->BlitCell(palette->IndexToCell(color2));

		palette->SelectCell(cell,grRED);
		color2 = palette->CellToIndex(cell);
		}

	palette->BlitCell(cell);
	}

//----------------------------------------------------------------------------
//
//		HANDLETHUMBINPUT
//
//----------------------------------------------------------------------------
void BitmapWindow::HandleThumbInput(Event& e)
	{
	// make sure the function doesn't respond to the mouse up event
	// otherwise single click makes a double jump
	if (!((e.mouse.event == mb_leftdn)||(e.mouse.event == mb_leftauto)||
			(e.mouse.event == mb_rightdn)||(e.mouse.event == mb_rightauto)))
		return;

	current_image_index = thumbnail_start_index+thumbnails->CellToIndex(thumbnails->WhichCell(e.mouse.where));

	CLIP1(current_image_index,0,number_images-1);
	DrawThumb();

	// if we're in a mode that responds to the current image then redraw
	if ((input_grid_mode != EXTENDED) && (input_grid_mode != EXTENDED_SINGLE))
		{
		DrawEdit();
		DrawTile();
		}
	}

//----------------------------------------------------------------------------
//
//		CLEARTILEDIMAGE
//
//----------------------------------------------------------------------------
void BitmapWindow::ClearTiledImage()
	{
	int x,y;

	for (y=0; y<MAX_TILES_DOWN; y++)
		for (x=0; x<MAX_TILES_ACROSS; x++)
			tile[x][y] = 0;

	tile_selected_x = 0;
	tile_selected_y = 0;

	tiles_across = 0;
	tiles_down = 0;

	tile_width = default_width;
	tile_height = default_height;

	any_tiles_selected = false;

	ResetExtendedImage();
	}

//----------------------------------------------------------------------------
//
//		MAKETILEDIMAGE
//
//----------------------------------------------------------------------------
void BitmapWindow::MakeTiledImage()
	{
	int		i,x,y,x1,y1,x2,y2, mw=0,mh=0;

	tiles_across = 0;
	tiles_down = 0;

	ResetExtendedImage();

	for (y=0;y<MAX_TILES_DOWN;y++)
		for (x=0;x<MAX_TILES_ACROSS;x++)
			{
			if (tile[x][y])
				{
				current_image = image[tile[x][y]-1];
				tile_width = current_image->hdr.width;
				tile_height = current_image->hdr.height;

				if (x>=tiles_across) tiles_across = x+1;
				if (y>=tiles_down) tiles_down = y+1;
				}
			}
	extended_image->hdr.width = tiles_across*tile_width;
	extended_image->hdr.height = tiles_down*tile_height;

	if ((tiles_across==0) || (tiles_down==0))
		any_tiles_selected = false;
	else
		{
		any_tiles_selected = true;

		for (y=0;y<tiles_down;y++)
			for (x=0;x<tiles_across;x++)
				{
				if (tile[x][y])
					{
					current_image = image[tile[x][y]-1];

					x1 = tile_width * x;
					y1 = tile_height * y;

					x2 = x1 + tile_width - 1;
					y2 = y1 + tile_height - 1;

					grPutImageVirtual(extended_image,x1,y1,x2,y2,current_image);
					}
				}
		}
	}

//----------------------------------------------------------------------------
//
//		USETILEDIMAGE
//
//----------------------------------------------------------------------------
void BitmapWindow::UseTiledImage()
	{
	if ((input_grid_mode != EXTENDED) && (input_grid_mode != EXTENDED_SINGLE))
		return;

	int x,y, x1,y1,x2,y2, width,height;

	// copy the extended image into its respective tiles
	for (y=0;y<tiles_down;y++)
		for (x=0;x<tiles_across;x++)
			{
			if (tile[x][y])
				{
				grImage* img = image[tile[x][y]-1];
				width = img->hdr.width;
				height = img->hdr.height;

				x1 = width * x;
				y1 = height * y;

				x2 = x1 + width - 1;
				y2 = y1 + height - 1;

				grGetImageVirtual(extended_image,x1,y1,x2,y2,img);
				}
			}
	}

//----------------------------------------------------------------------------
//
//		PUTPOINT
//
//----------------------------------------------------------------------------
void BitmapWindow::PutPoint(grCell cell, uchar color)
	{
	int	x1,y1,x2,y2, minx,miny,maxx,maxy;

	// just fill from the cell if fill mode is on solid
	if (output_fill_mode == SOLID)
		{
		grFloodFill(current_image,cell.x,cell.y,color);
		DrawEdit();

		return;
		}

	// if we're dragging the brush around
	if (dragison)
		{
		// the first point is the old point
		first_x = second_x;
		first_y = second_y;
		// and second point is the new position
		second_x = cell.x;
		second_y = cell.y;

		// and depending on the order translate those to image coordinates
		if (first_x<=second_x)
			{
			x1 = first_x;	y1 = first_y;
			x2 = second_x;	y2 = second_y;
			}
		else
			{
			x1 = second_x;	y1 = second_y;
			x2 = first_x;	y2 = first_y;
			}

		if (input_grid_mode == EXTENDED_SINGLE)
			{
			AdjustCoords(&x1,&y1);
			AdjustCoords(&x2,&y2);
			}

		// and draw a line on the image
		grDrawLineVirtual(current_image,x1,y1,x2,y2,color);

		// find the max bounds made by both new and old
		minx = second_x;
		if (first_x<minx) minx = first_x;

		maxx = second_x;
		if (first_x>maxx) maxx = first_x;

		miny = second_y;
		if (first_y<miny) miny = first_y;

		maxy = second_y;
		if (first_y>maxy) maxy = first_y;

		// and draw the region of grid that has changed
		DrawEdit(minx,miny,maxx,maxy);

		return;
		}

	int index;

	memcpy(working,current_image,grImageSize(0,0,current_image->hdr.width-1,current_image->hdr.height-1));

	// if the mouse has only just been pushed
	// put a point on the image
	index = edit_region->CellToIndex(cell);

	if (input_grid_mode == EXTENDED_SINGLE)
		{
		x1 = cell.x;
		y1 = cell.y;

		AdjustCoords(&x1,&y1);
		index = edit_region->CellToIndex(x1,y1);
		}

	if ((index>=0) && (index<current_image->hdr.width*current_image->hdr.height))
		{
		current_image->data[index] = color;

		// and the grid as well
		edit_region->FillCell(cell,color);
		edit_region->BlitCell(cell);
		}

	// and setup variables in case they want to drag the brush around
	second_x = cell.x;
	second_y = cell.y;

	dragison = true;
	dragging = BRUSH;
	}

//----------------------------------------------------------------------------
//
//		PUTLINE
//
//----------------------------------------------------------------------------
void BitmapWindow::PutLine(uchar color)
	{
	int	x1,y1,x2,y2, oldx1,oldy1,oldx2,oldy2, minx,miny,maxx,maxy;

	// depending on the order translate ends of line to image coordinates
	if (first_x<=second_x)
		{
		x1 = first_x;	y1 = first_y;
		x2 = second_x;	y2 = second_y;
		}
	else
		{
		x1 = second_x;	y1 = second_y;
		x2 = first_x;	y2 = first_y;
		}

	if (input_grid_mode == EXTENDED_SINGLE)
		{
		AdjustCoords(&x1,&y1);
		AdjustCoords(&x2,&y2);
		}

	// and draw a line on the image
	grDrawLineVirtual(current_image,x1,y1,x2,y2,color);

	// find the max bounds made by both new and old
	minx = second_x;
	if (old_second_x<minx) minx = old_second_x;
	if (first_x<minx) minx = first_x;

	maxx = second_x;
	if (old_second_x>maxx) maxx = old_second_x;
	if (first_x>maxx) maxx = first_x;

	miny = second_y;
	if (old_second_y<miny) miny = old_second_y;
	if (first_y<miny) miny = first_y;

	maxy = second_y;
	if (old_second_y>maxy) maxy = old_second_y;
	if (first_y>maxy) maxy = first_y;

	// and draw the region of grid that has changed
	DrawEdit(minx,miny,maxx,maxy);
	}

//----------------------------------------------------------------------------
//
//		PUTRECT
//
//----------------------------------------------------------------------------
void BitmapWindow::PutRect(uchar color)
	{
	int	x1,y1,x2,y2, minx,miny,maxx,maxy;

	// depending on the order translate ends of line to image coordinates
	if (first_x<=second_x)
		{
		x1 = first_x;	y1 = first_y;
		x2 = second_x;	y2 = second_y;
		}
	else
		{
		x1 = second_x;	y1 = second_y;
		x2 = first_x;	y2 = first_y;
		}

	if (input_grid_mode == EXTENDED_SINGLE)
		{
		AdjustCoords(&x1,&y1);
		AdjustCoords(&x2,&y2);
		}

	// and draw a rect/filledrect on the image
	if (output_fill_mode)
		grDrawFilledRectVirtual(current_image,x1,y1,x2,y2,color);
	else
		grDrawRectVirtual(current_image,x1,y1,x2,y2,color);

	// find the max bounds made by both new and old
	minx = second_x;
	if (old_second_x<minx) minx = old_second_x;
	if (first_x<minx) minx = first_x;

	maxx = second_x;
	if (old_second_x>maxx) maxx = old_second_x;
	if (first_x>maxx) maxx = first_x;

	miny = second_y;
	if (old_second_y<miny) miny = old_second_y;
	if (first_y<miny) miny = first_y;

	maxy = second_y;
	if (old_second_y>maxy) maxy = old_second_y;
	if (first_y>maxy) maxy = first_y;

	// and draw the region of grid that has changed
	DrawEdit(minx,miny,maxx,maxy);
	}

//----------------------------------------------------------------------------
//
//		PUTELLIPSE
//
//----------------------------------------------------------------------------
void BitmapWindow::PutEllipse(uchar color)
	{
	int	x1,y1,x2,y2, oldx2,oldy2, minx,miny,maxx,maxy;

	// depending on the order translate ends of line to image coordinates
	x1 = first_x;
	y1 = first_y;
	x2 = abs(second_x-first_x);
	y2 = abs(second_y-first_y);

	if (input_grid_mode == EXTENDED_SINGLE)
		AdjustCoords(&x1,&y1);

	// and draw a rect/filledrect on the image
	if (output_fill_mode)
		grDrawFilledEllipseVirtual(current_image,x1,y1,x2,y2,color);
	else
		grDrawEllipseVirtual(current_image,x1,y1,x2,y2,color);

	// work out where the last drag image was
	oldx2 = abs(old_second_x-first_x);
	oldy2 = abs(old_second_y-first_y);

	// find the max bounds made by both new and old
	minx = ((first_x-oldx2) < (first_x-x2)) ? (first_x-oldx2) : (first_x-x2);
	miny = ((first_y-oldy2) < (first_y-y2)) ? (first_y-oldy2) : (first_y-y2);
	maxx = ((first_x+oldx2) > (first_x+x2)) ? (first_x+oldx2) : (first_x+x2);
	maxy = ((first_y+oldy2) > (first_y+y2)) ? (first_y+oldy2) : (first_y+y2);

	// and draw the region of grid that has changed
	DrawEdit(minx,miny,maxx,maxy);
	}
