//	-------------------------------------------------------------------------
//	BMP.HPP
//	-------------------------------------------------------------------------
//	AUTHOR:			Andrew Carvel Evans
//	DESCRIPTION:	This is the Bitmap Editting Window
//	-------------------------------------------------------------------------

#ifndef __bmp_hpp
#define __bmp_hpp

#include	"stddefns.hpp"
#include	"onscreen.hpp"
#include	"strings.hpp"
#include	"pcxfile.hpp"
#include	"deffile.hpp"
#include	"graphics.hpp"
#include	"grid.hpp"

#define		SIZEOF_EDITREGION		380
#define		EXTRA_SPACE				4
#define		MAX_WIDTH				64
#define		MAX_HEIGHT				64
#define		MAX_TILES_ACROSS		3
#define		MAX_TILES_DOWN			3
#define		SIZEOF_TILEAREA			MAX_TILES_ACROSS*MAX_WIDTH

class BitmapWindow : public OnScreenHandler
	{
	public:

					BitmapWindow(String* file_name);
		void		InitialiseVariables();
					~BitmapWindow();

		// multiple images
		String*		definition_name;		// definition file details that we
		DEFFile*	definition_file;		// use to select images to edit
		String*		definition_str;			// list of files in definition file

		int			number_images;			// number of images in definition file

		PCXFile**	pcx_file;				// for loading the images
		grImage**	image;					// current images

		int			current_image_index;

		grImage*	current_image;
		grImage*	extended_image;

	// -------------------------------------------------------------------

		// standard functions
		virtual void	Draw();
		virtual void	HandleEvent(Event& e);

	private:

		Grid*		edit_region;

		int			image_selected;

		int			default_width,
					default_height;

		// variables used while dragging
		int			first_x,
					first_y,
					second_x, old_second_x,
					second_y, old_second_y;

		// currently selected colors
		uchar		color1,			// left mouse button
					color2;			// right mouse button

	// -------------------------------------------------------------------

		// States

		// mouse state for continuous movement
		bool		mouseisdown;

		// are we dragging domething now?
		bool		dragison;

		// what we are currently dragging
		int			dragging;

		// input mode, eg. NORMAL_INPUT,TILE_INPUT
		int			input_query_mode;

		// the way we edit the images, eg. SINGLE or EXTENDED images
		int			input_grid_mode;

		// output mode, eg. ERASE,DRAW,LINE,RECT,ELLIPSE
		int			output_draw_mode;

		// fill shapes
		int			output_fill_mode;

		// tile mode
		int			output_tile_mode;

		// whether we draw red lines all over the place
		bool		output_bounds_mode;

		String*		message;

	// -------------------------------------------------------------------

		// for the undo command
		grImage*	undo;
		// for working with when resizing, scaling, etc.
		grImage*	working;
		// for store and retrieve command
		grImage*	store;
		// for copying and pasting parts of an image
		grImage*	clipboard;

		bool		undoUsed;
		bool		storeUsed;
		bool		clipboardUsed;

	// -------------------------------------------------------------------

		// Thumbnail Selection Bar
		Grid*		thumbnails;

		int			thumbnail_start_index;		// offset of the first thumbnail image

	// -------------------------------------------------------------------

		// Palette Select Region
		Grid*		palette;

	// -------------------------------------------------------------------

		// a grid used to select tile positions
		Grid*		small_tile_grid;

		// an array to store what images are in what positions on tile
		int			tile[MAX_TILES_ACROSS][MAX_TILES_DOWN];

		int			tile_selected_x,
					tile_selected_y;

		int			tiles_across,
					tiles_down;

		int			tile_width,
					tile_height;

		bool		any_tiles_selected;

	// -------------------------------------------------------------------

		// Special Text Positions

		// where we display output_draw_mode and output_fill_mode
		int			mode_x, mode_y;

	// -------------------------------------------------------------------

		// startup functions
		void		CreateGrids();
		void		CreateButtons();
		void		SetupImages(String* name);

		// disk functions
		void		LoadDEFFile();
		void		FreeDEFFile();

		void		LoadPCXFiles();
		void		FreePCXFiles();

		bool		CheckImageOnDiskIsOK(PCXFile* pf);
		void		LoadImages();
		void		SaveImages();
		void		FreeImages();

		// functions to reset the default/maximum/minimum image sizes
		void		ResetCurrentImage();
		void		ResetExtendedImage();

		void		ResetUndo();
		void		ResetWorking();
		void		ResetStore();
		void		ResetClipboard();

		void		MakeImageDefaultSize(grImage* img);
		void		ResizeImageIfNecessary(grImage* img);

		void		ResizeEditRegion();
		void		AdjustCoords(int* x,int* y);

		// draw functions
		void		DrawMouseXYPos(grPoint m);
		void		DrawMode();
		void		DrawMessage(bool just_clear_it = false);
		void		DrawImageDetails();

		void		DrawEdit(int x1=0,int y1=0,int x2=10000,int y2=10000);
		void		DrawThumb(bool draw_current_image = false);
		void		DrawTile();
		void		DrawPalette();

		void		DrawTileSelect();

		// handle event functions
		bool		HandleMouseInput(Event& e);

		void		HandleEditInput(Event& e);
		void		HandleTileInput(Event& e);
		void		HandleThumbInput(Event& e);
		void		HandlePaletteInput(Event& e);

		void		ClearTiledImage();
		void		MakeTiledImage();
		void		UseTiledImage();

		void		PutPoint(grCell cell,uchar color);
		void		PutLine(uchar color);
		void		PutRect(uchar color);
		void		PutEllipse(uchar color);
	};

#define		edit_x1		(edit_region->TopLeft()).x
#define		edit_y1		(edit_region->TopLeft()).y
#define		edit_x2		(edit_region->BottomRight()).x
#define		edit_y2		(edit_region->BottomRight()).y
#define		max_edit_x2	(edit_region->TopLeft()).x+SIZEOF_EDITREGION
#define		max_edit_y2	(edit_region->TopLeft()).y+SIZEOF_EDITREGION

#define		thumb_x1	(thumbnails->TopLeft()).x
#define		thumb_y1	(thumbnails->TopLeft()).y
#define		thumb_x2	(thumbnails->BottomRight()).x
#define		thumb_y2	(thumbnails->BottomRight()).y

#define		palette_x1	(palette->TopLeft()).x
#define		palette_y1	(palette->TopLeft()).y
#define		palette_x2	(palette->BottomRight()).x
#define		palette_y2	(palette->BottomRight()).y

#define		tile_x1		(small_tile_grid->TopLeft()).x
#define		tile_y1		(small_tile_grid->TopLeft()).y
#define		tile_x2		(small_tile_grid->BottomRight()).x
#define		tile_y2		(small_tile_grid->BottomRight()).y
#define		max_tile_x2	(small_tile_grid->TopLeft()).x+SIZEOF_TILEAREA
#define		max_tile_y2	(small_tile_grid->TopLeft()).y+SIZEOF_TILEAREA

#endif
