//	-------------------------------------------------------------------------
//
//	FILEMGNT.CPP
//
//	-------------------------------------------------------------------------
//
//	AUTHOR:			Andrew Carvel Evans
//
//	DESCRIPTION:	This is the File Manager for Tool
//
//	-------------------------------------------------------------------------

#include "eventmgr.hpp"
#include "pcxfile.hpp"
#include "palfile.hpp"
#include "deffile.hpp"

#include "filemgnt.hpp"

int				evhno_FileManager;
FileManager*	filemgr;

// Note: icons_name and palette_name are global variables set by TOOL
FileManager::FileManager()
	: BackgroundTask (&evhno_FileManager,0)
	{
	SetupImages();
	SetupPalette();
	}

FileManager::~FileManager()
	{
	// PALETTE STUFF
	grDestroyPalette(palette);
	delete palette_file;

	// IMAGE STUFF
	if (icons_available)
		{
		FreeImages();
		FreePCXFiles();
		FreeDEFFile();
		}

	delete definition_file;
	}

void FileManager::HandleEvent(Event& e)
	{
	BackgroundTask::HandleEvent(e);
	}

//----------------------------------------------------------------------------
//
//		SETUPIMAGES
//
//----------------------------------------------------------------------------
void FileManager::SetupImages()
	{
	// set up the definition file which lists all the image to work with
	definition_file = new DEFFile(icons_name);

	LoadDEFFile();
	LoadPCXFiles();
	LoadImages();
	}

void FileManager::SetupPalette()
	{
	// load in the palette to edit
	palette_file = new PALFile(palette_name);
	palette_file->Open();

	if (!palette_file->Exists())
		{
		palette = grCreatePalette(256);
		grGetPalette(palette,0,255);
		}
	else
		palette = palette_file->LoadPalette();

	grSetPalette(palette,0,255);
	grGetPalette(grPALETTE,0,255);
	}

//----------------------------------------------------------------------------
//
//		DISK FUNCTIONS
//
//----------------------------------------------------------------------------

void FileManager::LoadDEFFile()
	{
	if (!definition_file->Exists())	icons_available = false;
	else							icons_available = true;

	if (icons_available)
		{
		// are we working with a pcx image
		if (icons_name->Contains(".PCX"))
			{
			definition_str = new String(icons_name->Contents());
			}
		else if (icons_name->Contains(".DEF"))
			{
			// load in the definition file
			definition_file->Open();
			definition_str = definition_file->RecordList();
			definition_file->Close();
			}
		else
			{
			grExitVesa();
			printf("\n%s is an unrecognised file type!!!",icons_name->Contents());
			printf("\nFiles must be either .PCX or .DEF\n\n");
			exit(1);
			}
		}

	number_images = definition_str->NumElems();
	}

void FileManager::FreeDEFFile()
	{
	delete definition_str;
	}

void FileManager::LoadPCXFiles()
	{
	pcx_file = (PCXFile**) malloc(sizeof(PCXFile*)*number_images);

	for (int i=0; i<number_images; i++)
		{
		pcx_file[i] = new PCXFile((definition_str->Elem(i))->Contents());
		pcx_file[i]->Open();

		DBG(i*10,"Opening >%s<",pcx_file[i]->FileName());
		}
	}

void FileManager::FreePCXFiles()
	{
	for (int i=number_images-1;i>=0;i--)
		{
		DBG(i*10,"Closing >%s<",pcx_file[i]->FileName());

		pcx_file[i]->Close();
		delete pcx_file[i];
		}

	free(pcx_file);
	}

bool FileManager::ImageOnDiskIsOK(PCXFile* pf)
	{
	bool isOK = true;

	pf->Open();
	current_image = pf->LoadImage();

	if (current_image)
		{
		// if any of the bounds are out then file is not suitable
		if (current_image->hdr.width > MAX_WIDTH) isOK = false;
		if (current_image->hdr.height > MAX_HEIGHT) isOK = false;

		free(current_image);
		}
	else
		isOK = false;

	pf->Close();

	return isOK;
	}

void FileManager::LoadImage(int i)
	{
	// make sure the image is within bounds and present
	if (ImageOnDiskIsOK(pcx_file[i]))
		{
		// then read the file off the disk
		pcx_file[i]->Open();
		pcx_file[i]->LoadImage(image[i]);
		pcx_file[i]->Close();

		// since we have found an image and loaded it
		// we'll use this one as a default for all the other ones
		default_width = image[i]->hdr.width;
		default_height = image[i]->hdr.height;
		}
	else
		{
		MakeImageDefaultSize(image[i]);
		}
	}

void FileManager::LoadImages()
	{
	if (!image)
		{
		image = (grImage**) malloc(sizeof(grImage*)*number_images);
		// initially make all image as large as they're allowed to be
		for (int j=0;j<number_images;j++)
			image[j] = grCreateImage(0,0, MAX_WIDTH-1,MAX_HEIGHT-1,true);
		}

	for (int i=0;i<number_images;i++) LoadImage(i);
	}

void FileManager::SaveImage(int i)
	{
	pcx_file[i]->Open();
	pcx_file[i]->SaveImage(image[i]);
	pcx_file[i]->Close();
	}

void FileManager::SaveImages()
	{
	for (int i=0;i<number_images;i++) SaveImage(i);
	}

void FileManager::FreeImage(int i)
	{
	free(image[i]);
	}

void FileManager::FreeImages()
	{
	// free all the image
	for (int i=number_images-1;i>=0;i--) FreeImage(i);

	free(image);
	}

//============================================================================

void FileManager::LoadPalette()
	{
	// load palette off disk into a temp, then put that into memory
	grPalette* pal = palette_file->LoadPalette();
	grSetPalette(pal,0,255);
	grDestroyPalette(pal);

	// and get it back into our global structue
	grGetPalette(palette,0,255);
	}

void FileManager::SavePalette()
	{
	// if file doesn't exist create it
	if (!palette_file->Exists())
		{
		palette_file->Create();
		palette_file->Open();
		}

	int err = palette_file->SavePalette(palette);
	if (err < 0)
		DBG(460,"SavePalette didn't work #%d",err);
	}
