//	-------------------------------------------------------------------------
//
//	TOOL.CPP
//
//	-------------------------------------------------------------------------
//
//	AUTHOR:			Andrew C. Evans
//
//	DESCRIPTION:	This program is a useful tool to create graphic
//					resources such as palettes, fonts, bitmaps and sprites
//					to be used by the `grGraphics Library'.
//
//	-------------------------------------------------------------------------

#include "strings.hpp"

// include program definition
#include "tool.hpp"

// include all program specific code
#include "tooldefn.hpp"		// program-specific definitions

#include "filemgnt.hpp"		// background task for load/save to disk

#include "editwin.hpp"		// windows
#include "basewin.hpp"

#include "thumbwin.hpp"
#include "rangewin.hpp"

#include "butwin.hpp"
#include "tilewin.hpp"
#include "palwin.hpp"

//	-------------------------------------------------------------------------
//
//	FUNC:		TOOL::Constructor()
//	DESC:		Allocates memory for map storage & since the program only
//				uses a single palette, sets this also.
//
//	-------------------------------------------------------------------------
TOOL::TOOL(char* bmpname,char* palname, char* Init_String, int options) : Program("Andrew's Tool",Init_String,options)
	{
	// set up our timer to work at 100 times a second
	timer_ChangeInterruptRate(timer_100Hz);
	mouse_EnableMoveEvents();

	icons_name = new String(bmpname);
	palette_name = new String(palname);

	grPALETTE = grCreatePalette(256);
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		TOOL::Destructor()
//	DESC:		Cleans up after TOOL::Constructor()
//
//	-------------------------------------------------------------------------
TOOL::~TOOL()
	{
	grDestroyPalette(grPALETTE);

	delete palette_name;
	delete icons_name;
	}

//----------------------------------------------------------------------------
//
//		Init & Exit GUI functions
//
//----------------------------------------------------------------------------
void TOOL::Init_GUI()
	{
	// load in the GUI pics
	GUI_file = new ResourceFile(PROGRAM_PATH,"resource\\TOOLGUI.RES");

	GUIbitmap = (grImage**) malloc(sizeof(grImage*)*GUI_file->NumRecords());

	for (int i=0;i<GUI_file->NumRecords();i++)
		GUIbitmap[i] = (grImage*) GUI_file->ReadRecord(i);
	}

void TOOL::Exit_GUI()
	{
	for (int i=(GUI_file->NumRecords())-1;i>=0;i--)
		free(GUIbitmap[i]);

	free(GUIbitmap);

	delete GUI_file;
	}

//----------------------------------------------------------------------------
//
//		INITIALISE
//
//----------------------------------------------------------------------------
void TOOL::Init_Variables()
	{
	current_image_index = 0;
	default_width = MAX_WIDTH;
	default_height = MAX_HEIGHT;

	extended_image = grCreateImage(0, 0, MAX_WIDTH*MAX_TILES_ACROSS-1, MAX_HEIGHT*MAX_TILES_DOWN-1, true);
	undo = grCreateImage(0, 0, MAX_WIDTH*MAX_TILES_ACROSS-1, MAX_HEIGHT*MAX_TILES_DOWN-1, true);
	working = grCreateImage(0, 0, MAX_WIDTH*MAX_TILES_ACROSS-1, MAX_HEIGHT*MAX_TILES_DOWN-1, true);
	store = grCreateImage(0, 0, MAX_WIDTH*MAX_TILES_ACROSS-1, MAX_HEIGHT*MAX_TILES_DOWN-1, true);
	clipboard = grCreateImage(0, 0, MAX_WIDTH*MAX_TILES_ACROSS-1, MAX_HEIGHT*MAX_TILES_DOWN-1, true);
	undoUsed = false;
	storeUsed = false;
	clipboardUsed = false;

	mouseisdown = false;
	dragison = false;
	dragging = BRUSH;
	color1 = 15;
	color2 = 0;

	input_query_mode = NORMAL_INPUT;
	input_grid_mode = SINGLE;
	output_draw_mode = DRAW;
	output_fill_mode = NONE;
	output_tile_mode = SINGLE;
	output_bounds_mode = true;

	tilewin->ClearTiledImage();

	mode_x = tile_x1;
	mode_y = tile_y2 + gap;
	}

void TOOL::Exit_Variables()
	{
	free(clipboard);
	free(store);
	free(working);
	free(undo);
	free(extended_image);
	}

//----------------------------------------------------------------------------
//
//		HANDLERS
//
//----------------------------------------------------------------------------
void TOOL::Init_Handlers()
	{
	// create all handlers
	filemgr = (FileManager*) EVENTMGR->AddEventHandler(new FileManager());
	if (!icons_available) screen_mode = GUI_PALETTE;

	editwin = (EditRegion*) EVENTMGR->AddEventHandler(new EditRegion());
	basewin = (BaseRegion*) EVENTMGR->AddEventHandler(new BaseRegion());
	thumbwin = (ThumbRegion*) EVENTMGR->AddEventHandler(new ThumbRegion());
	fillrngwin = (FillRangeRegion*) EVENTMGR->AddEventHandler(new FillRangeRegion());
	palwin = (PaletteRegion*) EVENTMGR->AddEventHandler(new PaletteRegion());
	tilewin = (TileRegion*) EVENTMGR->AddEventHandler(new TileRegion());
	bmpbutwin = (BmpButtonRegion*) EVENTMGR->AddEventHandler(new BmpButtonRegion());
	palbutwin = (PalButtonRegion*) EVENTMGR->AddEventHandler(new PalButtonRegion());
	}

void TOOL::Exit_Handlers()
	{
	// delete all handlers
	EVENTMGR->DelAllEventHandlers();
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		MainLoop()
//	DESC:       This is the main loop of the program.
//
//	-------------------------------------------------------------------------
void TOOL::MainLoop()
	{
	// clear the screen to the default background colour
	grClearScreen(grBACKGROUND);
	grBlitScreen();

	Init_GUI();
	Init_Handlers();
	Init_Variables();

	// draw all the relevant handlers then execute the eventmgr
	grClearScreen(grBACKGROUND);
	EVENTMGR->DrawAllHandlers();
	grBlitScreen();
	EVENTMGR->Execute();

	Exit_Variables();
	Exit_Handlers();
	Exit_GUI();

	// clear the screen to black
	grClearScreen(grBLACK);
	grBlitScreen();
	}

//	-------------------------------------------------------------------------
//
//	FUNC:		main()
//	DESC:		This is the function called at program execution.
//				At the moment it simply creates the program, executes it,
//				then deletes it, exiting normally
//
//	-------------------------------------------------------------------------
void main(int argc, char* argv[])
	{
	// set up some default file strings
	String* defstr = new String(255);
	String* bmpstr = new String("MAKERES.DEF",255);
	String* palstr = new String("\\GAME\\UTILS\\TOOL\\TOOL.PAL",255);

	screen_mode = GUI_ICONS;

	// get the first two arguments from command line
	if (argc>1)
		{
		defstr->Set(argv[1]);
		defstr->ToUpper();

		if (defstr->Contains(".DEF") || defstr->Contains(".PCX"))
			bmpstr->Set(defstr->Contents());

		if (defstr->Contains(".PAL"))
			{
			palstr->Set(defstr->Contents());
			screen_mode = GUI_PALETTE;
			}
		}

	if (argc>2)
		{
		defstr->Set(argv[2]);
		defstr->ToUpper();

		if (defstr->Contains(".DEF") || defstr->Contains(".PCX"))
			bmpstr->Set(defstr->Contents());

		if (defstr->Contains(".PAL"))
			palstr->Set(defstr->Contents());
		}

	// create our program
	TOOL* theProgram = new TOOL(bmpstr->Contents(),palstr->Contents(),argv[0],prg_InstallSystemFont);

	// run it's intro, then setup, then mainloop, then finale
	theProgram->Execute();

	// and clear up after ourselves
	delete theProgram;

	delete palstr;
	delete bmpstr;
	delete defstr;
	}
